Package: tidyllm
Title: Tidy Integration of Large Language Models
Version: 0.5.0
Authors@R: c(
    person("Eduard", "Brüll", NULL, "eduard.bruell@zew.de", role = c("aut", "cre")),
    person("Jia", "Zhang", NULL, NULL, role = "ctb")
  )
Description: A tidy interface for integrating large language model (LLM) APIs such as 'Claude', 'OpenAI', 'Gemini', 'Mistral', and local models via 'Ollama' into R workflows. The package supports text, image, audio, video, and document interactions; a unified media interface for attaching inline files or uploading to provider file stores; batch request APIs for cost-efficient large-scale processing; and a pipeline-oriented interface for seamless integration into data workflows. Web services are available at <https://www.anthropic.com>, <https://openai.com>, <https://aistudio.google.com/>, <https://mistral.ai/> and <https://ollama.com>.
License: MIT + file LICENSE
URL: https://edubruell.github.io/tidyllm/
BugReports: https://github.com/edubruell/tidyllm/issues
Encoding: UTF-8
RoxygenNote: 7.3.3
VignetteBuilder: knitr
Suggests: knitr, rmarkdown, testthat (>= 3.0.0), tidyverse, httptest2,
        httpuv, ellmer
Imports: S7 (>= 0.2.0), base64enc, glue, jsonlite, curl, httr2 (>=
        1.1.1), lubridate, purrr, rlang, stringr, grDevices, pdftools,
        tibble, cli, png, lifecycle
Collate: 'tidyllm-package.R' 'utilites.R' 'embedding_helpers.R'
        'LLMMessage.R' 'APIProvider.R' 'llm_message.R' 'llm_verbs.R'
        'media.R' 'message_retrieval.R' 'perform_api_requests.R'
        'rate_limits.R' 'tidyllm_schema.R' 'tools.R'
        'api_chat_completions.R' 'api_openai.R' 'api_azure_openai.R'
        'api_groq.R' 'api_mistral.R' 'api_perplexity.R'
        'api_deepseek.R' 'api_voyage.R' 'api_ollama.R' 'api_claude.R'
        'api_gemini.R' 'api_ellmer.R' 'api_openrouter.R'
        'api_llamacpp.R' 'pdfbatch.R' 'zzz.R'
Depends: R (>= 4.2.0)
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-29 20:30:12 UTC; ebr
Author: Eduard Brüll [aut, cre],
  Jia Zhang [ctb]
Maintainer: Eduard Brüll <eduard.bruell@zew.de>
Repository: CRAN
Date/Publication: 2026-04-30 08:30:08 UTC
Built: R 4.5.2; ; 2026-05-01 18:28:36 UTC; unix

Package: tidyllm
Title: Tidy Integration of Large Language Models
Version: 0.5.0
Authors@R: c(
    person("Eduard", "Brüll", NULL, "eduard.bruell@zew.de", role = c("aut", "cre")),
    person("Jia", "Zhang", NULL, NULL, role = "ctb")
  )
Description: A tidy interface for integrating large language model (LLM) APIs such as 'Claude', 'OpenAI', 'Gemini', 'Mistral', and local models via 'Ollama' into R workflows. The package supports text, image, audio, video, and document interactions; a unified media interface for attaching inline files or uploading to provider file stores; batch request APIs for cost-efficient large-scale processing; and a pipeline-oriented interface for seamless integration into data workflows. Web services are available at <https://www.anthropic.com>, <https://openai.com>, <https://aistudio.google.com/>, <https://mistral.ai/> and <https://ollama.com>.
License: MIT + file LICENSE
URL: https://edubruell.github.io/tidyllm/
BugReports: https://github.com/edubruell/tidyllm/issues
Encoding: UTF-8
RoxygenNote: 7.3.3
VignetteBuilder: knitr
Suggests: knitr, rmarkdown, testthat (>= 3.0.0), tidyverse, httptest2,
        httpuv, ellmer
Imports: S7 (>= 0.2.0), base64enc, glue, jsonlite, curl, httr2 (>=
        1.1.1), lubridate, purrr, rlang, stringr, grDevices, pdftools,
        tibble, cli, png, lifecycle
Collate: 'tidyllm-package.R' 'utilites.R' 'embedding_helpers.R'
        'LLMMessage.R' 'APIProvider.R' 'llm_message.R' 'llm_verbs.R'
        'media.R' 'message_retrieval.R' 'perform_api_requests.R'
        'rate_limits.R' 'tidyllm_schema.R' 'tools.R'
        'api_chat_completions.R' 'api_openai.R' 'api_azure_openai.R'
        'api_groq.R' 'api_mistral.R' 'api_perplexity.R'
        'api_deepseek.R' 'api_voyage.R' 'api_ollama.R' 'api_claude.R'
        'api_gemini.R' 'api_ellmer.R' 'api_openrouter.R'
        'api_llamacpp.R' 'pdfbatch.R' 'zzz.R'
Depends: R (>= 4.2.0)
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-29 20:30:12 UTC; ebr
Author: Eduard Brüll [aut, cre],
  Jia Zhang [ctb]
Maintainer: Eduard Brüll <eduard.bruell@zew.de>
Repository: CRAN
Date/Publication: 2026-04-30 08:30:08 UTC
Built: R 4.6.0; ; 2026-05-01 21:42:57 UTC; unix

# Generated by roxygen2: do not edit by hand

export(.onLoad)
export(LLMMessage)
export(audio_file)
export(azure_openai)
export(azure_openai_chat)
export(azure_openai_embedding)
export(cancel_openai_batch)
export(chat)
export(chat_completions)
export(chat_completions_chat)
export(chat_ellmer)
export(chatgpt)
export(check_azure_openai_batch)
export(check_batch)
export(check_claude_batch)
export(check_gemini_batch)
export(check_groq_batch)
export(check_job)
export(check_mistral_batch)
export(check_openai_batch)
export(claude)
export(claude_chat)
export(claude_delete_file)
export(claude_file_metadata)
export(claude_list_files)
export(claude_list_models)
export(claude_upload_file)
export(claude_websearch)
export(deep_research)
export(deepseek)
export(deepseek_chat)
export(delete_file)
export(df_llm_message)
export(ellmer)
export(ellmer_tool)
export(embed)
export(fetch_azure_openai_batch)
export(fetch_batch)
export(fetch_claude_batch)
export(fetch_gemini_batch)
export(fetch_groq_batch)
export(fetch_job)
export(fetch_mistral_batch)
export(fetch_openai_batch)
export(field_chr)
export(field_dbl)
export(field_fct)
export(field_lgl)
export(field_object)
export(file_info)
export(gemini)
export(gemini_chat)
export(gemini_delete_file)
export(gemini_embedding)
export(gemini_file_metadata)
export(gemini_list_files)
export(gemini_list_models)
export(gemini_upload_file)
export(get_logprobs)
export(get_metadata)
export(get_reply)
export(get_reply_data)
export(get_user_message)
export(groq)
export(groq_chat)
export(groq_list_models)
export(groq_transcribe)
export(img)
export(last_metadata)
export(last_reply)
export(last_reply_data)
export(last_user_message)
export(list_azure_openai_batches)
export(list_batches)
export(list_claude_batches)
export(list_files)
export(list_gemini_batches)
export(list_groq_batches)
export(list_hf_gguf_files)
export(list_mistral_batches)
export(list_models)
export(list_openai_batches)
export(llamacpp)
export(llamacpp_chat)
export(llamacpp_delete_model)
export(llamacpp_download_model)
export(llamacpp_embedding)
export(llamacpp_health)
export(llamacpp_list_local_models)
export(llamacpp_list_models)
export(llamacpp_rerank)
export(llm_message)
export(mistral)
export(mistral_chat)
export(mistral_embedding)
export(mistral_list_models)
export(ollama)
export(ollama_chat)
export(ollama_delete_model)
export(ollama_download_model)
export(ollama_embedding)
export(ollama_list_models)
export(openai)
export(openai_chat)
export(openai_check_research)
export(openai_code_interpreter)
export(openai_deep_research)
export(openai_embedding)
export(openai_fetch_research)
export(openai_list_models)
export(openai_websearch)
export(openrouter)
export(openrouter_chat)
export(openrouter_credits)
export(openrouter_embedding)
export(openrouter_generation)
export(openrouter_list_models)
export(pdf_file)
export(pdf_page_batch)
export(perplexity)
export(perplexity_chat)
export(perplexity_check_research)
export(perplexity_deep_research)
export(perplexity_fetch_research)
export(rate_limit_info)
export(send_azure_openai_batch)
export(send_batch)
export(send_claude_batch)
export(send_gemini_batch)
export(send_groq_batch)
export(send_mistral_batch)
export(send_ollama_batch)
export(send_openai_batch)
export(tidyllm_schema)
export(tidyllm_tool)
export(upload_file)
export(video_file)
export(voyage)
export(voyage_embedding)
export(voyage_rerank)
if (getRversion() < "4.3.0") importFrom("S7", "@")
import(S7)
import(base64enc)
import(pdftools)
importFrom(base64enc,base64encode)
importFrom(glue,glue)
importFrom(grDevices,dev.copy)
importFrom(grDevices,dev.off)
importFrom(grDevices,png)
importFrom(jsonlite,toJSON)
importFrom(rlang,"%||%")
importFrom(tibble,as_tibble)

# tidyllm <a href="https://edubruell.github.io/tidyllm/"><img src="man/figures/logo.png" align="right" height="139" alt="tidyllm website" /></a>

[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)
[![CRAN Status](https://www.r-pkg.org/badges/version/tidyllm)](https://cran.r-project.org/package=tidyllm)


**tidyllm** is an R package designed to access various large language model APIs, including **Anthropic Claude**, **OpenAI**,**Google Gemini**, **Perplexity**,**Groq**, **Mistral**, and local models via **Ollama** or OpenAI-compatible APIs. Built for simplicity and functionality, it helps you generate text, analyze media, and integrate model feedback into your data workflows with ease.

## Features

- **Multiple Model Support**: Seamlessly switch between various model providers using the best of what each has to offer.
- **Media Handling**: Extract and process text from PDFs and capture console outputs for messaging. Upload imagefiles or the last plotpane to multimodal models. For the Gemini API even video and audio inputs are supported.
- **Interactive Messaging History**: Manage an ongoing conversation with models, maintaining a structured history of messages and media interactions, which are automatically formatted for each API
- **Batch processing:** Efficiently handle large workloads with Anthropic, OpenAI or Mistral batch processing APIs, reducing costs by up to 50%.
- **Tidy Workflow**: Use R's functional programming features for a side-effect-free, pipeline-oriented operation style.


## Installation

To install **tidyllm** from CRAN, use:

```r
install.packages("tidyllm")
```

Or for the development version from GitHub:
```r
# Install devtools if not already installed
if (!requireNamespace("devtools", quietly = TRUE)) {
  install.packages("devtools")
}
devtools::install_github("edubruell/tidyllm")
```

## Basic Example

Here’s a quick  example using tidyllm to describe an image using the Claude model 
to and follow up with local open-source models:

```R
library("tidyllm")

# Describe an image with  claude
conversation <- llm_message("Describe this image", 
                              .imagefile = here("image.png")) |>
  chat(claude())

# Use the description to query further with groq
conversation |>
  llm_message("Based on the previous description,
  what could the research in the figure be about?") |>
  chat(ollama(.model = "gemma2"))
```

For more examples and advanced usage, check the [Get Started vignette](https://edubruell.github.io/tidyllm/articles/tidyllm.html).

Please note: To use **tidyllm**, you need either an installation of **ollama** or an active API key for one of the supported providers (e.g., Claude, ChatGPT). See the [Get Started vignette](https://edubruell.github.io/tidyllm/articles/tidyllm.html) for setup instructions.

## Interface-change in 0.3.0

The CRAN release of **tidyllm** 0.3.0,  introduced a major interface change to provide a more intuitive user experience. Previously, provider-specific functions like `claude()`, `openai()`, and others were directly used for chat-based workflows. They specified both an API-provider and performed a chat-interaction. Now, these functions primarily serve as provider configuration for more general verbs like `chat()`,`embed()` or `send_batch()`. A combination of a general verb and a provider will always route requests to a provider-specific function like `openai_chat()`. Read the [Changelog](https://edubruell.github.io/tidyllm/news/) or the [package vignette](https://edubruell.github.io/tidyllm/articles/tidyllm.html) for more information. 

For backward compatibility, the old use of functions like `openai()` or `claude()` directly for chat requests still works but now but issues deprecation warnings. It is recommended to either use the verb-based interface:
```r
llm_message("Hallo") |> chat(openai(.model="gpt-4o"))
```
or to use the more verbose provider-specific functions directly:
```r
llm_message("Hallo") |> openai_chat(.model="gpt-4o")
```

## Learn More

For detailed instructions and advanced features, see:

- [Get Started with tidyllm](https://edubruell.github.io/tidyllm/articles/tidyllm.html)
- [Changelog](https://edubruell.github.io/tidyllm/news/)
- [Documentation](https://edubruell.github.io/tidyllm/)
- Use-case oriented articles:
  - [Classifying Texts with tidyllm](https://edubruell.github.io/tidyllm/articles/tidyllm_classifiers.html)
  - [Structured Question Answering from PDFs](https://edubruell.github.io/tidyllm/articles/tidyllm-pdfquestions.html)
  - [Embedding Models in tidyllm](https://edubruell.github.io/tidyllm/articles/tidyllm_embed.html)
  - [Video and Audio Data with the Gemini API](https://edubruell.github.io/tidyllm/articles/tidyllm_video.html)
  - [Local Models with tidyllm](https://edubruell.github.io/tidyllm/articles/tidyllm_local_models.html)
  
## Similar packages
The are some similar R packages for working with LLMs:

  - [ellmer](https://ellmer.tidyverse.org/) is especially great for asynchronous chats, chatbots in Shiny and advanced tool-calling capabilities. Its schema functions offer robust support for complex structured data extraction, making it a great choice for applications that require highly interactive or structured LLM interactions. While **ellmer**’s feature set overlaps with **tidyllm** in some areas, its interface and design philosophy are very different.
  - [rollama](https://jbgruber.github.io/rollama/) is specifically designed to support the Ollama API, enabling seamless interaction with local LLM models. A key strength of **rollama** lies in its specialized Ollama API functionalities, such as `copy` and `create` which are not currently available in **tidyllm**. These features make **rollama** particularly suited for workflows requiring model management or deployment within the Ollama ecosystem.
  
## Contributing

We welcome contributions! Feel free to open issues or submit pull requests on [GitHub](https://github.com/edubruell/tidyllm).

## License

This project is licensed under the MIT License - see the [LICENSE](https://opensource.org/licenses/MIT) file for details.

# tidyllm <a href="https://edubruell.github.io/tidyllm/"><img src="man/figures/logo.png" align="right" height="139" alt="tidyllm website" /></a>

[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)
[![CRAN Status](https://www.r-pkg.org/badges/version/tidyllm)](https://cran.r-project.org/package=tidyllm)


**tidyllm** is an R package for working with large language model APIs in data analysis workflows. It supports **Anthropic Claude**, **OpenAI**, **Google Gemini**, **Mistral**, **Groq**, **Perplexity**, **DeepSeek**, **OpenRouter**, local models via **Ollama** and **llama.cpp**, and more — all through a single consistent interface.

## Features

- **Multiple providers**: Switch between cloud and local models using the same verb + provider pattern.
- **Unified media system**: Send images, audio, video, and PDFs to any provider that supports them via `.media`. Upload files to provider servers for reuse via `.files` and `upload_file()`.
- **Interactive message history**: Manage multi-turn conversations with structured history automatically formatted for each API.
- **Batch processing**: Handle large workloads with Anthropic, OpenAI, Mistral, Groq, and Gemini batch APIs, reducing costs by up to 50%.
- **Tidy workflow**: Pipeline-oriented, side-effect-free design that integrates naturally with tidyverse data workflows.


## Installation

To install **tidyllm** from CRAN, use:

```r
install.packages("tidyllm")
```

Or for the development version from GitHub:
```r
devtools::install_github("edubruell/tidyllm")
```

## Basic Example

```r
library(tidyllm)

# Describe an image with Claude, continue with a local model
conversation <- llm_message("Describe this image.",
                             .media = img("photo.jpg")) |>
  chat(claude())

conversation |>
  llm_message("Based on that description, what research topic could this figure relate to?") |>
  chat(ollama(.model = "qwen3.5:4b"))
```

For more examples and advanced usage, see the [Get Started vignette](https://edubruell.github.io/tidyllm/articles/tidyllm.html).

Please note: To use **tidyllm** you need either a local Ollama or llama.cpp installation, or an active API key for one of the supported cloud providers. See the [Get Started vignette](https://edubruell.github.io/tidyllm/articles/tidyllm.html) for setup instructions.

## What's new in 0.5.0

**Unified media and files system.** All non-text content now attaches to messages, not to chat() call arguments:

```r
# Inline binary — works with any provider that supports the type
llm_message("Transcribe this audio.", .media = audio_file("recording.mp3")) |>
  chat(gemini())

llm_message("Summarize this paper.", .media = pdf_file("paper.pdf")) |>
  chat(claude())

# Multiple images in one message
llm_message("Compare these two charts.",
            .media = list(img("chart_a.png"), img("chart_b.png"))) |>
  chat(openai())

# Provider Files API — upload once, reuse across requests
report <- upload_file(claude(), .path = "annual_report.pdf")
llm_message("What are the key risks?", .files = report) |>
  chat(claude())
```

Audio and video are supported by Gemini, OpenRouter (Gemini, Gemma 4, Qwen 3.6, and other models), Mistral (Voxtral), and llama.cpp (Ultravox, Gemma 4, Qwen2.5-Omni). The `.imagefile` and old provider-specific upload functions still work but now emit deprecation warnings.

OpenAI now uses the Responses API (`POST /v1/responses`) with reasoning effort control, built-in web search, and background deep research:

```r
llm_message("What are the latest developments in fusion energy?") |>
  chat(openai(), .tools = openai_websearch())

llm_message("Write a detailed report on EU AI regulation.") |>
  deep_research(openai())
```

Read the [Changelog](https://edubruell.github.io/tidyllm/news/) for the full list of changes.

## Learn More

- [Get Started with tidyllm](https://edubruell.github.io/tidyllm/articles/tidyllm.html)
- [Changelog](https://edubruell.github.io/tidyllm/news/)
- [Documentation](https://edubruell.github.io/tidyllm/)
- Use-case oriented articles:
  - [Classifying Texts with tidyllm](https://edubruell.github.io/tidyllm/articles/tidyllm_classifiers.html)
  - [Structured Question Answering from PDFs](https://edubruell.github.io/tidyllm/articles/tidyllm-pdfquestions.html)
  - [Embedding Models in tidyllm](https://edubruell.github.io/tidyllm/articles/tidyllm_embed.html)
  - [Working with Files and Media](https://edubruell.github.io/tidyllm/articles/tidyllm_video.html)
  - [Local Models with tidyllm](https://edubruell.github.io/tidyllm/articles/tidyllm_local_models.html)

## Similar packages

- [ellmer](https://ellmer.tidyverse.org/) is especially well-suited for asynchronous chats, chatbots in Shiny, and advanced tool-calling workflows. Its design philosophy differs from tidyllm: ellmer targets interactive agents while tidyllm targets data pipelines and batch operations.
- [rollama](https://jbgruber.github.io/rollama/) is purpose-built for the Ollama API with specialized model management features not currently in tidyllm.

## Contributing

We welcome contributions! Feel free to open issues or submit pull requests on [GitHub](https://github.com/edubruell/tidyllm).

## License

This project is licensed under the MIT License — see the [LICENSE](https://opensource.org/licenses/MIT) file for details.

## ----eval=FALSE---------------------------------------------------------------
# install.packages("tidyllm")

## ----eval=FALSE---------------------------------------------------------------
# devtools::install_github("edubruell/tidyllm")

## ----eval=FALSE---------------------------------------------------------------
# Sys.setenv(OPENAI_API_KEY = "your-key-here")

## ----eval=FALSE---------------------------------------------------------------
# ollama_download_model("qwen3.5:4b")

## ----convo1, eval=FALSE-------------------------------------------------------
# library(tidyllm)
# 
# conversation <- llm_message("What is the capital of France?") |>
#   chat(claude())
# 
# conversation

## ----convo1_out, eval=TRUE, echo=FALSE, message=FALSE-------------------------
library(tidyllm)

conversation <- llm_message("What is the capital of France?") |>
  llm_message("The capital of France is Paris.", .role = "assistant")

conversation

## ----convo2, eval=FALSE-------------------------------------------------------
# # Continue the conversation with a different provider
# conversation <- conversation |>
#   llm_message("What's a famous landmark in this city?") |>
#   chat(openai())
# 
# get_reply(conversation)

## ----convo2_out, eval=TRUE, echo=FALSE----------------------------------------
c("A famous landmark in Paris is the Eiffel Tower.")

## ----echo=FALSE, out.width="70%", fig.alt="A photograph showing lake Garda and the scenery near Torbole, Italy."----
knitr::include_graphics("picture.jpeg")

## ----images, eval=FALSE-------------------------------------------------------
# # Single image
# image_description <- llm_message("Describe this picture. Can you guess where it was taken?",
#                                   .media = img("picture.jpeg")) |>
#   chat(openai(.model = "gpt-5.4"))
# 
# get_reply(image_description)

## ----images_out, eval=TRUE, echo=FALSE----------------------------------------
c("The picture shows a beautiful landscape with a lake, mountains, and a town nestled below. The area appears lush and green, with agricultural fields visible. This scenery is reminiscent of northern Italy, particularly around Lake Garda.")

## ----eval=FALSE---------------------------------------------------------------
# llm_message("Compare these two charts.",
#             .media = list(img("chart_a.png"), img("chart_b.png"))) |>
#   chat(claude())
# 
# # Mixed types in one message
# llm_message("Does this figure match the results described in the paper?",
#             .media = list(img("figure.png"), pdf_file("paper.pdf", pages = 1:5))) |>
#   chat(gemini())

## ----eval=FALSE---------------------------------------------------------------
# # Audio transcription or analysis
# llm_message("Transcribe and summarise this recording.",
#             .media = audio_file("interview.mp3")) |>
#   chat(gemini())
# 
# # Video understanding
# llm_message("Describe what happens in this clip.",
#             .media = video_file("demo.mp4")) |>
#   chat(gemini())

## ----pdf, eval=FALSE----------------------------------------------------------
# llm_message("Summarize the key points from this document.",
#             .media = pdf_file("die_verwandlung.pdf")) |>
#   chat(claude())

## ----pdf_out, eval=TRUE, echo=FALSE-------------------------------------------
llm_message("Summarize the key points from this document.",
            .media = pdf_file("die_verwandlung.pdf")) |>
  llm_message("The story centres on Gregor Samsa, who wakes up transformed into a giant insect. Unable to work, he becomes isolated while his family struggles. Eventually Gregor dies, and his relieved family looks ahead to a better future.",
              .role = "assistant")

## ----eval=FALSE---------------------------------------------------------------
# pdf_file("report.pdf", pages = 1:5)

## ----eval=FALSE---------------------------------------------------------------
# report <- upload_file(claude(), .path = "annual_report.pdf")
# 
# llm_message("What are the key risks mentioned?", .files = report) |>
#   chat(claude())
# 
# # Manage uploaded files
# list_files(claude())
# file_info(claude(), report)
# delete_file(claude(), report)

## ----routputs_base, eval=TRUE, echo=TRUE, message=FALSE, fig.alt="Scatter plot of car weight (x-axis) versus miles per gallon (y-axis) from the mtcars dataset, with a fitted linear regression line showing a negative relationship between weight and fuel efficiency."----
library(tidyverse)

ggplot(mtcars, aes(wt, mpg)) +
  geom_point() +
  geom_smooth(method = "lm", formula = "y ~ x") +
  labs(x = "Weight", y = "Miles per gallon")

## ----routputs_base_out, eval=FALSE--------------------------------------------
# llm_message("Analyze this plot and data summary:",
#             .capture_plot = TRUE,
#             .f = ~{summary(mtcars)}) |>
#   chat(claude())

## ----routputs_fake, eval=TRUE, echo=FALSE-------------------------------------
llm_message("Analyze this plot and data summary:",
            .imagefile = "file1568f6c1b4565.png",
            .f = ~{summary(mtcars)}) |>
  llm_message("The scatter plot shows a clear negative correlation between weight and fuel efficiency. Heavier cars consistently achieve lower mpg, with the linear trend confirming this relationship. Variability around the line suggests other factors, engine size and transmission, also play a role.", .role = "assistant")

## ----eval=FALSE---------------------------------------------------------------
# conversation <- llm_message("Imagine a German address.") |>
#   chat(groq()) |>
#   llm_message("Imagine another address.") |>
#   chat(claude())
# 
# conversation

## ----eval=TRUE, echo=FALSE----------------------------------------------------
conversation <- llm_message("Imagine a German address.") |>
  llm_message("Herr Müller\nMusterstraße 12\n53111 Bonn", .role = "assistant") |>
  llm_message("Imagine another address.") |>
  llm_message("Frau Schmidt\nFichtenweg 78\n42103 Wuppertal", .role = "assistant")
conversation

## ----standard_last_reply, eval=TRUE-------------------------------------------
conversation |> get_reply(1)   # first reply
conversation |> get_reply()    # last reply (default)

## ----as_tibble, eval=TRUE-----------------------------------------------------
conversation |> as_tibble()

## ----get_metadata, eval=FALSE-------------------------------------------------
# conversation |> get_metadata()

## ----get_metadata_out, eval=TRUE, echo=FALSE----------------------------------
tibble::tibble(
  model             = c("groq-model", "claude-sonnet-4-6-20251001"),
  timestamp         = as.POSIXct(c("2025-11-08 14:25:43", "2025-11-08 14:26:02")),
  prompt_tokens     = c(20L, 80L),
  completion_tokens = c(45L, 40L),
  total_tokens      = c(65L, 120L),
  api_specific      = list(list(), list())
) |> print(width = 60)

## ----list_models, eval=FALSE--------------------------------------------------
# list_models(openai())
# list_models(openrouter())   # 300+ models across providers
# list_models(ollama())       # models installed locally

## ----schema, eval=FALSE-------------------------------------------------------
# person_schema <- tidyllm_schema(
#   first_name = field_chr("A male first name"),
#   last_name  = field_chr("A common last name"),
#   occupation = field_chr("A quirky occupation"),
#   address    = field_object(
#     "The person's home address",
#     street  = field_chr("Street name"),
#     number  = field_dbl("House number"),
#     city    = field_chr("A large city"),
#     zip     = field_dbl("Postal code"),
#     country = field_fct("Country", .levels = c("Germany", "France"))
#   )
# )
# 
# profile <- llm_message("Imagine a person profile matching the schema.") |>
#   chat(openai(), .json_schema = person_schema)
# 
# profile

## ----schema_out, eval=TRUE, echo=FALSE----------------------------------------
profile <- llm_message("Imagine a person profile matching the schema.") |>
  llm_message("{\"first_name\":\"Julien\",\"last_name\":\"Martin\",\"occupation\":\"Gondola Repair Specialist\",\"address\":{\"street\":\"Rue de Rivoli\",\"number\":112,\"city\":\"Paris\",\"zip\":75001,\"country\":\"France\"}}",
              .role = "assistant")
profile@message_history[[3]]$json <- TRUE
profile

## ----get_reply_data-----------------------------------------------------------
profile |> get_reply_data() |> str()

## ----fobj, eval=FALSE---------------------------------------------------------
# llm_message("Imagine five people.") |>
#   chat(openai(),
#        .json_schema = tidyllm_schema(
#          person = field_object(
#            first_name = field_chr(),
#            last_name  = field_chr(),
#            occupation = field_chr(),
#            .vector    = TRUE
#          )
#        )) |>
#   get_reply_data() |>
#   bind_rows()

## ----fobjout, echo=FALSE, eval=TRUE-------------------------------------------
data.frame(
  first_name = c("Alice", "Robert", "Maria", "Liam", "Sophia"),
  last_name  = c("Johnson", "Anderson", "Gonzalez", "O'Connor", "Lee"),
  occupation = c("Software Developer", "Graphic Designer", "Data Scientist",
                 "Mechanical Engineer", "Content Writer")
)

## ----temperature, eval=FALSE--------------------------------------------------
# # Common args in the verb
# llm_message("Write a haiku about tidyllm.") |>
#   chat(ollama(), .temperature = 0)
# 
# # Provider-specific args in the provider
# llm_message("Hello") |>
#   chat(ollama(.ollama_server = "http://my-server:11434"), .temperature = 0)

## ----eval=FALSE---------------------------------------------------------------
# get_current_time <- function(tz, format = "%Y-%m-%d %H:%M:%S") {
#   format(Sys.time(), tz = tz, format = format, usetz = TRUE)
# }
# 
# time_tool <- tidyllm_tool(
#   .f          = get_current_time,
#   .description = "Returns the current time in a specified timezone.",
#   tz     = field_chr("The timezone identifier, e.g. 'Europe/Berlin'."),
#   format = field_chr("strftime format string. Default: '%Y-%m-%d %H:%M:%S'.")
# )
# 
# llm_message("What time is it in Stuttgart right now?") |>
#   chat(openai(), .tools = time_tool)

## ----eval=TRUE, echo=FALSE----------------------------------------------------
llm_message("What time is it in Stuttgart right now?") |>
  llm_message("The current time in Stuttgart (Europe/Berlin) is 2025-03-03 09:51:22 CET.",
              .role = "assistant")

## ----eval=FALSE---------------------------------------------------------------
# library(ellmer)
# btw_tool <- ellmer_tool(btw::btw_tool_files_list_files)
# llm_message("List files in the R/ folder.") |>
#   chat(claude(), .tools = btw_tool)

## ----embed, eval=FALSE--------------------------------------------------------
# c("What is the meaning of life?",
#   "How much wood would a woodchuck chuck?",
#   "How does the brain work?") |>
#   embed(ollama())

## ----embed_output, eval=TRUE, echo=FALSE--------------------------------------
tibble::tibble(
  input      = c("What is the meaning of life?",
                 "How much wood would a woodchuck chuck?",
                 "How does the brain work?"),
  embeddings = purrr::map(1:3, ~runif(min = -1, max = 1, n = 384))
)

## ----eval=FALSE---------------------------------------------------------------
# list("tidyllm logo", img("docs/logo.png")) |>
#   embed(voyage())

## ----eval=FALSE---------------------------------------------------------------
# voyage_rerank(
#   "best R package for LLMs",
#   c("tidyllm", "ellmer", "httr2", "rvest")
# )

## ----eval=FALSE---------------------------------------------------------------
# # Submit a batch and save the job handle to disk
# glue::glue("Write a poem about {x}", x = c("cats", "dogs", "hamsters")) |>
#   purrr::map(llm_message) |>
#   send_batch(claude()) |>
#   saveRDS("claude_batch.rds")

## ----eval=FALSE---------------------------------------------------------------
# # Check status
# readRDS("claude_batch.rds") |>
#   check_batch(claude())

## ----echo=FALSE---------------------------------------------------------------
tibble::tribble(
  ~batch_id,                          ~status, ~created_at,                      ~expires_at,                      ~req_succeeded, ~req_errored, ~req_expired, ~req_canceled,
  "msgbatch_02A1B2C3D4E5F6G7H8I9J",   "ended", as.POSIXct("2025-11-01 10:30:00"), as.POSIXct("2025-11-02 10:30:00"), 3L,             0L,           0L,           0L
)

## ----eval=FALSE---------------------------------------------------------------
# # Fetch completed results
# conversations <- readRDS("claude_batch.rds") |>
#   fetch_batch(claude())
# 
# poems <- purrr::map_chr(conversations, get_reply)

## ----eval=FALSE---------------------------------------------------------------
# # Blocking: waits for completion, returns an LLMMessage
# result <- llm_message("Compare Rust and Go for systems programming in 2025.") |>
#   deep_research(perplexity())
# 
# get_reply(result)
# get_metadata(result)$api_specific[[1]]$citations

## ----eval=FALSE---------------------------------------------------------------
# # Background mode: returns a job handle immediately
# job <- llm_message("Summarise recent EU AI Act developments.") |>
#   deep_research(perplexity(), .background = TRUE)
# 
# check_job(job)           # poll status
# result <- fetch_job(job) # retrieve when complete

## ----eval=FALSE---------------------------------------------------------------
# llm_message("Write a short poem about R.") |>
#   chat(claude(), .stream = TRUE)

## ----eval=FALSE---------------------------------------------------------------
# options(tidyllm_chat_default   = openai(.model = "gpt-5.4"))
# options(tidyllm_embed_default  = ollama())
# options(tidyllm_sbatch_default = claude(.temperature = 0))
# options(tidyllm_cbatch_default = claude())
# options(tidyllm_fbatch_default = claude())
# options(tidyllm_lbatch_default = claude())
# 
# # Now the provider argument can be omitted
# llm_message("Hello!") |> chat()
# 
# c("text one", "text two") |> embed()

#' @title TOOL Class
#' @description A class representing a tool for Language Model function calling
#' 
#' @slot description Character string describing what the tool does
#' @slot input_schema List of parameter schemas for the tool (empty for builtin tools)
#' @slot func Function to be called by the LLM (dummy function for builtin tools that raises an error)
#' @slot name Character string name of the tool
#' @slot builtin List containing provider-native builtin tool definitions (empty for custom tools)
#'
#' @noRd 
TOOL <- new_class("TOOL", properties = list(
  description  = class_character,
  input_schema = class_list,
  func         = class_function,
  name         = class_character,
  builtin      = class_list
))


#' Create a Tool Definition for tidyllm
#'
#' @description
#' Creates a tool definition for use with Language Model API calls that support function calling.
#' This function wraps an existing R function with schema information for LLM interaction.
#'
#' @param .f The function to wrap as a tool 
#' @param .description Character string describing what the tool does
#' @param ... Named arguments providing schema definitions for each function parameter using tidyllm_fields
#'
#' @return A `TOOL` class object that can be used with tidyllm `chat()` functions
#'
#' @details
#' Each parameter schema in `...` should correspond to a parameter in the wrapped function.
#' All required function parameters must have corresponding schema definitions.
#'
#' @examples
#' get_weather <- function(location){}
#' weather_tool <- tidyllm_tool(
#'   get_weather,
#'   "Get the current weather in a given location",
#'   location = field_chr("The city and state, e.g., San Francisco, CA")
#' )
#'
#' @export
tidyllm_tool <- function(.f, .description = character(0), ...) {
  # Convert formula to function if needed
  formula_flag <- FALSE
  if (rlang::is_formula(.f)) {
    .f <- rlang::as_function(.f)
    formula_flag <- TRUE
  }
  
  # Check if function is anonymous
  fn_name <- tryCatch(
    rlang::as_name(rlang::ensym(.f)), 
    error = function(e) NULL
  )
  
  # Assign default name if function is anonymous
  if (is.null(fn_name) || fn_name == "") {
    fn_hash <- substr(rlang::hash(.f), 1, 8)  # Shorten hash
    .name <- paste0("anonymous_", fn_hash)
  } else {
    .name <- fn_name
  }
  
  # Extract function arguments and required ones
  fn_args <- rlang::fn_fmls(.f)
  required_args <- names(purrr::keep(fn_args, ~ identical(., rlang::missing_arg())))
  if (formula_flag==TRUE) {
    required_args <- ".x"
  }
  # Capture provided schema definitions
  schema_args <- list(...)
  
  # Ensure all required function args are in schema
  missing_args <- setdiff(required_args, names(schema_args))
  if (length(missing_args) > 0) {
    stop("Missing schema definitions for required arguments: ", paste(missing_args, collapse = ", "))
  }
  
  # Ensure all schema arguments are valid function arguments
  invalid_args <- setdiff(names(schema_args), names(fn_args))
  if (length(invalid_args) > 0) {
    stop("Schema defines arguments not present in function: ", paste(invalid_args, collapse = ", "))
  }
  
  obj <- TOOL(
    description = .description,
    input_schema = schema_args,
    func = .f,
    name = .name,
    builtin = list()
  )
  
  obj
}

is_ellmer_class <- function(classes, class_name) {
  any(classes %in% c(paste0("ellmer::", class_name), paste0("ellmer_", class_name)))
}

#' Convert an ellmer Tool to a tidyllm TOOL
#'
#' @description
#' Converts an ellmer `ToolDef` or `ToolBuiltIn` object to a tidyllm `TOOL` object,
#' allowing seamless integration of ellmer-defined tools and builtin provider tools
#' with tidyllm workflows.
#'
#' @param .ellmer_tool An ellmer `ToolDef` object created via `ellmer::tool()`,
#'   or a builtin tool like `ellmer::claude_tool_web_search()`
#'
#' @return A `TOOL` class object that can be used with tidyllm `chat()` functions
#'
#' @details
#' This function supports two types of ellmer tools:
#'
#' **Custom ToolDef objects**: Extracts the function, description, and argument
#' schemas from an ellmer tool and converts them to tidyllm's internal representation.
#' Ellmer type objects are automatically converted to tidyllm field descriptors.
#'
#' **Builtin ToolBuiltIn objects**: Converts provider-native builtin tools (like
#' Claude's web search) to tidyllm format. For builtin tools, the tool definition
#' is passed through as-is to the provider's API, which handles the tool execution
#' natively.
#'
#' @examples
#' \dontrun{
#' library(ellmer)
#'
#' # Example 1: Custom tool
#' tool_rnorm <- ellmer::tool(
#'   rnorm,
#'   description = "Draw numbers from a random normal distribution",
#'   arguments = list(
#'     n = ellmer::type_integer("The number of observations"),
#'     mean = ellmer::type_number("The mean value"),
#'     sd = ellmer::type_number("The standard deviation")
#'   )
#' )
#'
#' tidyllm_tool_rnorm <- ellmer_tool(tool_rnorm)
#'
#' llm_message("Generate 100 random numbers") |> 
#'   chat(openai(), .tools = tidyllm_tool_rnorm)
#'
#' # Example 2: Builtin tool
#' web_search <- ellmer_tool(ellmer::claude_tool_web_search())
#' llm_message("What are the latest AI developments?") |>
#'   chat(claude(), .tools = web_search)
#' }
#' @export
ellmer_tool <- function(.ellmer_tool) {
  if (!requireNamespace("ellmer", quietly = TRUE)) {
    stop("ellmer package required for ellmer_tool(). Install with: install.packages('ellmer')")
  }
  
  tool_classes <- class(.ellmer_tool)
  is_tool_def <- is_ellmer_class(tool_classes, "ToolDef")
  is_builtin  <- is_ellmer_class(tool_classes, "ToolBuiltIn")
  
  if (!is_tool_def && !is_builtin) {
    stop("Input must be an ellmer ToolDef or ToolBuiltIn object created via ellmer::tool() or ellmer::*_tool_*()")
  }
  
  fn_name <- .ellmer_tool@name %||% "unknown"
  
  if (is_builtin) {
    dummy_fn <- function() {
      stop(sprintf("Builtin tool '%s' is executed by the provider, not locally", fn_name))
    }
    
    TOOL(
      description = sprintf("Builtin tool: %s", fn_name),
      input_schema = list(),
      func = dummy_fn,
      name = fn_name,
      builtin = list(.ellmer_tool@json)
    )
  } else {
    fn_desc <- .ellmer_tool@description %||% ""
    arguments_obj <- .ellmer_tool@arguments
    
    if (!is_ellmer_class(class(arguments_obj), "TypeObject")) {
      stop("ellmer tool @arguments must be a TypeObject")
    }
    
    properties <- arguments_obj@properties
    
    if (length(properties) == 0) {
      input_schema <- list()
    } else {
      input_schema <- purrr::map(properties, function(arg_type) {
        convert_ellmer_type_to_field(arg_type)
      })
    }
    
    TOOL(
      description = fn_desc,
      input_schema = input_schema,
      func = .ellmer_tool,
      name = fn_name,
      builtin = list()
    )
  }
}


#' Convert an ellmer Type to a tidyllm Field
#'
#' @description Internal helper to convert ellmer type objects to tidyllm field descriptors
#'
#' @param .ellmer_type An ellmer type object (TypeBasic, TypeEnum, TypeArray, TypeObject)
#'
#' @return A `tidyllm_field` object
#'
#' @noRd
convert_ellmer_type_to_field <- function(.ellmer_type) {
  type_class <- class(.ellmer_type)
  
  if (is_ellmer_class(type_class, "TypeBasic")) {
    ellmer_type_name <- .ellmer_type@type
    
    type_map <- c(
      "string" = "string",
      "integer" = "number",
      "number" = "number",
      "boolean" = "boolean"
    )
    
    tidyllm_type <- type_map[ellmer_type_name] %||% "string"
    
    tidyllm_field(
      type = tidyllm_type,
      description = .ellmer_type@description %||% "",
      enum = character(0),
      vector = FALSE,
      schema = list()
    )
  } else if (is_ellmer_class(type_class, "TypeEnum")) {
    tidyllm_field(
      type = "string",
      description = .ellmer_type@description %||% "",
      enum = as.character(.ellmer_type@values),
      vector = FALSE,
      schema = list()
    )
  } else if (is_ellmer_class(type_class, "TypeArray")) {
    inner_field <- convert_ellmer_type_to_field(.ellmer_type@items)
    inner_field@vector <- TRUE
    inner_field
  } else if (is_ellmer_class(type_class, "TypeObject")) {
    properties <- .ellmer_type@properties
    nested_fields <- purrr::map(properties, convert_ellmer_type_to_field)
    nested_schema <- build_schema(nested_fields)
    
    tidyllm_field(
      type = "object",
      description = .ellmer_type@description %||% "",
      enum = character(0),
      vector = FALSE,
      schema = nested_schema
    )
  } else {
    stop("Unsupported ellmer type class: ", paste(type_class, collapse = ", "))
  }
}

#Generics for tools
tools_to_api <- new_generic("tools_to_api", c(".api", ".tools"))
run_tool_calls <- new_generic("run_tool_calls", c(".api",".tool_calls",".tools"))
send_tool_results <- new_generic("send_tool_results", c(".api",".request",".request_body"))

#Generics for the unified tool loop
has_tool_calls <- new_generic("has_tool_calls", c(".api", ".response"))
extract_tool_calls <- new_generic("extract_tool_calls", c(".api", ".response"))
append_tool_messages <- new_generic("append_tool_messages", c(".api", ".request_body", ".response", ".tool_results"))

method(has_tool_calls, list(APIProvider, class_any)) <- function(.api, .response) FALSE

print.TOOL  <- new_external_generic("base", "print", "x")

#' Print method for a Tool Definition for tidyllm
#' @noRd
method(print.TOOL,TOOL) <- function(x, ...){
  cat("<tidyllm_tool> ", x@name, "\n", sep = "")
  cat("  Description: ", x@description, "\n", sep = "")
  cat("  Arguments: \n")
  purrr::iwalk(x@input_schema, ~ cat("    -", .y, ": ", .x@type, "\n"))
}

field_to_param_schema <- function(param) {
  schema <- parse_field(param)
  if (length(param@description) > 0) {
    schema$description <- param@description
  }
  schema
}

#' Generic method to convert a tidyllm TOOL definition for a generic API
#' @noRd
method(tools_to_api, list(APIProvider, class_list)) <- function(.api, .tools) {
  purrr::map(.tools, function(tool) {
    if (length(tool@builtin) > 0) {
      tool@builtin[[1]]
    } else {
      list(
        type = "function",
        `function` = list(
          name = tool@name,
          description = tool@description,
          parameters = list(
            type = "object",
            properties = purrr::map(tool@input_schema, field_to_param_schema),
            required = as.list(names(tool@input_schema))
          )
        )
      )
    }
  })
}

process_tool_loop <- function(.api, .response, .tools_def, .request_body,
                               .request, .timeout, .max_tries, .max_tool_rounds = 10) {
  round <- 0
  while (has_tool_calls(.api, .response) && round < .max_tool_rounds) {
    round <- round + 1
    tool_calls   <- extract_tool_calls(.api, .response)
    tool_results <- run_tool_calls(.api, tool_calls, .tools_def)
    .request_body <- append_tool_messages(.api, .request_body, .response, tool_results)
    .request  <- httr2::req_body_json(.request, data = .request_body)
    .response <- perform_chat_request(.request, .api, FALSE, .timeout, .max_tries)
  }
  if (round >= .max_tool_rounds && has_tool_calls(.api, .response)) {
    stop(sprintf("Maximum tool rounds (%d) reached with pending tool calls", .max_tool_rounds))
  }
  .response
}

structure(list(method = "POST", url = "https://api.openai.com/v1/chat/completions", 
    status_code = 200L, headers = structure(list(date = "Wed, 16 Oct 2024 00:51:41 GMT", 
        `content-type` = "application/json", `access-control-expose-headers` = "X-Request-ID", 
        `openai-organization` = "user-j2dlcmauhfkzofrqjehceoc7", 
        `openai-processing-ms` = "754", `openai-version` = "2020-10-01", 
        `x-ratelimit-limit-requests` = "5000", `x-ratelimit-limit-tokens` = "80000", 
        `x-ratelimit-remaining-requests` = "4999", `x-ratelimit-remaining-tokens` = "78939", 
        `x-ratelimit-reset-requests` = "12ms", `x-ratelimit-reset-tokens` = "795ms", 
        `x-request-id` = "req_b8d107b9a41b1b85cb73cac2620f1305", 
        `strict-transport-security` = "max-age=31536000; includeSubDomains; preload", 
        `cf-cache-status` = "DYNAMIC", `set-cookie` = "REDACTED", 
        `x-content-type-options` = "nosniff", `set-cookie` = "REDACTED", 
        server = "cloudflare", `cf-ray` = "8d340fd1aaa1dc68-FRA", 
        `content-encoding` = "gzip", `alt-svc` = "h3=\":443\"; ma=86400"), class = "httr2_headers"), 
    body = charToRaw("{\n  \"id\": \"chatcmpl-AImYG5bv8QAitb0aR8EiK2PDDMmup\",\n  \"object\": \"chat.completion\",\n  \"created\": 1729039900,\n  \"model\": \"gpt-4-0613\",\n  \"choices\": [\n    {\n      \"index\": 0,\n      \"message\": {\n        \"role\": \"assistant\",\n        \"content\": \"A famous landmark in Paris is the Eiffel Tower.\",\n        \"refusal\": null\n      },\n      \"logprobs\": null,\n      \"finish_reason\": \"stop\"\n    }\n  ],\n  \"usage\": {\n    \"prompt_tokens\": 51,\n    \"completion_tokens\": 12,\n    \"total_tokens\": 63,\n    \"prompt_tokens_details\": {\n      \"cached_tokens\": 0\n    },\n    \"completion_tokens_details\": {\n      \"reasoning_tokens\": 0\n    }\n  },\n  \"system_fingerprint\": null\n}\n"), 
    cache = new.env(parent = emptyenv())), class = "httr2_response")

#' Retrieve Assistant Reply as Text
#'
#' Extracts the plain text content of the assistant's reply from an `LLMMessage` object.
#' Use [get_reply_data()] for structured replies in JSON format.
#'
#' This function is the core utility for retrieving assistant replies by index. 
#' For convenience, [last_reply()] is provided as a wrapper to retrieve the 
#' latest assistant reply.
#'
#' @param .llm An `LLMMessage` object containing the message history.
#' @param .index A positive integer indicating the index of the assistant reply to retrieve.
#'   Defaults to `NULL`, which retrieves the last reply.
#' @return Returns a character string containing the assistant's reply, or `NA_character_` if no reply exists.
#' @seealso [get_reply_data()], [last_reply()]
#' @export
#' @rdname get_reply
get_reply <- function(.llm, .index = NULL) {
  # Validate that .llm is an LLMMessage object and index is within bounds if provided
  validate_inputs(c(
    "Input .llm must be an LLMMessage object" = S7_inherits(.llm, LLMMessage),
    "Index must be a positive integer within bounds" = is.null(.index) || 
      (is.numeric(.index) && .index > 0 && .index <= length(Filter(function(x) x$role == "assistant", .llm@message_history)))
  ))
  
  # Retrieve assistant replies
  assistant_replies <-  filter_roles(.llm@message_history,"assistant")
  
  # Check if any assistant replies are available
  if (length(assistant_replies) == 0) {
    warning("No assistant replies available in the message history.")
    return(NA_character_)
  }
  
  # Select reply by index or use the last reply if .index is NULL
  selected_reply <- if (is.null(.index)) assistant_replies[[length(assistant_replies)]] else assistant_replies[[.index]]
  
  selected_reply$content
}

#' @export
#' @rdname get_reply
last_reply <- function(.llm) {
  get_reply(.llm = .llm, .index = NULL)
}


#' Retrieve Assistant Reply as Structured Data
#'
#' Parses the assistant's reply as JSON and returns the corresponding structured data.
#' If the reply is not marked as JSON, attempts to extract and parse JSON content from the text.
#'
#' For convenience, [last_reply_data()] is provided as a wrapper to retrieve the 
#' latest assistant reply's data.
#'
#' @param .llm An `LLMMessage` object containing the message history.
#' @param .index A positive integer indicating the index of the assistant reply to retrieve.
#'   Defaults to `NULL`, which retrieves the last reply.
#' @return Returns the parsed data from the assistant's reply, or `NULL` if parsing fails.
#' @seealso [get_reply()], [last_reply_data()]
#' @export
#' @rdname get_reply_data
get_reply_data <- function(.llm, .index = NULL) {
  # Validate inputs for .llm and .index
  validate_inputs(c(
    "Input .llm must be an LLMMessage object" = S7_inherits(.llm, LLMMessage)
  ))
  
  # Retrieve assistant replies with JSON flag
  assistant_replies <- filter_roles(.llm@message_history, "assistant")
  
  # Check if any assistant replies are available
  if (length(assistant_replies) == 0) {
    warning("No assistant replies available in the message history.")
    return(NULL)
  }
  
  # Validate and select reply based on index
  if (is.null(.index)) {
    selected_reply <- assistant_replies[[length(assistant_replies)]]
  } else {
    validate_inputs(c(
      "Index must be a positive integer within bounds" = is.numeric(.index) && .index > 0 && .index <= length(assistant_replies)
    ))
    selected_reply <- assistant_replies[[.index]]
  }
  
  # Regular expression to detect and extract JSON content across multiple lines
  json_extract_pattern <- "\\{[\\s\\S]*\\}|\\[[\\s\\S]*\\]"
  
  # Check if the selected reply is marked as JSON or contains JSON-like structure
  if (selected_reply$json) {
    json_content <- selected_reply$content
  } else {
    warning("The reply is not explicitly marked as JSON. Trying to extract JSON.")
    json_content <- stringr::str_extract(selected_reply$content, json_extract_pattern)
    if (is.na(json_content)) {
      warning("The selected reply does not appear to contain valid JSON. Returning NULL.")
      return(NULL)
    }
  }
  
  # Attempt to parse extracted JSON content
  tryCatch(
    jsonlite::fromJSON(json_content),
    error = function(e) {
      warning("Failed to parse JSON content. Returning NULL.")
      NULL
    }
  )
}

#' @export
#' @rdname get_reply_data
last_reply_data <- function(.llm) {
  get_reply_data(.llm = .llm, .index = NULL)
}



#' Retrieve Metadata from Assistant Replies
#'
#' Retrieves metadata from assistant replies within an `LLMMessage` object. 
#' It returns the metadata as a tibble.
#'
#' Metadata columns may include:
#' - `model`: The model used for generating the reply.
#' - `timestamp`: The time when the reply was generated.
#' - `prompt_tokens`: The number of tokens in the input prompt.
#' - `completion_tokens`: The number of tokens in the assistant's reply.
#' - `total_tokens`: The total number of tokens (prompt + completion).
#' - `api_specific`: A list column with API-specific metadata.
#' 
#' For convenience, [last_metadata()] is provided to retrieve the metadata for the last message.
#'
#' @param .llm An `LLMMessage` object containing the message history.
#' @param .index A positive integer specifying which assistant reply's metadata to extract.
#'   If `NULL` (default), metadata for all replies is returned.
#' @return A tibble containing metadata for the specified assistant reply or all replies.
#' @seealso [last_metadata()]
#' @export
#' @rdname get_metadata
get_metadata <- function(.llm, .index = NULL) {
  # Validate input
  validate_inputs(c(
    "Input .llm must be an LLMMessage object" = S7_inherits(.llm, LLMMessage),
    "Index must be a positive integer within bounds" = is.null(.index) || 
      (is.numeric(.index) && .index > 0 && .index <= length(filter_roles(.llm@message_history, "assistant")))
  ))
  
  # Extract assistant replies
  assistant_replies <- filter_roles(.llm@message_history, "assistant")
  
  # Check if any assistant replies are available
  if (length(assistant_replies) == 0) {
    warning("No assistant replies available in the message history.")
    return(tibble::tibble())
  }
  
  # Select the specified reply or all replies
  selected_replies <- if (is.null(.index)) assistant_replies else list(assistant_replies[[.index]])
  
  # Extract metadata
  purrr::map_dfr(selected_replies, function(reply) {
    meta <- reply$meta
    
    tibble::tibble(
      model = if (!is.null(meta$model)) meta$model else NA_character_,
      timestamp = if (!is.null(meta$timestamp)) meta$timestamp else NA_character_,
      prompt_tokens = if (!is.null(meta$prompt_tokens)) meta$prompt_tokens else NA_integer_,
      completion_tokens = if (!is.null(meta$completion_tokens)) meta$completion_tokens else NA_integer_,
      total_tokens = if (!is.null(meta$total_tokens)) meta$total_tokens else NA_integer_,
      stream  = if (!is.null(meta$stream)) meta$stream else NA,
      api_specific = list(meta$specific_metadata)
    )
  })
}

#' @export
#' @rdname get_metadata
last_metadata <- function(.llm) {
  get_metadata(.llm = .llm)
}


#' Retrieve Log Probabilities from Assistant Replies
#'
#' Extracts token log probabilities from assistant replies within an `LLMMessage` object. 
#' Each row represents a token with its log probability and top alternative tokens.
#'
#' An empty tibble is output if no logprobs were requested. Works with `openai_chat()`, `llamacpp_chat()`, and other providers that support logprobs.
#'
#' Columns include:
#' - `reply_index`: The index of the assistant reply in the message history.
#' - `token`: The generated token.
#' - `logprob`: The log probability of the generated token.
#' - `bytes`: The byte-level encoding of the token.
#' - `top_logprobs`: A list column containing the top alternative tokens with their log probabilities.
#'
#' @param .llm An `LLMMessage` object containing the message history.
#' @param .index A positive integer specifying which assistant reply's log probabilities to extract.
#'   If `NULL` (default), log probabilities for all replies are returned.
#' @return A tibble containing log probabilities for the specified assistant reply or all replies.
#' @seealso [get_metadata()]
#' @export
#' @rdname get_logprobs
get_logprobs <- function(.llm, .index = NULL) {
  # Validate input
  validate_inputs(c(
    "Input .llm must be an LLMMessage object" = S7_inherits(.llm, LLMMessage),
    "Index must be a positive integer within bounds" = is.null(.index) || 
      (is.numeric(.index) && .index > 0 && .index <= length(filter_roles(.llm@message_history, "assistant")))
  ))
  
  # Extract assistant replies
  assistant_replies <- filter_roles(.llm@message_history, "assistant")
  
  # Check if any assistant replies are available
  if (length(assistant_replies) == 0) {
    warning("No assistant replies available in the message history.")
    return(tibble::tibble())
  }
  
  # Select the specified reply or all replies
  selected_indices <- if (is.null(.index)) seq_along(assistant_replies) else .index
  selected_replies <- assistant_replies[selected_indices]
  
  # Extract logprobs
  purrr::imap_dfr(selected_replies, function(reply, idx) {
    if (is.null(reply$logprobs)) {
      return(tibble::tibble(
        reply_index = idx,
        token = character(),
        logprob = numeric(),
        bytes = list(),
        top_logprobs = list()
      ))
    }
    
    # Extract each token's logprobs
    purrr::map_dfr(reply$logprobs, function(token_data) {
      tibble::tibble(
        reply_index = idx,  # Include the reply index
        token = token_data$token,
        logprob = token_data$logprob,
        bytes = list(token_data$bytes),
        top_logprobs = list(token_data$top_logprobs)
      )
    })
  })
}




#' Retrieve a User Message by Index
#'
#' Extracts the content of a user's message from an `LLMMessage` object at a specific index.
#' 
#' For convenience, [last_user_message()] is provided as a wrapper to retrieve the
#' latest user message without specifying an index.
#'
#' @param .llm An `LLMMessage` object.
#' @param .index A positive integer indicating which user message to retrieve. Defaults to `NULL`, which retrieves the last message.
#' @return Returns the content of the user's message at the specified index. If no messages are found, returns `NA_character_`.
#' @seealso [last_user_message()]
#' @export
#' @rdname get_user_message
get_user_message <- function(.llm, .index = NULL) {
  # Validate inputs
  c(
    "Input .llm must be an LLMMessage object" = S7_inherits(.llm, LLMMessage)
  ) |> validate_inputs()
  
  # Filter to get all user messages
  user_messages <- filter_roles(.llm@message_history,"user")
  
  # Return NULL if there are no user messages
  if (length(user_messages) == 0) {
    warning("No user messages available in the message history.")
    return(NA_character_)
  }
  
  # If .index is NULL, get the last user message
  if (is.null(.index)) {
    selected_message <- user_messages[[length(user_messages)]]
  } else {
    # Validate .index
    c(
      "Index must be a positive integer" = is.numeric(.index) && .index > 0,
      "Index is out of bounds" = .index <= length(user_messages)
    ) |> validate_inputs()
    selected_message <- user_messages[[.index]]
  }
  
  return(selected_message$content)
}

#' @export
#' @rdname get_user_message
last_user_message <- function(.llm) {
  get_user_message(.llm = .llm, .index = NULL)
}

test_that("get_reply works with raw assistant message", {
  llm <- llm_message("Hello assistant") |>
    llm_message("Hello user",.role="assistant") 
  
  reply <- get_reply(llm)
  expect_equal(reply, "Hello user")
})

test_that("get_reply works with assistant JSON message", {
  llm <-llm_message("Provide data in JSON format")
  json_content <- '{"name": "Alice", "age": 30}'
  llm <-add_message(.llm = llm,
                    .role = "assistant", 
                    .content = json_content, 
                    .json = TRUE)
  
  reply_string <- get_reply(llm)
  expect_equal(reply_string, json_content)
})

test_that("get_reply_data works with assistant JSON message", {
  llm <-llm_message("Provide data in JSON format")
  json_content <- '{"name": "Alice", "age": 30}'
  llm <-add_message(.llm = llm,
                    .role = "assistant", 
                    .content = json_content, 
                    .json = TRUE)
  parsed_data <- get_reply_data(llm)
  expect_equal(parsed_data, jsonlite::fromJSON(json_content))
})

test_that("get_reply_data warns about corrupt JSON in assistant message marked as JSON", {
  llm <-llm_message("Provide data in JSON format")
  corrupt_json_content <- 'Garbage before JSON {"name": "Alice", "age": 30}'
  llm <-add_message(.llm = llm,
                    .role = "assistant", 
                    .content = corrupt_json_content, 
                    .json = TRUE)
  
  expect_warning(parsed_data <- get_reply_data(llm), "Failed to parse JSON content")
  expect_null(parsed_data)
})


test_that("get_reply_data extracts  JSON in assistant message not marked as JSON with warning", {
  llm <-llm_message("Provide data in JSON format")
  corrupt_json_content <- 'Garbage before JSON {"name": "Alice", "age": 30}'
  llm <-add_message(.llm = llm,
                    .role = "assistant", 
                    .content = corrupt_json_content, 
                    .json = FALSE)
  
  expect_warning(parsed_data <- get_reply_data(llm), "The reply is not explicitly marked as JSON. Trying to extract JSON.")
  expect_equal(parsed_data, jsonlite::fromJSON('{"name": "Alice", "age": 30}'))
})


test_that("last_reply returns NA_character_ when there are no assistant replies", {
  llm <- llm_message("Hello assistant")
  expect_warning(reply <- last_reply(llm),"No assistant replies available in the message history.")
  expect_equal(reply, NA_character_)
})

test_that("last_reply returns the most recent assistant reply", {
  llm <- llm_message("Hello") 
  llm <- add_message(.llm = llm,
                     .role = "assistant", 
                     .content = "First reply", 
                     .json = FALSE)
  llm <- add_message(.llm = llm,
                     .role = "assistant", 
                     .content = '{"data": "Second reply"}', 
                     .json = TRUE)
  
  reply <- last_reply(llm)
  expect_equal(reply, '{"data": "Second reply"}')
})

test_that("last_reply ignores non-assistant messages", {
  llm <- llm_message("User message")|>
    llm_message(.role="assistant",.prompt = "Assistant message")
  reply <- last_reply(llm)
  expect_equal(reply, "Assistant message")
})


test_that("get_reply retrieves assistant reply by index", {
 llm <- llm_message("User message")|>
    llm_message(.role="assistant",.prompt = "Assistant reply 1") |>
    llm_message(.role="assistant",.prompt = "Assistant reply 2")
  

  reply1 <- get_reply(llm, .index = 1)
  expect_equal(reply1, "Assistant reply 1")
  reply2 <- get_reply(llm, .index = 2)
  expect_equal(reply2, "Assistant reply 2")
})

test_that("get_reply returns an error for out-of-bounds index", {
  llm <- llm_message("User message")|>
    llm_message(.role="assistant",.prompt = "Assistant reply 1") 
  expect_error(reply <- get_reply(llm, .index = 2), "Index must be a positive integer within bounds")
})

test_that("get_user_message retrieves user message by index", {
  llm <- llm_message("User message 1")|>
    llm_message(.role="assistant",.prompt = "Assistant reply 1") |>
    llm_message(.role="user",.prompt = "User message 2")
  
  msg1 <- get_user_message(llm, .index = 1)
  expect_equal(msg1, "User message 1")
  msg2 <- get_user_message(llm, .index = 2)
  expect_equal(msg2, "User message 2")
})

test_that("get_user_message returns error for out-of-bounds index", {
  llm <- llm_message("User message 1")|>
    llm_message(.role="assistant",.prompt = "Assistant reply 1")
  expect_error(reply <- get_reply(llm, .index = 2), "Index must be a positive integer within bounds")
})

test_that("last_user_message retrieves the most recent user message", {
  llm <- llm_message("User message 1")|>
    llm_message(.role="assistant",.prompt = "Assistant reply 1") |>
    llm_message(.role="user",.prompt = "User message 2")
  last_msg <- last_user_message(llm)
  expect_equal(last_msg, "User message 2")
})

test_that("get_metadata works for multiple assistant replies", {
  llm <- llm_message("User message")
  llm <- add_message(.llm=llm,
                     .role = "assistant", 
                     .content = "First assistant message", 
                     .meta = list(model = "gpt-3", 
                                timestamp = "2024-11-08T10:00:00Z", 
                                prompt_tokens = 5, 
                                completion_tokens = 15, 
                                total_tokens = 20)
                     )
  llm <- add_message(.llm=llm,
                     .role = "assistant", 
                     .content = "Second assistant message", 
                     .meta = list(model = "gpt-4", 
                                 timestamp = "2024-11-08T10:00:00Z", 
                                 prompt_tokens = 10, 
                                 completion_tokens = 20, 
                                 total_tokens = 30)
  )
  

  metadata <- get_metadata(llm)
  expect_equal(nrow(metadata), 2)
  expect_equal(metadata$model, c("gpt-3", "gpt-4"))
  expect_equal(metadata$total_tokens, c(20, 30))
})

test_that("get_metadata handles missing with an NA row", {
  llm <- llm_message("User message") |>
    llm_message("Assistant message without meta",.role="assistant")
  metadata <- get_metadata(llm)
  expect_equal(nrow(metadata), 1)
  #All standard metadata fields are NA. The one with the api_specific metadata is a list with NULL
  expect_true(all(is.na(metadata)|is.list(metadata)))
})

test_that("get_metadata retrieves metadata for a specific index", {
  llm <- llm_message("User message")
  llm <- add_message(.llm=llm,
                     .role = "assistant", 
                     .content = "First assistant message", 
                     .meta = list(model = "gpt-3", 
                                 timestamp = "2024-11-08T10:00:00Z", 
                                 prompt_tokens = 5, 
                                 completion_tokens = 15, 
                                 total_tokens = 20)
  )
  llm <- add_message(.llm=llm,
                     .role = "assistant", 
                     .content = "Second assistant message", 
                     .meta = list(model = "gpt-4", 
                                 timestamp = "2024-11-08T10:00:00Z", 
                                 prompt_tokens = 10, 
                                 completion_tokens = 20, 
                                 total_tokens = 30)
  )
  
  metadata <- get_metadata(llm, .index = 2)
  expect_equal(nrow(metadata), 1)
  expect_equal(metadata$model, "gpt-4")
  expect_equal(metadata$total_tokens, 30)
})

test_that("get_metadata returns error for out-of-bounds index", {
  llm <- llm_message("User message")
  llm <- add_message(.llm=llm,
                     .role = "assistant", 
                     .content = "First assistant message", 
                     .meta = list(model = "gpt-3", 
                                 timestamp = "2024-11-08T10:00:00Z", 
                                 prompt_tokens = 5, 
                                 completion_tokens = 15, 
                                 total_tokens = 20)
  )
  
  expect_error(get_metadata(llm, .index = 2), "Index must be a positive integer within bounds")
})

test_that("get_metadata returns empty tibble and warns when no assistant messages", {
  llm <- llm_message("User message")
  expect_warning(metadata <- get_metadata(llm), "No assistant replies available in the message history.")
  expect_equal(nrow(metadata), 0)
})

testthat::skip_if_not_installed("httptest2")
library(httptest2)

# Test request construction and dry run
test_that("deepseek function constructs a correct request and dry runs it", {
  llm <- llm_message("Write a poem about a (stochastic) parrot")
  
  request <- llm |> chat(deepseek, .dry_run = TRUE)
  dry_run <- request |> httr2::req_dry_run(redact_headers = TRUE, quiet = TRUE)
  
  expect_equal(dry_run$method, "POST")
  expect_true(grepl("/chat/completions", dry_run$path))
  
  headers <- dry_run$headers
  expect_type(headers, "list")
  expect_true("authorization" %in% names(headers))
  expect_true("content-type" %in% names(headers))
  expect_equal(headers$`content-type`, "application/json")
  
  body_json <- request$body |> jsonlite::toJSON() |> as.character()
  expect_true(grepl("deepseek-chat", body_json))
})

# Test response handling
test_that("deepseek returns expected response", {
  with_mock_dir("deepseek", expr = {
    
    if (Sys.getenv("DEEPSEEK_API_KEY") == "") {
      Sys.setenv(DEEPSEEK_API_KEY = "DUMMY_KEY_FOR_TESTING")
    }
    
    result <- deepseek_chat(
      .llm = llm_message("Hello, world"),
      .temperature = 0,
      .stream = FALSE
    )

    reply <- result |>
      get_reply()
    
    if (Sys.getenv("DEEPSEEK_API_KEY") == "DUMMY_KEY_FOR_TESTING") {
      Sys.setenv(DEEPSEEK_API_KEY = "")
    }
    
    expect_true(S7_inherits(result, LLMMessage))
    expect_equal(reply, "Hello! How can I assist you today?")

  }, simplify = FALSE)
})

#' @export
.onLoad <- function(libname, pkgname) {
  options(tidyllm_print_metadata = FALSE) # Default is to not print metadata
  S7::methods_register()
  
  # Initialize the parent environment for rate limits on package load
  if (!exists(".tidyllm_rate_limit_env", envir = .GlobalEnv)) {
    .GlobalEnv$.tidyllm_rate_limit_env <- new.env(parent = emptyenv())
  }
}

# enable usage of <S7_object>@name in package code
#' @rawNamespace if (getRversion() < "4.3.0") importFrom("S7", "@")
NULL

#' The Azure OpenAI API provider class
#'
#'Inherit most of the functionality from vanilla OpenAI API
#'
#' @noRd
api_azure_openai <- new_class("Azure OpenAI", api_openai)


#' Extract rate limit info from  Azure Openai API-Headers
#'
#' @noRd
method(ratelimit_from_header, list(api_azure_openai,new_S3_class("httr2_headers"))) <- function(.api, .headers) {
  request_time <- strptime(.headers["date"]$date, 
                           format="%a, %d %b %Y %H:%M:%S", tz="GMT")
  
  # Extract remaining requests and tokens
  ratelimit_requests_remaining <- as.integer(
    .headers["x-ratelimit-remaining-requests"]$`x-ratelimit-remaining-requests`)
  ratelimit_tokens_remaining <- as.integer(
    .headers["x-ratelimit-remaining-tokens"]$`x-ratelimit-remaining-tokens`)
  
  # Assuming reset occurs every 60 seconds (at least I got minutes in my azure console)
  reset_interval <- 60         
  
  ratelimit_requests_reset_time <- request_time + reset_interval
  ratelimit_tokens_reset_time <- request_time + reset_interval
  
  list(
    this_request_time = request_time,
    ratelimit_requests = NA,
    ratelimit_requests_remaining = ratelimit_requests_remaining,
    ratelimit_requests_reset_time = ratelimit_requests_reset_time,
    ratelimit_tokens = NA,
    ratelimit_tokens_remaining = ratelimit_tokens_remaining,
    ratelimit_tokens_reset_time = ratelimit_tokens_reset_time
  )
}


#' A chat parsing method for Azure Openai to extract the assitant response
#'
#' @noRd
method(parse_chat_response, list(api_azure_openai,class_list)) <- function(.api,.content) {
  api_label <- .api@long_name 
  if("error" %in% names(.content)){
    sprintf("%s returned an Error:\nType: %s\nMessage: %s",
            api_label,
            .content$error$code,
            .content$error$message) |>
      stop()
  }
  
  if (length(.content$choices) == 0) {
    paste0("Received empty response from ",api_label) |>
      stop()
  }
  .content$choices[[1]]$message$content  
}



#' Send LLM Messages to an Azure OpenAI Chat Completions endpoint 
#'
#' @description
#' This function sends a message history to the Azure OpenAI Chat Completions API and returns the assistant's reply.
#'
#' @param .llm An `LLMMessage` object containing the conversation history.
#' @param .endpoint_url Base URL for the API (default: Sys.getenv("AZURE_ENDPOINT_URL")).
#' @param .deployment The identifier of the model that is deployed (default: "gpt-4o-mini").
#' @param .api_version Which version of the API is deployed (default: "2024-08-01-preview").
#' @param .max_completion_tokens An upper bound for the number of tokens that can be generated for a completion.
#' @param .reasoning_effort How long should reasoning models reason (can either be "low","medium" or "high").
#' @param .frequency_penalty Number between -2.0 and 2.0. Positive values penalize new tokens based on their existing frequency.
#' @param .logit_bias A named list modifying the likelihood of specified tokens appearing in the completion.
#' @param .presence_penalty Number between -2.0 and 2.0. Positive values penalize new tokens based on whether they appear in the text so far.
#' @param .seed If specified, the system will make a best effort to sample deterministically.
#' @param .stop Up to 4 sequences where the API will stop generating further tokens.
#' @param .stream If set to TRUE, the answer will be streamed to console as it comes (default: FALSE).
#' @param .temperature What sampling temperature to use, between 0 and 2. Higher values make the output more random.
#' @param .top_p An alternative to sampling with temperature, called nucleus sampling.
#' @param .timeout Request timeout in seconds (default: 60).
#' @param .verbose Should additional information be shown after the API call (default: FALSE).
#' @param .json_schema A JSON schema object provided by tidyllm schema or ellmer schemata.
#' @param .max_tries Maximum retries to perform request.
#' @param .dry_run If TRUE, perform a dry run and return the request object (default: FALSE).
#' @param .logprobs If TRUE, get the log probabilities of each output token (default: NULL).
#' @param .top_logprobs If specified, get the top N log probabilities of each output token (0-5, default: NULL).
#' @param .tools Either a single TOOL object or a list of TOOL objects representing the available functions for tool calls.
#' @param .tool_choice A character string specifying the tool-calling behavior; valid values are "none", "auto", or "required".
#' @param .max_tool_rounds Integer specifying the maximum number of tool use iterations (default: 10).
#'
#' @return A new `LLMMessage` object containing the original messages plus the assistant's response.
#' @export
azure_openai_chat <- function(
    .llm,
    .endpoint_url = Sys.getenv("AZURE_ENDPOINT_URL"),
    .deployment = "gpt-4o-mini",
    .api_version = "2024-08-01-preview",
    .max_completion_tokens = NULL,
    .reasoning_effort = NULL,
    .frequency_penalty = NULL,
    .logit_bias = NULL,
    .presence_penalty = NULL,
    .seed = NULL,
    .stop = NULL,
    .stream = FALSE,
    .temperature = NULL,
    .top_p = NULL,
    .timeout = 60,
    .verbose = FALSE,
    .json_schema = NULL,
    .max_tries = 3,
    .dry_run = FALSE,
    .logprobs = NULL,       
    .top_logprobs = NULL,
    .tools = NULL,
    .tool_choice = NULL,
    .max_tool_rounds = 10
) {
  #Check endpoint
  if (.endpoint_url == "" && .dry_run == FALSE) {
    stop("No valid Azure endpoint defined. Please set it either as input to this function or with: Sys.setenv(AZURE_ENDPOINT_URL = \"https://endpoint.openai.azure.com/\")")
  }
  
  # Validate inputs
  c(
    "Input .llm must be an LLMMessage object" = S7_inherits(.llm, LLMMessage),
    "Input .deployment must be a string" = is.character(.deployment),
    "Input .max_completion_tokens must be NULL or a positive integer" = is.null(.max_completion_tokens) | (is_integer_valued(.max_completion_tokens) & .max_completion_tokens > 0),
    "Input .reasoning_effort must be NULL or one of 'low', 'medium', 'high'" = is.null(.reasoning_effort) | (.reasoning_effort %in% c("low", "medium", "high")),   
    "Input .frequency_penalty must be numeric or NULL" = is.null(.frequency_penalty) | is.numeric(.frequency_penalty),
    "Input .logit_bias must be a list or NULL" = is.null(.logit_bias) | is.list(.logit_bias),
    "Input .presence_penalty must be numeric or NULL" = is.null(.presence_penalty) | is.numeric(.presence_penalty),
    "Input .seed must be NULL or an integer" = is.null(.seed) | is_integer_valued(.seed),
    "Input .stop must be NULL or a character vector or string" = is.null(.stop) | is.character(.stop),
    "Input .stream must be logical or NULL" = is.null(.stream) | is.logical(.stream),
    "Input .temperature must be numeric or NULL" = is.null(.temperature) | is.numeric(.temperature),
    "Input .top_p must be numeric or NULL" = is.null(.top_p) | is.numeric(.top_p),
    "Input .timeout must be integer-valued numeric" = is_integer_valued(.timeout),
    "Input .verbose must be logical" = is.logical(.verbose),
    "Input .json_schema must be NULL or a list or an ellmer type object" = is.null(.json_schema) | is.list(.json_schema) | is_ellmer_type(.json_schema),
    "Input .max_tries must be integer-valued numeric" = is_integer_valued(.max_tries),
    "Input .dry_run must be logical" = is.logical(.dry_run),
    "Input .logprobs must be NULL or a logical" = is.null(.logprobs) | is.logical(.logprobs),
    "Input .top_logprobs must be NULL or an integer between 0 and 5" = is.null(.top_logprobs) | (is_integer_valued(.top_logprobs) && .top_logprobs >= 0 && .top_logprobs <= 5),
    "Input .tools must be NULL, a TOOL object, or a list of TOOL objects" = is.null(.tools) || S7_inherits(.tools, TOOL) || (is.list(.tools) && all(purrr::map_lgl(.tools, ~ S7_inherits(.x, TOOL)))),
    "Input .tool_choice must be NULL or a character (one of 'none', 'auto', 'required')" = is.null(.tool_choice) || (is.character(.tool_choice) && .tool_choice %in% c("none", "auto", "required")),
    "Streaming is not supported for requests with tool calls" = is.null(.tools) || !isTRUE(.stream)
  ) |> validate_inputs()
  
  # Create API object
  api_obj <- api_azure_openai(
    short_name = "azure_openai",
    long_name = "Azure OpenAI",
    api_key_env_var = "AZURE_OPENAI_API_KEY"
  )
  
  # Get API key
  api_key <- get_api_key(api_obj, .dry_run)
  
  # Use the helper function to prepare request components
  # We'll use .model parameter internally to pass the .deployment value
  request_data <- prepare_openai_request(
    .llm = .llm,
    .api = api_obj,
    .model = .deployment,  # Use deployment as model internally
    .max_completion_tokens = .max_completion_tokens,
    .reasoning_effort = .reasoning_effort,
    .frequency_penalty = .frequency_penalty,
    .logit_bias = .logit_bias,
    .presence_penalty = .presence_penalty,
    .seed = .seed,
    .stop = .stop,
    .temperature = .temperature,
    .top_p = .top_p,
    .json_schema = .json_schema,
    .logprobs = .logprobs,
    .top_logprobs = .top_logprobs
  )
  
  # Get components from the request data
  request_body <- request_data$request_body
  json <- request_data$json
  
  # Handle tools
  tools_def <- if (!is.null(.tools)) {
    if (S7_inherits(.tools, TOOL)) list(.tools) else .tools
  } else {
    NULL
  }
  
  # Add tools to request body if provided
  if (!is.null(tools_def)) {
    request_body$tools <- tools_to_api(api_obj, tools_def)
    request_body$tool_choice <- .tool_choice
  }
  
  # Add streaming options if requested
  if (.stream == TRUE) {
    request_body <- request_body |>
      append(list(
        stream = TRUE,
        stream_options = list(include_usage = TRUE)
      ))
  }
  
  # Build the request with Azure-specific URL path
  request <- httr2::request(.endpoint_url) |>
    httr2::req_url_path_append(paste0("openai/deployments/", .deployment, "/chat/completions")) |>
    httr2::req_url_query(`api-version` = .api_version) |>
    httr2::req_headers(
      `Content-Type` = "application/json",
      `api-key` = api_key
    ) |>
    httr2::req_body_json(data = request_body)
  
  # Return only the request object in a dry run
  if (.dry_run) {
    return(request)
  }
  
  # Perform the request
  response <- perform_chat_request(request, api_obj, .stream, .timeout, .max_tries)
  
  if (.stream == FALSE && !is.null(tools_def)) {
    response <- process_tool_loop(
      .api = api_obj,
      .response = response,
      .tools_def = tools_def,
      .request_body = request_body,
      .request = request,
      .timeout = .timeout,
      .max_tries = .max_tries,
      .max_tool_rounds = .max_tool_rounds
    )
  }
  
  # Extract assistant reply
  assistant_reply <- response$assistant_reply
  
  # Check for log probabilities
  logprobs <- parse_logprobs(api_obj, response$raw)
  
  # Track rate limit
  track_rate_limit(api_obj, response$headers, .verbose)
  
  # Update the LLMMessage with the assistant's response
  add_message(
    .llm = .llm,
    .role = "assistant",
    .content = assistant_reply,
    .json = json,
    .meta = response$meta,
    .logprobs = logprobs
  )
}

#' Generate Embeddings Using OpenAI API on Azure
#'
#' @param .input A character vector of texts to embed or an `LLMMesssage`object
#' @param .deployment The embedding model identifier (default: "text-embedding-3-small").
#' @param .endpoint_url Base URL for the API (default:  Sys.getenv("AZURE_ENDPOINT_URL")).
#' @param .truncate Whether to truncate inputs to fit the model's context length (default: TRUE).
#' @param .timeout Timeout for the API request in seconds (default: 120).
#' @param .dry_run If TRUE, perform a dry run and return the request object.
#' @param .max_tries Maximum retry attempts for requests (default: 3).
#' @param .api_version What API-Version othe Azure OpenAI API should be used (default: "2023-05-15")
#' @return A tibble with two columns: `input` and `embeddings`. 
#' The `input` column contains the texts sent to embed, and the `embeddings` column 
#' is a list column where each row contains an embedding vector of the sent input.
#' @export
azure_openai_embedding <- function(.input,
                                   .deployment = "text-embedding-3-small",
                                   .endpoint_url = Sys.getenv("AZURE_ENDPOINT_URL"),
                                   .api_version = "2023-05-15",
                                   .truncate = TRUE,
                                   .timeout = 120,
                                   .dry_run = FALSE,
                                   .max_tries = 3) {

  # Validate the inputs
  c(
    "Input .input must be an LLMMessage object or a character vector" = S7_inherits(.input, LLMMessage) | is.character(.input),
    "Input .deployment must be a string" = is.character(.deployment),
    "Input .truncate must be logical" = is.logical(.truncate),
    "Input .timeout must be an integer-valued numeric (seconds till timeout)" = is.numeric(.timeout) && .timeout > 0,
    ".dry_run must be logical" = is.logical(.dry_run)
  ) |> validate_inputs()
  
  # Get the Azure OpenAI API key
  api_key <- Sys.getenv("AZURE_OPENAI_API_KEY")
  if ((api_key == "")& .dry_run==FALSE){
    stop("API key is not set. Please set it with: Sys.setenv(AZURE_OPENAI_API_KEY = \"YOUR-KEY-GOES-HERE\")")
  }
  
  input_texts <- parse_embedding_input(.input)
  # Prepare the request body
  request_body <- list(
    input = input_texts
  )
  
  # Build the request
  request <- httr2::request(.endpoint_url) |>
    httr2::req_url_path_append(paste0("openai/deployments/", .deployment,"/embeddings")) |>
    httr2::req_url_query(`api-version` = .api_version) |>
    httr2::req_headers(
      `Content-Type` = "application/json",
      `api-key` = api_key,
    )  |>
    httr2::req_body_json(data = request_body)
  
  # Return the request object if it's a dry run
  if (.dry_run) {
    return(request)
  }
  
  extract_embeddings_fn <- function(response_content,error,response_headers){
    if(error){
      paste0("API error response - ", response_content$error$message) |>
        stop()
    }

    response_content$data |> 
      purrr::map("embedding") |>
      purrr::map(unlist)
  }
  
  # Perform a standard embedding API request
  perform_embedding_request(.request = request,
                            .timeout = .timeout,
                            .max_tries = 3,
                            .input_texts = input_texts, 
                            .fn_extract_embeddings = extract_embeddings_fn)
}


#' Send a Batch of Messages to Azure OpenAI Batch API
#'
#' This function creates and submits a batch of messages to the Azure OpenAI Batch API for asynchronous processing.
#'
#' @param .llms A list of LLMMessage objects containing conversation histories.
#' @param .endpoint_url Base URL for the API (default: Sys.getenv("AZURE_ENDPOINT_URL")).
#' @param .deployment The identifier of the model that is deployed (default: "gpt-4o-mini").
#' @param .api_version Which version of the API is deployed (default: "2024-10-01-preview")
#' @param .max_completion_tokens Integer specifying the maximum tokens per response (default: NULL).
#' @param .reasoning_effort How long should reasoning models reason (can either be "low","medium" or "high")
#' @param .frequency_penalty Number between -2.0 and 2.0. Positive values penalize new tokens based on their existing frequency.
#' @param .logit_bias A named list modifying the likelihood of specified tokens appearing in the completion.
#' @param .presence_penalty Number between -2.0 and 2.0. Positive values penalize new tokens based on whether they appear in the text so far.
#' @param .seed If specified, the system will make a best effort to sample deterministically.
#' @param .stop Up to 4 sequences where the API will stop generating further tokens.
#' @param .temperature What sampling temperature to use, between 0 and 2. Higher values make the output more random.
#' @param .top_p An alternative to sampling with temperature, called nucleus sampling.
#' @param .logprobs If TRUE, get the log probabilities of each output token (default: NULL).
#' @param .top_logprobs If specified, get the top N log probabilities of each output token (0-5, default: NULL).
#' @param .dry_run Logical; if TRUE, returns the prepared request object without executing it (default: FALSE).
#' @param .overwrite Logical; if TRUE, allows overwriting an existing batch ID (default: FALSE).
#' @param .max_tries Maximum number of retries to perform the request (default: 3).
#' @param .timeout Integer specifying the request timeout in seconds (default: 60).
#' @param .verbose Logical; if TRUE, additional info about the requests is printed (default: FALSE).
#' @param .json_schema A JSON schema object provided by tidyllm schema or ellmer schemata (default: NULL).
#' @param .id_prefix Character string to specify a prefix for generating custom IDs when names in `.llms` are missing (default: "tidyllm_azure_openai_req_").
#' 
#' @return An updated and named list of `.llms` with identifiers that align with batch responses, including a `batch_id` attribute.
#' @export
send_azure_openai_batch <- function(.llms,
                                    .endpoint_url = Sys.getenv("AZURE_ENDPOINT_URL"),
                                    .deployment = "gpt-4o-mini",
                                    .api_version = "2024-10-01-preview",
                                    .max_completion_tokens = NULL,
                                    .reasoning_effort = NULL,
                                    .frequency_penalty = NULL,
                                    .logit_bias = NULL,
                                    .presence_penalty = NULL,
                                    .seed = NULL,
                                    .stop = NULL,
                                    .temperature = NULL,
                                    .top_p = NULL,
                                    .logprobs = NULL,       
                                    .top_logprobs = NULL,
                                    .dry_run = FALSE,
                                    .overwrite = FALSE,
                                    .max_tries = 3,
                                    .timeout = 60,
                                    .verbose = FALSE,
                                    .json_schema = NULL,
                                    .id_prefix = "tidyllm_azure_openai_req_") {
  # Check endpoint
  if (.endpoint_url == "" && .dry_run == FALSE) {
    stop("No valid Azure endpoint defined. Please set it either as input to this function or with: Sys.setenv(AZURE_ENDPOINT_URL = \"https://endpoint.openai.azure.com/\")")
  }
  
  # Input validation
  c(
    ".llms must be a list of LLMMessage objects" = is.list(.llms) && all(sapply(.llms, S7_inherits, LLMMessage)),
    "Input .deployment must be a string" = is.character(.deployment),
    ".max_completion_tokens must be NULL or a positive integer" = is.null(.max_completion_tokens) | (is_integer_valued(.max_completion_tokens) & .max_completion_tokens > 0),
    ".reasoning_effort must be NULL or one of 'low', 'medium', 'high'" = is.null(.reasoning_effort) | (.reasoning_effort %in% c("low", "medium", "high")),
    ".frequency_penalty must be numeric or NULL" = is.null(.frequency_penalty) | is.numeric(.frequency_penalty),
    ".logit_bias must be a list or NULL" = is.null(.logit_bias) | is.list(.logit_bias),
    ".presence_penalty must be numeric or NULL" = is.null(.presence_penalty) | is.numeric(.presence_penalty),
    ".seed must be NULL or an integer" = is.null(.seed) | is_integer_valued(.seed),
    ".stop must be NULL or a character vector or string" = is.null(.stop) | is.character(.stop),
    ".temperature must be numeric or NULL" = is.null(.temperature) | is.numeric(.temperature),
    ".top_p must be numeric or NULL" = is.null(.top_p) | is.numeric(.top_p),
    ".logprobs must be NULL or a logical" = is.null(.logprobs) | is.logical(.logprobs),
    ".json_schema must be NULL or a list or an ellmer type object" = is.null(.json_schema) | is.list(.json_schema) | is_ellmer_type(.json_schema),
    ".top_logprobs must be NULL or an integer between 0 and 5" = is.null(.top_logprobs) | (is_integer_valued(.top_logprobs) && .top_logprobs >= 0 && .top_logprobs <= 5),
    ".dry_run must be logical" = is.logical(.dry_run),
    ".verbose must be logical" = is.logical(.verbose),
    ".overwrite must be logical" = is.logical(.overwrite),
    ".id_prefix must be a character vector of length 1" = is.character(.id_prefix),
    ".max_tries must be integer-valued numeric" = is_integer_valued(.max_tries),
    ".timeout must be integer-valued numeric" = is_integer_valued(.timeout)
  ) |> validate_inputs()
  
  # Create API object
  api_obj <- api_azure_openai(
    short_name = "azure_openai",
    long_name = "Azure OpenAI",
    api_key_env_var = "AZURE_OPENAI_API_KEY"
  )
  
  # Get API key
  api_key <- get_api_key(api_obj, .dry_run)
  
  # Prepare LLMs for batch processing
  prepared_llms <- prepare_llms_for_batch(
    api_obj,
    .llms = .llms,
    .id_prefix = .id_prefix,
    .overwrite = .overwrite
  )
  
  # Prepare the request lines
  request_lines <- lapply(seq_along(prepared_llms), function(i) { 
    custom_id <- names(prepared_llms)[i]
    
    # Use prepare_openai_request to set up common request parameters
    request_data <- prepare_openai_request(
      .llm = prepared_llms[[i]],
      .api = api_obj,
      .model = .deployment,  # Use deployment as model internally
      .max_completion_tokens = .max_completion_tokens,
      .reasoning_effort = .reasoning_effort,
      .frequency_penalty = .frequency_penalty,
      .logit_bias = .logit_bias,
      .presence_penalty = .presence_penalty,
      .seed = .seed,
      .stop = .stop,
      .temperature = .temperature,
      .top_p = .top_p,
      .json_schema = .json_schema,
      .logprobs = .logprobs,
      .top_logprobs = .top_logprobs
    )
    
    # Create the request line as JSON
    request_line <- list(
      custom_id = custom_id,
      method = "POST",
      url = "/chat/completions",
      body = request_data$request_body
    )
    
    # Convert to JSON
    jsonlite::toJSON(request_line, auto_unbox = TRUE)
  })
  
  # Write the request lines to a temporary .jsonl file
  temp_file <- tempfile(fileext = ".jsonl")
  writeLines(unlist(request_lines), con = temp_file)
  
  if (.dry_run) {
    # Return the prepared .jsonl file path
    return(readLines(temp_file))
  }
  
  # Upload the .jsonl file via OpenAI's Files API
  upload_request <- httr2::request(.endpoint_url) |> 
    httr2::req_url_path_append("openai/files") |> 
    httr2::req_url_query(`api-version` = .api_version) |>
    httr2::req_headers(
      `Content-Type` = "multipart/form-data",
      `api-key` = api_key
    ) |>
    httr2::req_body_multipart(
      purpose = "batch",
      file = curl::form_file(temp_file),
      type = 'application/json'
    )
  
  upload_response <- upload_request |>
    perform_generic_request(.timeout = .timeout, .max_tries = .max_tries)
  
  if (.verbose) {
    message("Batch request file uploaded via files API")
  }
  
  if ("error" %in% names(upload_response$content)) {
    sprintf("Azure OpenAI API returned an error during batch creation:\nType: %s\nMessage: %s",
            upload_response$content$error$code,
            upload_response$content$error$message) |>
      stop()
  }
  
  input_file_id <- upload_response$content$id
  
  # Now, create the batch
  batch_request_body <- list(
    input_file_id = input_file_id,
    endpoint = "/chat/completions",
    completion_window = "24h"
  )
  
  batch_request <- httr2::request(.endpoint_url) |> 
    httr2::req_url_path("openai/batches") |>
    httr2::req_url_query(`api-version` = .api_version) |>
    httr2::req_headers(
      `Content-Type` = "application/json",
      `api-key` = api_key
    ) |>
    httr2::req_body_json(batch_request_body)
  
  batch_response <- batch_request |>
    perform_generic_request(.timeout = .timeout, .max_tries = .max_tries)
  
  batch_response_body <- batch_response$content
  
  if (.verbose) {
    message("Batch request for file sent")
  }
  
  if ("error" %in% names(batch_response_body)) {
    sprintf("Azure OpenAI API returned an Error during batch creation:\nType: %s\nMessage: %s",
            batch_response_body$error$type,
            batch_response_body$error$message) |>
      stop()
  }
  
  # Attach batch_id as an attribute to prepared_llms
  batch_id <- batch_response_body$id
  attr(prepared_llms, "batch_id") <- batch_id
  attr(prepared_llms, "json") <- if (!is.null(.json_schema)) TRUE else FALSE
  
  # Optionally, remove the temporary file
  unlink(temp_file)
  
  return(prepared_llms)
}


#' Check Batch Processing Status for Azure OpenAI Batch API
#'
#' This function retrieves the processing status and other details of a specified Azure OpenAI batch ID
#' from the Azure OpenAI Batch API.
#'
#' @param .llms A list of LLMMessage objects.
#' @param .batch_id A manually set batch ID.
#' @param .dry_run Logical; if TRUE, returns the prepared request object without executing it (default: FALSE).
#' @param .max_tries Maximum retries to perform the request (default: 3).
#' @param .endpoint_url Base URL for the API (default:  Sys.getenv("AZURE_ENDPOINT_URL")).
#' @param .timeout Integer specifying the request timeout in seconds (default: 60).
#'
#' @return A tibble with information about the status of batch processing.
#' @export
check_azure_openai_batch <- function(.llms = NULL, 
                                     .endpoint_url = Sys.getenv('AZURE_ENDPOINT_URL'), 
                               .batch_id = NULL,
                               .dry_run = FALSE,
                               .max_tries = 3,
                               .timeout = 60) {
  # Extract batch_id
  if (is.null(.batch_id)) {
    if (!is.null(.llms)) {
      .batch_id <- attr(.llms, "batch_id")
      if (is.null(.batch_id)) {
        stop("No batch_id attribute found in the provided list.")
      }
    } else {
      stop("Either .llms or .batch_id must be provided.")
    }
  }
  
  # Retrieve API key
  api_key <- Sys.getenv("AZURE_OPENAI_API_KEY")
  if ((api_key == "") & !.dry_run){
    stop("API key is not set.")
  }
  
  # Build request
  request <- httr2::request(.endpoint_url) |> 
    httr2::req_url_path(paste0('openai/batches/', .batch_id)) |> 
    httr2::req_url_query(`api-version` = "2024-10-01-preview") |>
    httr2::req_headers(
      `api-key` = api_key,
    )
  
  # If .dry_run is TRUE, return the request object for inspection
  if (.dry_run) {
    return(request)
  }
  
  # Perform request with retries and error handling
  response <- request |>
    perform_generic_request(.timeout=.timeout,
                            .max_tries = .max_tries)
  
  # Parse response
  response_body <- response$content
  if("error" %in% names(response_body)){
    sprintf("Azure OpenAI API returned an Error:\nType: %s\nMessage: %s",
            response_body$error$code,
            response_body$error$message) |>
      stop()
  }
  
  # Create tibble with batch details
  result_tbl <- tibble::tibble(
    batch_id = response_body$id,
    status = response_body$status,
    created_at = lubridate::as_datetime(if(is.null(response_body$created_at)) NA_real_ else response_body$created_at),
    expires_at = lubridate::as_datetime(if(is.null(response_body$expires_at)) NA_real_ else response_body$expires_at),
    total_requests = response_body$request_counts$total,
    completed_requests = response_body$request_counts$completed,
    failed_requests = response_body$request_counts$failed
  )
  
  return(result_tbl)
}

#' List Azure OpenAI Batch Requests
#'
#' Retrieves batch request details from the Azure OpenAI Batch API.
#'
#' @param .limit Maximum number of batches to retrieve (default: 20).
#' @param .max_tries Maximum retry attempts for requests (default: 3).
#' @param .timeout Request timeout in seconds (default: 60).
#' @param .endpoint_url Base URL for the API (default:  Sys.getenv("AZURE_ENDPOINT_URL")).
#'
#' @return A tibble with batch details: batch ID, status, creation time, expiration time, 
#' and request counts (total, completed, failed).
#'
#' @export
list_azure_openai_batches <- function(.endpoint_url = Sys.getenv('AZURE_ENDPOINT_URL'), 
                                      .limit = 20,
                                      .max_tries = 3,
                                      .timeout = 60) {
  # Retrieve API key
  api_key <- Sys.getenv("AZURE_OPENAI_API_KEY")
  if (api_key == "") {
    stop("API key is not set. Please set it with: Sys.setenv(AZURE_OPENAI_API_KEY = \"YOUR-KEY-GOES-HERE\").")
  }
  
  # Set up request URL with query parameters
  request <- httr2::request(.endpoint_url) |> 
    httr2::req_url_path('openai/batches') |> 
    httr2::req_url_query(`api-version` = "2024-10-01-preview") |>
    httr2::req_headers(
      `api-key` = api_key
    ) |>
    httr2::req_url_query(limit = .limit)
  
  # Perform the request with retries and error handling
  response <- request |>
    perform_generic_request(.timeout=.timeout,
                            .max_tries = .max_tries)
  
  # Parse response
  response_body <- response$content
  if ("error" %in% names(response_body)) {
    sprintf("OpenAI API returned an Error:\nType: %s\nMessage: %s",
            response_body$error$code,
            response_body$error$message) |>
      stop()  
  }
  
  # Extract batch data and format as tibble
  batch_data <- response_body$data
  batch_tibble <- purrr::map_dfr(batch_data, ~ tibble::tibble(
    batch_id = .x$id,
    status = .x$status,
    created_at = lubridate::as_datetime(if(is.null(response_body$created_at)) NA_real_ else response_body$created_at),
    expires_at = lubridate::as_datetime(if(is.null(response_body$expires_at)) NA_real_ else response_body$expires_at),
    request_total = .x$request_counts$total,
    request_completed = .x$request_counts$completed,
    request_failed = .x$request_counts$failed
  ))
  
  return(batch_tibble)
}

#' Fetch Results for an Azure OpenAI Batch
#'
#' This function retrieves the results of a completed Azure OpenAI batch and updates
#' the provided list of `LLMMessage` objects with the responses. It aligns each
#' response with the original request using the `custom_id`s generated in `send_azure_openai_batch()`.
#'
#' @param .llms A list of `LLMMessage` objects that were part of the batch.
#' @param .batch_id Character; the unique identifier for the batch. By default this is NULL
#'                  and the function will attempt to use the `batch_id` attribute from `.llms`.
#' @param .dry_run Logical; if `TRUE`, returns the constructed request without executing it (default: `FALSE`).
#' @param .max_tries Integer; maximum number of retries if the request fails (default: `3`).
#' @param .timeout Integer; request timeout in seconds (default: `60`).
#' @param .endpoint_url Base URL for the API (default:  Sys.getenv("AZURE_ENDPOINT_URL")).
#'
#' @return A list of updated `LLMMessage` objects, each with the assistant's response added if successful.
#' @export


fetch_azure_openai_batch <- function(.llms,
                               .endpoint_url = Sys.getenv('AZURE_ENDPOINT_URL'),
                               .batch_id = NULL,
                               .dry_run = FALSE,
                               .max_tries = 3,
                               .timeout = 60) {
  c(
    ".llms must be a list of LLMMessage objects with names as custom IDs" = is.list(.llms) && all(sapply(.llms, S7_inherits, LLMMessage)),
    ".batch_id must be a non-empty character string or NULL" = is.null(.batch_id) || (is.character(.batch_id) && nzchar(.batch_id)),
    ".dry_run must be logical" = is.logical(.dry_run),
    ".max_tries must be integer-valued numeric" = is_integer_valued(.max_tries),
    ".timeout must be integer-valued numeric" = is_integer_valued(.timeout)
  ) |> validate_inputs()
  
  # Preserve original names
  original_names <- names(.llms)
  
  # Retrieve batch_id from .llms if not provided
  if (is.null(.batch_id)) {
    .batch_id <- attr(.llms, "batch_id")
    if (is.null(.batch_id)) {
      stop("No batch_id provided and no batch_id attribute found in the provided list.")
    }
  }
  
  .json <- attr(.llms, "json")
  if (is.null(.json)) {.json <- FALSE}
  
  api_obj <- api_azure_openai(short_name = "azure_openai",
                              long_name  = "Azure OpenAI",
                              api_key_env_var = "AZURE_OPENAI_API_KEY")
  
  api_key <- get_api_key(api_obj,.dry_run)
  
  request <- httr2::request(.endpoint_url) |> 
    httr2::req_url_path(paste0('openai/batches/', .batch_id)) |> 
    httr2::req_url_query(`api-version` = "2024-10-01-preview") |>
    httr2::req_headers(
      `api-key` = api_key,
    )
  
  # If .dry_run is TRUE, return the request object for inspection
  if (.dry_run) {
    return(request)
  }
  
  response <- request |>
    perform_generic_request(.timeout=.timeout,
                            .max_tries = .max_tries)
  
  response_body <- response$content
  if ("error" %in% names(response_body)) {
    sprintf("Azure OpenAI API returned an Error:\nType: %s\nMessage: %s",
            response_body$error$type,
            response_body$error$message) |>
      stop()  
  }
  
  # Check if batch has completed processing
  if (response_body$status != "completed") {
    stop("Batch processing has not completed yet. Please check again later.")
  }
  
  # Retrieve the output_file_id
  output_file_id <- response_body$output_file_id
  if (is.null(output_file_id)) {
    stop("No output_file_id found in the batch details.")
  }
  
  # Download the output file
  results_request <- httr2::request(.endpoint_url) |> 
    httr2::req_url_path(paste0('openai/files/', output_file_id, '/content')) |> 
    httr2::req_url_query(`api-version` = "2024-10-01-preview") |>
    httr2::req_headers(
      `api-key` = api_key,
    )
  
  results_response <- results_request |>
    httr2::req_timeout(.timeout) |>
    httr2::req_error(is_error = function(resp) FALSE) |>
    httr2::req_retry(
      max_tries = .max_tries,
      retry_on_failure = TRUE,
      is_transient = function(resp) httr2::resp_status(resp) %in% c(429, 503)
    ) |>
    httr2::req_perform()
  
  # Parse JSONL response and map results by custom_id
  results_lines <- strsplit(httr2::resp_body_string(results_response), "\n")[[1]]
  results_list <- lapply(results_lines, function(line) {
    if (nzchar(line)) jsonlite::fromJSON(line) else NULL
  })
  results_list <- Filter(Negate(is.null), results_list)
  
  results_by_custom_id <- purrr::set_names(results_list, sapply(results_list, function(x) x$custom_id))
  
  # Map results back to the original .llms list using names as custom IDs
  updated_llms <- lapply(names(.llms), function(custom_id) {
    result <- results_by_custom_id[[custom_id]]
    
    if (!is.null(result) && is.null(result$error) && result$response$status_code == 200) {
      assistant_reply <- result$response$body$choices$message$content
      meta_data <- extract_metadata(api_obj,result$response$body)
      llm <- add_message(.llm = .llms[[custom_id]],
                         .role = "assistant", 
                         .content =  assistant_reply,
                         .json = .json,
                         .meta = meta_data)
      return(llm)
    } else {
      warning(sprintf("Result for custom_id %s was unsuccessful or not found", custom_id))
      return(.llms[[custom_id]])
    }
  })
  
  # Restore original names
  names(updated_llms) <- original_names
  
  # Remove batch_id attribute before returning to avoid reuse conflicts
  attr(updated_llms, "batch_id") <- NULL
  attr(updated_llms, "json") <- NULL
  
  
  return(updated_llms)
}



#' Azure OpenAI Endpoint Provider Function
#'
#' The `azure_openai()` function acts as an interface for interacting with the Azure OpenAI API 
#' through main `tidyllm` verbs.
#' 
#' `azure_openai()` currently routes messages only to `azure_openai_chat()` when used with `chat()`.
#' 
#' `send_batch()`. It dynamically routes requests to OpenAI-specific functions 
#' like `azure_openai_chat()` and `azure_openai_embedding()` based on the context of the call.
#'
#' @param ... Parameters to be passed to the Azure OpenAI API specific function, 
#'   such as model configuration, input text, or API-specific options.
#' @param .called_from An internal argument that specifies which action (e.g., 
#'   `chat`) the function is being invoked from. 
#'   This argument is automatically managed and should not be modified by the user.
#'
#' @return The result of the requested action, depending on the specific function invoked 
#'   (currently, only an updated `LLMMessage` object for `azure_openai_chat()`).
#'
#' @export
azure_openai <- create_provider_function(
  .name = "azure_openai",
  chat = azure_openai_chat,
  embed = azure_openai_embedding, 
  send_batch = send_azure_openai_batch, 
  check_batch = check_azure_openai_batch,
  list_batches = list_azure_openai_batches,
  fetch_batch = fetch_azure_openai_batch
)

#' The Claude API provider class
#'
#' @noRd
api_claude <- new_class("Claude", APIProvider)


#' Convert LLMMessage to Claude API-Compatible Format
#'
#' Converts the `message_history` of an `LLMMessage` object into the
#' one needed for the Anthropic Claude API.
#'
#' @noRd
method(to_api_format, list(LLMMessage, api_claude)) <- function(.llm, 
                                                                .api) {
  
  claude_history <- filter_roles(.llm@message_history, c("user", "assistant"))
  lapply(claude_history, function(m) {
    formatted_message <- format_message(m)
    if (!is.null(formatted_message$image)) {
      list(role = m$role, content = list(
        list(type = "image", source = formatted_message$image),
        list(type = "text", text = formatted_message$content)
      ))
    } else {
      list(role = m$role, content = list(
        list(type = "text", text = formatted_message$content)
      ))
    }
  })
}

#' Extract rate limit info from  Claude API-Headers
#'
#' @noRd
method(ratelimit_from_header, list(api_claude,new_S3_class("httr2_headers"))) <- function(.api,.headers){
  list(
    this_request_time = strptime(.headers["date"], 
                                 format="%a, %d %b %Y %H:%M:%S", tz="GMT"),
    ratelimit_requests = as.integer(
      .headers["anthropic-ratelimit-requests-limit"]),
    ratelimit_requests_remaining = as.integer(
      .headers["anthropic-ratelimit-requests-remaining"]),
    ratelimit_requests_reset_time = as.POSIXct(
      .headers["anthropic-ratelimit-requests-reset"]$`anthropic-ratelimit-requests-reset`,
      format="%Y-%m-%dT%H:%M:%SZ", tz="UTC"),
    ratelimit_tokens = as.integer(
      .headers["anthropic-ratelimit-tokens-limit"]),
    ratelimit_tokens_remaining = as.integer(
      .headers["anthropic-ratelimit-tokens-remaining"]),
    ratelimit_tokens_reset_time = as.POSIXct(
      .headers["anthropic-ratelimit-tokens-reset"]$`anthropic-ratelimit-tokens-reset`,
      format="%Y-%m-%dT%H:%M:%SZ", tz="UTC")
  )
}

#' A chat parsing method for Openai to extract the assitant response 
#'
#' @noRd
method(parse_chat_response, list(api_claude,class_list)) <- function(.api,.content) {
  api_label <- .api@long_name 
  if("error" %in% names(.content)){
    sprintf("%s returned an Error:\nType: %s\nMessage: %s",
            api_label,
            .content$error$type,
            .content$error$message) |>
      stop()
  }
  
  if (!"content" %in% names(.content) || length(.content$content) == 0) {
    paste0("Received empty response from ",api_label) |>
      stop()
  }
  
  if(r_has_name(.content,"thinking")) return(.content$content[[2]]$text)
  .content$content[[1]]$text
}


#' A function to get metadata from Claude responses
#'
#' @noRd
method(extract_metadata, list(api_claude,class_list))<- function(.api,.response) {
  list(
    model             = .response$model,
    timestamp         = lubridate::as_datetime(lubridate::now()),
    prompt_tokens     = .response$usage$input_tokens,
    completion_tokens = .response$usage$output_tokens,
    total_tokens      = .response$usage$input_tokens + .response$usage$output_tokens,
    stream            = FALSE,
    specific_metadata = list(
      stop_reason        = .response$stop_reason,
      id                 = .response$id,
      stop_sequence      = .response$stop_sequence,
      thinking           = if(r_has_name(.response,"thinking")) .response$content[[1]]$thinking else NULL,
      signature          = if(r_has_name(.response,"thinking")) .response$content[[1]]$signature else NULL
    )
  )
}  

#' A function to get metadata from claude streaming responses
#'
#' @noRd
method(extract_metadata_stream, list(api_claude,class_list))<- function(.api,.stream_raw_data) {
  start_message <- .stream_raw_data |> 
    purrr::keep(~.x$type=="message_start") |>
    unlist(recursive = FALSE)
  
  last_message <- .stream_raw_data[[length(.stream_raw_data) - 1]] |> 
    unlist(recursive = FALSE)

  list(
    model             = start_message$message$model,
    timestamp         = lubridate::as_datetime(lubridate::now()),
    prompt_tokens     = start_message$message$usage$input_tokens,
    completion_tokens = last_message$usage.output_tokens,
    total_tokens      = start_message$message$usage$input_tokens + last_message$usage.output_tokens,
    stream            = TRUE,
    specific_metadata = list(warning="Specific Metadata is not yet implemented for Claude streaming requests") 
  )
}  


#Claude-specific method to format tool calls for the API
method(tools_to_api, list(api_claude, class_list)) <- function(.api, .tools) {
  purrr::map(.tools, function(tool) {
    if (length(tool@builtin) > 0) {
      tool@builtin[[1]]
    } else {
      list(
        name = tool@name,
        description = tool@description,
        input_schema = list(
          type = "object",
          properties = purrr::map(tool@input_schema, field_to_param_schema),
          required = as.list(names(tool@input_schema))
        )
      )
    }
  })
}

method(has_tool_calls, list(api_claude, class_any)) <- function(.api, .response)
  isTRUE(.response$raw$content$stop_reason == "tool_use")

method(extract_tool_calls, list(api_claude, class_any)) <- function(.api, .response)
  purrr::keep(.response$raw$content$content, ~ .x$type == "tool_use")

method(append_tool_messages, list(api_claude, class_any, class_any, class_any)) <-
  function(.api, .request_body, .response, .tool_results) {
    .request_body$messages <- .request_body$messages |>
      append(list(list(role = "assistant", content = .response$raw$content$content))) |>
      append(list(.tool_results))
    .request_body
  }

#' A method to run tool calls on Claude and create the expected response
#'
#' @noRd
method(run_tool_calls, list(api_claude, class_list, class_list)) <- function(.api, .tool_calls, .tools) {
  # Iterate over tool calls
  tool_results <- purrr::map(.tool_calls, function(tool_call) {
    # Extract name, input, and id from the tool call
    tool_name    <- tool_call$name
    tool_args    <- tool_call$input
    tool_call_id <- tool_call$id
    
    # Find the corresponding tool in the tools list
    matching_tool <- purrr::keep(.tools, ~ .x@name == tool_name)
    if (length(matching_tool) == 0) {
      warning(sprintf("No matching tool found for: %s", tool_name))
      return(NULL)
    }
    
    # Get the function for the tool and execute it with the provided arguments
    tool_function <- matching_tool[[1]]@func
    
    tool_result <-  utils::capture.output(
                        do.call(tool_function, as.list(tool_args))
                        , file = NULL) |> 
      stringr::str_c(collapse = "\n")
    
    
    # Return a content block as required by Claude:
    # type "tool_result", the tool_use_id, and the output (as a JSON string)
    list(
      type = "tool_result",
      tool_use_id = tool_call_id,
      content = jsonlite::toJSON(tool_result, auto_unbox = TRUE)
    )
  })
  
  # Remove any NULLs for tool calls that did not have matching tools
  tool_results <- purrr::compact(tool_results)
  
  # Wrap the tool results in a Claude message with role "user"
  list(
    role = "user",
    content = tool_results
  )
}

#' A method to handle streaming requests for Anthropic
#' request
#'
#' @noRd
method(handle_stream,list(api_claude,new_S3_class("httr2_response"))) <- function(.api,.stream_response) {
  stream_text <- ""
  stream_data <- list()
  repeat {
    stream_chunk <- httr2::resp_stream_sse(.stream_response)
    
    # Skip empty chunks
    if (is.null(stream_chunk$data) || !nzchar(stream_chunk$data)) {
      next  
    }
    

    # Try to parse the JSON content
    parsed_event <- tryCatch(
      jsonlite::fromJSON(stream_chunk$data, simplifyVector = FALSE, simplifyDataFrame = FALSE),
      error = function(e) {
        message("Failed to parse JSON: ", e$message)
        return(NULL)
      }
    )
    
    if (!is.null(parsed_event)) {
      stream_data <- append(stream_data,list(parsed_event))
      if (parsed_event$type == "message_stop") {
        close(.stream_response)
        message("\n---------\nStream finished\n---------\n")
        break
      }
      
      delta_content <- parsed_event$delta$text
      if (!is.null(delta_content)) {
          stream_text <- paste0(stream_text, delta_content)
          cat(delta_content)
          utils::flush.console()
        }
      }
  }
  
  list(
    reply = stream_text,
    raw_data = stream_data
  )
}




#' Collapse Claude content blocks into a single character string
#'
#' @noRd
collapse_claude_blocks <- function(blocks) {
  if (is.null(blocks)) {
    return("")
  }
  
  txt <- purrr::map_chr(blocks, function(b) {
    if (!is.list(b) || is.null(b$type)) return("")
    if (b$type == "text") return(b$text)
    ""
  })
  
  paste(txt[nzchar(txt)], collapse = "")
}


#' Inject files into Claude message contents (auto-detect block type)
#'
#' @param .claude_messages List of messages as output from to_api_format(.llm, api_claude)
#' @param .file_ids Character vector of Claude file IDs (from claude_upload_file)
#' @return Updated list of messages with appropriate file content blocks added to last user message
#' @noRd
claude_inject_files <- function(.claude_messages, .file_ids) {
  if (is.null(.file_ids) || length(.file_ids) == 0) return(.claude_messages)
  
  user_idxs <- which(sapply(.claude_messages, function(msg) msg$role == "user"))
  if (length(user_idxs) == 0) stop("No user message found to inject files into.")
  last_user_idx <- user_idxs[length(user_idxs)]
  msg <- .claude_messages[[last_user_idx]]
  
  # For each file, get metadata and inject correct block type
  file_blocks <- lapply(.file_ids, function(fid) {
    meta <- claude_file_metadata(fid)
    mime <- meta$mime_type[1] # always a tibble, so use [1]
    
    if (mime %in% c("application/pdf", "text/plain")) {
      # Document block for PDF or text
      list(
        type = "document",
        source = list(
          type = "file",
          file_id = fid
        )
      )
    } else if (grepl("^image/", mime)) {
      # Image block for images
      list(
        type = "image",
        source = list(
          type = "file",
          file_id = fid
        )
      )
    } else {
      stop(sprintf(
        "Unsupported file type for Claude message: %s (%s). Only PDF, plaintext, and images are supported.",
        fid, mime
      ))
    }
  })
  
  msg$content <- c(msg$content, file_blocks)
  .claude_messages[[last_user_idx]] <- msg
  .claude_messages
}


#' Interact with Claude AI models via the Anthropic API
#'
#' @param .llm An LLMMessage object containing the conversation history and system prompt.
#' @param .model Character string specifying the Claude model version (default: "claude-sonnet-4-6").
#' @param .max_tokens Integer specifying the maximum number of tokens in the response (default: 1024).
#' @param .temperature Numeric between 0 and 1 controlling response randomness.
#' @param .top_k Integer controlling diversity by limiting the top K tokens.
#' @param .top_p Numeric between 0 and 1 for nucleus sampling.
#' @param .metadata List of additional metadata to include with the request.
#' @param .stop_sequences Character vector of sequences that will halt response generation.
#' @param .tools List of additional tools or functions the model can use.
#' @param .json_schema A schema to enforce an output structure
#' @param .api_url Base URL for the Anthropic API (default: "https://api.anthropic.com/").
#' @param .verbose Logical; if TRUE, displays additional information about the API call (default: FALSE).
#' @param .max_tries Maximum retries to peform request
#' @param .timeout Integer specifying the request timeout in seconds (default: 60).
#' @param .stream Logical; if TRUE, streams the response piece by piece (default: FALSE).
#' @param .dry_run Logical; if TRUE, returns the prepared request object without executing it (default: FALSE).
#' @param .file_ids Character; A vector of file IDs for files that were uploaded to Anthropics Servers
#' @param .thinking Logical; if TRUE, enables Claude's thinking mode for complex reasoning tasks (default: FALSE).
#' @param .thinking_budget Integer specifying the maximum tokens Claude can spend on thinking (default: 1024). Must be at least 1024.
#' @param .max_tool_rounds Integer specifying the maximum number of tool use iterations (default: 10). 
#'   Set to 1 for single-round tool use, or higher for multi-turn agentic loops.
#'
#' @return A new LLMMessage object containing the original messages plus Claude's response.
#' @examples
#' \dontrun{
#' # Basic usage
#' msg <- llm_message("What is R programming?")
#' result <- claude_chat(msg)
#' 
#' # With custom parameters
#' result2 <- claude_chat(msg, 
#'                  .temperature = 0.7, 
#'                  .max_tokens = 1000)
#' }
#'
#' @export
claude_chat <- function(.llm,
                        .model = "claude-sonnet-4-6",
                        .max_tokens = 2048,
                        .temperature = NULL,
                        .top_k = NULL,
                        .top_p = NULL,
                        .metadata = NULL,
                        .stop_sequences = NULL,
                        .tools = NULL,
                        .json_schema = NULL,
                        .file_ids = NULL,
                        .api_url = "https://api.anthropic.com/",
                        .verbose = FALSE,
                        .max_tries = 3,
                        .timeout = 60,
                        .stream = FALSE,
                        .dry_run = FALSE,
                        .thinking = FALSE,
                        .thinking_budget = 1024,
                        .max_tool_rounds = 10) {  
  # Validate inputs to the Claude function
  c(
    ".llm must be an LLMMessage object" = S7_inherits(.llm, LLMMessage),
    ".max_tokens must be an integer" = is_integer_valued(.max_tokens),
    ".timeout must be an integer-valued numeric (seconds till timeout)" = is_integer_valued(.timeout),
    ".temperature must be numeric between 0 and 1 if provided" = 
      is.null(.temperature) | (is.numeric(.temperature) && .temperature >= 0 && .temperature <= 1),
    ".top_k must be a positive integer if provided" = 
      is.null(.top_k) | (is.numeric(.top_k) && .top_k > 0 && floor(.top_k) == .top_k),
    ".top_p must be numeric between 0 and 1 if provided" = 
      is.null(.top_p) | (is.numeric(.top_p) && .top_p >= 0 && .top_p <= 1),
    "Only one of .temperature or .top_p should be specified" = 
      is.null(.temperature) | is.null(.top_p),
    ".stop_sequences must be a character vector" = 
      is.null(.stop_sequences) | is.character(.stop_sequences),
    ".tools must be NULL, a TOOL object, or a list of TOOL objects" = is.null(.tools) || S7_inherits(.tools, TOOL) || (is.list(.tools) && all(purrr::map_lgl(.tools, ~ S7_inherits(.x, TOOL)))),
    ".verbose must be logical" = is.logical(.verbose),
    ".stream must be logical" = is.logical(.stream),
    ".max_tries must be integer-valued numeric" = is_integer_valued(.max_tries),
    ".dry_run must be logical" = is.logical(.dry_run),
    "Streaming is not supported for requests with tool calls" = is.null(.tools) || !isTRUE(.stream),
    ".json_schema must be NULL or a list or an ellmer type object" = is.null(.json_schema) | is.list(.json_schema) | is_ellmer_type(.json_schema),
    "Streaming is not supported for requests with structured outputs" = is.null(.json_schema) || !isTRUE(.stream),
    ".thinking must be logical" = is.logical(.thinking),
    ".thinking_budget must be a positive integer larger than 1024" = is_integer_valued(.thinking_budget) && .thinking_budget >= 1024,
    ".file_ids must be a character vector" =
      is.null(.file_ids) | is.character(.file_ids),
    ".max_tool_rounds must be a positive integer" = is_integer_valued(.max_tool_rounds) && .max_tool_rounds >= 1
  ) |>
    validate_inputs()
  
  json <- FALSE
  output_format <- NULL
  if(!is.null(.json_schema)){
    json <- TRUE
    if (requireNamespace("ellmer", quietly = TRUE)) {
      if(S7_inherits(.json_schema,ellmer::TypeObject)){
        .json_schema = to_schema(.json_schema)
      }
    }
    .json_schema <- add_no_extra_fields(.json_schema)
    output_format <- list(
      type = "json_schema",
      schema = .json_schema
    )
  }
  
  api_obj <- api_claude(short_name = "claude",
                        long_name  = "Anthropic Claude",
                        api_key_env_var = "ANTHROPIC_API_KEY")
  
  api_key <- get_api_key(api_obj,.dry_run)
  
  # Format message list for Claude model
  messages <- to_api_format(.llm,api_obj)
  messages <- claude_inject_files(messages, .file_ids)
  
  #Put a single tool into a list if only one is provided. 
  tools_def <- if (!is.null(.tools)) {
    if (S7_inherits(.tools, TOOL))  list(.tools) else .tools
  } else {
    NULL
  }
  
  request_body <- list(
    model = .model,
    max_tokens = .max_tokens,
    messages = messages,
    system = .llm@system_prompt,
    temperature = .temperature,
    top_k = .top_k,
    top_p = .top_p,
    metadata = .metadata,
    stop_sequences = .stop_sequences,
    stream = .stream,
    tools = if(!is.null(tools_def)) tools_to_api(api_obj, tools_def) else NULL,
    thinking = if(.thinking) list(type = "enabled", budget_tokens = .thinking_budget) else NULL,
    output_format = output_format
  ) |> purrr::compact()

  # Build beta headers - combine features as needed
  beta_features <- "files-api-2025-04-14"
  if (!is.null(output_format)) {
    beta_features <- paste(beta_features, "structured-outputs-2025-11-13", sep = ",")
  }
  
  # Build request with httr2
  request <- httr2::request(.api_url) |>
    httr2::req_url_path("/v1/messages") |>
    httr2::req_headers(
      `x-api-key` = Sys.getenv("ANTHROPIC_API_KEY"),
      `anthropic-beta` = beta_features,
      `anthropic-version` = "2023-06-01",
      `content-type` = "application/json; charset=utf-8",
      .redact = "x-api-key"
    ) |>
    httr2::req_body_json(data = request_body)  
  
  # Return only the request object 
  if (.dry_run) {
    return(request)  
  }
  
  response <- perform_chat_request(request, api_obj, .stream, .timeout)
  
  if (.stream == FALSE && !is.null(tools_def)) {
    response <- process_tool_loop(
      .api = api_obj,
      .response = response,
      .tools_def = tools_def,
      .request_body = request_body,
      .request = request,
      .timeout = .timeout,
      .max_tries = .max_tries,
      .max_tool_rounds = .max_tool_rounds
    )
  }
  
  assistant_reply <- if (.stream) {
    response$assistant_reply
  } else {
    collapse_claude_blocks(response$raw$content$content)
  }
  
  track_rate_limit(api_obj, response$headers, .verbose)
  
  add_message(
    .llm     = .llm,
    .role    = "assistant",
    .content = assistant_reply,
    .json    = json,
    .meta    = response$meta
  )
}  

#' Send a Batch of Messages to Claude API
#'
#' This function creates and submits a batch of messages to the Claude API for asynchronous processing.
#'
#' @param .llms A list of LLMMessage objects containing conversation histories.
#' @param .model Character string specifying the Claude model version (default: "claude-sonnet-4-6").
#' @param .max_tokens Integer specifying the maximum tokens per response (default: 1024).
#' @param .temperature Numeric between 0 and 1 controlling response randomness.
#' @param .top_k Integer for diversity by limiting the top K tokens.
#' @param .top_p Numeric between 0 and 1 for nucleus sampling.
#' @param .stop_sequences Character vector of sequences that halt response generation.
#' @param .api_url Base URL for the Claude API (default: "https://api.anthropic.com/").
#' @param .verbose Logical; if TRUE, prints a message with the batch ID (default: FALSE).
#' @param .overwrite Logical; if TRUE, allows overwriting an existing batch ID associated with the request (default: FALSE).
#' @param .max_tries Maximum number of retries to perform the request.
#' @param .timeout Integer specifying the request timeout in seconds (default: 60).
#' @param .dry_run Logical; if TRUE, returns the prepared request object without executing it (default: FALSE).
#' @param .id_prefix Character string to specify a prefix for generating custom IDs when names in `.llms` are missing.
#' @param .json_schema A schema to enforce an output structure
#' @param .thinking Logical; if TRUE, enables Claude's thinking mode for complex reasoning tasks (default: FALSE).
#' @param .thinking_budget Integer specifying the maximum tokens Claude can spend on thinking (default: 1024). Must be at least 1024.
#'
#'   Defaults to "tidyllm_claude_req_".
#' 
#' @return An updated and named list of `.llms` with identifiers that align with batch responses, including a `batch_id` attribute.
#' @export
send_claude_batch <- function(.llms, 
                              .model = "claude-sonnet-4-6", 
                              .max_tokens = 1024, 
                              .temperature = NULL, 
                              .top_k = NULL, 
                              .top_p = NULL, 
                              .stop_sequences = NULL, 
                              .json_schema = NULL,
                              .thinking = FALSE,
                              .thinking_budget = 1024,
                              .api_url = "https://api.anthropic.com/", 
                              .verbose = FALSE,
                              .dry_run = FALSE,
                              .overwrite = FALSE,
                              .max_tries = 3,
                              .timeout = 60,
                              .id_prefix = "tidyllm_claude_req_"
                              ) {
  # Input validation
  c(
    ".llms must be a list of LLMMessage objects" = is.list(.llms) && all(sapply(.llms, S7_inherits, LLMMessage)),
    ".max_tokens must be an integer" = is_integer_valued(.max_tokens),
    ".temperature must be numeric between 0 and 1 if provided" = is.null(.temperature) || (.temperature >= 0 && .temperature <= 1),
    ".top_k must be a positive integer if provided" = is.null(.top_k) || (is.numeric(.top_k) && .top_k > 0 && floor(.top_k) == .top_k),
    ".top_p must be numeric between 0 and 1 if provided" = is.null(.top_p) || (.top_p >= 0 && .top_p <= 1),
    "Only one of .temperature or .top_p should be specified" = is.null(.temperature) || is.null(.top_p),
    ".stop_sequences must be a character vector" = is.null(.stop_sequences) || is.character(.stop_sequences),
    ".verbose must be logical" = is.logical(.verbose),
    ".dry_run must be logical" = is.logical(.dry_run),
    ".overwrite must be logical" = is.logical(.overwrite),
    ".json_schema must be NULL or a list or an ellmer type object" = is.null(.json_schema) | is.list(.json_schema) | is_ellmer_type(.json_schema),
    ".id_prefix must be a character vector of length 1" = is.character(.id_prefix),
    ".max_tries must be integer-valued numeric" = is_integer_valued(.max_tries),
    ".timeout must be an integer" = is_integer_valued(.timeout)
  ) |> validate_inputs()
  
  
  api_obj <- api_claude(short_name = "claude",
                        long_name  = "Anthropic Claude",
                        api_key_env_var = "ANTHROPIC_API_KEY")
  
  api_key <- get_api_key(api_obj,.dry_run)
  
  prepared_llms  <- prepare_llms_for_batch(api_obj,
                                           .llms=.llms,
                                           .id_prefix=.id_prefix,
                                           .overwrite = .overwrite)
  
  json <- FALSE
  output_format <- NULL
  if(!is.null(.json_schema)){
    json <- TRUE
    if (requireNamespace("ellmer", quietly = TRUE)) {
      if(S7_inherits(.json_schema,ellmer::TypeObject)){
        .json_schema = to_schema(.json_schema)
      }
    }
    .json_schema <- add_no_extra_fields(.json_schema)
    output_format <- list(
      type = "json_schema",
      schema = .json_schema
    )
  }
  
  requests_list <- lapply(seq_along(prepared_llms), function(i) { 
    messages <- to_api_format(.llms[[i]], api_obj)
    
    custom_id <- names(prepared_llms)[i]
    list(
      custom_id = custom_id,
      params = list(
        model = .model,
        max_tokens = .max_tokens,
        messages = messages,
        temperature = .temperature,
        top_k = .top_k,
        top_p = .top_p,
        system = .llms[[i]]@system_prompt,
        stop_sequences = .stop_sequences,
        thinking = if(.thinking) list(type = "enabled", budget_tokens = .thinking_budget) else NULL,
        output_format = output_format
      ) |> purrr::compact()
    )
  })
  
  beta_features <- "message-batches-2024-09-24"
  if (!is.null(output_format)) {
    beta_features <- paste(beta_features, "structured-outputs-2025-11-13", sep = ",")
  }
  
  request <- httr2::request(.api_url) |>
    httr2::req_url_path("/v1/messages/batches") |>
    httr2::req_headers(
      `x-api-key` = api_key,
      `anthropic-version` = "2023-06-01",
      `anthropic-beta` = beta_features,
      `content-type` = "application/json; charset=utf-8",
      .redact = "x-api-key"
    ) |>
    httr2::req_body_json(list(requests = requests_list))
  
  if (.dry_run) {
    return(request)
  }
  
  response <- request |>
    perform_generic_request(.timeout=.timeout,
                            .max_tries = .max_tries)
  

  if("error" %in% names(response$content)){
    sprintf("Anthropic API returned an Error:\nType: %s\nMessage: %s",
            response$content$error$type,
            response$content$error$message) |>
      stop()
  }
  
  # Attach batch_id as an attribute to .llms
  batch_id <- response$content$id
  attr(prepared_llms, "batch_id") <- batch_id
  attr(prepared_llms, "json") <- json
  
  
  if (.verbose) {
    message("Batch submitted successfully. Batch ID: ", batch_id)
  }
  
  return(prepared_llms)
}


#' Check Batch Processing Status for Claude API
#'
#' This function retrieves the processing status and other details of a specified Claude batch ID
#' from the Claude API.
#'
#' @param .llms A list of LLMMessage objects
#' @param .batch_id A manually set batchid
#' @param .api_url Character; base URL of the Claude API (default: "https://api.anthropic.com/").
#' @param .max_tries Maximum retries to peform request
#' @param .timeout Integer specifying the request timeout in seconds (default: 60).
#' @param .dry_run Logical; if TRUE, returns the prepared request object without executing it (default: FALSE).
#' @return A tibble with information about the status of batch processing 
#' @export
check_claude_batch <- function(.llms = NULL, 
                               .batch_id = NULL, 
                               .api_url = "https://api.anthropic.com/",
                               .dry_run = FALSE,
                               .max_tries = 3,
                               .timeout = 60) {
  # Extract batch_id
  if (is.null(.batch_id)) {
    if (!is.null(.llms)) {
      .batch_id <- attr(.llms, "batch_id")
      if (is.null(.batch_id)) {
        stop("No batch_id attribute found in the provided list.")
      }
    } else {
      stop("Either .llms or .batch_id must be provided.")
    }
  }
  
  # Retrieve API key
  api_key <- Sys.getenv("ANTHROPIC_API_KEY")
  if ((api_key == "") & !.dry_run){
    stop("API key is not set.")
  }
  
  # Build request
  request_url <- paste0(.api_url, "v1/messages/batches/", .batch_id)
  request <- httr2::request(request_url) |>
    httr2::req_headers(
      `x-api-key` = api_key,
      `anthropic-version` = "2023-06-01",
      `anthropic-beta` = "message-batches-2024-09-24",
      .redact = "x-api-key"
    )
  
  # Perform request with retries and error handling
  response <- request  |>
    perform_generic_request(.timeout=.timeout,
                            .max_tries = .max_tries)
  
  # Parse response
  response_body <- response$content
  if("error" %in% names(response_body)){
    sprintf("Anthropic API returned an Error:\nType: %s\nMessage: %s",
            response_body$error$type,
            response_body$error$message) |>
      stop()
  }
  
  # Create tibble with batch details
  tibble::tibble(
    batch_id = response_body$id,
    status = response_body$processing_status,
    created_at = lubridate::ymd_hms(response_body$created_at, tz = "UTC"),
    expires_at = lubridate::ymd_hms(response_body$expires_at, tz = "UTC"),
    req_succeeded  = response_body$request_counts$succeeded,
    req_errored    = response_body$request_counts$errored,
    req_expired    = response_body$request_counts$expired,
    req_canceled   = response_body$request_counts$canceled
  )
}


#' Fetch Results for a Claude Batch
#'
#' This function retrieves the results of a completed Claude batch and updates
#' the provided list of `LLMMessage` objects with the responses. It aligns each
#' response with the original request using the `custom_id`s generated in `send_claude_batch()`.
#'
#' @param .llms A list of `LLMMessage` objects that were part of the batch. The list should
#'              have names (custom IDs) set by `send_claude_batch()` to ensure correct alignment.
#' @param .batch_id Character; the unique identifier for the batch. By default this is NULL
#'                  and the function will attempt to use the `batch_id` attribute from `.llms`.
#' @param .api_url Character; the base URL for the Claude API (default: "https://api.anthropic.com/").
#' @param .dry_run Logical; if `TRUE`, returns the constructed request without executing it (default: `FALSE`).
#' @param .max_tries Integer; maximum number of retries if the request fails (default: `3`).
#' @param .timeout Integer; request timeout in seconds (default: `60`).
#'
#' @return A list of updated `LLMMessage` objects, each with the assistant's response added if successful.
#' @export
fetch_claude_batch <- function(.llms, 
                               .batch_id = NULL, 
                               .api_url = "https://api.anthropic.com/",
                               .dry_run = FALSE,
                               .max_tries = 3,
                               .timeout = 60) {
  c(
    ".llms must be a list of LLMMessage objects with names as custom IDs" = is.list(.llms) && all(sapply(.llms, S7_inherits, LLMMessage)),
    ".batch_id must be a non-empty character string or NULL" = is.null(.batch_id) || (is.character(.batch_id) && nzchar(.batch_id)),
    ".api_url must be a non-empty character string" = is.character(.api_url) && nzchar(.api_url),
    ".dry_run must be logical" = is.logical(.dry_run),
    ".max_tries must be an integer" = is_integer_valued(.max_tries),
    ".timeout must be an integer" = is_integer_valued(.timeout)
  ) |> validate_inputs()
  
  # Preserve original names
  original_names <- names(.llms)
  if(!is.null(attr(.llms,"json"))) json <- attr(.llms,"json") else json <- FALSE
  
  # Retrieve batch_id from .llms if not provided
  if (is.null(.batch_id)) {
    .batch_id <- attr(.llms, "batch_id")
    if (is.null(.batch_id)) {
      stop("No batch_id provided and no batch_id attribute found in the provided list.")
    }
  }
  
  api_obj <- api_claude(short_name = "claude",
                        long_name  = "Anthropic Claude",
                        api_key_env_var = "ANTHROPIC_API_KEY")
  
  api_key <- get_api_key(api_obj,.dry_run)
  
  # Construct request URL to get batch details
  batch_details_url <- paste0(.api_url, "v1/messages/batches/", .batch_id)
  
  request <- httr2::request(batch_details_url) |>
    httr2::req_headers(
      `x-api-key` = api_key,
      `anthropic-version` = "2023-06-01",
      `anthropic-beta` = "message-batches-2024-09-24",
      .redact = "x-api-key"
    )
  
  # If .dry_run is TRUE, return the request object for inspection
  if (.dry_run) {
    return(request)
  }
  
  response <- request |>
    perform_generic_request(.timeout=.timeout,
                            .max_tries = .max_tries)
  
  response_body <- response$content
  if ("error" %in% names(response_body)) {
    sprintf("Anthropic API returned an Error:\nType: %s\nMessage: %s",
            response_body$error$type,
            response_body$error$message) |>
      stop()  
  }
  
  # Check if batch has completed processing
  if (response_body$processing_status != "ended") {
    stop("Batch processing has not ended yet. Please check again later.")
  }
  
  # Retrieve the results from results_url
  results_url <- response_body$results_url
  results_request <- httr2::request(results_url) |>
    httr2::req_headers(
      `x-api-key` = api_key,
      `anthropic-version` = "2023-06-01",
      `anthropic-beta` = "message-batches-2024-09-24",
      .redact = "x-api-key"
    )
  
  results_response <- results_request |>
    httr2::req_timeout(.timeout) |>
    httr2::req_error(is_error = function(resp) FALSE) |>
    httr2::req_retry(
      max_tries = .max_tries,
      retry_on_failure = TRUE,
      is_transient = function(resp) httr2::resp_status(resp) %in% c(429, 503)
    ) |>
    httr2::req_perform()
  
  # Parse JSONL response and map results by custom_id
  results_lines <- strsplit(httr2::resp_body_string(results_response), "\n")[[1]]
  results_list <- lapply(results_lines, function(line) {
    if (nzchar(line)) jsonlite::fromJSON(line) else NULL
  })
  results_list <- Filter(Negate(is.null), results_list)
  
  results_by_custom_id <- purrr::set_names(results_list, sapply(results_list, function(x) x$custom_id))
  
  # Map results back to the original .llms list using names as custom IDs
  updated_llms <- lapply(names(.llms), function(custom_id) {
    result <- results_by_custom_id[[custom_id]]
    
    if (!is.null(result) && result$result$type == "succeeded") {
      assistant_reply <- result$result$message$content$text
      llm <- add_message(.llm = .llms[[custom_id]],
                         .role = "assistant", 
                         .content = assistant_reply,
                         .json = json,
                         .meta = extract_metadata(api_obj, result$result$message))
      return(llm)
    } else {
      warning(sprintf("Result for custom_id %s was unsuccessful or not found", custom_id))
      return(.llms[[custom_id]])
    }
  })
  
  # Restore original names
  names(updated_llms) <- original_names
  
  # Remove batch_id attribute before returning to avoid reuse conflicts
  attr(updated_llms, "batch_id") <- NULL
  attr(updated_llms, "json") <- NULL
  
  return(updated_llms)
}

#' List Claude Batch Requests
#'
#' Retrieves batch request details from the Claude API.
#'
#' @param .api_url Base URL for the Claude API (default: "https://api.anthropic.com/").
#' @param .limit Maximum number of batches to retrieve (default: 20).
#' @param .max_tries Maximum retry attempts for requests (default: 3).
#' @param .timeout Request timeout in seconds (default: 60).
#'
#' @return A tibble with batch details: batch ID, status, creation time, expiration time, 
#' and request counts (succeeded, errored, expired, canceled).
#'
#' @export
list_claude_batches <- function(.api_url = "https://api.anthropic.com/", 
                                .limit = 20, 
                                .max_tries = 3,
                                .timeout = 60) {
  # Retrieve API key
  api_key <- Sys.getenv("ANTHROPIC_API_KEY")
  if (api_key == "") {
    stop("API key is not set. Please set it with: Sys.setenv(ANTHROPIC_API_KEY = \"YOUR-KEY-GOES-HERE\").")
  }
  
  # Set up request URL with query parameters
  request <- httr2::request(.api_url) |>
    httr2::req_url_path("/v1/messages/batches") |>
    httr2::req_headers(
      `x-api-key` = api_key,
      `anthropic-version` = "2023-06-01",
      `anthropic-beta` = "message-batches-2024-09-24",
      .redact = "x-api-key"
    ) |>
    httr2::req_url_query(limit = .limit)
  
  # Perform the request with retries and error handling
  response <- request |>
    perform_generic_request(.timeout=.timeout,
                            .max_tries = .max_tries)
  
  # Parse response and handle any errors
  if ("error" %in% names(response$content)) {
    stop(sprintf("Anthropic API Error: %s - %s", response$content$error$type, response$content$error$message))
  }
  
  # Extract batch list details with parsed dates
  batch_list <- purrr::map_dfr(response$content$data, function(batch) {
    tibble::tibble(
      batch_id = batch$id,
      status = batch$processing_status,
      created_at = lubridate::ymd_hms(batch$created_at, tz = "UTC"),
      expires_at = lubridate::ymd_hms(batch$expires_at, tz = "UTC"),
      req_succeeded = batch$request_counts$succeeded,
      req_errored = batch$request_counts$errored,
      req_expired = batch$request_counts$expired,
      req_canceled = batch$request_counts$canceled
    )
  })
  

  return(batch_list)
}

#' List Available Models from the Anthropic Claude API
#'
#' @param .api_url Base URL for the API (default: "https://api.anthropic.com").
#' @param .timeout Request timeout in seconds (default: 60).
#' @param .max_tries Maximum number of retries for the API request (default: 3).
#' @param .dry_run Logical; if TRUE, returns the prepared request object without executing it.
#' @param .verbose Logical; if TRUE, prints additional information about the request.
#'
#' @return A tibble containing model information (columns include `type`,`id`, `display_name`, and `created_at`),
#'   or NULL if no models are found.
#'
#' @export
claude_list_models <- function(.api_url = "https://api.anthropic.com",
                               .timeout = 60,
                               .max_tries = 3,
                               .dry_run = FALSE,
                               .verbose = FALSE) {

  # Create an API object for Claude using the tidyllm helper
  api_obj <- api_claude(short_name = "claude",
                        long_name  = "Anthropic Claude",
                        api_key_env_var = "ANTHROPIC_API_KEY")
  
  # Retrieve the API key (will error if not set, unless in dry run mode)
  api_key <- get_api_key(api_obj, .dry_run)
  
  # Build the request to the /v1/models endpoint
  request <- httr2::request(.api_url) |>
    httr2::req_url_path("/v1/models") |>
    httr2::req_headers(
      `x-api-key` = api_key,
      `anthropic-version` = "2023-06-01",
      `content-type` = "application/json; charset=utf-8"
    )
  
  # If dry run is requested, return the constructed request object
  if (.dry_run) {
    return(request)
  }
  
  # Perform the request with specified timeout and retry logic
  response <- request |>
    httr2::req_timeout(.timeout) |>
    httr2::req_retry(max_tries = .max_tries) |>
    httr2::req_perform() |>
    httr2::resp_body_json()
  
  if (.verbose) {
    message("Retrieved response from Anthropic Claude: ", response$object)
  }
  
  # Check if the "data" field exists and contains models
  if (!is.null(response$data)) {
    models <- response$data
    
    # Create a tibble with selected model information
    model_info <- models |> 
      purrr::map_dfr(tibble::as_tibble)
    
    return(model_info)
  } else {
    return(NULL)
  }
}

#' Upload a File to Claude API
#'
#' Uploads a file to the Claude API and returns its metadata as a tibble.
#'
#' @param .file_path The local file path of the file to upload.
#' @param .api_url Base URL for the Claude API (default: "https://api.anthropic.com/").
#' @param .timeout Request timeout in seconds (default: 60).
#' @param .max_tries Maximum retry attempts for requests (default: 3).
#' @param .dry_run Logical; if TRUE, returns the prepared request object without executing it.
#' @return A tibble containing metadata about the uploaded file, including its file_id, name, and size.
#' @export
claude_upload_file <- function(.file_path, 
                               .api_url = "https://api.anthropic.com/",
                               .timeout = 60,
                               .max_tries = 3,
                               .dry_run = FALSE) {
  
  # Validate inputs
  c(
    ".file_path must be a character string" = is.character(.file_path) && length(.file_path) == 1,
    "File must exist" = file.exists(.file_path),
    ".api_url must be a character string" = is.character(.api_url),
    ".timeout must be an integer" = is_integer_valued(.timeout),
    ".max_tries must be an integer" = is_integer_valued(.max_tries),
    ".dry_run must be logical" = is.logical(.dry_run)
  ) |> validate_inputs()
  
  api_obj <- api_claude(short_name = "claude",
                        long_name  = "Anthropic Claude",
                        api_key_env_var = "ANTHROPIC_API_KEY")
  
  api_key <- get_api_key(api_obj, .dry_run)
  
  # Build the request
  request <- httr2::request(.api_url) |>
    httr2::req_url_path("/v1/files") |>
    httr2::req_headers(
      `x-api-key` = api_key,
      `anthropic-version` = "2023-06-01",
      `anthropic-beta` = "files-api-2025-04-14",
      .redact = "x-api-key"
    ) |>
    httr2::req_body_multipart(
      file = curl::form_file(.file_path)
    )
  
  if (.dry_run) {
    return(request)
  }
  
  # Perform the request
  response <- request |>
    perform_generic_request(.timeout = .timeout, .max_tries = .max_tries)
  
  # Handle errors
  if ("error" %in% names(response$content)) {
    sprintf("Claude API returned an Error:\nType: %s\nMessage: %s",
            response$content$error$type,
            response$content$error$message) |>
      stop()
  }
  
  # Return metadata as tibble
  tibble::tibble(
    file_id = response$content$id,
    filename = response$content$filename,
    size_bytes = as.numeric(response$content$size_bytes),
    mime_type = response$content$mime_type,
    created_at = response$content$created_at
  )
}

#' Retrieve Metadata for a File from Claude API
#'
#' Retrieves metadata for a specific file uploaded to the Claude API.
#'
#' @param .file_id The file ID to retrieve metadata for.
#' @param .api_url Base URL for the Claude API (default: "https://api.anthropic.com/").
#' @param .timeout Request timeout in seconds (default: 60).
#' @param .max_tries Maximum retry attempts for requests (default: 3).
#' @param .dry_run Logical; if TRUE, returns the prepared request object without executing it.
#' @return A tibble containing metadata fields such as file_id, filename, size, and MIME type.
#' @export
claude_file_metadata <- function(.file_id,
                                 .api_url = "https://api.anthropic.com/",
                                 .timeout = 60,
                                 .max_tries = 3,
                                 .dry_run = FALSE) {
  
  # Validate inputs
  c(
    ".file_id must be a character string" = is.character(.file_id) && length(.file_id) == 1,
    ".api_url must be a character string" = is.character(.api_url),
    ".timeout must be an integer" = is_integer_valued(.timeout),
    ".max_tries must be an integer" = is_integer_valued(.max_tries),
    ".dry_run must be logical" = is.logical(.dry_run)
  ) |> validate_inputs()
  
  api_obj <- api_claude(short_name = "claude",
                        long_name  = "Anthropic Claude",
                        api_key_env_var = "ANTHROPIC_API_KEY")
  
  api_key <- get_api_key(api_obj, .dry_run)
  
  # Build the request
  request <- httr2::request(.api_url) |>
    httr2::req_url_path(paste0("/v1/files/", .file_id)) |>
    httr2::req_headers(
      `x-api-key` = api_key,
      `anthropic-version` = "2023-06-01",
      `anthropic-beta` = "files-api-2025-04-14",
      .redact = "x-api-key"
    )
  
  if (.dry_run) {
    return(request)
  }
  
  # Perform the request
  response <- request |>
    perform_generic_request(.timeout = .timeout, .max_tries = .max_tries)
  
  # Handle errors
  if ("error" %in% names(response$content)) {
    sprintf("Claude API returned an Error:\nType: %s\nMessage: %s",
            response$content$error$type,
            response$content$error$message) |>
      stop()
  }
  
  # Return metadata as tibble
  tibble::tibble(
    file_id = response$content$id,
    filename = response$content$filename,
    size_bytes = as.numeric(response$content$size_bytes),
    mime_type = response$content$mime_type,
    created_at = response$content$created_at
  )
}

#' List Files in Claude API
#'
#' Lists metadata for files uploaded to the Claude API, supporting pagination.
#'
#' @param .limit The maximum number of files to return (default: 20).
#' @param .order Order of results, either "asc" or "desc" (default: "desc").
#' @param .api_url Base URL for the Claude API (default: "https://api.anthropic.com/").
#' @param .timeout Request timeout in seconds (default: 60).
#' @param .max_tries Maximum retry attempts for requests (default: 3).
#' @param .dry_run Logical; if TRUE, returns the prepared request object without executing it.
#' @return A tibble containing metadata for each file, including file_id, filename, size, and MIME type.
#' @export
claude_list_files <- function(.limit = 20,
                              .order = "desc",
                              .api_url = "https://api.anthropic.com/",
                              .timeout = 60,
                              .max_tries = 3,
                              .dry_run = FALSE) {

  # Validate inputs
  c(
    ".limit must be a positive integer" = is_integer_valued(.limit) && .limit > 0,
    ".order must be 'asc' or 'desc'" = .order %in% c("asc", "desc"),
    ".api_url must be a character string" = is.character(.api_url),
    ".timeout must be an integer" = is_integer_valued(.timeout),
    ".max_tries must be an integer" = is_integer_valued(.max_tries),
    ".dry_run must be logical" = is.logical(.dry_run)
  ) |> validate_inputs()
  
  api_obj <- api_claude(short_name = "claude",
                        long_name  = "Anthropic Claude",
                        api_key_env_var = "ANTHROPIC_API_KEY")
  
  api_key <- get_api_key(api_obj, .dry_run)
  
  # Build the request
  request <- httr2::request(.api_url) |>
    httr2::req_url_path("/v1/files") |>
    httr2::req_url_query(
      limit = .limit,
      order = .order
    ) |>
    httr2::req_headers(
      `x-api-key` = api_key,
      `anthropic-version` = "2023-06-01",
      `anthropic-beta` = "files-api-2025-04-14",
      .redact = "x-api-key"
    )
  
  if (.dry_run) {
    return(request)
  }
  
  # Perform the request
  response <- request |>
    perform_generic_request(.timeout = .timeout, .max_tries = .max_tries)
  
  # Handle errors
  if ("error" %in% names(response$content)) {
    sprintf("Claude API returned an Error:\nType: %s\nMessage: %s",
            response$content$error$type,
            response$content$error$message) |>
      stop()
  }
  
  # Convert to tibble
  files <- response$content$data
  if (length(files) == 0) {
    return(tibble::tibble(
      file_id = character(0),
      filename = character(0),
      size_bytes = numeric(0),
      mime_type = character(0),
      created_at = character(0)
    ))
  }
  
  purrr::map_dfr(files, ~ tibble::tibble(
    file_id = .x$id,
    filename = .x$filename,
    size_bytes = as.numeric(.x$size_bytes),
    mime_type = .x$mime_type,
    created_at = .x$created_at
  ))
}

#' Delete a File from Claude API
#'
#' Deletes a specific file from the Claude API using its file ID.
#'
#' @param .file_id The file ID to delete.
#' @param .api_url Base URL for the Claude API (default: "https://api.anthropic.com/").
#' @param .timeout Request timeout in seconds (default: 60).
#' @param .max_tries Maximum retry attempts for requests (default: 3).
#' @param .dry_run Logical; if TRUE, returns the prepared request object without executing it.
#' @return Invisibly returns NULL. Prints a confirmation message upon successful deletion.
#' @export
claude_delete_file <- function(.file_id,
                               .api_url = "https://api.anthropic.com/",
                               .timeout = 60,
                               .max_tries = 3,
                               .dry_run = FALSE) {
  
  # Validate inputs
  c(
    ".file_id must be a character string" = is.character(.file_id) && length(.file_id) == 1,
    ".api_url must be a character string" = is.character(.api_url),
    ".timeout must be an integer" = is_integer_valued(.timeout),
    ".max_tries must be an integer" = is_integer_valued(.max_tries),
    ".dry_run must be logical" = is.logical(.dry_run)
  ) |> validate_inputs()
  
  api_obj <- api_claude(short_name = "claude",
                        long_name  = "Anthropic Claude",
                        api_key_env_var = "ANTHROPIC_API_KEY")
  
  api_key <- get_api_key(api_obj, .dry_run)
  
  # Build the request
  request <- httr2::request(.api_url) |>
    httr2::req_url_path(paste0("/v1/files/", .file_id)) |>
    httr2::req_method("DELETE") |>
    httr2::req_headers(
      `x-api-key` = api_key,
      `anthropic-version` = "2023-06-01",
      `anthropic-beta` = "files-api-2025-04-14",
      .redact = "x-api-key"
    )
  
  if (.dry_run) {
    return(request)
  }
  
  # Perform the request
  response <- request |>
    perform_generic_request(.timeout = .timeout, .max_tries = .max_tries)
  
  # Handle errors (though DELETE typically returns 204 with no content)
  if ("error" %in% names(response$content)) {
    sprintf("Claude API returned an Error:\nType: %s\nMessage: %s",
            response$content$error$type,
            response$content$error$message) |>
      stop()
  }
  
  message("File ", .file_id, " has been successfully deleted.")
  invisible(NULL)
}


#' Builtin Claude Web Search Tool
#'
#' Returns a TOOL object for Claude's builtin web_search tool.
#'
#' @export
claude_websearch <- function() {
  TOOL(
    name = "web_search",
    description = "Builtin tool: web_search",
    input_schema = list(),           # no input schema needed
    func = function() NULL,          # not used for builtin tools
    builtin = list(list(
      name = "web_search",
      type = "web_search_20250305"
    ))
  )
}



#' Provider Function for Claude models on the Anthropic API
#'
#' The `claude()` function acts as an interface for interacting with the Anthropic API 
#' through main `tidyllm` verbs such as `chat()`, `embed()`, and 
#' `send_batch()`. It dynamically routes requests to Claude-specific functions 
#' like `claude_chat()` and `send_claude_batch()` based on the context of the call.
#'
#' @param ... Parameters to be passed to the appropriate OpenAI-specific function, 
#'   such as model configuration, input text, or API-specific options.
#' @param .called_from An internal argument that specifies which action (e.g., 
#'   `chat`, `send_batch`) the function is being invoked from. 
#'   This argument is automatically managed and should not be modified by the user.
#'
#' @return The result of the requested action, depending on the specific function invoked 
#'   (e.g., an updated `LLMMessage` object for `chat()`, or a matrix for `embed()`).
#' 
#' @export
claude <- create_provider_function(
  .name = "claude",
  chat = claude_chat,
  send_batch = send_claude_batch,
  check_batch = check_claude_batch,
  list_batches = list_claude_batches,
  fetch_batch = fetch_claude_batch,
  list_models = claude_list_models
)

#' Is an input an integer valued numeric (needed to validate inputs)
#'
#' This internal function is a small helper used to validate inputs in functions
#'
#' @param .x An input that is checked to see whether it is an interger valued numeric
#' @return A logical
#' @noRd
is_integer_valued <- function(.x) {
  # Check if the input is numeric
  if (is.numeric(.x)) {
    # Further check if the numeric input is effectively an integer
    if (.x == as.integer(.x)) {
      return(TRUE)
    } else {
      return(FALSE)
    }
  } else {
    return(FALSE)
  }
}

#' Validate Input Conditions for R Functions
#'
#' This internal function validates specified conditions for function inputs and stops the function execution if any condition is not met. It uses a named vector of predicates where each name is the error message associated with the predicate condition.
#'
#' @param .predicates A named vector where each element is a logical condition and the name of each element is the corresponding error message to be displayed if the condition is FALSE.
#' @return None; the function will stop execution and throw an error if a validation fails.
#' @examples
#' validate_inputs(c(
#'   "Input must be numeric" = is.numeric(5),
#'   "Input must be integer" = 5 == as.integer(5)
#' ))
#' @noRd
validate_inputs <- function(.predicates) {
  # Use lapply to iterate over predicates and stop on the first failure
  results <- lapply(names(.predicates), function(error_msg) {
    if (!.predicates[[error_msg]]) {
      stop(error_msg)
    }
  })
}

#' Small helper function to recursively find whether a list contains a name
#'
#' @param .x A list to search within.
#' @param .name A character string of the name to search for.
#' @noRd
r_has_name <- function(.x, .name) {
  if (is.list(.x)) {
    return(.name %in% names(.x) || any(sapply(.x, r_has_name, .name = .name)))
  }
  FALSE
}

#' Helper to determine if some object is an ellmer type
#'
#' @param .x An input
#' @noRd
is_ellmer_type <- function(.x) {
  if (!requireNamespace("ellmer", quietly = TRUE)) return(FALSE)
  inherits(.x, c("ellmer::TypeBasic", "ellmer::TypeEnum", 
                "ellmer::TypeObject", "ellmer::TypeArray"))
}

# Helper function to filter mmessages by roles
#'
#' @param .message_hisotry A message history in the format used within LLMMessage
#' @param .roles A vector of roles (default: `c("user", "assistant")`)
#' @return A filtered message_history
#' @noRd
filter_roles = function(.message_history,
                        .roles = c("user", "assistant")) {
  Filter(function(x) "role" %in% names(x) && x$role %in% .roles, .message_history)
}

#' @title Guess MIME Type from File Extension
#' @description Internal utility function to guess the MIME type of a file based on its extension.
#' This function provides a mapping of common file extensions to their respective MIME types.
#' 
#' @param file_path A string representing the path to the file.
#' 
#' @return A character vector of possible MIME types, or "application/octet-stream" if the extension is unknown.
#' 
#' @importFrom rlang %||%
#' @noRd
guess_mime_type <- function(file_path) {
  # Extract the file extension
  ext <- tools::file_ext(file_path)
  
  # Define a mapping of extensions to MIME types
  mime_types <- list(
    pdf = "application/pdf",
    txt = "text/plain",
    html = "text/html",
    css = "text/css",
    md = "text/md",
    csv = "text/csv",
    xml = "text/xml",
    rtf = "text/rtf",
    js =  "text/javascript",
    py =  "text/x-python",
    png = "image/png",
    jpeg = "image/jpeg",
    jpg = "image/jpeg",
    webp = "image/webp",
    heic = "image/heic",
    heif = "image/heif",
    mp4 = "video/mp4",
    mpeg = "video/mpeg",
    mov = "video/mov",
    avi = "video/avi",
    flv = "video/x-flv",
    mpg = "video/mpg",
    webm = "video/webm",
    wmv = "video/wmv",
    `3gpp` = "video/3gpp",
    wav = "audio/wav",
    mp3 = "audio/mp3",
    aiff = "audio/aiff",
    aac = "audio/aac",
    ogg = "audio/ogg",
    flac = "audio/flac"
  )
  
  # Return the corresponding MIME type(s) or a default value
  mime_types[[tolower(ext)]] %||% "application/octet-stream"
}


#' Recursively enforces Anthropic's structured output requirement that
#' every object-type node in a JSON schema must explicitly declare that
#' no unspecified fields are allowed. 
#' The function walks the schema, identifies nodes with type set to "object", adds
#' additionalProperties set to FALSE, and then applies the same logic
#' to all nested properties. Used only inside the Claude backend to
#' adjust provider-agnostic schemata generated by tidyllm_schema or
#' by ellmer before sending them to the API.
#' @noRd
add_no_extra_fields <- function(node) {
  if (!is.list(node)) return(node)
  
  if (identical(node$type, "object")) {
    node$additionalProperties <- FALSE
  }
  
  if (!is.null(node$properties)) {
    node$properties <- lapply(node$properties, add_no_extra_fields)
  }

  if (!is.null(node$items)) {
    node$items <- add_no_extra_fields(node$items)
  }

  node
}

#' The Perplexity API provider class (inherits from OpenAI)
#'
#' @noRd
api_perplexity <- new_class("Perplexity", api_openai)


#' A function to get metadata from Perplexity responses
#'
#' @noRd
method(extract_metadata, list(api_perplexity,class_list))<- function(.api,.response) {
  list(
    model             = .response$model,
    timestamp         = lubridate::as_datetime(.response$created),
    prompt_tokens     = .response$usage$prompt_tokens,
    completion_tokens = .response$usage$completion_tokens,
    total_tokens      = .response$usage$total_tokens,
    stream            = FALSE,
    specific_metadata = list(
      id        = .response$id,
      citations = .response$citations
    ) 
  )
}  

#' A function to get metadata from Openai streaming responses
#'
#' @noRd
method(extract_metadata_stream, list(api_perplexity,class_list))<- function(.api,.stream_raw_data) {
  final_stream_chunk <- .stream_raw_data[[length(.stream_raw_data)]]
  list(
    model             = final_stream_chunk$model,
    timestamp         = lubridate::as_datetime(final_stream_chunk$created),
    prompt_tokens     = final_stream_chunk$usage$prompt_tokens,
    completion_tokens = final_stream_chunk$usage$completion_tokens,
    total_tokens      = final_stream_chunk$usage$total_tokens,
    stream            = TRUE,
    specific_metadata = list(
      id        = final_stream_chunk$id,
      citations = final_stream_chunk$citations
    ) 
  )
}  


#' A method to handle streaming requests 
#' Perplixity is different than vanilla openai here. 
#'
#' @noRd
method(handle_stream,list(api_perplexity,new_S3_class("httr2_response"))) <- function(.api,.stream_response) {
  stream_text <- ""
  stream_data <- list()
  repeat {
    stream_chunk <- httr2::resp_stream_sse(.stream_response)
    # Skip empty chunks
    if (is.null(stream_chunk$data) || !nzchar(stream_chunk$data)) {
      next  
    }
    

    # Try to parse the JSON content
    parsed_event <- tryCatch(
      jsonlite::fromJSON(stream_chunk$data, simplifyVector = FALSE, simplifyDataFrame = FALSE),
      error = function(e) {
        message("Failed to parse JSON: ", e$message)
        return(NULL)
      }
    )
    
    if (!is.null(parsed_event)) {
      if(length(parsed_event$choices) >= 1) {
        stream_data <- append(stream_data,list(parsed_event))
        if(!is.null(parsed_event$choices[[1]]$finish_reason)){
          finished <- parsed_event$choices[[1]]$finish_reason
          if (finished == "stop") {
            close(.stream_response)
            message("\n---------\nStream finished\n---------\n")
            break
          }
          
        }
        
        delta_content <- parsed_event$choices[[1]]$delta$content
        if (!is.null(delta_content)) {
          stream_text <- paste0(stream_text, delta_content)
          cat(delta_content)
          utils::flush.console()
        }
      }
    }
  }
  
  list(
    reply = stream_text,
    raw_data = stream_data
  )
}


#' Send LLM Messages to the Perplexity Chat API
#'
#' @description
#' Sends a chat message history to the Perplexity Chat API, supporting all documented API parameters.
#'
#' @param .llm An `LLMMessage` object containing the conversation history.
#' @param .model Model name to use (default: "sonar").
#' @param .max_tokens Max completion tokens (default: 1024).
#' @param .temperature Controls response randomness (0 < x < 2).
#' @param .top_p Nucleus sampling threshold (0 < x < 1).
#' @param .frequency_penalty Number > 0. Penalizes frequent tokens.
#' @param .presence_penalty Numeric between -2 and 2. Penalizes present tokens.
#' @param .stop Stop sequence(s), string or character vector/list.
#' @param .search_domain_filter Character vector of domains to allowlist/denylist (max 10; prefix with "-" to denylist).
#' @param .search_language_filter ISO 639-1 language code to restrict search results (e.g. "en", "de", "fr").
#' @param .language_preference ISO 639-1 code for preferred response language (e.g. "en").
#' @param .return_images Logical; if TRUE, returns images from search.
#' @param .image_domain_filter Character vector of domains to restrict image results to.
#' @param .image_format_filter Character vector of image formats to include (e.g. c("png", "jpg")).
#' @param .search_recency_filter Restrict search to recent results: "hour", "day", "week", "month", or "year".
#' @param .search_mode Search index to use: "web" (default), "academic", or "sec".
#' @param .search_after_date_filter Only include content published after this date (MM/DD/YYYY).
#' @param .search_before_date_filter Only include content published before this date (MM/DD/YYYY).
#' @param .last_updated_after_filter Only include content last updated after this date (MM/DD/YYYY).
#' @param .last_updated_before_filter Only include content last updated before this date (MM/DD/YYYY).
#' @param .disable_search Logical; if TRUE, disables web search entirely (default: FALSE).
#' @param .enable_search_classifier Logical; if TRUE, lets the model decide whether to search (default: FALSE).
#' @param .reasoning_effort Reasoning level: "low", "medium", or "high".
#' @param .return_related_questions Logical; if TRUE, returns related questions.
#' @param .user_location Named list for geographic search personalisation. Accepted fields:
#'   \code{country} (ISO 3166-1 alpha-2), \code{city}, \code{region}, \code{latitude}, \code{longitude}.
#'   Example: \code{list(country = "DE", city = "Berlin")}.
#' @param .search_context_size Amount of search context to include: "low", "medium" (default), or "high".
#' @param .search_type Search quality preference inside \code{web_search_options}: "fast", "pro", or "auto".
#' @param .stream_mode Response format: "full" (default) or "concise".
#' @param .json_schema A tidyllm schema created with `tidyllm_schema()` for structured JSON output (optional).
#' @param .top_k Top-k token sampling (integer, 0 disables).
#' @param .web_search_options Named list with raw web_search_options overrides. Values set here take
#'   precedence over the dedicated parameters above.
#' @param .api_url API endpoint (default: "https://api.perplexity.ai/").
#' @param .timeout Timeout in seconds (default: 60).
#' @param .stream If TRUE, streams response.
#' @param .verbose If TRUE, prints additional info.
#' @param .max_tries Max request retries (default: 3).
#' @param .dry_run If TRUE, returns constructed request instead of sending.
#'
#' @return An updated `LLMMessage` object with the assistant's reply and metadata, including citations and search_results.
#' @export
perplexity_chat <- function(
    .llm,
    .model = "sonar",
    .max_tokens = 1024,
    .temperature = NULL,
    .top_p = NULL,
    .frequency_penalty = NULL,
    .presence_penalty = NULL,
    .stop = NULL,
    .search_domain_filter = NULL,
    .search_language_filter = NULL,
    .language_preference = NULL,
    .return_images = FALSE,
    .image_domain_filter = NULL,
    .image_format_filter = NULL,
    .search_recency_filter = NULL,
    .search_mode = "web",
    .search_after_date_filter = NULL,
    .search_before_date_filter = NULL,
    .last_updated_after_filter = NULL,
    .last_updated_before_filter = NULL,
    .disable_search = FALSE,
    .enable_search_classifier = FALSE,
    .reasoning_effort = NULL,
    .return_related_questions = FALSE,
    .user_location = NULL,
    .search_context_size = NULL,
    .search_type = NULL,
    .stream_mode = NULL,
    .json_schema = NULL,
    .top_k = NULL,
    .web_search_options = NULL,
    .api_url = "https://api.perplexity.ai/",
    .timeout = 60,
    .stream = FALSE,
    .verbose = FALSE,
    .max_tries = 3,
    .dry_run = FALSE
) {
  c(
    "Input .llm must be an LLMMessage object" = S7_inherits(.llm, LLMMessage),
    "Input .max_tokens must be a positive integer" = is_integer_valued(.max_tokens) & .max_tokens > 0,
    "Input .model must be a non-empty string" = is.character(.model) & nzchar(.model),
    "Input .api_url must be a non-empty string" = is.character(.api_url) & nzchar(.api_url),
    "Input .timeout must be positive integer" = is_integer_valued(.timeout) & .timeout > 0,
    "Input .temperature must be NULL or 0 < x < 2" = is.null(.temperature) | (.temperature > 0 & .temperature < 2),
    "Input .top_p must be NULL or 0 < x < 1" = is.null(.top_p) | (.top_p > 0 & .top_p < 1),
    "Input .frequency_penalty must be NULL or > 0" = is.null(.frequency_penalty) | (.frequency_penalty > 0),
    "Input .presence_penalty must be NULL or -2 <= x <= 2" = is.null(.presence_penalty) | (.presence_penalty >= -2 & .presence_penalty <= 2),
    "Input .stop must be NULL, character, or list" = is.null(.stop) | is.character(.stop) | is.list(.stop),
    "Input .search_domain_filter must be NULL or character vector" = is.null(.search_domain_filter) | is.character(.search_domain_filter),
    "If .search_domain_filter provided, must be <= 10 domains" = is.null(.search_domain_filter) | (length(.search_domain_filter) <= 10),
    "Input .search_language_filter must be NULL or a single string" = is.null(.search_language_filter) | (is.character(.search_language_filter) & length(.search_language_filter) == 1),
    "Input .language_preference must be NULL or a single string" = is.null(.language_preference) | (is.character(.language_preference) & length(.language_preference) == 1),
    "Input .return_images must be logical" = is.logical(.return_images),
    "Input .image_domain_filter must be NULL or character vector" = is.null(.image_domain_filter) | is.character(.image_domain_filter),
    "Input .image_format_filter must be NULL or character vector" = is.null(.image_format_filter) | is.character(.image_format_filter),
    "Input .return_related_questions must be logical" = is.logical(.return_related_questions),
    "Input .search_recency_filter must be NULL or one of 'hour','day','week','month','year'" = is.null(.search_recency_filter) | .search_recency_filter %in% c("hour", "day", "week", "month", "year"),
    "Input .search_mode must be 'web', 'academic', or 'sec'" = .search_mode %in% c("web", "academic", "sec"),
    "Input .disable_search must be logical" = is.logical(.disable_search),
    "Input .enable_search_classifier must be logical" = is.logical(.enable_search_classifier),
    "Input .reasoning_effort must be NULL or one of 'low', 'medium', 'high'" = is.null(.reasoning_effort) | .reasoning_effort %in% c("low", "medium", "high"),
    "Input .search_after_date_filter must be NULL or character" = is.null(.search_after_date_filter) | is.character(.search_after_date_filter),
    "Input .search_before_date_filter must be NULL or character" = is.null(.search_before_date_filter) | is.character(.search_before_date_filter),
    "Input .last_updated_after_filter must be NULL or character" = is.null(.last_updated_after_filter) | is.character(.last_updated_after_filter),
    "Input .last_updated_before_filter must be NULL or character" = is.null(.last_updated_before_filter) | is.character(.last_updated_before_filter),
    "Input .user_location must be NULL or a named list" = is.null(.user_location) | (is.list(.user_location) & !is.null(names(.user_location))),
    "Input .search_context_size must be NULL or one of 'low', 'medium', 'high'" = is.null(.search_context_size) | .search_context_size %in% c("low", "medium", "high"),
    "Input .search_type must be NULL or one of 'fast', 'pro', 'auto'" = is.null(.search_type) | .search_type %in% c("fast", "pro", "auto"),
    "Input .stream_mode must be NULL or one of 'full', 'concise'" = is.null(.stream_mode) | .stream_mode %in% c("full", "concise"),
    "Input .top_k must be NULL or a single non-negative integer" = is.null(.top_k) | (is.numeric(.top_k) & length(.top_k) == 1 & .top_k >= 0),
    "Input .web_search_options must be NULL or named list" = is.null(.web_search_options) | (is.list(.web_search_options) & !is.null(names(.web_search_options))),
    "Input .json_schema must be NULL or a list" = is.null(.json_schema) | is.list(.json_schema),
    "Input .verbose must be logical" = is.logical(.verbose),
    "Input .max_tries must be positive integer" = is_integer_valued(.max_tries) & .max_tries > 0,
    "Input .dry_run must be logical" = is.logical(.dry_run)
  ) |>
    validate_inputs()

  api_obj <- api_perplexity(short_name = "perplexity",
                            long_name  = "Perplexity",
                            api_key_env_var = "PERPLEXITY_API_KEY")

  messages <- to_api_format(.llm, api_obj, TRUE)
  api_key <- get_api_key(api_obj, .dry_run)

  search_domain_filter  <- if (!is.null(.search_domain_filter))  as.list(.search_domain_filter)  else NULL
  image_domain_filter   <- if (!is.null(.image_domain_filter))   as.list(.image_domain_filter)   else NULL
  image_format_filter   <- if (!is.null(.image_format_filter))   as.list(.image_format_filter)   else NULL

  web_search_options <- c(
    list(search_context_size = .search_context_size,
         search_type         = .search_type,
         user_location       = .user_location),
    .web_search_options
  ) |> purrr::compact()
  web_search_options <- if (length(web_search_options) == 0) NULL else web_search_options

  json <- FALSE
  response_format <- NULL
  if (!is.null(.json_schema)) {
    json <- TRUE
    schema_name <- attr(.json_schema, "name") %||% "schema"
    if (is.null(.json_schema$additionalProperties)) {
      .json_schema$additionalProperties <- FALSE
    }
    response_format <- list(
      type = "json_schema",
      json_schema = list(
        schema = .json_schema,
        name = schema_name,
        strict = TRUE
      )
    )
  }

  request_body <- list(
    model = .model,
    messages = messages,
    max_tokens = .max_tokens,
    temperature = .temperature,
    top_p = .top_p,
    frequency_penalty = .frequency_penalty,
    presence_penalty = .presence_penalty,
    stop = .stop,
    search_domain_filter = search_domain_filter,
    search_language_filter = .search_language_filter,
    language_preference = .language_preference,
    return_images = .return_images,
    image_domain_filter = image_domain_filter,
    image_format_filter = image_format_filter,
    search_recency_filter = .search_recency_filter,
    search_mode = .search_mode,
    search_after_date_filter = .search_after_date_filter,
    search_before_date_filter = .search_before_date_filter,
    last_updated_after_filter = .last_updated_after_filter,
    last_updated_before_filter = .last_updated_before_filter,
    disable_search = if (.disable_search) TRUE else NULL,
    enable_search_classifier = if (.enable_search_classifier) TRUE else NULL,
    reasoning_effort = .reasoning_effort,
    return_related_questions = .return_related_questions,
    stream_mode = .stream_mode,
    top_k = .top_k,
    web_search_options = web_search_options,
    stream = .stream
  ) |> purrr::compact()

  if (!is.null(response_format)) {
    request_body$response_format <- response_format
  }

  request <- httr2::request(.api_url) |>
    httr2::req_url_path("/chat/completions") |>
    httr2::req_headers(
      `Authorization` = sprintf("Bearer %s", api_key),
      `Content-Type` = "application/json"
    ) |>
    httr2::req_body_json(data = request_body)
  
  # Dry run returns request object for inspection
  if (.dry_run) return(request)
  
  # --- Perform Request (may stream) ---
  response <- perform_chat_request(request, api_obj, .stream, .timeout, .max_tries)
  assistant_reply <- response$assistant_reply
  
  # --- Extract metadata including search_results if present ---
  meta <- response$meta
  if (!is.null(response$search_results)) {
    meta$specific_metadata$search_results <- response$search_results
  }
  
  add_message(.llm     = .llm,
              .role    = "assistant",
              .content = assistant_reply,
              .meta    = meta,
              .json    = json)
}


#' Submit a Deep Research Request to Perplexity
#'
#' @param .llm An `LLMMessage` object containing the research question.
#' @param .background Logical; if TRUE, returns a `tidyllm_research_job` immediately without waiting (default: FALSE).
#' @param .reasoning_effort Reasoning level: "low", "medium" (default), or "high".
#' @param .search_context_size Amount of search context: "low", "medium" (default), or "high".
#' @param .search_domain_filter Character vector of domains to allowlist/denylist (max 10; prefix with "-" to denylist).
#' @param .search_language_filter ISO 639-1 language code to restrict search results (e.g. "en", "de").
#' @param .language_preference ISO 639-1 code for preferred response language.
#' @param .search_recency_filter Restrict search to recent results: "hour", "day", "week", "month", or "year".
#' @param .search_mode Search index to use: "web" (default), "academic", or "sec".
#' @param .search_after_date_filter Only include content published after this date (MM/DD/YYYY).
#' @param .search_before_date_filter Only include content published before this date (MM/DD/YYYY).
#' @param .last_updated_after_filter Only include content last updated after this date (MM/DD/YYYY).
#' @param .last_updated_before_filter Only include content last updated before this date (MM/DD/YYYY).
#' @param .user_location Named list for geographic search personalisation (fields: country, city, region, latitude, longitude).
#' @param .json_schema A tidyllm schema created with `tidyllm_schema()` for structured JSON output (optional).
#' @param .idempotency_key Optional string; unique key to prevent duplicate submissions.
#' @param .api_key Character; Perplexity API key (default: from environment variable).
#' @param .timeout Integer; request timeout in seconds for blocking polling (default: 300).
#' @param .max_tries Integer; maximum retries (default: 3).
#'
#' @return If `.background = FALSE`, an updated `LLMMessage` with the research reply.
#'   If `.background = TRUE`, a `tidyllm_research_job` object.
#' @export
perplexity_deep_research <- function(.llm,
                                     .background = FALSE,
                                     .reasoning_effort = "medium",
                                     .search_context_size = "medium",
                                     .search_domain_filter = NULL,
                                     .search_language_filter = NULL,
                                     .language_preference = NULL,
                                     .search_recency_filter = NULL,
                                     .search_mode = NULL,
                                     .search_after_date_filter = NULL,
                                     .search_before_date_filter = NULL,
                                     .last_updated_after_filter = NULL,
                                     .last_updated_before_filter = NULL,
                                     .user_location = NULL,
                                     .json_schema = NULL,
                                     .idempotency_key = NULL,
                                     .api_key = Sys.getenv("PERPLEXITY_API_KEY"),
                                     .timeout = 300,
                                     .max_tries = 3) {
  c(
    "Input .llm must be an LLMMessage object" = S7_inherits(.llm, LLMMessage),
    "Input .background must be logical" = is.logical(.background),
    "Input .reasoning_effort must be 'low', 'medium', or 'high'" = .reasoning_effort %in% c("low", "medium", "high"),
    "Input .search_context_size must be 'low', 'medium', or 'high'" = .search_context_size %in% c("low", "medium", "high"),
    "Input .search_domain_filter must be NULL or character vector" = is.null(.search_domain_filter) | is.character(.search_domain_filter),
    "If .search_domain_filter provided, must be <= 10 domains" = is.null(.search_domain_filter) | (length(.search_domain_filter) <= 10),
    "Input .search_language_filter must be NULL or a single string" = is.null(.search_language_filter) | (is.character(.search_language_filter) & length(.search_language_filter) == 1),
    "Input .language_preference must be NULL or a single string" = is.null(.language_preference) | (is.character(.language_preference) & length(.language_preference) == 1),
    "Input .search_recency_filter must be NULL or one of 'hour','day','week','month','year'" = is.null(.search_recency_filter) | .search_recency_filter %in% c("hour", "day", "week", "month", "year"),
    "Input .search_mode must be NULL or one of 'web', 'academic', 'sec'" = is.null(.search_mode) | .search_mode %in% c("web", "academic", "sec"),
    "Input .search_after_date_filter must be NULL or character" = is.null(.search_after_date_filter) | is.character(.search_after_date_filter),
    "Input .search_before_date_filter must be NULL or character" = is.null(.search_before_date_filter) | is.character(.search_before_date_filter),
    "Input .last_updated_after_filter must be NULL or character" = is.null(.last_updated_after_filter) | is.character(.last_updated_after_filter),
    "Input .last_updated_before_filter must be NULL or character" = is.null(.last_updated_before_filter) | is.character(.last_updated_before_filter),
    "Input .user_location must be NULL or a named list" = is.null(.user_location) | (is.list(.user_location) & !is.null(names(.user_location))),
    "Input .json_schema must be NULL or a list" = is.null(.json_schema) | is.list(.json_schema),
    "Input .idempotency_key must be NULL or a single string" = is.null(.idempotency_key) | (is.character(.idempotency_key) & length(.idempotency_key) == 1),
    "Input .api_key must be non-empty" = nzchar(.api_key),
    "Input .timeout must be a positive integer" = is_integer_valued(.timeout) && .timeout > 0,
    "Input .max_tries must be a positive integer" = is_integer_valued(.max_tries) && .max_tries > 0
  ) |> validate_inputs()

  api_obj <- api_perplexity(short_name = "perplexity",
                            long_name  = "Perplexity",
                            api_key_env_var = "PERPLEXITY_API_KEY")

  messages <- to_api_format(.llm, api_obj, TRUE)

  search_domain_filter <- if (!is.null(.search_domain_filter)) as.list(.search_domain_filter) else NULL

  web_search_options <- list(
    search_context_size = .search_context_size,
    user_location       = .user_location
  ) |> purrr::compact()

  json <- FALSE
  response_format <- NULL
  if (!is.null(.json_schema)) {
    json <- TRUE
    schema_name <- attr(.json_schema, "name") %||% "schema"
    if (is.null(.json_schema$additionalProperties)) {
      .json_schema$additionalProperties <- FALSE
    }
    response_format <- list(
      type = "json_schema",
      json_schema = list(
        schema = .json_schema,
        name = schema_name,
        strict = TRUE
      )
    )
  }

  inner_request <- list(
    model = "sonar-deep-research",
    messages = messages,
    reasoning_effort = .reasoning_effort,
    search_domain_filter = search_domain_filter,
    search_language_filter = .search_language_filter,
    language_preference = .language_preference,
    search_recency_filter = .search_recency_filter,
    search_mode = .search_mode,
    search_after_date_filter = .search_after_date_filter,
    search_before_date_filter = .search_before_date_filter,
    last_updated_after_filter = .last_updated_after_filter,
    last_updated_before_filter = .last_updated_before_filter,
    web_search_options = web_search_options,
    response_format = response_format
  ) |> purrr::compact()

  request_body <- list(
    request = inner_request,
    idempotency_key = .idempotency_key
  ) |> purrr::compact()

  response <- httr2::request("https://api.perplexity.ai") |>
    httr2::req_url_path("/async/chat/completions") |>
    httr2::req_headers(
      `Authorization` = sprintf("Bearer %s", .api_key),
      `Content-Type` = "application/json"
    ) |>
    httr2::req_body_json(data = request_body) |>
    httr2::req_retry(max_tries = .max_tries) |>
    httr2::req_perform() |>
    httr2::resp_body_json()

  job_id <- response$id
  job <- structure(list(job_id = job_id, message = .llm, json = json), class = "tidyllm_research_job")

  if (.background) return(job)

  perplexity_fetch_research(perplexity_poll_research(job, .api_key, .timeout, .max_tries))
}


#' Poll a Perplexity Deep Research Job Until Completion
#'
#' @param .job A `tidyllm_research_job` object.
#' @param .api_key Character; Perplexity API key.
#' @param .timeout Integer; total seconds to wait before giving up (default: 300).
#' @param .max_tries Integer; maximum retries per poll (default: 3).
#' @return An updated `tidyllm_research_job` with status information.
#' @noRd
perplexity_poll_research <- function(.job, .api_key = Sys.getenv("PERPLEXITY_API_KEY"),
                                     .timeout = 300, .max_tries = 3) {
  start_time <- proc.time()[["elapsed"]]
  repeat {
    status_response <- httr2::request("https://api.perplexity.ai") |>
      httr2::req_url_path(sprintf("/async/chat/completions/%s", .job$job_id)) |>
      httr2::req_headers(`Authorization` = sprintf("Bearer %s", .api_key)) |>
      httr2::req_retry(max_tries = .max_tries) |>
      httr2::req_perform() |>
      httr2::resp_body_json()

    if (identical(status_response$status, "COMPLETED")) {
      .job$response <- status_response
      return(.job)
    }

    elapsed <- proc.time()[["elapsed"]] - start_time
    if (elapsed > .timeout) {
      stop(sprintf("Perplexity deep research job '%s' did not complete within %d seconds.", .job$job_id, .timeout))
    }
    Sys.sleep(5)
  }
}


#' Check the Status of a Perplexity Deep Research Job
#'
#' @param .job A `tidyllm_research_job` object returned by `perplexity_deep_research(.background = TRUE)`.
#' @param .api_key Character; Perplexity API key (default: from environment).
#' @param .max_tries Integer; maximum retries (default: 3).
#' @return An updated `tidyllm_research_job` with a `$status` field and `$response` if completed.
#' @export
perplexity_check_research <- function(.job,
                                      .api_key = Sys.getenv("PERPLEXITY_API_KEY"),
                                      .max_tries = 3) {
  c(
    "Input .job must be a tidyllm_research_job" = inherits(.job, "tidyllm_research_job"),
    "Input .api_key must be non-empty" = nzchar(.api_key)
  ) |> validate_inputs()

  status_response <- httr2::request("https://api.perplexity.ai") |>
    httr2::req_url_path(sprintf("/async/chat/completions/%s", .job$job_id)) |>
    httr2::req_headers(`Authorization` = sprintf("Bearer %s", .api_key)) |>
    httr2::req_retry(max_tries = .max_tries) |>
    httr2::req_perform() |>
    httr2::resp_body_json()

  .job$status <- status_response$status
  if (identical(status_response$status, "COMPLETED")) {
    .job$response <- status_response
  }
  .job
}


#' Fetch Results from a Completed Perplexity Deep Research Job
#'
#' @param .job A `tidyllm_research_job` object. Must have status "completed" or will poll once.
#' @param .api_key Character; Perplexity API key (default: from environment).
#' @param .max_tries Integer; maximum retries (default: 3).
#' @return An updated `LLMMessage` with the research reply appended.
#' @export
perplexity_fetch_research <- function(.job,
                                      .api_key = Sys.getenv("PERPLEXITY_API_KEY"),
                                      .max_tries = 3) {
  c(
    "Input .job must be a tidyllm_research_job" = inherits(.job, "tidyllm_research_job"),
    "Input .api_key must be non-empty" = nzchar(.api_key)
  ) |> validate_inputs()

  if (is.null(.job$response)) {
    .job <- perplexity_check_research(.job, .api_key, .max_tries)
    if (!identical(.job$status, "COMPLETED")) {
      stop(sprintf("Perplexity deep research job '%s' is not yet completed (status: %s).", .job$job_id, .job$status))
    }
  }

  api_obj <- api_perplexity(short_name = "perplexity",
                            long_name  = "Perplexity",
                            api_key_env_var = "PERPLEXITY_API_KEY")

  response_content <- .job$response$response
  assistant_reply <- parse_chat_response(api_obj, response_content)
  meta <- extract_metadata(api_obj, response_content)

  add_message(.llm     = .job$message,
              .role    = "assistant",
              .content = assistant_reply,
              .meta    = meta,
              .json    = isTRUE(.job$json))
}


#' Perplexity Provider Function
#'
#' The `perplexity()` function acts as a provider interface for interacting with the Perplexity API 
#' through `tidyllm`'s `chat()` verb.
#' It dynamically routes requests to Perplxeity-specific function. At the moment this is only
#' `perplexity_chat()`
#'
#' @param ... Parameters to be passed to the appropriate Perplexity-specific function, 
#'   such as model configuration, input text, or API-specific options.
#' @param .called_from An internal argument specifying which action (e.g., 
#'   `chat`, `embed`) the function is invoked from. 
#'   This argument is automatically managed by the `tidyllm` verbs and should not be modified by the user.
#'
#' @return The result of the requested action, depending on the specific function invoked 
#'   (e.g., an updated `LLMMessage` object for `chat()`).
#'
#' @export
perplexity <- create_provider_function(
  .name = "perplexity",
  chat = perplexity_chat,
  deep_research = perplexity_deep_research
)

#' Perform a generic request with retry and backoff
#'
#' @param .request The httr2 request object.
#' @param .timeout Request timeout in seconds.
#' @param .max_tries Maximum retry attempts for requests (default: 3).
#'
#' @return A list containing the assistant's reply and response headers.
#' @noRd
perform_generic_request <- function(.request, 
                                    .timeout = 60, 
                                    .max_tries = 3) {
  response <- .request |>
    httr2::req_timeout(.timeout) |>
    httr2::req_error(is_error = function(resp) FALSE) |>
    httr2::req_retry(
      max_tries = .max_tries,
      retry_on_failure = TRUE,
      is_transient = function(resp) httr2::resp_status(resp) %in% c(429, 503)
    ) |>
    httr2::req_perform()
  
  list(
    content = httr2::resp_body_json(response),
    headers = httr2::resp_headers(response),
    status  = httr2::resp_status(response)
  )
}



#' Perform a Chat API request to interact with language models
#'
#' @param .request The httr2 request object.
#' @param .api An api provider object
#' @param .stream Stream the response if TRUE.
#' @param .timeout Request timeout in seconds.
#' @param .max_tries Maximum retry attempts for requests (default: 3).
#'
#' @return A list containing the assistant's reply and response headers.
#' @noRd
perform_chat_request <- function(.request, 
                                .api, 
                                .stream = FALSE, 
                                .timeout = 60, 
                                .max_tries = 3) {

  api_name <- .api@long_name
  if (.stream == TRUE) {
    message("\n---------\nStart ", api_name, " streaming: \n---------\n")
    # Perform the streaming request
    response <- httr2::req_perform_connection(.request, blocking = FALSE)
    stream_response <- handle_stream(.api,response)
    assistant_reply <- stream_response$reply
    response_data   <- stream_response$raw_data 
    metadata <- extract_metadata_stream(.api,stream_response$raw_data)
    # Capture response headers for rate limiting information
    response_headers <- httr2::resp_headers(response)
    
  } else {
    # Non-streaming mode
    response_data <- perform_generic_request(.request, .timeout, .max_tries)
    #parse_chat_response <- parse_chat_function(.api)
    assistant_reply <- parse_chat_response(.api,response_data$content)
    # Capture response headers for rate limiting information
    response_headers <- response_data$headers
    metadata <- extract_metadata(.api,response_data$content)
    #metadata <- extract_response_metadata(response_data$content)
  }
  

  list(assistant_reply  = assistant_reply, 
       headers          = response_headers,
       meta             = metadata,
       raw              = response_data)
}


#' Perform Embedding Request
#'
#' A utility function to perform embedding requests.
#'
#' @param .request The prepared httr2 request object.
#' @param .timeout Timeout for the API request in seconds.
#' @param .max_tries Maximum number of retry attempts.
#' @param .fn_extract_embeddings A functiont to extract embeddings
#' @return An embedding response tibble
#' @keywords internal
#' @noRd
perform_embedding_request <- function(.request, .timeout, .max_tries, .input_texts, .fn_extract_embeddings) {
  response_data <- perform_generic_request(.request, .timeout, .max_tries)
  
  embeddings <- .fn_extract_embeddings(
    response_data$content,
    response_data$status >= 400,  # Check if the status indicates an error
    response_data$headers
  )
  
  if (is.null(embeddings) || length(embeddings) == 0) {
    stop("No embeddings returned in the response.")
  }
  
  tibble::tibble(
    input = .input_texts,
    embeddings = embeddings
  )
}

#' Initialize or Retrieve API-specific Environment
#'
#' This function initializes a named environment for storing rate limit information
#' specific to an API. It ensures that each API's rate limit data is stored separately.
#'
#' @param .api_name The name of the API for which to initialize or retrieve the environment
#' @noRd
initialize_api_env <- function(.api_name) {
  if (!exists(.api_name, envir = .tidyllm_rate_limit_env)) {
    .tidyllm_rate_limit_env[[.api_name]] <- new.env()
  }
  invisible(NULL)
}


#' Track the standard API rate limit info in the hidden .tidyllm_rate_limit_env environment
#'
#' @param .api_name The name of the API for which to initialize or retrieve the environment.
#' @param .headers The response headers from an API-answer.
#' @param .verbose Logical indicating whether to print rate limit details.
#' 
#' @noRd
track_rate_limit <- function(.api,.headers,.verbose) {
  api            <- .api@short_name
  api_long_name  <- .api@long_name
  tryCatch({
    rl <- ratelimit_from_header(.api, .headers)
    
    initialize_api_env(api)
    .tidyllm_rate_limit_env[[api]]$last_request        <- rl$this_request_time
    .tidyllm_rate_limit_env[[api]]$requests_remaining  <- rl$ratelimit_requests_remaining
    .tidyllm_rate_limit_env[[api]]$requests_reset_time <- rl$ratelimit_requests_reset_time
    .tidyllm_rate_limit_env[[api]]$tokens_remaining    <- rl$ratelimit_tokens_remaining
    .tidyllm_rate_limit_env[[api]]$tokens_reset_time   <- rl$ratelimit_tokens_reset_time
    
    # Show rate limit info if verbose
    if (.verbose) {
      glue::glue(
        "{api_long_name} answer received at {rl$this_request_time}.
        Remaining requests rate limit: {rl$ratelimit_requests_remaining}/{rl$ratelimit_requests}
        Requests rate limit reset at: {rl$ratelimit_requests_reset_time}
        Remaining tokens rate limit: {rl$ratelimit_tokens_remaining}/{rl$ratelimit_tokens}
        Tokens rate limit reset at: {rl$ratelimit_tokens_reset_time}\n"
      ) |> message()
    }
    
  }, error = function(e) {
    warning(glue::glue("Failed to track rate limit for {api_long_name}: {e$message}"))
  })
  
  invisible(NULL)
}

#'
#'This internal function parses duration strings as returned by the OpenAI API
#'
#' @param .duration_str A duration string. 
#' @return A numeric number of seconds 
#' @noRd
parse_duration_to_seconds <- function(.duration_str) {
  if(grepl("ms", .duration_str)) {
    # Format is in milliseconds, e.g., "176ms"
    duration_sec <- as.numeric(sub("ms", "", .duration_str)) / 1000
  } else if(grepl("m", .duration_str) && grepl("s", .duration_str)) {
    # Format is in minutes and seconds, e.g., "20m47.458s"
    minutes <- as.numeric(sub("m.*", "", .duration_str))
    seconds <- as.numeric(sub(".*m", "", sub("s", "", .duration_str)))
    duration_sec <- minutes * 60 + seconds
  } else if(grepl("s", .duration_str)) {
    # Format is just in seconds, e.g., "47s"
    duration_sec <- as.numeric(sub("s", "", .duration_str))
  } else {
    # Default to zero if format is unrecognized
    duration_sec <- 0
  }
  return(duration_sec)
}


#' Get the current rate limit information for all or a specific API
#'
#' This function retrieves the rate limit details for the specified API,
#' or for all APIs stored in the .tidyllm_rate_limit_env if no API is specified.
#'
#' @param .api_name (Optional) The name of the API whose rate limit info you want to get
#'                   If not provided, the rate limit info for all APIs in the environment will be returned
#' @return A tibble containing the rate limit information.
#' @export
rate_limit_info <- function(.api_name = NULL) {
  # If no API name is provided, print for all APIs in .tidyllm_rate_limit_env
  api_names <- ls(envir = .tidyllm_rate_limit_env)
  
  if (length(api_names) == 0) {
    message("No rate limit information available in .tidyllm_rate_limit_env.")
    return(tibble::tibble())
  }
  
  if (!is.null(.api_name)) {
    # If a specific API is provided, check if it exists
    if (!exists(.api_name, envir = .tidyllm_rate_limit_env)) {
      stop(glue::glue("No rate limit information found for the API: {.api_name}."))
    }
    api_names <- .api_name  # Restrict to the provided API
  }
  
  # Use lapply to iterate over all relevant APIs and print their rate limits
  rl_tibble <- purrr::map_dfr(api_names, function(api) {
    rl_info <- .tidyllm_rate_limit_env[[api]]
    
    # Check for incomplete rate limit data
    if (is.null(rl_info$last_request) ||
        is.null(rl_info$requests_remaining) ||
        is.null(rl_info$tokens_remaining)) {
      warning(glue::glue("Incomplete rate limit data for API: {api}."))
    }
    
    
    tibble::tibble(api =api, 
                   last_request = rl_info$last_request,
                   requests_remaining  = rl_info$requests_remaining,
                   requests_reset_time = rl_info$requests_reset_time,
                   tokens_remaining    = rl_info$tokens_remaining,
                   tokens_reset_time   = rl_info$tokens_reset_time)
  })
  rl_tibble
}

testthat::skip_if_not_installed("httptest2")
library(httptest2)

test_that("groq function constructs a correct request and dry runs it", {
  # Make sure the environment starts clean
  if (exists("groq", envir = .tidyllm_rate_limit_env)) {
    .tidyllm_rate_limit_env[["groq"]] <- NULL
  }
  
  # Call groq with .dry_run = TRUE and perform the dry run
  request <- llm_message("Write a poem about meerkats") |> chat(groq,.dry_run = TRUE)
  
  dry_run <- request |>
    httr2::req_dry_run(redact_headers = TRUE, quiet = TRUE)
  
  # Check that the method is POST
  expect_equal(dry_run$method, "POST")
  
  # Check that the path is correct
  expect_equal(dry_run$path, "/openai/v1/chat/completions")
  
  # Inspect headers
  headers <- dry_run$headers
  expect_type(headers, "list")
  
  # Check that the required headers are present
  expect_true("authorization" %in% names(headers))
  expect_true("content-type" %in% names(headers))
  
  # Check that the content-type is JSON
  expect_equal(headers$`content-type`, "application/json")
  
  # Now check the body content to ensure the JSON is constructed as expected
  body_json <- request$body |> jsonlite::toJSON() |> as.character()
  
  expected_json <- "{\"data\":{\"model\":[\"moonshotai/kimi-k2-instruct-0905\"],\"max_tokens\":[1024],\"messages\":[{\"role\":[\"user\"],\"content\":[\"Write a poem about meerkats \"]}],\"stream\":[false]},\"type\":[\"json\"],\"content_type\":[\"application/json\"],\"params\":{\"auto_unbox\":[true],\"digits\":[22],\"null\":[\"null\"]}}"
  # Check if the JSON matches the expected JSON
  expect_equal(body_json, expected_json)
})

test_that("groq returns expected response", {
  with_mock_dir("groq",expr = {
    
    # Make sure the environment starts clean
    if (exists("groq", envir = .tidyllm_rate_limit_env)) {
      .tidyllm_rate_limit_env[["groq"]] <- NULL
    }
    
    llm <- llm_message("Hello, world")
    
    # Set a dummy key if none exists
    if (Sys.getenv("GROQ_API_KEY") == "") {
      Sys.setenv(GROQ_API_KEY = "DUMMY_KEY_FOR_TESTING")
    }
    
    result <- groq_chat(
      .llm = llm,
      .max_tokens = 1024,
      .temperature = 0,
    )
    
    # Set a dummy key if none exists
    if (Sys.getenv("GROQ_API_KEY") == "DUMMY_KEY_FOR_TESTING") {
      Sys.setenv(GROQ_API_KEY = "")
    }
    
    # Assertions based on the message in the captured mock response
    expect_true(S7_inherits(result, LLMMessage))
    expect_equal(result@message_history[[length(result@message_history)]]$role, "assistant")
    
    # Now, check that the rate limit environment has been populated with correct values
    expect_true(exists("groq", envir = .tidyllm_rate_limit_env))
    
    # Retrieve rate limit info for chatgpt
    rl_info <- rate_limit_info("groq")
    
    # Assertions for rate limit values based on the mocked response
    expect_equal(rl_info$requests_remaining, 499999.0)
    expect_equal(rl_info$tokens_remaining, 249969.0)
  },simplify = FALSE)
})

# Skip file if ellmer is not installed
if (!requireNamespace("ellmer", quietly = TRUE)) return(invisible(NULL))

api_ellmer <- new_class("Ellmer", APIProvider)

method(to_api_format, list(LLMMessage, api_ellmer)) <- function(.llm, 
                                                                .api) {
  filtered_history <- filter_roles(.llm@message_history, c("user", "assistant", "system"))
  
  lapply(filtered_history, function(m) {
    formatted_message <- format_message(m)
    
    if (m$role == "system") {
      NULL
    } else if (m$role == "user") {
      contents <- list(ellmer::ContentText(formatted_message$content))
      
      if (!is.null(formatted_message$image)) {
        image_content <- ellmer::ContentImageInline(
          type = formatted_message$image$media_type,
          data = formatted_message$image$data
        )
        contents <- c(contents, list(image_content))
      }
      
      ellmer::Turn(role = "user", contents = contents)
    } else if (m$role == "assistant") {
      ellmer::Turn(role = "assistant", contents = list(ellmer::ContentText(formatted_message$content)))
    }
  }) |> 
    Filter(Negate(is.null), x = _)
}


method(parse_chat_response, list(api_ellmer, class_list)) <- function(.api, .content) {
  if (!is.character(.content) || length(.content) == 0) {
    stop("Received empty response from ellmer chat object")
  }
  .content
}


method(extract_metadata, list(api_ellmer, class_list)) <- function(.api, .response) {
  tokens <- if (!is.null(.response$tokens)) {
    .response$tokens
  } else {
    c(input = NA_integer_, output = NA_integer_, cached_input = NA_integer_)
  }
  
  list(
    model             = .response$model %||% "unknown",
    timestamp         = lubridate::as_datetime(lubridate::now()),
    prompt_tokens     = as.integer(tokens["input"] %||% NA),
    completion_tokens = as.integer(tokens["output"] %||% NA),
    total_tokens      = as.integer((tokens["input"] %||% 0) + (tokens["output"] %||% 0)),
    stream            = FALSE,
    specific_metadata = list(
      cost = .response$cost,
      duration = .response$duration
    ) 
  )
}


#' Send LLM Messages to Ellmer Chat Object
#'
#' @description
#' This function converts an `LLMMessage` to the turns ellmer chat object.
#' The ellmer object is cloned and its turn history is cleared internatlly.
#' This maintains tidyllms stateless style. 
#'
#' @param .llm An `LLMMessage` object containing the conversation history.
#' @param .ellmer_chat An ellmer chat object (e.g., from `ellmer::chat_anthropic()`).
#' @param .stream Logical; if TRUE, streams the response piece by piece (default: FALSE).
#' @param .timeout Request timeout in seconds (default: 60).
#' @param .verbose If TRUE, displays additional information about the chat (default: FALSE).
#' @param .dry_run If TRUE, returns the constructed request object without executing it (default: FALSE).
#' @param .tools Dummy for error message if called from chat with tools (default: NULL).
#' @param .max_tries Maximum retries to perform the request (default: 3).
#'
#' @return A new `LLMMessage` object containing the original messages plus the assistant's response.
#'
#' @export
chat_ellmer <- function(.llm,
                        .ellmer_chat,
                        .stream = FALSE,
                        .timeout = 60,
                        .verbose = FALSE,
                        .dry_run = FALSE,
                        .tools  = NULL,
                        .max_tries = 3) {
  # Validate inputs
  c(
    "Input .llm must be an LLMMessage object" = S7_inherits(.llm, LLMMessage),
    "Input .ellmer_chat must be an ellmer chat object" = inherits(.ellmer_chat, "Chat"),
    "Input .stream must be logical" = is.logical(.stream),
    "Input .timeout must be a positive integer" = is_integer_valued(.timeout) & .timeout > 0,
    "Input .verbose must be logical" = is.logical(.verbose),
    "Input .dry_run must be logical" = is.logical(.dry_run),
    "Tidyllm tools are not supported in the ellmer backend. Set ellmer tools directly in the ellmer chat object instead" = is.null(.tools),
    "Input .max_tries must be integer-valued numeric" = is_integer_valued(.max_tries)
  ) |> validate_inputs()
  
  api_obj <- api_ellmer(short_name = "ellmer",
                        long_name  = "Ellmer")
  
  if (.dry_run) {
    return(list(
      action = "ellmer_chat",
      llm_message = .llm,
      ellmer_object = .ellmer_chat
    ))
  }
  
  turn_data <- to_api_format(.llm, api_obj)
  
  cloned_ellmer <- .ellmer_chat$clone()
  cloned_ellmer[[".__enclos_env__"]][["private"]][[".turns"]] <- turn_data
  
  last_user_turn <- turn_data[[length(turn_data)]]
  last_user_text <- ellmer::contents_text(last_user_turn)
  
  assistant_reply <- cloned_ellmer$chat(last_user_text,echo=FALSE)
  
  last_turn <- cloned_ellmer$last_turn()
  
  assistant_reply
  metadata <- extract_metadata(api_obj, list(
    model = cloned_ellmer$get_model(),
    tokens = last_turn@tokens,
    cost = last_turn@cost,
    duration = last_turn@duration
  ))
  
  add_message(.llm     = .llm,
              .role    = "assistant", 
              .content = assistant_reply, 
              .json    = FALSE,
              .meta    = metadata)
}


#' Alias for the Ellmer Provider Function
#'
#' The `chat_ellmer()` function acts as a provider interface for interacting with
#' ellmer chat objects through `tidyllm`'s  verb interface
#'
#' This function allows you to use any ellmer chat object (e.g., from
#' `ellmer::chat_anthropic()`, `ellmer::chat_openai()`, etc.) as a stateless
#' backend for tidyllm. The ellmer object is cloned for each interaction to
#' maintain tidyllm's stateless approach. 
#'
#' @param ... Additional parameters to pass through (for consistency with other providers).
#' @param .called_from An internal argument specifying which action (e.g.,
#'   `chat`) the function is invoked from.
#'
#' @return A provider function that can be used with `tidyllm::chat()`.
#'
#' @export
ellmer <- create_provider_function(
  .name = "ellmer",
  chat = chat_ellmer
)

structure(list(method = "POST", url = "https://api.perplexity.ai/chat/completions", 
    status_code = 200L, headers = structure(list(date = "Thu, 21 Aug 2025 12:41:56 GMT", 
        `content-type` = "application/json", `content-encoding` = "gzip", 
        `cf-cache-status` = "DYNAMIC", `set-cookie` = "REDACTED", 
        `strict-transport-security` = "max-age=15552000; includeSubDomains; preload", 
        server = "cloudflare", `cf-ray` = "972a350c8e486937-FRA"), redact = character(0), class = "httr2_headers"), 
    body = charToRaw("{\"id\": \"d632bdec-4fa8-4674-abfe-e181a736fa62\", \"model\": \"sonar\", \"created\": 1755780116, \"usage\": {\"prompt_tokens\": 8, \"completion_tokens\": 19, \"total_tokens\": 27, \"search_context_size\": \"low\", \"cost\": {\"input_tokens_cost\": 0.0, \"output_tokens_cost\": 0.0, \"request_cost\": 0.005, \"total_cost\": 0.005}}, \"citations\": [\"https://www.m1-project.com/blog/what-is-perplexity-ai-and-how-it-works\", \"https://web.stanford.edu/class/archive/cs/cs224n/cs224n.1224/posters/116768967.pdf\", \"https://www.techtarget.com/searchenterpriseai/definition/Perplexity-AI\", \"https://www.poetrysoup.com/poems/perplexity\", \"https://www.digitalocean.com/resources/articles/what-is-perplexity-ai\"], \"search_results\": [{\"title\": \"What Is Perplexity AI and How It Works - M1-Project\", \"url\": \"https://www.m1-project.com/blog/what-is-perplexity-ai-and-how-it-works\", \"date\": \"2025-07-24\", \"last_updated\": \"2025-07-24\"}, {\"title\": \"Haiku Generation models\", \"url\": \"https://web.stanford.edu/class/archive/cs/cs224n/cs224n.1224/posters/116768967.pdf\", \"date\": null, \"last_updated\": null}, {\"title\": \"What is Perplexity AI? | Definition from TechTarget\", \"url\": \"https://www.techtarget.com/searchenterpriseai/definition/Perplexity-AI\", \"date\": \"2025-03-04\", \"last_updated\": \"2025-08-14\"}, {\"title\": \"Examples of Poems about Perplexity\", \"url\": \"https://www.poetrysoup.com/poems/perplexity\", \"date\": \"2020-07-27\", \"last_updated\": \"2025-08-04\"}, {\"title\": \"What is Perplexity AI? A Smarter Way to Search | DigitalOcean\", \"url\": \"https://www.digitalocean.com/resources/articles/what-is-perplexity-ai\", \"date\": \"2025-03-13\", \"last_updated\": \"2025-08-01\"}], \"object\": \"chat.completion\", \"choices\": [{\"index\": 0, \"finish_reason\": \"stop\", \"message\": {\"role\": \"assistant\", \"content\": \"Silent web unfolds,  \\nPerplexity finds the truth\\u2014  \\nAnswers bloom like spring.\"}, \"delta\": {\"role\": \"assistant\", \"content\": \"\"}}]}"), 
    cache = new.env(parent = emptyenv())), class = "httr2_response")

#' The Grow API provider class (inherits from OpenAI)
#'
#' @noRd
api_groq <- new_class("Groq", api_openai)

#' A function to get metadata from Openai responses
#'
#' @noRd
method(extract_metadata, list(api_groq,class_list))<- function(.api, .response) {
  list(
    model             = .response$model,
    timestamp         = lubridate::as_datetime(.response$created),
    prompt_tokens     = .response$usage$prompt_tokens,
    completion_tokens = .response$usage$completion_tokens,
    total_tokens      = .response$usage$total_tokens,
    specific_metadata = list(
      system_fingerprint        = .response$system_fingerprint,
      completion_time           = .response$usage$completion_time,
      total_time                = .response$usage$total_time,
      groq_id                   = .response$x_groq$id
    ) 
  )
}  


#' Send LLM Messages to the Groq Chat API
#'
#' @description
#' This function sends a message history to the Groq Chat API and returns the assistant's reply.
#'
#' @param .llm An `LLMMessage` object containing the conversation history.
#' @param .model The identifier of the model to use (default: "llama-3.2-11b-vision-preview").
#' @param .max_tokens The maximum number of tokens that can be generated in the response (default: 1024).
#' @param .temperature Controls the randomness in the model's response. Values between 0 and 2 are allowed, where higher values increase randomness (optional).
#' @param .top_p Nucleus sampling parameter that controls the proportion of probability mass considered. Values between 0 and 1 are allowed (optional).
#' @param .frequency_penalty Number between -2.0 and 2.0. Positive values penalize repeated tokens, reducing likelihood of repetition (optional).
#' @param .presence_penalty Number between -2.0 and 2.0. Positive values encourage new topics by penalizing tokens that have appeared so far (optional).
#' @param .stop One or more sequences where the API will stop generating further tokens. Can be a string or a list of strings (optional).
#' @param .tools Either a single TOOL object or a list of TOOL objects representing the available functions for tool calls (optional).
#' @param .tool_choice A character string specifying the tool-calling behavior; valid values are "none", "auto", or "required" (optional).
#' @param .seed An integer for deterministic sampling. If specified, attempts to return the same result for repeated requests with identical parameters (optional).
#' @param .api_url Base URL for the Groq API (default: "https://api.groq.com/").
#' @param .json_schema A list or tidyllm schema created with `tidyllm_schema()` for structured JSON output (optional).
#' @param .timeout Request timeout in seconds (default: 60).
#' @param .stream Logical; if TRUE, streams the response piece by piece (default: FALSE).
#' @param .verbose If TRUE, displays additional information after the API call, including rate limit details (default: FALSE).
#' @param .max_tries Maximum retries to peform request
#' @param .dry_run If TRUE, performs a dry run and returns the constructed request object without executing it (default: FALSE).
#' @param .max_tool_rounds Integer specifying the maximum number of tool use iterations (default: 10). 
#'   Set to 1 for single-round tool use, or higher for multi-turn agentic loops.
#'
#' @return A new `LLMMessage` object containing the original messages plus the assistant's response.
#' 
#' @examples
#' \dontrun{
#' # Basic usage
#' msg <- llm_message("What is Groq?")
#' result <- groq_chat(msg)
#' 
#' # With custom parameters
#' result2 <- groq_chat(msg, 
#'                .model = "llama-3.2-vision",
#'                .temperature = 0.5, 
#'                .max_tokens = 512)
#' }
#'
#' @export
groq_chat <- function(.llm,
                 .model = "moonshotai/kimi-k2-instruct-0905",
                 .max_tokens = 1024,
                 .temperature = NULL,
                 .top_p = NULL,
                 .frequency_penalty = NULL,
                 .presence_penalty = NULL,
                 .stop = NULL,
                 .seed = NULL,
                 .tools = NULL,
                 .tool_choice = NULL,
                 .api_url = "https://api.groq.com/",
                 .json_schema = NULL,
                 .timeout = 60,
                 .verbose = FALSE,
                 .stream = FALSE,
                 .dry_run = FALSE,
                 .max_tries = 3,
                 .max_tool_rounds = 10) {

  # Validate inputs
  c(
    "Input .llm must be an LLMMessage object" = S7_inherits(.llm, LLMMessage),
    "Input .max_tokens must be an integer" = is_integer_valued(.max_tokens) & .max_tokens > 0,
    "Input .model must be a non-empty string" = is.character(.model) & nzchar(.model),
    "Input .api_url must be a valid URL" = is.character(.api_url) & nzchar(.api_url),
    "Input .timeout must be an integer-valued numeric greater than 0" = is_integer_valued(.timeout) & .timeout > 0,
    "Input .temperature must be numeric between 0 and 2 if provided" = is.null(.temperature) | (.temperature >= 0 & .temperature <= 2),
    "Input .top_p must be numeric between 0 and 1 if provided" = is.null(.top_p) | (.top_p >= 0 & .top_p <= 1),
    "Input .frequency_penalty must be numeric between -2 and 2 if provided" = is.null(.frequency_penalty) | (.frequency_penalty >= -2 & .frequency_penalty <= 2),
    "Input .presence_penalty must be numeric between -2 and 2 if provided" = is.null(.presence_penalty) | (.presence_penalty >= -2 & .presence_penalty <= 2),
    "Input .stop must be a charachter vector or a list of character vectors, or NULL" = is.null(.stop) | is.character(.stop) | is.list(.stop),
    "Input .seed must be an integer if provided" = is.null(.seed) | is_integer_valued(.seed),
    "Input .tools must be NULL, a TOOL object, or a list of TOOL objects" = is.null(.tools) || S7_inherits(.tools, TOOL) || (is.list(.tools) && all(purrr::map_lgl(.tools, ~ S7_inherits(.x, TOOL)))),
    "Input .tool_choice must be NULL or a character (one of 'none', 'auto', 'required')" = is.null(.tool_choice) || (is.character(.tool_choice) && .tool_choice %in% c("none", "auto", "required")),
    "Input .json_schema must be NULL or a list or an ellmer type object" = is.null(.json_schema) | is.list(.json_schema) | is_ellmer_type(.json_schema),
    "Input .verbose must be logical" = is.logical(.verbose),
    "Input .max_tries must be integer-valued numeric" = is_integer_valued(.max_tries),
    "Input .dry_run must be logical" = is.logical(.dry_run),
    ".max_tool_rounds must be a positive integer" = is_integer_valued(.max_tool_rounds) && .max_tool_rounds >= 1,
    "Streaming is not supported for requests with tool calls" = is.null(.tools) || !isTRUE(.stream)
  ) |>
    validate_inputs()
  
  api_obj <- api_groq(short_name = "groq",
                    long_name  = "Groq",
                    api_key_env_var = "GROQ_API_KEY")

  # Get formatted message list for Groq models
  messages <- to_api_format(.llm,api_obj,TRUE)

  api_key <- get_api_key(api_obj,.dry_run)

  #Put a single tool into a list if only one is provided.
  tools_def <- if (!is.null(.tools)) {
    if (S7_inherits(.tools, TOOL))  list(.tools) else .tools
  } else {
    NULL
  }

  json <- FALSE
  response_format <- NULL
  if (!is.null(.json_schema)) {
    json <- TRUE
    schema_name <- "empty"
    if (requireNamespace("ellmer", quietly = TRUE)) {
      if (S7_inherits(.json_schema, ellmer::TypeObject)) {
        .json_schema <- to_schema(.json_schema)
        schema_name <- "ellmer_schema"
      }
    }
    if (schema_name != "ellmer_schema") {
      schema_name <- attr(.json_schema, "name") %||% "schema"
    }
    if (is.null(.json_schema$additionalProperties)) {
      .json_schema$additionalProperties <- FALSE
    }
    response_format <- list(
      type = "json_schema",
      json_schema = list(
        schema = .json_schema,
        name = schema_name,
        strict = TRUE
      )
    )
  }

  # Fill the request body
  request_body <- list(
    model = .model,
    max_tokens = .max_tokens,
    messages = messages,
    temperature = .temperature,
    top_p = .top_p,
    frequency_penalty = .frequency_penalty,
    presence_penalty = .presence_penalty,
    stop = .stop,
    seed = .seed,
    stream = .stream,
    tools = if(!is.null(tools_def)) tools_to_api(api_obj,tools_def) else NULL,
    tool_choice = .tool_choice
  ) |> purrr::compact()

  if (!is.null(response_format)) {
    request_body$response_format <- response_format
  }
  
  request <- httr2::request(.api_url) |>
    httr2::req_url_path("/openai/v1/chat/completions") |>
    httr2::req_headers(
      `Authorization` = sprintf("Bearer %s", api_key),
      `Content-Type` = "application/json"
    ) |>
    httr2::req_body_json(data = request_body)
  
  # Return only the request object in a dry run.
  if (.dry_run) {
    return(request)  
  }
  
  response <- perform_chat_request(request,api_obj,.stream,.timeout,.max_tries)
  
  # Handle tool calls with multi-turn support
  if (.stream == FALSE && !is.null(tools_def)) {
    response <- process_tool_loop(
      .api = api_obj,
      .response = response,
      .tools_def = tools_def,
      .request_body = request_body,
      .request = request,
      .timeout = .timeout,
      .max_tries = .max_tries,
      .max_tool_rounds = .max_tool_rounds
    )
  }
  
  # Extract assistant reply and rate limiting info from response headers
  assistant_reply <- response$assistant_reply
  track_rate_limit(api_obj,response$headers,.verbose)
  
  # Add model's message to the history of the LLMMessage object
  add_message(.llm     = .llm,
              .role    = "assistant",
              .content = assistant_reply,
              .json    = json,
              .meta    = response$meta)
}



#' Transcribe an Audio File Using Groq transcription API 
#'
#' @description
#' This function reads an audio file and sends it to the Groq transcription API for transcription.
#'
#' @param .audio_file The path to the audio file (required). Supported formats include flac, mp3, mp4, mpeg, mpga, m4a, ogg, wav, or webm.
#' @param .model The model to use for transcription (default: "whisper-large-v3").
#' @param .language The language of the input audio, in ISO-639-1 format (optional).
#' @param .prompt A prompt to guide the transcription style. It should match the audio language (optional).
#' @param .temperature Sampling temperature, between 0 and 1, with higher values producing more randomness (default: 0).
#' @param .api_url Base URL for the API (default: "https://api.groq.com/openai/v1/audio/transcriptions").
#' @param .verbose Logical; if TRUE, outputs additional information (default: FALSE).
#' @param .dry_run Logical; if TRUE, performs a dry run and returns the request object without making the API call (default: FALSE).
#' @param .verbose Logical; if TRUE, rate limiting info is displayed after the API request (default: FALSE).
#' @param .max_tries Maximum retries to peform request
#' @examples
#' \dontrun{
#' # Basic usage
#' groq_transcribe(.audio_file = "example.mp3")
#' }
#'
#' @return A character vector containing the transcription.
#' @export
groq_transcribe <- function(
    .audio_file,
    .model = "playai-tts",
    .language = NULL,
    .prompt = NULL,
    .temperature = 0,
    .api_url = "https://api.groq.com/openai/v1/audio/transcriptions",
    .dry_run = FALSE,
    .verbose = FALSE,
    .max_tries = 3) {

  # Validate audio file path
  if (!file.exists(.audio_file)) stop("Audio file does not exist.")
  api_key <- Sys.getenv("GROQ_API_KEY")
  if ((api_key == "")& .dry_run==FALSE){
    stop("API key is not set. Please set it with: Sys.setenv(GROQ_API_KEY = \"YOUR-KEY-GOES-HERE\")")
  }
  
  # Validate inputs
  c(
    "Input .model must be a character vector" = is.character(.model),
    "Input .language must be a character vector if specified" = is.null(.language) || is.character(.language),
    "Input .prompt must be a character vector if specified" = is.null(.prompt) || is.character(.prompt),
    "Input .temperature" = is.numeric(.temperature) && .temperature >= 0 && .temperature <= 1,
    "Input .max_tries must be integer-valued numeric" = is_integer_valued(.max_tries)
  ) |>
    validate_inputs()
  
  # Build the request
  request <- httr2::request(.api_url) |>
    httr2::req_headers(
      Authorization = sprintf("Bearer %s", Sys.getenv("GROQ_API_KEY"))
    ) |>
    httr2::req_body_multipart(
      file = curl::form_file(.audio_file),
      model = .model,
      prompt = .prompt,
      response_format ="json",
      language = .language,
      temperature = as.character(.temperature)
    )
  
  # Handle dry run
  if (.dry_run) return(request)
  

  #Perform the request
  response <- request |>
               httr2::req_error(is_error = function(resp) FALSE) |>
               httr2::req_retry(max_tries = .max_tries,  
                                retry_on_failure = TRUE,
                                 is_transient = \(resp) httr2::resp_status(resp) %in% c(429, 503)) |>
               httr2::req_perform()
  
  content <- httr2::resp_body_json(response)
  if("error" %in% names(content)){
    sprintf("Groq API returned an Error:\nCode: %s\nMessage: %s",
            content$error$code,
            content$error$message) |>
      stop()
  }
  
  groq_audio_headers <- httr2::resp_headers(response)
  
  #Parse the headers from the response
  request_time <- strptime(groq_audio_headers[["date"]], 
                           format="%a, %d %b %Y %H:%M:%S", tz="GMT")
  rl_total_sec     <- as.integer(groq_audio_headers[["x-ratelimit-limit-audio-seconds"]])
  rl_total_req     <- as.integer(groq_audio_headers[["x-ratelimit-limit-requests"]])
  rl_remaining_sec <- as.integer(groq_audio_headers[["x-ratelimit-remaining-audio-seconds"]])
  rl_remaining_req <- as.integer(groq_audio_headers[["x-ratelimit-remaining-requests"]])
  rl_durartion_sec <- parse_duration_to_seconds(groq_audio_headers[["x-ratelimit-reset-audio-seconds"]])
  rl_durartion_req <- parse_duration_to_seconds(groq_audio_headers[["x-ratelimit-reset-requests"]])
  
  #Compute reset times
  rl_reset_req <- request_time + rl_durartion_req
  rl_reset_sec <- request_time + rl_durartion_sec
  
  # Show rate limit info if verbose
  if (.verbose == TRUE) {
    glue::glue(
     "Groq Transcribe API answer received at {request_time}.
      Remaining requests rate limit: {rl_remaining_req}/{rl_total_req}
      Requests rate limit reset at: {rl_reset_req}
      Remaining Audio seconds rate limit: {rl_remaining_sec}/{rl_total_sec}
      Audio seconds rate limit reset at: {rl_reset_sec}\n"
    ) |> cat("\n")
  }
  
  
  return(content$text)
}

#' List Available Models from the Groq API
#'
#' @param .api_url Base URL for the API (default: "https://api.groq.com").
#' @param .timeout Request timeout in seconds (default: 60).
#' @param .max_tries Maximum number of retries for the API request (default: 3).
#' @param .dry_run Logical; if TRUE, returns the prepared request object without executing it.
#' @param .verbose Logical; if TRUE, prints additional information about the request.
#'
#' @return A tibble containing model information (columns include `id`, `created`, `owned_by`, and `context_window`),
#'   or NULL if no models are found.
#'
#' @export
groq_list_models <- function(.api_url = "https://api.groq.com",
                             .timeout = 60,
                             .max_tries = 3,
                             .dry_run = FALSE,
                             .verbose = FALSE) {
  # Create an API object for Groq using the tidyllm helper
  api_obj <- api_openai(short_name = "groq",
                        long_name  = "Groq",
                        api_key_env_var = "GROQ_API_KEY")
  
  # Retrieve the API key (will error if not set, unless in dry run mode)
  api_key <- get_api_key(api_obj, .dry_run)
  
  # Build the request to the /openai/v1/models endpoint
  request <- httr2::request(.api_url) |>
    httr2::req_url_path("/openai/v1/models") |>
    httr2::req_headers(
      Authorization = sprintf("Bearer %s", api_key),
      `Content-Type` = "application/json"
    )
  
  # If dry run is requested, return the constructed request object
  if (.dry_run) {
    return(request)
  }
  
  # Perform the request with specified timeout and retry logic
  response <- request |>
    httr2::req_timeout(.timeout) |>
    httr2::req_retry(max_tries = .max_tries) |>
    httr2::req_perform() |>
    httr2::resp_body_json()
  
  if (.verbose) {
    message("Retrieved response from Groq: ", response$object)
  }
  
  # Check if the "data" field exists and contains models
  if (!is.null(response$data)) {
    models <- response$data
    
    # Create a tibble with selected model information
    model_info <- tibble::tibble(
      id = vapply(models, function(model) model$id, character(1)),
      created = vapply(models, function(model) {
        as.character(strptime(
          format(as.POSIXct(model$created, origin = "1970-01-01", tz = "GMT"),
                 format = "%a, %d %b %Y %H:%M:%S"),
          format = "%a, %d %b %Y %H:%M:%S", tz = "GMT"))
      }, character(1)),
      owned_by = vapply(models, function(model) model$owned_by, character(1)),
      context_window = vapply(models, function(model) model$context_window, numeric(1))
    )
    
    return(model_info)
  } else {
    return(NULL)
  }
}

#' Send a Batch of Messages to the Groq API
#'
#' This function creates and submits a batch of messages to the Groq API for asynchronous processing.
#'
#' @param .llms A list of LLMMessage objects containing conversation histories.
#' @param .model Character string specifying the model to use (default: "deepseek-r1-distill-llama-70b").
#' @param .max_tokens Integer specifying the maximum tokens per response (default: 1024).
#' @param .temperature Numeric between 0 and 2 controlling response randomness.
#' @param .top_p Numeric between 0 and 1 for nucleus sampling.
#' @param .frequency_penalty Number between -2.0 and 2.0 to penalize repetition.
#' @param .presence_penalty Number between -2.0 and 2.0 to encourage new topics.
#' @param .stop One or more sequences where the API will stop generating further tokens.
#' @param .seed An integer for deterministic sampling.
#' @param .api_url Base URL for the Groq API (default: "https://api.groq.com/").
#' @param .json_schema A list or tidyllm schema created with `tidyllm_schema()` for structured JSON output (optional).
#' @param .completion_window Character string for the batch completion window (default: "24h").
#' @param .verbose Logical; if TRUE, prints a message with the batch ID (default: FALSE).
#' @param .overwrite Logical; if TRUE, allows overwriting an existing batch ID (default: FALSE).
#' @param .max_tries Maximum number of retries to perform the request.
#' @param .timeout Integer specifying the request timeout in seconds (default: 60).
#' @param .dry_run Logical; if TRUE, returns the prepared request objects without executing (default: FALSE).
#' @param .id_prefix Character string to specify a prefix for generating custom IDs when names in `.llms` are missing.
#' 
#' @return An updated and named list of `.llms` with identifiers that align with batch responses, including a `batch_id` attribute.
#' @export
send_groq_batch <- function(.llms,
                            .model = "openai/gpt-oss-120b",
                            .max_tokens = 1024,
                            .temperature = NULL,
                            .top_p = NULL,
                            .frequency_penalty = NULL,
                            .presence_penalty = NULL,
                            .stop = NULL,
                            .seed = NULL,
                            .api_url = "https://api.groq.com/",
                            .json_schema = NULL,
                            .completion_window = "24h",
                            .verbose = FALSE,
                            .dry_run = FALSE,
                            .overwrite = FALSE,
                            .max_tries = 3,
                            .timeout = 60,
                            .id_prefix = "tidyllm_groq_req_") {

  # Input validation
  c(
    ".llms must be a list of LLMMessage objects" = is.list(.llms) && all(sapply(.llms, S7_inherits, LLMMessage)),
    ".max_tokens must be an integer" = is_integer_valued(.max_tokens),
    ".model must be a non-empty string" = is.character(.model) & nzchar(.model),
    ".temperature must be numeric between 0 and 2 if provided" = is.null(.temperature) || (.temperature >= 0 && .temperature <= 2),
    ".top_p must be numeric between 0 and 1 if provided" = is.null(.top_p) || (.top_p >= 0 && .top_p <= 1),
    ".frequency_penalty must be numeric between -2 and 2 if provided" = is.null(.frequency_penalty) || (.frequency_penalty >= -2 && .frequency_penalty <= 2),
    ".presence_penalty must be numeric between -2 and 2 if provided" = is.null(.presence_penalty) || (.presence_penalty >= -2 && .presence_penalty <= 2),
    ".stop must be a character vector or NULL" = is.null(.stop) || is.character(.stop) || is.list(.stop),
    ".seed must be an integer if provided" = is.null(.seed) || is_integer_valued(.seed),
    ".completion_window must be a valid window format" = is.character(.completion_window) && grepl("^\\d+[hd]$", .completion_window),
    ".verbose must be logical" = is.logical(.verbose),
    ".dry_run must be logical" = is.logical(.dry_run),
    ".overwrite must be logical" = is.logical(.overwrite),
    ".id_prefix must be a character vector of length 1" = is.character(.id_prefix),
    ".max_tries must be integer-valued numeric" = is_integer_valued(.max_tries),
    ".timeout must be an integer" = is_integer_valued(.timeout),
    ".json_schema must be NULL or a list or an ellmer type object" = is.null(.json_schema) | is.list(.json_schema) | is_ellmer_type(.json_schema)
  ) |> validate_inputs()

  response_format <- NULL
  if (!is.null(.json_schema)) {
    schema_name <- "empty"
    if (requireNamespace("ellmer", quietly = TRUE)) {
      if (S7_inherits(.json_schema, ellmer::TypeObject)) {
        .json_schema <- to_schema(.json_schema)
        schema_name <- "ellmer_schema"
      }
    }
    if (schema_name != "ellmer_schema") {
      schema_name <- attr(.json_schema, "name") %||% "schema"
    }
    if (is.null(.json_schema$additionalProperties)) {
      .json_schema$additionalProperties <- FALSE
    }
    response_format <- list(
      type = "json_schema",
      json_schema = list(
        schema = .json_schema,
        name = schema_name,
        strict = TRUE
      )
    )
  }

  api_obj <- api_groq(short_name = "groq",
                      long_name  = "Groq",
                      api_key_env_var = "GROQ_API_KEY")
  
  api_key <- get_api_key(api_obj, .dry_run)
  
  prepared_llms <- prepare_llms_for_batch(api_obj,
                                          .llms = .llms,
                                          .id_prefix = .id_prefix,
                                          .overwrite = .overwrite)
  
  # Create JSONL content for batch submission
  jsonl_lines <- lapply(seq_along(prepared_llms), function(i) {
    # Get messages from each LLMMessage object
    messages <- to_api_format(prepared_llms[[i]], api_obj, TRUE)
    
    custom_id <- names(prepared_llms)[i]
    request_body <- list(
      model = .model,
      max_tokens = .max_tokens,
      messages = messages,
      temperature = .temperature,
      top_p = .top_p,
      frequency_penalty = .frequency_penalty,
      presence_penalty = .presence_penalty,
      stop = .stop,
      seed = .seed
    ) |> purrr::compact()

    if (!is.null(response_format)) {
      request_body$response_format <- response_format
    }

    # Create the batch request entry
    batch_entry <- list(
      custom_id = custom_id,
      method = "POST",
      url = "/v1/chat/completions",
      body = request_body
    )
    
    # Convert to JSON string
    jsonlite::toJSON(batch_entry, auto_unbox = TRUE)
  })
  
  # Combine all lines into a single JSONL string
  jsonl_content <- paste(jsonl_lines, collapse = "\n")
  
  # Create a temporary file to hold the JSONL content
  temp_file <- tempfile(fileext = ".jsonl")
  writeLines(jsonl_content, temp_file)
  
  if (.dry_run) {
    return(readLines(temp_file))
  }
  
  # 1. Upload the JSONL file
  upload_request <- httr2::request(paste0(.api_url, "openai/v1/files")) |>
    httr2::req_headers(
      Authorization = sprintf("Bearer %s", api_key),
      .redact = "Authorization"
    ) |>
    httr2::req_body_multipart(
      file = curl::form_file(temp_file),
      purpose = "batch"
    )
  
  upload_response <- upload_request |>
    perform_generic_request(.timeout = .timeout, .max_tries = .max_tries)
  
  if ("error" %in% names(upload_response$content)) {
    stop(sprintf("Groq API returned an Error while uploading file:\nMessage: %s",
                 upload_response$content$error$message))
  }
  
  file_id <- upload_response$content$id
  
  # 2. Create the batch job
  batch_request <- httr2::request(paste0(.api_url, "openai/v1/batches")) |>
    httr2::req_headers(
      Authorization = sprintf("Bearer %s", api_key),
      "Content-Type" = "application/json",
      .redact = "Authorization"
    ) |>
    httr2::req_body_json(list(
      input_file_id = file_id,
      endpoint = "/v1/chat/completions",
      completion_window = .completion_window
    ))
  
  batch_response <- batch_request |>
    perform_generic_request(.timeout = .timeout, .max_tries = .max_tries)
  
  if ("error" %in% names(batch_response$content)) {
    stop(sprintf("Groq API returned an Error while creating batch:\nMessage: %s",
                 batch_response$content$error$message))
  }
  
  # Attach batch_id as an attribute to .llms
  batch_id <- batch_response$content$id
  attr(prepared_llms, "batch_id") <- batch_id
  attr(prepared_llms, "file_id") <- file_id
  attr(prepared_llms, "json") <- !is.null(.json_schema)
  
  if (.verbose) {
    message("Batch submitted successfully. Batch ID: ", batch_id)
  }
  
  return(prepared_llms)
}

#' Check Batch Processing Status for Groq API
#'
#' This function retrieves the processing status and other details of a specified Groq batch.
#'
#' @param .llms A list of LLMMessage objects with a batch_id attribute.
#' @param .batch_id A character string with the batch ID to check.
#' @param .api_url Character; base URL of the Groq API (default: "https://api.groq.com/").
#' @param .max_tries Maximum retries to perform request.
#' @param .timeout Integer specifying the request timeout in seconds (default: 60).
#' @param .dry_run Logical; if TRUE, returns the prepared request object without executing (default: FALSE).
#' 
#' @return A tibble with information about the status of batch processing.
#' @export
check_groq_batch <- function(.llms = NULL, 
                             .batch_id = NULL, 
                             .api_url = "https://api.groq.com/",
                             .dry_run = FALSE,
                             .max_tries = 3,
                             .timeout = 60) {
  # Extract batch_id
  if (is.null(.batch_id)) {
    if (!is.null(.llms)) {
      .batch_id <- attr(.llms, "batch_id")
      if (is.null(.batch_id)) {
        stop("No batch_id attribute found in the provided list.")
      }
    } else {
      stop("Either .llms or .batch_id must be provided.")
    }
  }
  
  # Retrieve API key
  api_key <- Sys.getenv("GROQ_API_KEY")
  if ((api_key == "") & !.dry_run) {
    stop("API key is not set.")
  }
  
  # Build request
  request_url <- paste0(.api_url, "openai/v1/batches/", .batch_id)
  request <- httr2::request(request_url) |>
    httr2::req_headers(
      Authorization = sprintf("Bearer %s", api_key),
      .redact = "Authorization"
    )
  
  if (.dry_run) {
    return(request)
  }
  
  # Perform request with retries and error handling
  response <- request |>
    perform_generic_request(.timeout = .timeout, .max_tries = .max_tries)
  
  # Parse response
  response_body <- response$content
  if ("error" %in% names(response_body)) {
    stop(sprintf("Groq API returned an Error:\nMessage: %s",
                 response_body$error$message))
  }
  
  # Create tibble with batch details
  tibble::tibble(
    batch_id = response_body$id,
    status = response_body$status,
    created_at = lubridate::as_datetime(response_body$created_at, origin = "1970-01-01"),
    expires_at = lubridate::as_datetime(response_body$expires_at, origin = "1970-01-01"),
    req_total = response_body$request_counts$total,
    req_completed = response_body$request_counts$completed,
    req_failed = response_body$request_counts$failed,
    input_file_id = response_body$input_file_id,
    output_file_id = response_body$output_file_id,
    error_file_id = response_body$error_file_id
  )
}

#' Fetch Results for a Groq Batch
#'
#' This function retrieves the results of a completed Groq batch and updates
#' the provided list of `LLMMessage` objects with the responses.
#'
#' @param .llms A list of `LLMMessage` objects that were part of the batch.
#' @param .batch_id Character; the unique identifier for the batch.
#' @param .api_url Character; the base URL for the Groq API (default: "https://api.groq.com/").
#' @param .dry_run Logical; if `TRUE`, returns the constructed request without executing it (default: `FALSE`).
#' @param .max_tries Integer; maximum number of retries if the request fails (default: `3`).
#' @param .timeout Integer; request timeout in seconds (default: `60`).
#'
#' @return A list of updated `LLMMessage` objects, each with the assistant's response added if successful.
#' @export
fetch_groq_batch <- function(.llms, 
                             .batch_id = NULL, 
                             .api_url = "https://api.groq.com/",
                             .dry_run = FALSE,
                             .max_tries = 3,
                             .timeout = 60) {
  c(
    ".llms must be a list of LLMMessage objects with names as custom IDs" = is.list(.llms) && all(sapply(.llms, S7_inherits, LLMMessage)),
    ".batch_id must be a non-empty character string or NULL" = is.null(.batch_id) || (is.character(.batch_id) && nzchar(.batch_id)),
    ".api_url must be a non-empty character string" = is.character(.api_url) && nzchar(.api_url),
    ".dry_run must be logical" = is.logical(.dry_run),
    ".max_tries must be an integer" = is_integer_valued(.max_tries),
    ".timeout must be an integer" = is_integer_valued(.timeout)
  ) |> validate_inputs()
  
  # Preserve original names
  original_names <- names(.llms)
  
  # Retrieve batch_id from .llms if not provided
  if (is.null(.batch_id)) {
    .batch_id <- attr(.llms, "batch_id")
    if (is.null(.batch_id)) {
      stop("No batch_id provided and no batch_id attribute found in the provided list.")
    }
  }
  
  api_obj <- api_groq(short_name = "groq",
                      long_name  = "Groq",
                      api_key_env_var = "GROQ_API_KEY")
  
  api_key <- get_api_key(api_obj, .dry_run)
  
  batch_status <- check_groq_batch(.batch_id = .batch_id)
  
  # Check if batch has completed processing
  if (batch_status$status != "completed") {
    stop("Batch processing has not completed yet. Current status: ", batch_status$status)
  }
  
  if (is.null(batch_status$output_file_id)) {
    stop("Batch completed but no output file is available.")
  }
  
  # Download the results file
  output_file_id <- batch_status$output_file_id
  file_download_url <- paste0(.api_url, "openai/v1/files/", output_file_id, "/content")
  file_download_request <- httr2::request(file_download_url) |>
    httr2::req_headers(
      Authorization = sprintf("Bearer %s", api_key),
      .redact = "Authorization"
    )
  
  file_download_response <- file_download_request |>
    httr2::req_timeout(.timeout) |>
    httr2::req_retry(max_tries = .max_tries) |>
    httr2::req_perform()
  
  # Parse the JSONL response
  results_content <- httr2::resp_body_string(file_download_response)
  results_lines <- strsplit(results_content, "\n")[[1]]
  results_list <- lapply(results_lines, function(line) {
    if (nzchar(line)) jsonlite::fromJSON(line) else NULL
  })
  results_list <- Filter(Negate(is.null), results_list)
  
  # Map results by custom_id
  results_by_custom_id <- purrr::set_names(
    results_list, 
    sapply(results_list, function(x) x$custom_id)
  )
  
  .json <- attr(.llms, "json")
  if (is.null(.json)) .json <- FALSE

  # Map results back to the original .llms list using names as custom IDs
  updated_llms <- lapply(names(.llms), function(custom_id) {
    result <- results_by_custom_id[[custom_id]]

    if (!is.null(result) && result$response$status_code == 200) {
      assistant_reply <- result$response$body$choices$message$content
      metadata        <- extract_metadata(api_obj, result$response$body)

      llm <- add_message(
        .llm = .llms[[custom_id]],
        .role = "assistant",
        .content = assistant_reply,
        .json = .json,
        .meta = metadata
      )

      return(llm)
    } else {
      if (!is.null(result$error)) {
        warning(sprintf("Error for custom_id %s: %s", custom_id, result$error$message))
      } else {
        warning(sprintf("Result for custom_id %s was unsuccessful or not found", custom_id))
      }
      return(.llms[[custom_id]])
    }
  })

  # Restore original names
  names(updated_llms) <- original_names

  # Remove batch attributes before returning to avoid reuse conflicts
  attr(updated_llms, "batch_id") <- NULL
  attr(updated_llms, "file_id") <- NULL
  attr(updated_llms, "json") <- NULL
  
  return(updated_llms)
}

#' List Groq Batch Requests
#'
#' Retrieves batch request details from the Groq API.
#'
#' @param .api_url Base URL for the Groq API (default: "https://api.groq.com/").
#' @param .limit Maximum number of batches to retrieve (default: 20).
#' @param .max_tries Maximum retry attempts for requests (default: 3).
#' @param .timeout Request timeout in seconds (default: 60).
#'
#' @return A tibble with batch details including batch ID, status, creation time, and request counts.
#'
#' @export
list_groq_batches <- function(.api_url = "https://api.groq.com/", 
                              .limit = 20, 
                              .max_tries = 3,
                              .timeout = 60) {
  # Retrieve API key
  api_key <- Sys.getenv("GROQ_API_KEY")
  if (api_key == "") {
    stop("API key is not set. Please set it with: Sys.setenv(GROQ_API_KEY = \"YOUR-KEY-GOES-HERE\").")
  }
  
  # Set up request URL with query parameters
  request <- httr2::request(.api_url) |>
    httr2::req_url_path("openai/v1/batches") |>
    httr2::req_headers(
      Authorization = sprintf("Bearer %s", api_key),
      .redact = "Authorization"
    ) |>
    httr2::req_url_query(limit = .limit)
  
  # Perform the request with retries and error handling
  response <- request |>
    perform_generic_request(.timeout = .timeout, .max_tries = .max_tries)
  
  # Parse response and handle any errors
  if ("error" %in% names(response$content)) {
    stop(sprintf("Groq API Error: %s", response$content$error$message))
  }
  
  # Extract batch list details
  batches <- response$content$data
  if (length(batches) == 0) {
    return(tibble::tibble())
  }
  
  batch_list <- purrr::map_dfr(batches, function(batch) {
    tibble::tibble(
      batch_id = batch$id,
      status = batch$status,
      created_at = lubridate::as_datetime(batch$created_at, origin = "1970-01-01"),
      expires_at = lubridate::as_datetime(batch$expires_at, origin = "1970-01-01"),
      req_total = batch$request_counts$total,
      req_completed = batch$request_counts$completed,
      req_failed = batch$request_counts$failed,
      input_file_id = batch$input_file_id,
      output_file_id = batch$output_file_id,
      error_file_id = batch$error_file_id
    )
  })
  
  return(batch_list)
}

#' Groq API Provider Function
#'
#' The `groq()` function acts as an interface for interacting with the Groq API 
#' through `tidyllm`'s main verbs. Currently, Groq only supports `groq_chat()` 
#' for chat-based interactions and `groq_transcribe()` for transcription tasks.
#'
#' Since `groq_transcribe()` is unique to Groq and does not have a general verb counterpart, 
#' `groq()` currently routes messages only to `groq_chat()` when used with verbs like `chat()`.
#'
#' @param ... Parameters to be passed to the Groq-specific function, 
#'   such as model configuration, input text, or API-specific options.
#' @param .called_from An internal argument that specifies which action (e.g., 
#'   `chat`) the function is being invoked from. 
#'   This argument is automatically managed and should not be modified by the user.
#'
#' @return The result of the requested action, depending on the specific function invoked 
#'   (currently, only an updated `LLMMessage` object for `groq_chat()`).
#'
#' @export
groq <- create_provider_function(
  .name = "groq",
  chat = groq_chat,
  list_models = groq_list_models,
  send_batch = send_groq_batch,
  check_batch = check_groq_batch,
  fetch_batch = fetch_groq_batch,
  list_batches = list_groq_batches
)

tidyllm_image <- S7::new_class(
  name = "tidyllm_image",
  properties = list(
    imagepath   = class_character,
    imagename   = class_character,
    imagebase64 = class_character
  )
)

tidyllm_audio <- S7::new_class(
  name = "tidyllm_audio",
  properties = list(
    audiopath = class_character,
    audioname = class_character,
    audiomime = class_character
  )
)

tidyllm_video <- S7::new_class(
  name = "tidyllm_video",
  properties = list(
    videopath = class_character,
    videoname = class_character,
    videomime = class_character
  )
)

tidyllm_pdf <- S7::new_class(
  name = "tidyllm_pdf",
  properties = list(
    pdfpath      = class_character,
    pdfname      = class_character,
    pdftext      = class_character,
    pages        = class_any,
    text_extract = class_logical
  )
)

tidyllm_file <- S7::new_class(
  name = "tidyllm_file",
  properties = list(
    id        = class_character,
    provider  = class_character,
    mime_type = class_character,
    filename  = class_character,
    uri       = class_character
  )
)

print.tidyllm_image <- S7::new_external_generic("base", "print", "x")
method(print.tidyllm_image, tidyllm_image) <- function(x, ...) {
  size_kb <- round(file.size(x@imagepath) / 1024, 1)
  cat(sprintf("<tidyllm_image> %s  (%.1f KB)\n", x@imagename, size_kb))
  invisible(x)
}

print.tidyllm_audio <- S7::new_external_generic("base", "print", "x")
method(print.tidyllm_audio, tidyllm_audio) <- function(x, ...) {
  size_kb <- round(file.size(x@audiopath) / 1024, 1)
  cat(sprintf("<tidyllm_audio> %s  [%s]  (%.1f KB)\n",
              x@audioname, x@audiomime, size_kb))
  invisible(x)
}

print.tidyllm_video <- S7::new_external_generic("base", "print", "x")
method(print.tidyllm_video, tidyllm_video) <- function(x, ...) {
  size_mb <- round(file.size(x@videopath) / 1024^2, 1)
  cat(sprintf("<tidyllm_video> %s  [%s]  (%.1f MB)\n",
              x@videoname, x@videomime, size_mb))
  invisible(x)
}

print.tidyllm_pdf <- S7::new_external_generic("base", "print", "x")
method(print.tidyllm_pdf, tidyllm_pdf) <- function(x, ...) {
  size_kb <- round(file.size(x@pdfpath) / 1024, 1)
  pages_str <- if (is.null(x@pages)) "all pages" else paste0("pages ", paste(range(x@pages), collapse = "-"))
  mode_str  <- if (x@text_extract) "text" else "binary"
  cat(sprintf("<tidyllm_pdf> %s  %s  [%s]  (%.1f KB)\n",
              x@pdfname, pages_str, mode_str, size_kb))
  invisible(x)
}

print.tidyllm_file <- S7::new_external_generic("base", "print", "x")
method(print.tidyllm_file, tidyllm_file) <- function(x, ...) {
  cat(sprintf("<tidyllm_file [%s] id=%s mime=%s file=%s>\n",
              x@provider, x@id, x@mime_type, x@filename))
  invisible(x)
}


#' Extract  Media Content
#'
#' This internal function extracts and formats media content, specifically for Images or text formats. For PDF, Input files and RConsole types,
#' it formats these items into tagged strings suitable for inclusion in message content.
#'
#' @param .media_list A list of media items, each either a tidyllm S7 object or a legacy named list with 'type', 'filename', and 'content' keys.
#' @param .type What kind of media should be extracted (text or image)
#' @return A list of media items with formatted strings based on the media type.
#' @importFrom glue glue
#' @noRd
extract_media <- function(.media_list, .type) {
  if (is.null(.media_list) || length(.media_list) == 0) return(list())

  if (.type == "text") {
    out <- Filter(Negate(is.null), lapply(.media_list, function(x) {
      if (S7_inherits(x, tidyllm_pdf)) {
        tagged <- glue::glue("<pdf filename=\"{x@pdfname}\">{x@pdftext}</pdf>\n")
        return(tagged)
      }
      if (is.list(x) && !is.null(x$type)) {
        if (x$type == "PDF") {
          return(glue::glue("<pdf filename=\"{x$filename}\">{x$content}</pdf>\n"))
        }
        if (x$type == "RConsole") {
          return(glue::glue("<ROutput>{x$content}</ROutput>\n"))
        }
        if (x$type == "TextFile") {
          return(glue::glue("<file filename=\"{x$filename}\">{x$content}</file>\n"))
        }
      }
      NULL
    }))
  } else if (.type == "image") {
    out <- Filter(Negate(is.null), lapply(.media_list, function(x) {
      if (S7_inherits(x, tidyllm_image)) return(x)
      if (is.list(x) && !is.null(x$type) && x$type == "Image") return(x)
      NULL
    }))
  } else if (.type == "audio") {
    out <- Filter(Negate(is.null), lapply(.media_list, function(x) {
      if (S7_inherits(x, tidyllm_audio)) return(x)
      NULL
    }))
  } else if (.type == "video") {
    out <- Filter(Negate(is.null), lapply(.media_list, function(x) {
      if (S7_inherits(x, tidyllm_video)) return(x)
      NULL
    }))
  } else {
    out <- list()
  }

  return(out)
}

#' Create an Image Object
#'
#' This function reads an image file from disk, encodes it in base64,
#' and returns a `tidyllm_image` object that can be used in multimodal
#' embedding requests or attached to messages via `.media`.
#'
#' @param .path The path to the image file on disk.
#'
#' @return A `tidyllm_image` object.
#' @export
img <- function(.path) {
  if (!file.exists(.path)) {
    stop("File does not exist: ", .path)
  }

  raw_file <- readBin(.path, what = "raw", n = file.size(.path))
  b64      <- base64enc::base64encode(raw_file)

  mime_type <- if (grepl("\\.jpe?g$", .path, ignore.case = TRUE)) {
    "image/jpeg"
  } else if (grepl("\\.png$", .path, ignore.case = TRUE)) {
    "image/png"
  } else {
    "application/octet-stream"
  }

  b64_with_header <- paste0("data:", mime_type, ";base64,", b64)

  tidyllm_image(
    imagepath   = .path,
    imagename   = basename(.path),
    imagebase64 = b64_with_header
  )
}

#' Create an Audio Object
#'
#' Stores a reference to a local audio file for use in multimodal messages.
#' The file is not encoded until the message is formatted for a provider.
#'
#' @param .path The path to the audio file on disk.
#' @return A `tidyllm_audio` object.
#' @export
audio_file <- function(.path) {
  if (!file.exists(.path)) {
    stop("File does not exist: ", .path)
  }
  tidyllm_audio(
    audiopath = .path,
    audioname = basename(.path),
    audiomime = guess_mime_type(.path)
  )
}

#' Create a Video Object
#'
#' Stores a reference to a local video file for use in multimodal messages.
#' The file is not encoded until the message is formatted for a provider.
#'
#' @param .path The path to the video file on disk.
#' @return A `tidyllm_video` object.
#' @export
video_file <- function(.path) {
  if (!file.exists(.path)) {
    stop("File does not exist: ", .path)
  }
  tidyllm_video(
    videopath = .path,
    videoname = basename(.path),
    videomime = guess_mime_type(.path)
  )
}

#' Create a PDF Object
#'
#' Stores a reference to a local PDF file along with a pdftools text extraction.
#' Binary encoding is deferred to format time; providers that support binary PDF
#' encode on demand, others use the stored text fallback.
#'
#' @param .path The path to the PDF file on disk.
#' @param pages An integer vector of page numbers to include (e.g. `1:3`). NULL uses all pages.
#' @param .text_extract Logical; if TRUE always use pdftools text extraction regardless of provider
#'   (equivalent to the old `.pdf` parameter behaviour). Default FALSE uses binary-first.
#' @param engine Optional engine name (e.g. `"mistral-ocr"`); a warning is issued if the provider
#'   does not support binary PDF and the engine choice cannot be honoured.
#' @return A `tidyllm_pdf` object.
#' @export
pdf_file <- function(.path, pages = NULL, .text_extract = FALSE, engine = NULL) {
  if (!file.exists(.path)) {
    stop("File does not exist: ", .path)
  }
  if (!requireNamespace("pdftools", quietly = TRUE)) {
    stop("The 'pdftools' package is required for pdf_file(). Please install it.")
  }

  if (!is.null(pages)) {
    all_text <- pdftools::pdf_text(.path)
    valid_pages <- pages[pages >= 1 & pages <= length(all_text)]
    pdftext <- paste(all_text[valid_pages], collapse = "\n")
  } else {
    pdftext <- paste(pdftools::pdf_text(.path), collapse = "\n")
  }

  tidyllm_pdf(
    pdfpath      = .path,
    pdfname      = basename(.path),
    pdftext      = pdftext,
    pages        = pages,
    text_extract = .text_extract
  )
}

#' Parse Embedding Input
#'
#' This function processes an .input for embedding functions, extracting and combining message content
#' and any associated text media. If the input is a character vector, it is returned as is. If the input mixes 
#' text and image objects it is returned as a multimodal embedding object.
#' 
#' @param .input The input object, either a character vector or a list containing message history.
#' @return A character vector containing the combined message texts.
#' @noRd
parse_embedding_input <- function(.input) {
  if (!is.character(.input)) {
    history <- Filter(function(x) {
      if ("role" %in% names(x)) {
        return(x$role %in% c("user", "assistant"))
      } else {
        return(FALSE)
      }
    }, .input@message_history)
    
    # Extract messages and combine content and text media
    message_texts <- purrr::map_chr(history, function(m) {
      # The basic text content supplied with the message
      base_content <- m$content
      
      # Get the relevant media for the current message
      media_list <- m$media
      
      # Extract the text content from media
      text_media <- extract_media(media_list, "text")
      text_media_combined <- paste(unlist(text_media), collapse = " ")
      
      # Combine base content and text media
      combined_text <- paste(base_content, text_media_combined, sep = " ")
      combined_text
    })
  } else {
    message_texts <- .input
  }
  
  return(message_texts)
}

structure(list(method = "POST", url = "https://api.groq.com/openai/v1/chat/completions", 
    status_code = 200L, headers = structure(list(date = "Wed, 16 Oct 2024 00:52:02 GMT", 
        `content-type` = "application/json", `cache-control` = "private, max-age=0, no-store, no-cache, must-revalidate", 
        vary = "Origin", `x-ratelimit-limit-requests` = "7000", 
        `x-ratelimit-limit-tokens` = "7000", `x-ratelimit-remaining-requests` = "6999", 
        `x-ratelimit-remaining-tokens` = "6977", `x-ratelimit-reset-requests` = "12.342857142s", 
        `x-ratelimit-reset-tokens` = "197.142857ms", `x-request-id` = "req_01ja9cyx5qfphbv3cz76bkjs2s", 
        via = "1.1 google", `alt-svc` = "h3=\":443\"; ma=86400", 
        `cf-cache-status` = "DYNAMIC", `set-cookie` = "REDACTED", 
        server = "cloudflare", `cf-ray` = "8d34105a3a00d34e-FRA", 
        `content-encoding` = "gzip"), class = "httr2_headers"), 
    body = charToRaw("{\"id\":\"chatcmpl-903407ce-fcd4-46c3-880f-27d1f7268118\",\"object\":\"chat.completion\",\"created\":1729039922,\"model\":\"llama-3.2-90b-text-preview\",\"choices\":[{\"index\":0,\"message\":{\"role\":\"assistant\",\"content\":\"A typical German address would look like this:\\n\\nHerrn Peter Müller\\nMünchner Straße 23a\\n80803 München\\n\\nIn this example:\\n\\n- \\\"Herrn\\\" is the title (Mr.) \\n- \\\"Peter Müller\\\" is the recipient's name\\n- \\\"Münchner Straße\\\" is the street name\\n- \\\"23a\\\" is the house number\\n- \\\"80803\\\" is the postal code\\n- \\\"München\\\" is the city name\"},\"logprobs\":null,\"finish_reason\":\"stop\"}],\"usage\":{\"queue_time\":0.005171867999999998,\"prompt_tokens\":46,\"prompt_time\":0.011150101,\"completion_tokens\":99,\"completion_time\":0.396019894,\"total_tokens\":145,\"total_time\":0.407169995},\"system_fingerprint\":\"fp_5c5d1b5cfb\",\"x_groq\":{\"id\":\"req_01ja9cyx5qfphbv3cz76bkjs2s\"}}\n"), 
    cache = new.env(parent = emptyenv())), class = "httr2_response")

#' The OpenRouter API provider class (inherits from OpenAI)
#'
#' @noRd
api_openrouter <- new_class("OpenRouter", api_openai)


#' Extract metadata from OpenRouter chat responses
#'
#' @noRd
method(extract_metadata, list(api_openrouter, class_list)) <- function(.api, .response) {
  list(
    model             = .response$model,
    timestamp         = lubridate::as_datetime(.response$created),
    prompt_tokens     = .response$usage$prompt_tokens,
    completion_tokens = .response$usage$completion_tokens,
    total_tokens      = .response$usage$total_tokens,
    stream            = FALSE,
    specific_metadata = list(
      id               = .response$id,
      native_tokens_prompt     = .response$usage$native_tokens_prompt,
      native_tokens_completion = .response$usage$native_tokens_completion,
      cost             = .response$usage$cost
    )
  )
}


#' Send LLM Messages to the OpenRouter Chat API
#'
#' @description
#' Sends a message history to the OpenRouter API, which provides access to hundreds
#' of models from different providers through a single OpenAI-compatible endpoint.
#'
#' @param .llm An `LLMMessage` object containing the conversation history.
#' @param .model The model identifier to use (default: `"google/gemini-2.5-flash"`).
#'   Any model available on OpenRouter can be used.
#' @param .max_tokens Maximum number of tokens in the response (default: 2048).
#' @param .temperature Controls randomness (0–2, optional).
#' @param .top_p Nucleus sampling parameter (0–1, optional).
#' @param .frequency_penalty Penalizes repeated tokens (-2 to 2, optional).
#' @param .presence_penalty Encourages new topics (-2 to 2, optional).
#' @param .stop One or more stop sequences (optional).
#' @param .stream Logical; if TRUE, streams the response (default: FALSE).
#' @param .tools Either a single TOOL object or a list of TOOL objects for tool calls.
#' @param .tool_choice Tool-calling behavior: `"none"`, `"auto"`, or `"required"` (optional).
#' @param .provider A named list of OpenRouter provider preferences, e.g.
#'   `list(order = c("Anthropic", "AWS Bedrock"), allow_fallbacks = TRUE)` (optional).
#' @param .route OpenRouter routing strategy; `"fallback"` routes to the next model if
#'   the primary is unavailable (optional).
#' @param .models A character vector of model IDs to use as fallbacks when the primary
#'   model is unavailable (optional). Used together with `.route = "fallback"`.
#' @param .api_url Base URL for the OpenRouter API (default: `"https://openrouter.ai"`).
#' @param .timeout Request timeout in seconds (default: 60).
#' @param .verbose If TRUE, displays additional information after the API call (default: FALSE).
#' @param .dry_run If TRUE, returns the request object without executing it (default: FALSE).
#' @param .max_tries Maximum retries (default: 3).
#' @param .max_tool_rounds Maximum number of tool use iterations (default: 10).
#' @param .json_schema A JSON schema object for structured output (default: NULL).
#'
#' @return A new `LLMMessage` object containing the original messages plus the assistant's response.
#'
#' @export
openrouter_chat <- function(.llm,
                            .model = "anthropic/claude-sonnet-4-6",
                            .max_tokens = 2048,
                            .temperature = NULL,
                            .top_p = NULL,
                            .frequency_penalty = NULL,
                            .presence_penalty = NULL,
                            .stop = NULL,
                            .stream = FALSE,
                            .json_schema = NULL,
                            .tools = NULL,
                            .tool_choice = NULL,
                            .provider = NULL,
                            .route = NULL,
                            .models = NULL,
                            .api_url = "https://openrouter.ai",
                            .timeout = 60,
                            .verbose = FALSE,
                            .dry_run = FALSE,
                            .max_tries = 3,
                            .max_tool_rounds = 10) {

  c(
    "Input .llm must be an LLMMessage object" = S7_inherits(.llm, LLMMessage),
    "Input .model must be a non-empty string" = is.character(.model) & nzchar(.model),
    "Input .max_tokens must be a positive integer" = is_integer_valued(.max_tokens) & .max_tokens > 0,
    "Input .temperature must be between 0 and 2 if provided" = is.null(.temperature) | (.temperature >= 0 & .temperature <= 2),
    "Input .top_p must be between 0 and 1 if provided" = is.null(.top_p) | (.top_p >= 0 & .top_p <= 1),
    "Input .frequency_penalty must be between -2 and 2 if provided" = is.null(.frequency_penalty) | (.frequency_penalty >= -2 & .frequency_penalty <= 2),
    "Input .presence_penalty must be between -2 and 2 if provided" = is.null(.presence_penalty) | (.presence_penalty >= -2 & .presence_penalty <= 2),
    "Input .stop must be NULL, a string, or a list of strings" = is.null(.stop) | is.character(.stop) | is.list(.stop),
    "Input .stream must be logical" = is.logical(.stream),
    "Input .json_schema must be NULL or a list or an ellmer type object" = is.null(.json_schema) | is.list(.json_schema) | is_ellmer_type(.json_schema),
    "Streaming is not supported for requests with structured outputs" = is.null(.json_schema) || !isTRUE(.stream),
    "Input .tools must be NULL, a TOOL object, or a list of TOOL objects" = is.null(.tools) || S7_inherits(.tools, TOOL) || (is.list(.tools) && all(purrr::map_lgl(.tools, ~ S7_inherits(.x, TOOL)))),
    "Input .tool_choice must be NULL or one of 'none', 'auto', 'required'" = is.null(.tool_choice) | (.tool_choice %in% c("none", "auto", "required")),
    "Streaming is not supported for requests with tool calls" = is.null(.tools) || !isTRUE(.stream),
    "Input .provider must be NULL or a named list" = is.null(.provider) | is.list(.provider),
    "Input .route must be NULL or a character string" = is.null(.route) | is.character(.route),
    "Input .models must be NULL or a character vector" = is.null(.models) | is.character(.models),
    "Input .api_url must be a valid URL" = is.character(.api_url) & nzchar(.api_url),
    "Input .timeout must be a positive integer" = is_integer_valued(.timeout) & .timeout > 0,
    "Input .verbose must be logical" = is.logical(.verbose),
    "Input .dry_run must be logical" = is.logical(.dry_run),
    "Input .max_tries must be integer-valued numeric" = is_integer_valued(.max_tries),
    ".max_tool_rounds must be a positive integer" = is_integer_valued(.max_tool_rounds) && .max_tool_rounds >= 1
  ) |> validate_inputs()

  api_obj <- api_openrouter(short_name = "openrouter",
                            long_name  = "OpenRouter",
                            api_key_env_var = "OPENROUTER_API_KEY")

  messages <- to_api_format(.llm, api_obj, FALSE)

  api_key <- get_api_key(api_obj, .dry_run)

  tools_def <- if (!is.null(.tools)) {
    if (S7_inherits(.tools, TOOL)) list(.tools) else .tools
  } else {
    NULL
  }

  json <- FALSE
  response_format <- NULL
  if (!is.null(.json_schema)) {
    json <- TRUE
    if (requireNamespace("ellmer", quietly = TRUE) && S7_inherits(.json_schema, ellmer::TypeObject)) {
      .json_schema <- to_schema(.json_schema)
    }
    schema_name <- attr(.json_schema, "name") %||% "tidyllm_schema"
    response_format <- list(
      type = "json_schema",
      json_schema = list(name = schema_name, schema = .json_schema)
    )
  }

  request_body <- list(
    model = .model,
    messages = messages,
    max_tokens = .max_tokens,
    temperature = .temperature,
    top_p = .top_p,
    frequency_penalty = .frequency_penalty,
    presence_penalty = .presence_penalty,
    stop = .stop,
    stream = .stream,
    response_format = response_format,
    tools = if (!is.null(tools_def)) tools_to_api(api_obj, tools_def) else NULL,
    tool_choice = .tool_choice,
    provider = .provider,
    route = .route,
    models = if (!is.null(.models)) as.list(.models) else NULL
  ) |> purrr::compact()

  request <- httr2::request(.api_url) |>
    httr2::req_url_path("/api/v1/chat/completions") |>
    httr2::req_headers(
      `Authorization` = sprintf("Bearer %s", api_key),
      `Content-Type` = "application/json"
    ) |>
    httr2::req_body_json(data = request_body)

  if (.dry_run) {
    return(request)
  }

  response <- perform_chat_request(request, api_obj, .stream, .timeout, .max_tries)

  if (.stream == FALSE && !is.null(tools_def)) {
    response <- process_tool_loop(
      .api = api_obj,
      .response = response,
      .tools_def = tools_def,
      .request_body = request_body,
      .request = request,
      .timeout = .timeout,
      .max_tries = .max_tries,
      .max_tool_rounds = .max_tool_rounds
    )
  }

  add_message(.llm     = .llm,
              .role    = "assistant",
              .content = response$assistant_reply,
              .json    = json,
              .meta    = response$meta)
}


#' List Available Models on OpenRouter
#'
#' @description
#' Retrieves the list of models available through the OpenRouter API, including
#' pricing and context window information.
#'
#' @param .api_url Base URL for the OpenRouter API (default: `"https://openrouter.ai"`).
#' @param .timeout Request timeout in seconds (default: 60).
#' @param .max_tries Maximum retries (default: 3).
#' @param .dry_run If TRUE, returns the request object without executing it (default: FALSE).
#' @param .verbose If TRUE, displays additional information (default: FALSE).
#'
#' @return A tibble with columns `id`, `name`, `context_length`,
#'   `prompt_price_per_million`, and `completion_price_per_million`,
#'   or NULL if no models are found.
#'
#' @export
openrouter_list_models <- function(.api_url = "https://openrouter.ai",
                                   .timeout = 60,
                                   .max_tries = 3,
                                   .dry_run = FALSE,
                                   .verbose = FALSE) {
  api_obj <- api_openrouter(short_name = "openrouter",
                            long_name  = "OpenRouter",
                            api_key_env_var = "OPENROUTER_API_KEY")

  api_key <- get_api_key(api_obj, .dry_run)

  request <- httr2::request(.api_url) |>
    httr2::req_url_path("/api/v1/models") |>
    httr2::req_headers(
      Authorization = sprintf("Bearer %s", api_key),
      `Content-Type` = "application/json"
    )

  if (.dry_run) {
    return(request)
  }

  response <- request |>
    httr2::req_timeout(.timeout) |>
    httr2::req_retry(max_tries = .max_tries) |>
    httr2::req_perform() |>
    httr2::resp_body_json()

  if (.verbose) {
    message("Retrieved ", length(response$data), " models from OpenRouter.")
  }

  if (is.null(response$data)) {
    return(NULL)
  }

  models <- response$data

  tibble::tibble(
    id = vapply(models, function(m) m$id %||% NA_character_, character(1)),
    name = vapply(models, function(m) m$name %||% NA_character_, character(1)),
    context_length = vapply(models, function(m) {
      v <- m$context_length
      if (is.null(v)) NA_real_ else as.numeric(v)
    }, numeric(1)),
    prompt_price_per_million = vapply(models, function(m) {
      v <- m$pricing$prompt
      if (is.null(v)) NA_real_ else as.numeric(v) * 1e6
    }, numeric(1)),
    completion_price_per_million = vapply(models, function(m) {
      v <- m$pricing$completion
      if (is.null(v)) NA_real_ else as.numeric(v) * 1e6
    }, numeric(1))
  )
}


#' Get OpenRouter Credit Balance
#'
#' @description
#' Returns the total credits purchased and total usage so far for the current
#' API key.
#'
#' @param .api_url Base URL for the OpenRouter API (default: `"https://openrouter.ai"`).
#' @param .timeout Request timeout in seconds (default: 60).
#' @param .max_tries Maximum retries (default: 3).
#'
#' @return A tibble with columns `total_credits` (USD purchased), `total_usage`
#'   (USD consumed so far), and `remaining` (USD remaining).
#'
#' @export
openrouter_credits <- function(.api_url = "https://openrouter.ai",
                               .timeout = 60,
                               .max_tries = 3) {
  api_obj <- api_openrouter(short_name = "openrouter",
                            long_name  = "OpenRouter",
                            api_key_env_var = "OPENROUTER_API_KEY")
  api_key <- get_api_key(api_obj, FALSE)

  response <- httr2::request(.api_url) |>
    httr2::req_url_path("/api/v1/credits") |>
    httr2::req_headers(Authorization = sprintf("Bearer %s", api_key)) |>
    httr2::req_timeout(.timeout) |>
    httr2::req_retry(max_tries = .max_tries) |>
    httr2::req_perform() |>
    httr2::resp_body_json()

  tibble::tibble(
    total_credits = response$data$total_credits,
    total_usage   = response$data$total_usage,
    remaining     = response$data$total_credits - response$data$total_usage
  )
}


#' Get Details for an OpenRouter Generation
#'
#' @description
#' Fetches cost, latency, and token details for a past generation using its ID.
#' The generation ID is available in `get_metadata(result)$api_specific$id`
#' after a `chat()` call.
#'
#' @param .id The generation ID string (e.g. `"gen-..."`) returned by OpenRouter.
#' @param .api_url Base URL for the OpenRouter API (default: `"https://openrouter.ai"`).
#' @param .timeout Request timeout in seconds (default: 60).
#' @param .max_tries Maximum retries (default: 3).
#'
#' @return A named list with generation details including `id`, `model`,
#'   `total_cost`, `tokens_prompt`, `tokens_completion`, and `latency`.
#'
#' @export
openrouter_generation <- function(.id,
                                  .api_url = "https://openrouter.ai",
                                  .timeout = 60,
                                  .max_tries = 3) {
  c(
    "Input .id must be a non-empty string" = is.character(.id) && nzchar(.id)
  ) |> validate_inputs()

  api_obj <- api_openrouter(short_name = "openrouter",
                            long_name  = "OpenRouter",
                            api_key_env_var = "OPENROUTER_API_KEY")
  api_key <- get_api_key(api_obj, FALSE)

  response <- httr2::request(.api_url) |>
    httr2::req_url_path("/api/v1/generation") |>
    httr2::req_url_query(id = .id) |>
    httr2::req_headers(Authorization = sprintf("Bearer %s", api_key)) |>
    httr2::req_timeout(.timeout) |>
    httr2::req_retry(max_tries = .max_tries) |>
    httr2::req_perform() |>
    httr2::resp_body_json()

  d <- response$data
  list(
    id                = d$id,
    model             = d$model,
    total_cost        = d$total_cost,
    tokens_prompt     = d$tokens_prompt,
    tokens_completion = d$tokens_completion,
    latency           = d$latency,
    provider          = d$provider_name,
    origin            = d$origin
  )
}


#' Generate Embeddings Using the OpenRouter API
#'
#' @description
#' Sends text to an embedding model accessible via OpenRouter and returns
#' embedding vectors. Note that embedding models are not listed in
#' `list_models(openrouter())` — specify the model ID directly.
#' Known supported models include `"openai/text-embedding-3-small"`,
#' `"openai/text-embedding-3-large"`, and `"mistralai/mistral-embed"`.
#'
#' @param .input An `LLMMessage` object or a character vector of texts to embed.
#' @param .model The embedding model ID (default: `"openai/text-embedding-3-small"`).
#' @param .timeout Request timeout in seconds (default: 120).
#' @param .dry_run If TRUE, returns the request object without executing it (default: FALSE).
#' @param .max_tries Maximum retries (default: 3).
#'
#' @return A tibble with columns `input` (text) and `embeddings` (list of numeric vectors).
#'
#' @export
openrouter_embedding <- function(.input,
                                 .model = "openai/text-embedding-3-small",
                                 .timeout = 120,
                                 .dry_run = FALSE,
                                 .max_tries = 3) {
  c(
    "Input .input must be an LLMMessage object or a character vector" = S7_inherits(.input, LLMMessage) | is.character(.input),
    "Input .model must be a non-empty string" = is.character(.model) && nzchar(.model),
    "Input .timeout must be a positive number" = is.numeric(.timeout) && .timeout > 0,
    "Input .dry_run must be logical" = is.logical(.dry_run)
  ) |> validate_inputs()

  api_obj <- api_openrouter(short_name = "openrouter",
                            long_name  = "OpenRouter",
                            api_key_env_var = "OPENROUTER_API_KEY")
  api_key <- get_api_key(api_obj, .dry_run)

  input_texts <- parse_embedding_input(.input)

  request <- httr2::request("https://openrouter.ai") |>
    httr2::req_url_path("/api/v1/embeddings") |>
    httr2::req_headers(
      Authorization  = sprintf("Bearer %s", api_key),
      `Content-Type` = "application/json"
    ) |>
    httr2::req_body_json(list(model = .model, input = input_texts))

  if (.dry_run) {
    return(request)
  }

  extract_embeddings_fn <- function(response_content, error, response_headers) {
    if (error) {
      paste0("OpenRouter embedding error: ", response_content$error$message) |> stop()
    }
    response_content$data |>
      purrr::map("embedding") |>
      purrr::map(unlist)
  }

  perform_embedding_request(
    .request             = request,
    .timeout             = .timeout,
    .max_tries           = .max_tries,
    .input_texts         = input_texts,
    .fn_extract_embeddings = extract_embeddings_fn
  )
}


#' OpenRouter Provider Function
#'
#' @description
#' The `openrouter()` function provides access to hundreds of AI models from
#' different providers through the OpenRouter API, using a single
#' OpenAI-compatible interface.
#'
#' @param ... Parameters passed to the appropriate OpenRouter-specific function.
#' @param .called_from An internal argument specifying which verb invoked this function.
#'   Managed automatically by tidyllm verbs; do not set manually.
#'
#' @return The result of the requested action (e.g., an updated `LLMMessage` for `chat()`).
#'
#' @export
openrouter <- create_provider_function(
  .name = "openrouter",
  chat = openrouter_chat,
  embed = openrouter_embedding,
  list_models = openrouter_list_models
)

structure(list(method = "POST", url = "https://api.openai.com/v1/responses",
    status_code = 200L, headers = structure(list(date = "Sun, 16 Nov 2025 13:05:38 GMT",
        `content-type` = "application/json",
        `openai-version` = "2020-10-01",
        `x-request-id` = "req_mock_schema_single",
        `cf-cache-status` = "DYNAMIC",
        `strict-transport-security` = "max-age=31536000; includeSubDomains; preload",
        `x-content-type-options` = "nosniff",
        server = "cloudflare"), redact = character(0), class = "httr2_headers"),
    body = charToRaw("{\n  \"id\": \"resp_mock_schema_single\",\n  \"object\": \"response\",\n  \"created_at\": 1763298337,\n  \"model\": \"gpt-5.5\",\n  \"output\": [\n    {\n      \"type\": \"message\",\n      \"role\": \"assistant\",\n      \"content\": [\n        {\n          \"type\": \"output_text\",\n          \"text\": \"{\\\"area\\\":42}\"\n        }\n      ]\n    }\n  ],\n  \"usage\": {\n    \"input_tokens\": 53,\n    \"output_tokens\": 18,\n    \"output_tokens_details\": {\n      \"reasoning_tokens\": 0\n    }\n  }\n}\n"),
    cache = new.env(parent = emptyenv())), class = "httr2_response")

structure(list(method = "POST", url = "https://api.openai.com/v1/chat/completions", 
    status_code = 200L, headers = structure(list(date = "Sun, 16 Nov 2025 13:05:38 GMT", 
        `content-type` = "application/json", `access-control-expose-headers` = "X-Request-ID", 
        `openai-organization` = "user-j2dlcmauhfkzofrqjehceoc7", 
        `openai-processing-ms` = "1099", `openai-project` = "proj_wN2FU9W9VUckqc7nz44xte9H", 
        `openai-version` = "2020-10-01", `x-envoy-upstream-service-time` = "1313", 
        `x-ratelimit-limit-requests` = "5000", `x-ratelimit-limit-tokens` = "800000", 
        `x-ratelimit-remaining-requests` = "4999", `x-ratelimit-remaining-tokens` = "799976", 
        `x-ratelimit-reset-requests` = "12ms", `x-ratelimit-reset-tokens` = "1ms", 
        `x-request-id` = "req_ea0f18fede984ed0bb949aa79be4c3b6", 
        `x-openai-proxy-wasm` = "v0.1", `cf-cache-status` = "DYNAMIC", 
        `set-cookie` = "REDACTED", `strict-transport-security` = "max-age=31536000; includeSubDomains; preload", 
        `x-content-type-options` = "nosniff", `set-cookie` = "REDACTED", 
        server = "cloudflare", `cf-ray` = "99f7336e2c45dc7d-FRA", 
        `content-encoding` = "gzip", `alt-svc` = "h3=\":443\"; ma=86400"), redact = character(0), class = "httr2_headers"), 
    body = charToRaw("{\n  \"id\": \"chatcmpl-CcWjhJ3qP05CLu9YpiImmkM8ZlYjs\",\n  \"object\": \"chat.completion\",\n  \"created\": 1763298337,\n  \"model\": \"gpt-5.1-chat-latest\",\n  \"choices\": [\n    {\n      \"index\": 0,\n      \"message\": {\n        \"role\": \"assistant\",\n        \"content\": \"{\\\"area\\\":42}\",\n        \"refusal\": null,\n        \"annotations\": []\n      },\n      \"finish_reason\": \"stop\"\n    }\n  ],\n  \"usage\": {\n    \"prompt_tokens\": 53,\n    \"completion_tokens\": 18,\n    \"total_tokens\": 71,\n    \"prompt_tokens_details\": {\n      \"cached_tokens\": 0,\n      \"audio_tokens\": 0\n    },\n    \"completion_tokens_details\": {\n      \"reasoning_tokens\": 0,\n      \"audio_tokens\": 0,\n      \"accepted_prediction_tokens\": 0,\n      \"rejected_prediction_tokens\": 0\n    }\n  },\n  \"service_tier\": \"default\",\n  \"system_fingerprint\": null\n}\n"), 
    cache = new.env(parent = emptyenv())), class = "httr2_response")

#' The Deepseek API provider class (inherits from OpenAI)
#'
#' @noRd
api_deepseek <- new_class("Deepseek", api_openai)


#' A function to get metadata from Perplexity responses
#'
#' @noRd
method(extract_metadata, list(api_deepseek,class_list))<- function(.api,.response) {
  list(
    model             = .response$model,
    timestamp         = lubridate::as_datetime(.response$created),
    prompt_tokens     = .response$usage$prompt_tokens,
    completion_tokens = .response$usage$completion_tokens,
    total_tokens      = .response$usage$total_tokens,
    stream            = FALSE,
    specific_metadata = list(
      id      = .response$id,
      thinking = .response$choices[[1]]$message$reasoning_content
    )
  )
}  

#' Send LLM Messages to the DeepSeek Chat API
#'
#' @description
#' This function sends a message history to the DeepSeek Chat API and returns the assistant's reply.
#' Currently tool calls cause problems on the DeepSeek API
#'
#' @param .llm An `LLMMessage` object containing the conversation history.
#' @param .model The identifier of the model to use (default: "deepseek-chat").
#' @param .max_tokens The maximum number of tokens that can be generated in the response (default: 2048).
#' @param .temperature Controls the randomness in the model's response. Values between 0 and 2 are allowed (optional).
#' @param .top_p Nucleus sampling parameter that controls the proportion of probability mass considered (optional).
#' @param .frequency_penalty Number between -2.0 and 2.0. Penalizes repeated tokens to reduce repetition (optional).
#' @param .presence_penalty Number between -2.0 and 2.0. Encourages new topics by penalizing tokens that have appeared so far (optional).
#' @param .stop One or more sequences where the API will stop generating further tokens (optional).
#' @param .stream Logical; if TRUE, streams the response piece by piece (default: FALSE).
#' @param .logprobs If TRUE, returns log probabilities of each output token (default: FALSE).
#' @param .top_logprobs Number between 0 and 5 specifying the number of top log probabilities to return (optional).
#' @param .tools Either a single TOOL object or a list of TOOL objects representing the available functions for tool calls.
#' @param .tool_choice A character string specifying the tool-calling behavior; valid values are `"none"`, `"auto"`, or `"required"` (optional).
#' @param .api_url Base URL for the DeepSeek API (default: `"https://api.deepseek.com/"`).
#' @param .timeout Request timeout in seconds (default: 60).
#' @param .verbose If TRUE, displays additional information after the API call (default: FALSE).
#' @param .dry_run If TRUE, returns the constructed request object without executing it (default: FALSE).
#' @param .max_tries Maximum retries to perform the request (default: 3).
#' @param .max_tool_rounds Integer specifying the maximum number of tool use iterations (default: 10).
#'   Set to 1 for single-round tool use, or higher for multi-turn agentic loops.
#' @param .thinking If TRUE, switches to the `deepseek-reasoner` model and captures the reasoning trace (default: NULL).
#'
#' @return A new `LLMMessage` object containing the original messages plus the assistant's response.
#'
#' @export
deepseek_chat <- function(.llm,
                          .model = "deepseek-chat",
                          .thinking = NULL,
                          .max_tokens = 2048,
                          .temperature = NULL,
                          .top_p = NULL,
                          .frequency_penalty = NULL,
                          .presence_penalty = NULL,
                          .stop = NULL,
                          .stream = FALSE,
                          .logprobs = NULL,
                          .top_logprobs = NULL,
                          .tools = NULL,
                          .tool_choice = NULL,
                          .api_url = "https://api.deepseek.com/",
                          .timeout = 60,
                          .verbose = FALSE,
                          .dry_run = FALSE,
                          .max_tries = 3,
                          .max_tool_rounds = 10) {

  # Validate inputs
  c(
    "Input .llm must be an LLMMessage object" = S7_inherits(.llm, LLMMessage),
    "Input .thinking must be NULL or logical" = is.null(.thinking) || is.logical(.thinking),
    "Input .model must be a non-empty string" = is.character(.model) & nzchar(.model),
    "Input .max_tokens must be a positive integer" = is_integer_valued(.max_tokens) & .max_tokens > 0,
    "Input .temperature must be between 0 and 2 if provided" = is.null(.temperature) | (.temperature >= 0 & .temperature <= 2),
    "Input .top_p must be between 0 and 1 if provided" = is.null(.top_p) | (.top_p >= 0 & .top_p <= 1),
    "Input .frequency_penalty must be between -2 and 2 if provided" = is.null(.frequency_penalty) | (.frequency_penalty >= -2 & .frequency_penalty <= 2),
    "Input .presence_penalty must be between -2 and 2 if provided" = is.null(.presence_penalty) | (.presence_penalty >= -2 & .presence_penalty <= 2),
    "Input .stop must be NULL, a string, or a list of strings" = is.null(.stop) | is.character(.stop) | is.list(.stop),
    "Input .stream must be logical" = is.logical(.stream),
    "Input .logprobs must be logical or NULL" = is.null(.logprobs) | is.logical(.logprobs),
    "Input .top_logprobs must be NULL or an integer between 0 and 5" = is.null(.top_logprobs) | (is_integer_valued(.top_logprobs) & .top_logprobs >= 0 & .top_logprobs <= 5),
    "Input .tools must be NULL, a TOOL object, or a list of TOOL objects" = is.null(.tools) || S7_inherits(.tools, TOOL) || (is.list(.tools) && all(purrr::map_lgl(.tools, ~ S7_inherits(.x, TOOL)))),
    "Input .tool_choice must be NULL or a character (one of 'none', 'auto', 'required')" = is.null(.tool_choice) || (is.character(.tool_choice) && .tool_choice %in% c("none", "auto", "required")),
    "Streaming is not supported for requests with tool calls" = is.null(.tools) || !isTRUE(.stream),
    "Streaming is not supported for requests with logprobs" = is.null(.logprobs) || !isTRUE(.stream),
    "Input .tool_choice must be NULL or one of 'none', 'auto', 'required'" = is.null(.tool_choice) | (.tool_choice %in% c("none", "auto", "required")),
    "Input .api_url must be a valid URL" = is.character(.api_url) & nzchar(.api_url),
    "Input .timeout must be a positive integer" = is_integer_valued(.timeout) & .timeout > 0,
    "Input .verbose must be logical" = is.logical(.verbose),
    "Input .dry_run must be logical" = is.logical(.dry_run),
    "Input .max_tries must be integer-valued numeric" = is_integer_valued(.max_tries),
    ".max_tool_rounds must be a positive integer" = is_integer_valued(.max_tool_rounds) && .max_tool_rounds >= 1
  ) |> validate_inputs()
  
  if (isTRUE(.thinking)) .model <- "deepseek-reasoner"

  api_obj <- api_deepseek(short_name = "deepseek",
                          long_name  = "DeepSeek",
                          api_key_env_var = "DEEPSEEK_API_KEY")
  
  messages <- to_api_format(.llm,api_obj,FALSE)
  
  api_key <- get_api_key(api_obj, .dry_run)
  
  #Put a single tool into a list if only one is provided. 
  tools_def <- if (!is.null(.tools)) {
    if (S7_inherits(.tools, TOOL))  list(.tools) else .tools
  } else {
    NULL
  }
  
  # Construct request body
  request_body <- list(
    model = .model,
    messages = messages,
    max_tokens = .max_tokens,
    temperature = .temperature,
    top_p = .top_p,
    frequency_penalty = .frequency_penalty,
    presence_penalty = .presence_penalty,
    stop = .stop,
    stream = .stream,
    logprobs = .logprobs,
    top_logprobs = .top_logprobs,
    tools = if(!is.null(tools_def)) tools_to_api(api_obj,tools_def) else NULL,
    tool_choice = .tool_choice
  ) |> purrr::compact()
  
  request <- httr2::request(.api_url) |>
    httr2::req_url_path("/chat/completions") |>
    httr2::req_headers(
      `Authorization` = sprintf("Bearer %s", api_key),
      `Content-Type` = "application/json"
    ) |>
    httr2::req_body_json(data = request_body)
  
  if (.dry_run) {
    return(request)
  }
  
  response <- perform_chat_request(request, api_obj, .stream, .timeout, .max_tries)
  
  # Handle tool calls with multi-turn support
  if (.stream == FALSE && !is.null(tools_def)) {
    response <- process_tool_loop(
      .api = api_obj,
      .response = response,
      .tools_def = tools_def,
      .request_body = request_body,
      .request = request,
      .timeout = .timeout,
      .max_tries = .max_tries,
      .max_tool_rounds = .max_tool_rounds
    )
  }
  
  assistant_reply <- response$assistant_reply

  logprobs  <- parse_logprobs(api_obj, response$raw)
  
  
  add_message(.llm     = .llm,
              .role    = "assistant", 
              .content = assistant_reply , 
              .json    = FALSE,
              .meta    = response$meta,
              .logprobs = logprobs)
}

#' Deepseek Provider Function
#'
#' The `deepseek()` function acts as a provider interface for interacting with the Deepseek API 
#' through `tidyllm`'s `chat()` verb.
#' It dynamically routes requests to deepseek-specific function. At the moment this is only
#' `deepseek_chat()`
#'
#' @param ... Parameters to be passed to the appropriate Deepseek-specific function, 
#'   such as model configuration, input text, or API-specific options.
#' @param .called_from An internal argument specifying which action (e.g., 
#'   `chat`, `embed`) the function is invoked from. 
#'   This argument is automatically managed by the `tidyllm` verbs and should not be modified by the user.
#'
#' @return The result of the requested action, depending on the specific function invoked 
#'   (e.g., an updated `LLMMessage` object for `chat()`).
#'
#' @export
deepseek <- create_provider_function(
  .name = "deepseek",
  chat = deepseek_chat
)

structure(list(method = "POST", url = "http://localhost:11434/api/chat", 
    status_code = 200L, headers = structure(list(`Content-Type` = "application/json; charset=utf-8", 
        Date = "Wed, 16 Oct 2024 00:52:07 GMT", `Content-Length` = "438"), class = "httr2_headers"), 
    body = charToRaw("{\"model\":\"llama3\",\"created_at\":\"2024-10-16T00:52:07.446423Z\",\"message\":{\"role\":\"assistant\",\"content\":\"{\\n\\\"street\\\": \\\"Mühlenstraße\\\",\\n\\\"houseNumber\\\": \\\"12-14\\\",\\n\\\"city\\\": \\\"Berlin\\\",\\n\\\"postalcode\\\": \\\"10117\\\",\\n\\\"country\\\": \\\"Deutschland\\\"\\n}\"},\"done_reason\":\"stop\",\"done\":true,\"total_duration\":4540468000,\"load_duration\":39213083,\"prompt_eval_count\":24,\"prompt_eval_duration\":2674874000,\"eval_count\":43,\"eval_duration\":1825018000}"), 
    cache = new.env(parent = emptyenv())), class = "httr2_response")

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @importFrom grDevices dev.copy
#' @importFrom grDevices dev.off
#' @importFrom grDevices png
#' @importFrom tibble as_tibble
#' @import S7
## usethis namespace: end
NULL

#' The OpenAI Responses API provider class
#'
#' @noRd
api_openai <- new_class("OpenAI", APIProvider)

#' Convert LLMMessage to OpenAI Responses API format
#'
#' System messages go to top-level `instructions`; all other messages go to `input`.
#'
#' @noRd
method(to_api_format, list(LLMMessage, api_openai)) <- function(.llm, .api, .no_system = FALSE) {
  history <- .llm@message_history

  # Collect system content into instructions string
  system_messages <- Filter(function(m) m$role == "system", history)
  instructions <- if (length(system_messages) > 0 && !.no_system) {
    paste(sapply(system_messages, function(m) format_message(m)$content), collapse = "\n")
  } else {
    NULL
  }

  non_system <- Filter(function(m) m$role != "system", history)

  input <- lapply(non_system, function(m) {
    formatted <- format_message(m)
    content_parts <- list()

    # Text part
    content_parts <- c(content_parts, list(list(type = "input_text", text = formatted$content)))

    # Inline images (multi-image)
    for (img_struct in formatted$images) {
      content_parts <- c(content_parts, list(list(
        type      = "input_image",
        image_url = glue::glue("data:{img_struct$media_type};base64,{img_struct$data}")
      )))
    }

    # Remote file references (tidyllm_file, provider == "openai")
    if (!is.null(m$files)) {
      for (f in m$files) {
        if (f@provider != "openai") next
        content_parts <- c(content_parts, list(list(type = "input_file", file_id = f@id)))
      }
    }

    if (length(content_parts) == 1 && content_parts[[1]]$type == "input_text") {
      list(role = m$role, content = content_parts[[1]]$text)
    } else {
      list(role = m$role, content = content_parts)
    }
  })

  list(instructions = instructions, input = input)
}

#' Extract rate limit info from OpenAI Responses API headers
#'
#' @noRd
method(ratelimit_from_header, list(api_openai, new_S3_class("httr2_headers"))) <- function(.api, .headers) {
  if (!any(grepl("x-ratelimit", names(.headers)))) return(NULL)

  request_time <- strptime(.headers["date"]$date, format = "%a, %d %b %Y %H:%M:%S", tz = "GMT")

  ratelimit_requests_reset_dt <- parse_duration_to_seconds(
    .headers["x-ratelimit-reset-requests"]$`x-ratelimit-reset-requests`)
  ratelimit_tokens_reset_dt <- parse_duration_to_seconds(
    .headers["x-ratelimit-reset-tokens"]$`x-ratelimit-reset-tokens`)

  list(
    this_request_time              = request_time,
    ratelimit_requests             = as.integer(.headers["x-ratelimit-limit-requests"]),
    ratelimit_requests_remaining   = as.integer(.headers["x-ratelimit-remaining-requests"]),
    ratelimit_requests_reset_time  = request_time + ratelimit_requests_reset_dt,
    ratelimit_tokens               = as.integer(.headers["x-ratelimit-limit-tokens"]),
    ratelimit_tokens_remaining     = as.integer(.headers["x-ratelimit-remaining-tokens"]),
    ratelimit_tokens_reset_time    = request_time + ratelimit_tokens_reset_dt
  )
}

#' Parse chat response from OpenAI Responses API
#'
#' @noRd
method(parse_chat_response, list(api_openai, class_list)) <- function(.api, .content) {
  if (!is.null(.content$error)) {
    sprintf("OpenAI returned an Error:\nType: %s\nMessage: %s",
            .content$error$type, .content$error$message) |> stop()
  }

  # Walk output[] for type == "message"; skip reasoning and function_call items
  message_items <- Filter(function(x) !is.null(x$type) && x$type == "message", .content$output)

  # When the model issued tool calls, output[] has no message item yet — return NULL
  # so the tool loop in process_tool_loop() can take over.
  if (length(message_items) == 0) return(NULL)

  content_blocks <- message_items[[1]]$content
  text_blocks <- Filter(function(b) !is.null(b$type) && b$type == "output_text", content_blocks)

  if (length(text_blocks) == 0) return(NULL)

  text_blocks[[1]]$text
}

#' Extract metadata from OpenAI Responses API response
#'
#' @noRd
method(extract_metadata, list(api_openai, class_list)) <- function(.api, .response) {
  usage <- .response$usage
  list(
    model             = .response$model,
    timestamp         = lubridate::as_datetime(lubridate::now()),
    prompt_tokens     = usage$input_tokens,
    completion_tokens = usage$output_tokens,
    total_tokens      = (usage$input_tokens %||% 0L) + (usage$output_tokens %||% 0L),
    stream            = FALSE,
    specific_metadata = list(
      response_id      = .response$id,
      reasoning_tokens = usage$output_tokens_details$reasoning_tokens
    )
  )
}

#' Extract metadata from OpenAI Responses API streaming data
#'
#' @noRd
method(extract_metadata_stream, list(api_openai, class_list)) <- function(.api, .stream_raw_data) {
  # The response.completed event contains the full response object
  completed <- purrr::keep(.stream_raw_data, ~ !is.null(.x$type) && .x$type == "response.completed")

  if (length(completed) == 0) {
    return(callNextMethod())
  }

  resp <- completed[[1]]$response
  usage <- resp$usage %||% list()

  list(
    model             = resp$model,
    timestamp         = lubridate::as_datetime(lubridate::now()),
    prompt_tokens     = usage$input_tokens,
    completion_tokens = usage$output_tokens,
    total_tokens      = (usage$input_tokens %||% 0L) + (usage$output_tokens %||% 0L),
    stream            = TRUE,
    specific_metadata = list(
      response_id      = resp$id,
      reasoning_tokens = usage$output_tokens_details$reasoning_tokens
    )
  )
}

#' Handle SSE streaming for OpenAI Responses API
#'
#' The Responses API embeds the event type as `type` inside the JSON data field
#' rather than using SSE `event:` headers.
#'
#' @noRd
method(handle_stream, list(api_openai, new_S3_class("httr2_response"))) <- function(.api, .stream_response) {
  stream_text <- ""
  stream_data <- list()

  repeat {
    stream_chunk <- httr2::resp_stream_sse(.stream_response)

    # NULL means no data available yet — keep reading
    if (is.null(stream_chunk)) next

    data_str <- stream_chunk$data
    if (is.null(data_str) || !nzchar(data_str)) next

    parsed <- tryCatch(
      jsonlite::fromJSON(data_str, simplifyVector = FALSE),
      error = function(e) NULL
    )
    if (is.null(parsed)) next

    event_type <- parsed$type %||% ""

    if (event_type == "response.output_text.delta") {
      delta <- parsed$delta
      if (!is.null(delta) && nzchar(delta)) {
        stream_text <- paste0(stream_text, delta)
        cat(delta)
        utils::flush.console()
      }
    } else if (event_type == "response.completed") {
      stream_data <- append(stream_data, list(parsed))
      close(.stream_response)
      message("\n---------\nStream finished\n---------\n")
      break
    } else if (event_type == "response.failed") {
      close(.stream_response)
      stop("OpenAI Responses API stream failed")
    }
  }

  list(reply = stream_text, raw_data = stream_data)
}

#' Convert TOOL list to OpenAI Responses API flat tool schema
#'
#' @noRd
method(tools_to_api, list(api_openai, class_list)) <- function(.api, .tools) {
  purrr::map(.tools, function(tool) {
    if (length(tool@builtin) > 0) {
      return(tool@builtin[[1]])
    }
    list(
      type        = "function",
      name        = tool@name,
      description = tool@description,
      parameters  = list(
        type                 = "object",
        properties           = purrr::map(tool@input_schema, field_to_param_schema),
        required             = as.list(names(tool@input_schema)),
        additionalProperties = FALSE
      ),
      strict      = TRUE
    )
  })
}

#' Check if Responses API response contains custom function_call items
#'
#' @noRd
method(has_tool_calls, list(api_openai, class_any)) <- function(.api, .response) {
  output <- .response$raw$content$output %||% list()
  any(purrr::map_lgl(output, ~ !is.null(.x$type) && .x$type == "function_call"))
}

#' Extract function_call items from Responses API output
#'
#' @noRd
method(extract_tool_calls, list(api_openai, class_any)) <- function(.api, .response) {
  output <- .response$raw$content$output %||% list()
  Filter(function(x) !is.null(x$type) && x$type == "function_call", output)
}

#' Execute tool calls and return function_call_output items
#'
#' @noRd
method(run_tool_calls, list(api_openai, class_list, class_list)) <- function(.api, .tool_calls, .tools) {
  parse_json_args <- function(json_str) {
    tryCatch(jsonlite::fromJSON(json_str, simplifyVector = TRUE), error = function(e) NULL)
  }

  purrr::map(.tool_calls, function(tc) {
    tool_name <- tc$name
    tool_args <- parse_json_args(tc$arguments)
    call_id   <- tc$call_id

    if (is.null(tool_args)) {
      warning(sprintf("Failed to parse arguments for tool: %s", tool_name))
      return(NULL)
    }

    matching <- purrr::keep(.tools, ~ .x@name == tool_name)
    if (length(matching) == 0) {
      warning(sprintf("No matching tool found for: %s", tool_name))
      return(NULL)
    }

    result <- do.call(matching[[1]]@func, as.list(tool_args))

    list(
      type    = "function_call_output",
      call_id = call_id,
      output  = jsonlite::toJSON(result, auto_unbox = TRUE)
    )
  }) |> purrr::compact()
}

#' Append tool messages to Responses API request input array
#'
#' Carries the model's full output[] (including server-tool items) and the
#' function_call_output items back into input for the next round.
#'
#' @noRd
method(append_tool_messages, list(api_openai, class_any, class_any, class_any)) <-
  function(.api, .request_body, .response, .tool_results) {
    model_output <- .response$raw$content$output %||% list()
    .request_body$input <- c(.request_body$input, model_output, .tool_results)
    .request_body
  }

#' Prepare OpenAI Responses API request body parameters
#'
#' @noRd
add_no_additional_properties <- function(schema) {
  if (is.list(schema) && identical(schema$type, "object")) {
    schema$additionalProperties <- FALSE
    if (!is.null(schema$properties)) {
      schema$properties <- lapply(schema$properties, add_no_additional_properties)
    }
  } else if (is.list(schema) && !is.null(schema$items)) {
    schema$items <- add_no_additional_properties(schema$items)
  }
  schema
}

last_openai_response_id <- function(.llm) {
  assistant_msgs <- Filter(function(m) m$role == "assistant", .llm@message_history)
  if (length(assistant_msgs) == 0) return(NULL)
  utils::tail(assistant_msgs, 1)[[1]]$meta$specific_metadata$response_id
}

prepare_responses_request <- function(.llm,
                                      .api,
                                      .model         = "gpt-5.5",
                                      .max_output_tokens = NULL,
                                      .temperature   = NULL,
                                      .reasoning_effort = NULL,
                                      .json_schema   = NULL,
                                      .seed          = NULL) {
  fmt <- to_api_format(.llm, .api)

  # Handle JSON schema → text.format
  text_format <- NULL
  json <- FALSE
  if (!is.null(.json_schema)) {
    json <- TRUE
    schema_name <- "tidyllm_schema"
    if (requireNamespace("ellmer", quietly = TRUE) && S7_inherits(.json_schema, ellmer::TypeObject)) {
      .json_schema <- to_schema(.json_schema)
      schema_name <- "ellmer_schema"
    }
    if (!is.null(attr(.json_schema, "name"))) schema_name <- attr(.json_schema, "name")
    .json_schema <- add_no_additional_properties(.json_schema)
    text_format <- list(
      format = list(
        type        = "json_schema",
        strict      = TRUE,
        name        = schema_name,
        schema      = .json_schema
      )
    )
  }

  request_body <- list(
    model              = .model,
    input              = fmt$input,
    instructions       = fmt$instructions,
    max_output_tokens  = .max_output_tokens,
    temperature        = .temperature,
    seed               = .seed,
    text               = text_format
  ) |> purrr::compact()

  if (!is.null(.reasoning_effort)) {
    request_body$reasoning <- list(effort = .reasoning_effort)
  }

  list(request_body = request_body, json = json)
}

#' Send LLM Messages to the OpenAI Responses API
#'
#' @description
#' Sends a message history to the OpenAI Responses API (`POST /v1/responses`) and
#' returns the assistant's reply. Supports streaming, tool use, structured output,
#' and reasoning models (o-series) via `.reasoning_effort`.
#'
#' @param .llm An `LLMMessage` object containing the conversation history.
#' @param .model The model identifier (default: `"gpt-5.5"`).
#' @param .max_output_tokens Maximum tokens to generate (caps reasoning + completion).
#' @param .temperature Sampling temperature (0-2).
#' @param .seed Seed for deterministic sampling.
#' @param .stream If TRUE, stream output to console (default: FALSE).
#' @param .timeout Request timeout in seconds (default: 60).
#' @param .verbose Print rate-limit info after the call (default: FALSE).
#' @param .json_schema A tidyllm schema or ellmer type for structured output.
#' @param .max_tries Maximum retry attempts (default: 3).
#' @param .dry_run If TRUE, return the request object without sending (default: FALSE).
#' @param .reasoning_effort For o-series models: `"low"`, `"medium"`, or `"high"`.
#' @param .tools A TOOL object or list of TOOL objects for function calling.
#' @param .tool_choice Tool selection behavior: `"none"`, `"auto"`, or `"required"`.
#' @param .max_tool_rounds Maximum tool-loop iterations (default: 10).
#' @param .stateful If `TRUE`, use `previous_response_id` to pass conversation context
#'   server-side instead of re-sending full message history. Falls back to full-history
#'   mode silently when no previous response ID is available (first turn or prior turn
#'   used a different provider), and with a warning if the server rejects the stored ID
#'   (e.g., expired context). Default: `FALSE`.
#'
#' @return A new `LLMMessage` object with the assistant's response appended.
#' @export
openai_chat <- function(
    .llm,
    .model               = "gpt-5.5",
    .max_output_tokens   = NULL,
    .temperature         = NULL,
    .seed                = NULL,
    .stream              = FALSE,
    .timeout             = 60,
    .verbose             = FALSE,
    .json_schema         = NULL,
    .max_tries           = 3,
    .dry_run             = FALSE,
    .reasoning_effort    = NULL,
    .tools               = NULL,
    .tool_choice         = NULL,
    .max_tool_rounds     = 10,
    .stateful            = FALSE
) {
  c(
    "Input .llm must be an LLMMessage object"                                             = S7_inherits(.llm, LLMMessage),
    "Input .model must be a string"                                                       = is.character(.model),
    "Input .max_output_tokens must be NULL or a positive integer"                         = is.null(.max_output_tokens) | (is_integer_valued(.max_output_tokens) & .max_output_tokens > 0),
    "Input .temperature must be numeric or NULL"                                          = is.null(.temperature) | is.numeric(.temperature),
    "Input .seed must be NULL or an integer"                                              = is.null(.seed) | is_integer_valued(.seed),
    "Input .stream must be logical"                                                       = is.logical(.stream),
    "Input .timeout must be integer-valued numeric"                                       = is_integer_valued(.timeout),
    "Input .verbose must be logical"                                                      = is.logical(.verbose),
    "Input .json_schema must be NULL, a list, or an ellmer type object"                   = is.null(.json_schema) | is.list(.json_schema) | is_ellmer_type(.json_schema),
    "Input .max_tries must be integer-valued numeric"                                     = is_integer_valued(.max_tries),
    "Input .dry_run must be logical"                                                      = is.logical(.dry_run),
    "Input .reasoning_effort must be NULL or one of 'low', 'medium', 'high'"              = is.null(.reasoning_effort) | (.reasoning_effort %in% c("low", "medium", "high")),
    "Input .tools must be NULL, a TOOL object, or a list of TOOL objects"                 = is.null(.tools) || S7_inherits(.tools, TOOL) || (is.list(.tools) && all(purrr::map_lgl(.tools, ~ S7_inherits(.x, TOOL)))),
    "Input .tool_choice must be NULL or one of 'none', 'auto', 'required'"                = is.null(.tool_choice) || (.tool_choice %in% c("none", "auto", "required")),
    ".max_tool_rounds must be a positive integer"                                         = is_integer_valued(.max_tool_rounds) && .max_tool_rounds >= 1,
    "Input .stateful must be logical"                                                     = is.logical(.stateful),
    "Streaming is not supported for requests with tool calls"                             = is.null(.tools) || !isTRUE(.stream)
  ) |> validate_inputs()

  api_obj <- api_openai(
    short_name      = "openai",
    long_name       = "OpenAI",
    api_key_env_var = "OPENAI_API_KEY"
  )

  api_key <- get_api_key(api_obj, .dry_run)

  request_data <- prepare_responses_request(
    .llm              = .llm,
    .api              = api_obj,
    .model            = .model,
    .max_output_tokens = .max_output_tokens,
    .temperature      = .temperature,
    .reasoning_effort = .reasoning_effort,
    .json_schema      = .json_schema,
    .seed             = .seed
  )

  request_body <- request_data$request_body
  json         <- request_data$json

  # Stateful mode: replace full input[] with only the last user message
  stateful_active <- FALSE
  if (isTRUE(.stateful)) {
    prev_id <- last_openai_response_id(.llm)
    if (!is.null(prev_id)) {
      user_msgs <- Filter(function(m) m$role == "user", .llm@message_history)
      if (length(user_msgs) > 0) {
        last_user <- utils::tail(user_msgs, 1)[[1]]
        formatted <- format_message(last_user)
        last_input <- if (!is.null(formatted$image)) {
          list(list(
            role    = "user",
            content = list(
              list(type = "input_text", text = formatted$content),
              list(type = "input_image",
                   image_url = glue::glue(
                     "data:{formatted$image$media_type};base64,{formatted$image$data}"))
            )
          ))
        } else {
          list(list(role = "user", content = formatted$content))
        }
        request_body$input                <- last_input
        request_body$instructions         <- NULL
        request_body$previous_response_id <- prev_id
        stateful_active <- TRUE
      }
    }
  }

  # Tools
  tools_def <- if (!is.null(.tools)) {
    if (S7_inherits(.tools, TOOL)) list(.tools) else .tools
  } else {
    NULL
  }

  if (!is.null(tools_def)) {
    request_body$tools        <- tools_to_api(api_obj, tools_def)
    request_body$tool_choice  <- .tool_choice
  }

  if (isTRUE(.stream)) {
    request_body$stream <- TRUE
  }

  request <- httr2::request("https://api.openai.com/v1/responses") |>
    httr2::req_headers(
      Authorization  = sprintf("Bearer %s", api_key),
      `Content-Type` = "application/json"
    ) |>
    httr2::req_body_json(data = request_body)

  if (.dry_run) return(request)

  response <- if (stateful_active) {
    tryCatch(
      perform_chat_request(request, api_obj, .stream, .timeout, .max_tries),
      error = function(e) {
        warning(
          "OpenAI stateful mode failed (stored context may have expired); ",
          "retrying with full message history. Original error: ", conditionMessage(e),
          call. = FALSE
        )
        full_body <- request_data$request_body
        if (!is.null(tools_def)) {
          full_body$tools       <- tools_to_api(api_obj, tools_def)
          full_body$tool_choice <- .tool_choice
        }
        if (isTRUE(.stream)) full_body$stream <- TRUE
        fallback_req <- httr2::request("https://api.openai.com/v1/responses") |>
          httr2::req_headers(
            Authorization  = sprintf("Bearer %s", api_key),
            `Content-Type` = "application/json"
          ) |>
          httr2::req_body_json(data = full_body)
        perform_chat_request(fallback_req, api_obj, .stream, .timeout, .max_tries)
      }
    )
  } else {
    perform_chat_request(request, api_obj, .stream, .timeout, .max_tries)
  }

  if (!isTRUE(.stream) && !is.null(tools_def)) {
    response <- process_tool_loop(
      .api           = api_obj,
      .response      = response,
      .tools_def     = tools_def,
      .request_body  = request_body,
      .request       = request,
      .timeout       = .timeout,
      .max_tries     = .max_tries,
      .max_tool_rounds = .max_tool_rounds
    )
  }

  track_rate_limit(api_obj, response$headers, .verbose)

  add_message(
    .llm     = .llm,
    .role    = "assistant",
    .content = response$assistant_reply,
    .json    = json,
    .meta    = response$meta
  )
}

#' OpenAI built-in web search tool (server-executed)
#'
#' @return A TOOL object declaring the web_search_preview built-in.
#' @export
openai_websearch <- function() {
  TOOL(
    name         = "web_search",
    description  = "Builtin: web_search_preview",
    input_schema = list(),
    func         = function() NULL,
    builtin      = list(list(type = "web_search_preview"))
  )
}

#' OpenAI built-in code interpreter tool (server-executed)
#'
#' @return A TOOL object declaring the code_interpreter built-in.
#' @export
openai_code_interpreter <- function() {
  TOOL(
    name         = "code_interpreter",
    description  = "Builtin: code_interpreter",
    input_schema = list(),
    func         = function() NULL,
    builtin      = list(list(type = "code_interpreter", container = list(type = "auto")))
  )
}

# openai_file_search() is intentionally not implemented.
#
# The file_search built-in tool queries OpenAI vector stores, but tidyllm has no
# vector store management: no way to create a store, add files to one, or list them.
# Without that, file_search would be a dangling feature requiring out-of-tidyllm setup.
# It will be added alongside vector store verbs (create_vector_store, add_to_vector_store,
# etc.) in a future release so the full workflow is self-contained.

#' Submit a Deep Research Request to OpenAI
#'
#' Sends a research request to OpenAI using the deep research models
#' (`o3-deep-research` or `o4-mini-deep-research`) via the Responses API with
#' `background: true`. The model autonomously searches the web and synthesises
#' a long-form answer, which can take 5-30 minutes.
#'
#' @param .llm An `LLMMessage` object containing the research question.
#' @param .model The deep research model to use (default: `"o4-mini-deep-research"`).
#' @param .background Logical; if `TRUE`, returns a `tidyllm_research_job` immediately
#'   without waiting for completion (default: `FALSE`).
#' @param .reasoning_effort Reasoning level for the model: `"low"`, `"medium"` (default),
#'   or `"high"`.
#' @param .json_schema A tidyllm schema for structured JSON output (optional).
#' @param .max_output_tokens Maximum tokens to generate (default: `NULL` for model default).
#' @param .timeout Seconds to wait in blocking mode before giving up (default: `1800`).
#' @param .max_tries Maximum retries per HTTP request (default: `3`).
#'
#' @return If `.background = FALSE`, an updated `LLMMessage` with the research reply.
#'   If `.background = TRUE`, a `tidyllm_research_job` for use with `check_job()`/`fetch_job()`.
#' @export
openai_deep_research <- function(.llm,
                                 .model              = "o4-mini-deep-research",
                                 .background         = FALSE,
                                 .reasoning_effort   = "medium",
                                 .json_schema        = NULL,
                                 .max_output_tokens  = NULL,
                                 .timeout            = 1800,
                                 .max_tries          = 3) {
  c(
    "Input .llm must be an LLMMessage object"                    = S7_inherits(.llm, LLMMessage),
    "Input .background must be logical"                          = is.logical(.background),
    "Input .reasoning_effort must be 'low', 'medium', or 'high'" = .reasoning_effort %in% c("low", "medium", "high"),
    "Input .json_schema must be NULL or a list"                  = is.null(.json_schema) | is.list(.json_schema),
    "Input .timeout must be a positive number"                   = is.numeric(.timeout) && .timeout > 0,
    "Input .max_tries must be a positive integer"                = is_integer_valued(.max_tries) && .max_tries > 0
  ) |> validate_inputs()

  api_key <- Sys.getenv("OPENAI_API_KEY")
  c("OPENAI_API_KEY environment variable is not set" = nzchar(api_key)) |> validate_inputs()

  api_obj <- api_openai(short_name = "openai", long_name = "OpenAI", api_key_env_var = "OPENAI_API_KEY")

  request_data <- prepare_responses_request(
    .llm               = .llm,
    .api               = api_obj,
    .model             = .model,
    .max_output_tokens = .max_output_tokens,
    .reasoning_effort  = .reasoning_effort,
    .json_schema       = .json_schema
  )

  request_body <- request_data$request_body
  request_body$background <- TRUE
  request_body$tools <- list(list(type = "web_search_preview"))

  response <- httr2::request("https://api.openai.com/v1/responses") |>
    httr2::req_headers(
      Authorization  = sprintf("Bearer %s", api_key),
      `Content-Type` = "application/json"
    ) |>
    httr2::req_body_json(data = request_body) |>
    httr2::req_retry(max_tries = .max_tries) |>
    httr2::req_perform() |>
    httr2::resp_body_json()

  job_id <- response$id
  job <- structure(
    list(job_id = job_id, message = .llm, json = isTRUE(request_data$json), provider = "openai"),
    class = "tidyllm_research_job"
  )

  if (.background) return(job)

  openai_fetch_research(openai_poll_research(job, api_key, .timeout, .max_tries))
}


#' Poll an OpenAI Background Response Until Completion
#'
#' @noRd
openai_poll_research <- function(.job, .api_key, .timeout = 1800, .max_tries = 3) {
  start_time <- proc.time()[["elapsed"]]
  repeat {
    status_response <- httr2::request(
      sprintf("https://api.openai.com/v1/responses/%s", .job$job_id)) |>
      httr2::req_headers(Authorization = sprintf("Bearer %s", .api_key)) |>
      httr2::req_retry(max_tries = .max_tries) |>
      httr2::req_perform() |>
      httr2::resp_body_json()

    if (identical(status_response$status, "completed")) {
      .job$response <- status_response
      return(.job)
    }

    if (status_response$status %in% c("failed", "cancelled", "incomplete")) {
      stop(sprintf("OpenAI deep research job '%s' ended with status: %s",
                   .job$job_id, status_response$status))
    }

    elapsed <- proc.time()[["elapsed"]] - start_time
    if (elapsed > .timeout) {
      stop(sprintf("OpenAI deep research job '%s' did not complete within %d seconds.",
                   .job$job_id, .timeout))
    }
    Sys.sleep(10)
  }
}


#' Check the Status of an OpenAI Background Research Job
#'
#' Polls the status of an OpenAI background response created by
#' `openai_deep_research(.background = TRUE)`.
#'
#' @param .job A `tidyllm_research_job` object returned by `openai_deep_research(.background = TRUE)`.
#' @param .max_tries Maximum retries per HTTP request (default: `3`).
#'
#' @return An updated `tidyllm_research_job` with `$status` set. If completed, `$response`
#'   is also populated and ready for `fetch_job()`.
#' @export
openai_check_research <- function(.job, .max_tries = 3) {
  c(
    "Input .job must be a tidyllm_research_job" = inherits(.job, "tidyllm_research_job"),
    "Input .job must be an OpenAI research job"  = identical(.job$provider, "openai")
  ) |> validate_inputs()

  api_key <- Sys.getenv("OPENAI_API_KEY")
  c("OPENAI_API_KEY environment variable is not set" = nzchar(api_key)) |> validate_inputs()

  status_response <- httr2::request(
    sprintf("https://api.openai.com/v1/responses/%s", .job$job_id)) |>
    httr2::req_headers(Authorization = sprintf("Bearer %s", api_key)) |>
    httr2::req_retry(max_tries = .max_tries) |>
    httr2::req_perform() |>
    httr2::resp_body_json()

  .job$status <- status_response$status
  if (identical(status_response$status, "completed")) {
    .job$response <- status_response
  }
  .job
}


#' Fetch Results from a Completed OpenAI Deep Research Job
#'
#' Extracts the assistant reply from a completed `tidyllm_research_job` returned
#' by `openai_deep_research(.background = TRUE)`.
#'
#' @param .job A `tidyllm_research_job` object. If not yet completed, polls once via
#'   `openai_check_research()` and errors if still incomplete.
#' @param .max_tries Maximum retries per HTTP request (default: `3`).
#'
#' @return An updated `LLMMessage` with the research reply appended.
#' @export
openai_fetch_research <- function(.job, .max_tries = 3) {
  c(
    "Input .job must be a tidyllm_research_job" = inherits(.job, "tidyllm_research_job"),
    "Input .job must be an OpenAI research job"  = identical(.job$provider, "openai")
  ) |> validate_inputs()

  if (is.null(.job$response)) {
    .job <- openai_check_research(.job, .max_tries)
    if (!identical(.job$status, "completed")) {
      stop(sprintf("OpenAI research job '%s' is not yet completed (status: %s).",
                   .job$job_id, .job$status))
    }
  }

  api_obj <- api_openai(short_name = "openai", long_name = "OpenAI", api_key_env_var = "OPENAI_API_KEY")
  assistant_reply <- parse_chat_response(api_obj, .job$response)
  meta            <- extract_metadata(api_obj, .job$response)

  add_message(
    .llm     = .job$message,
    .role    = "assistant",
    .content = assistant_reply,
    .meta    = meta,
    .json    = isTRUE(.job$json)
  )
}


#' OpenAI Provider Function
#'
#' Upload a file to OpenAI's Files API
#' @noRd
openai_upload_file <- function(.path, .purpose = "assistants", .called_from = NULL, ...) {
  api_obj <- api_openai(short_name = "openai", long_name = "OpenAI", api_key_env_var = "OPENAI_API_KEY")
  api_key <- get_api_key(api_obj)

  resp <- httr2::request("https://api.openai.com/v1/files") |>
    httr2::req_auth_bearer_token(api_key) |>
    httr2::req_body_multipart(
      purpose = .purpose,
      file = curl::form_file(.path)
    ) |>
    httr2::req_perform() |>
    httr2::resp_body_json()

  tidyllm_file(
    id        = resp$id,
    provider  = "openai",
    mime_type = guess_mime_type(.path),
    filename  = resp$filename %||% basename(.path),
    uri       = ""
  )
}

#' List files on OpenAI's Files API
#' @noRd
openai_list_files <- function(.called_from = NULL, ...) {
  api_obj <- api_openai(short_name = "openai", long_name = "OpenAI", api_key_env_var = "OPENAI_API_KEY")
  api_key <- get_api_key(api_obj)

  resp <- httr2::request("https://api.openai.com/v1/files") |>
    httr2::req_auth_bearer_token(api_key) |>
    httr2::req_perform() |>
    httr2::resp_body_json()

  if (is.null(resp$data) || length(resp$data) == 0) {
    return(tibble::tibble(file_id=character(), filename=character(), mime_type=character(),
                          size_bytes=numeric(), created_at=character(), expires_at=character()))
  }
  purrr::map_dfr(resp$data, function(f) tibble::tibble(
    file_id    = f$id %||% NA_character_,
    filename   = f$filename %||% NA_character_,
    mime_type  = guess_mime_type(f$filename %||% ""),
    size_bytes = as.numeric(f$bytes %||% NA),
    created_at = as.character(as.POSIXct(f$created_at %||% NA, origin = "1970-01-01", tz = "UTC")),
    expires_at = NA_character_
  ))
}

#' Get metadata for an OpenAI file
#' @noRd
openai_file_info <- function(.file_id, .called_from = NULL, ...) {
  api_obj <- api_openai(short_name = "openai", long_name = "OpenAI", api_key_env_var = "OPENAI_API_KEY")
  api_key <- get_api_key(api_obj)

  f <- httr2::request(paste0("https://api.openai.com/v1/files/", .file_id)) |>
    httr2::req_auth_bearer_token(api_key) |>
    httr2::req_perform() |>
    httr2::resp_body_json()

  tibble::tibble(
    file_id    = f$id %||% NA_character_,
    filename   = f$filename %||% NA_character_,
    mime_type  = guess_mime_type(f$filename %||% ""),
    size_bytes = as.numeric(f$bytes %||% NA),
    created_at = as.character(as.POSIXct(f$created_at %||% NA, origin = "1970-01-01", tz = "UTC")),
    expires_at = NA_character_
  )
}

#' Delete a file from OpenAI's Files API
#' @noRd
openai_delete_file <- function(.file_id, .called_from = NULL, ...) {
  api_obj <- api_openai(short_name = "openai", long_name = "OpenAI", api_key_env_var = "OPENAI_API_KEY")
  api_key <- get_api_key(api_obj)

  httr2::request(paste0("https://api.openai.com/v1/files/", .file_id)) |>
    httr2::req_auth_bearer_token(api_key) |>
    httr2::req_method("DELETE") |>
    httr2::req_perform()

  message("File ", .file_id, " has been successfully deleted.")
  invisible(NULL)
}


#' The `openai()` function acts as an interface for interacting with the OpenAI API
#' through main `tidyllm` verbs such as `chat()`, `embed()`, and `send_batch()`.
#' Chat uses the Responses API (`POST /v1/responses`); embeddings and batch operations
#' use the Chat Completions / Embeddings endpoints unchanged.
#'
#' @param ... Parameters passed to the appropriate OpenAI-specific function.
#' @param .called_from Internal routing argument; do not set manually.
#'
#' @return Result of the requested action.
#' @export
openai <- create_provider_function(
  .name          = "openai",
  chat           = openai_chat,
  embed          = openai_embedding,
  send_batch     = send_openai_batch,
  check_batch    = check_openai_batch,
  list_batches   = list_openai_batches,
  fetch_batch    = fetch_openai_batch,
  list_models    = openai_list_models,
  deep_research  = openai_deep_research,
  upload_file    = openai_upload_file,
  list_files     = openai_list_files,
  file_info      = openai_file_info,
  delete_file    = openai_delete_file
)

#' Alias for the OpenAI Provider Function
#'
#' @param ... Parameters passed to the appropriate OpenAI-specific function.
#' @param .called_from Internal routing argument; do not set manually.
#' @keywords internal
#' @export
chatgpt <- openai

#' The llama.cpp API provider class (inherits from OpenAI)
#'
#' @noRd
api_llamacpp <- new_class("LlamaCpp", api_openai)


#' Extract metadata from llama.cpp chat responses
#'
#' Extends the OpenAI metadata extractor to also capture logprobs and
#' reasoning_content (thinking traces from Qwen3 and similar models).
#'
#' @noRd
method(extract_metadata, list(api_llamacpp, class_list)) <- function(.api, .response) {
  list(
    model             = .response$model,
    timestamp         = lubridate::as_datetime(.response$created),
    prompt_tokens     = .response$usage$prompt_tokens,
    completion_tokens = .response$usage$completion_tokens,
    total_tokens      = .response$usage$total_tokens,
    stream            = FALSE,
    specific_metadata = list(
      system_fingerprint = .response$system_fingerprint,
      logprobs           = .response$choices[[1]]$logprobs,
      thinking           = .response$choices[[1]]$message$reasoning_content
    )
  )
}


#' Send LLM Messages to a llama.cpp Server
#'
#' @description
#' Sends a message history to a local llama.cpp server using the OpenAI-compatible
#' Chat Completions API. Supports BNF grammar constraints (a llama.cpp-specific
#' feature that enforces output format at the token sampling level) and token
#' logprobs for uncertainty quantification.
#'
#' @param .llm An `LLMMessage` object containing the conversation history.
#' @param .model The model name (default: `"local-model"`). llama.cpp ignores this
#'   value and serves whatever GGUF is loaded; the default is a clear placeholder.
#' @param .max_tokens Maximum tokens in the response (default: 1024).
#' @param .temperature Controls randomness (0–2, optional).
#' @param .top_p Nucleus sampling parameter (0–1, optional).
#' @param .stop One or more stop sequences (optional).
#' @param .stream Logical; if TRUE, streams the response (default: FALSE).
#' @param .tools Either a single TOOL object or a list of TOOL objects.
#' @param .tool_choice Tool-calling behavior: `"none"`, `"auto"`, or `"required"` (optional).
#' @param .json_schema A JSON schema as an R list to enforce structured output. Passed
#'   as `response_format = {type: "json_schema", ...}` — enforced at the sampler level.
#' @param .grammar A BNF grammar string to constrain sampling to any formal language
#'   (e.g. `"root ::= [0-9]{4}"` for a 4-digit year). Takes precedence over
#'   `.json_schema` when both are set; use one or the other.
#' @param .logprobs Logical; if TRUE, returns token log-probabilities (default: FALSE).
#' @param .top_logprobs Integer; number of top-alternative log-prob entries to return
#'   per token (1–20, optional). Requires `.logprobs = TRUE`.
#' @param .seed Integer; random seed for reproducible outputs (optional).
#' @param .thinking Logical; if TRUE, enables extended reasoning/thinking mode for models
#'   that support it (e.g. Qwen3). `NULL` (default) lets the server decide; `FALSE`
#'   explicitly disables it for faster responses; `TRUE` forces it on. The thinking
#'   trace is accessible via `get_metadata(result)$api_specific$thinking`.
#' @param .thinking_budget Integer; maximum tokens the model may use for thinking when
#'   `.thinking = TRUE`. `NULL` uses the server default (optional).
#' @param .server Base URL of the llama.cpp server. Defaults to the `LLAMACPP_SERVER`
#'   environment variable, falling back to `"http://localhost:8080"`.
#' @param .api_key API key for the llama.cpp server. Defaults to the `LLAMACPP_API_KEY`
#'   environment variable. Leave unset if the server was started without `--api-key`.
#' @param .timeout Request timeout in seconds (default: 120).
#' @param .verbose If TRUE, displays additional information (default: FALSE).
#' @param .dry_run If TRUE, returns the request object without executing it (default: FALSE).
#' @param .max_tries Maximum retries (default: 3).
#' @param .max_tool_rounds Maximum tool use iterations (default: 10).
#'
#' @return A new `LLMMessage` object containing the original messages plus the
#'   assistant's response.
#'
#' @export
llamacpp_chat <- function(.llm,
                          .model          = "local-model",
                          .max_tokens     = 1024,
                          .temperature    = NULL,
                          .top_p          = NULL,
                          .stop           = NULL,
                          .stream         = FALSE,
                          .tools          = NULL,
                          .tool_choice    = NULL,
                          .json_schema    = NULL,
                          .grammar        = NULL,
                          .logprobs        = FALSE,
                          .top_logprobs    = NULL,
                          .seed            = NULL,
                          .thinking        = NULL,
                          .thinking_budget = NULL,
                          .server          = Sys.getenv("LLAMACPP_SERVER", "http://localhost:8080"),
                          .api_key        = Sys.getenv("LLAMACPP_API_KEY", ""),
                          .timeout        = 120,
                          .verbose        = FALSE,
                          .dry_run        = FALSE,
                          .max_tries      = 3,
                          .max_tool_rounds = 10) {

  c(
    "Input .llm must be an LLMMessage object" = S7_inherits(.llm, LLMMessage),
    "Input .model must be a non-empty string" = is.character(.model) && nzchar(.model),
    "Input .max_tokens must be a positive integer" = is_integer_valued(.max_tokens) && .max_tokens > 0,
    "Input .temperature must be between 0 and 2 if provided" = is.null(.temperature) || (is.numeric(.temperature) && .temperature >= 0 && .temperature <= 2),
    "Input .top_p must be between 0 and 1 if provided" = is.null(.top_p) || (is.numeric(.top_p) && .top_p >= 0 && .top_p <= 1),
    "Input .stop must be NULL, a string, or a list" = is.null(.stop) || is.character(.stop) || is.list(.stop),
    "Input .stream must be logical" = is.logical(.stream),
    "Input .tools must be NULL, a TOOL object, or a list of TOOL objects" = is.null(.tools) || S7_inherits(.tools, TOOL) || (is.list(.tools) && all(purrr::map_lgl(.tools, ~ S7_inherits(.x, TOOL)))),
    "Input .tool_choice must be NULL or one of 'none', 'auto', 'required'" = is.null(.tool_choice) || (.tool_choice %in% c("none", "auto", "required")),
    "Streaming is not supported for requests with tool calls" = is.null(.tools) || !isTRUE(.stream),
    "Input .json_schema must be NULL or a list" = is.null(.json_schema) || is.list(.json_schema),
    "Input .grammar must be NULL or a non-empty string" = is.null(.grammar) || (is.character(.grammar) && nzchar(.grammar)),
    "Input .logprobs must be logical" = is.logical(.logprobs),
    "Input .top_logprobs must be NULL or an integer between 1 and 20" = is.null(.top_logprobs) || (is_integer_valued(.top_logprobs) && .top_logprobs >= 1 && .top_logprobs <= 20),
    "Input .seed must be an integer-valued numeric if provided" = is.null(.seed) || is_integer_valued(.seed),
    "Input .thinking must be NULL or logical" = is.null(.thinking) || is.logical(.thinking),
    "Input .thinking_budget must be NULL or a positive integer" = is.null(.thinking_budget) || (is_integer_valued(.thinking_budget) && .thinking_budget > 0),
    "Input .server must be a non-empty string" = is.character(.server) && nzchar(.server),
    "Input .timeout must be a positive integer" = is_integer_valued(.timeout) && .timeout > 0,
    "Input .verbose must be logical" = is.logical(.verbose),
    "Input .dry_run must be logical" = is.logical(.dry_run),
    "Input .max_tries must be integer-valued numeric" = is_integer_valued(.max_tries),
    ".max_tool_rounds must be a positive integer" = is_integer_valued(.max_tool_rounds) && .max_tool_rounds >= 1
  ) |> validate_inputs()

  api_obj <- api_llamacpp(short_name = "llamacpp",
                          long_name  = "llama.cpp",
                          api_key_env_var = "LLAMACPP_API_KEY")

  messages <- to_api_format(.llm, api_obj, FALSE)

  tools_def <- if (!is.null(.tools)) {
    if (S7_inherits(.tools, TOOL)) list(.tools) else .tools
  } else {
    NULL
  }

  json <- FALSE
  response_format <- NULL

  if (!is.null(.json_schema)) {
    json <- TRUE
    response_format <- list(
      type = "json_schema",
      json_schema = list(name = "output", schema = .json_schema, strict = TRUE)
    )
  }

  request_body <- list(
    model            = .model,
    messages         = messages,
    max_tokens       = .max_tokens,
    temperature      = .temperature,
    top_p            = .top_p,
    stop             = .stop,
    stream           = .stream,
    seed             = .seed,
    response_format  = response_format,
    grammar          = .grammar,
    logprobs         = if (.logprobs) TRUE else NULL,
    top_logprobs     = .top_logprobs,
    enable_thinking  = .thinking,
    thinking_budget  = if (!is.null(.thinking_budget) && isTRUE(.thinking)) .thinking_budget else NULL,
    tools            = if (!is.null(tools_def)) tools_to_api(api_obj, tools_def) else NULL,
    tool_choice      = .tool_choice
  ) |> purrr::compact()

  auth_header <- if (nzchar(.api_key)) {
    list(Authorization = sprintf("Bearer %s", .api_key))
  } else {
    list()
  }

  request <- httr2::request(.server) |>
    httr2::req_url_path("/v1/chat/completions") |>
    httr2::req_headers(!!!auth_header, `Content-Type` = "application/json") |>
    httr2::req_body_json(data = request_body)

  if (.dry_run) {
    return(request)
  }

  response <- perform_chat_request(request, api_obj, .stream, .timeout, .max_tries)

  if (.stream == FALSE && !is.null(tools_def)) {
    response <- process_tool_loop(
      .api             = api_obj,
      .response        = response,
      .tools_def       = tools_def,
      .request_body    = request_body,
      .request         = request,
      .timeout         = .timeout,
      .max_tries       = .max_tries,
      .max_tool_rounds = .max_tool_rounds
    )
  }

  logprobs <- parse_logprobs(api_obj, response$raw)

  add_message(.llm      = .llm,
              .role     = "assistant",
              .content  = response$assistant_reply,
              .json     = json,
              .meta     = response$meta,
              .logprobs = logprobs)
}


#' Generate Embeddings Using a llama.cpp Server
#'
#' @description
#' Sends text to the `/v1/embeddings` endpoint of a running llama.cpp server
#' and returns embedding vectors.
#'
#' @param .input A character vector of texts to embed, or an `LLMMessage` object.
#' @param .model The model name (default: `"local-model"`). llama.cpp ignores this
#'   and serves the loaded embedding model.
#' @param .server Base URL of the llama.cpp server. Defaults to the `LLAMACPP_SERVER`
#'   environment variable, falling back to `"http://localhost:8080"`.
#' @param .api_key API key for the server (default: `LLAMACPP_API_KEY` env var).
#' @param .timeout Request timeout in seconds (default: 120).
#' @param .dry_run If TRUE, returns the request object without executing it (default: FALSE).
#' @param .max_tries Maximum retries (default: 3).
#'
#' @return A tibble with columns `input` (text) and `embeddings` (list of numeric vectors).
#'
#' @export
llamacpp_embedding <- function(.input,
                               .model   = "local-model",
                               .server  = Sys.getenv("LLAMACPP_SERVER", "http://localhost:8080"),
                               .api_key = Sys.getenv("LLAMACPP_API_KEY", ""),
                               .timeout = 120,
                               .dry_run = FALSE,
                               .max_tries = 3) {
  c(
    "Input .input must be an LLMMessage object or a character vector" = S7_inherits(.input, LLMMessage) || is.character(.input),
    "Input .model must be a non-empty string" = is.character(.model) && nzchar(.model),
    "Input .server must be a non-empty string" = is.character(.server) && nzchar(.server),
    "Input .timeout must be a positive number" = is.numeric(.timeout) && .timeout > 0,
    "Input .dry_run must be logical" = is.logical(.dry_run)
  ) |> validate_inputs()

  input_texts <- parse_embedding_input(.input)

  auth_header <- if (nzchar(.api_key)) {
    list(Authorization = sprintf("Bearer %s", .api_key))
  } else {
    list()
  }

  request <- httr2::request(.server) |>
    httr2::req_url_path("/v1/embeddings") |>
    httr2::req_headers(!!!auth_header, `Content-Type` = "application/json") |>
    httr2::req_body_json(list(model = .model, input = input_texts))

  if (.dry_run) {
    return(request)
  }

  extract_embeddings_fn <- function(response_content, error, response_headers) {
    if (error) {
      paste0("llama.cpp embedding error: ", response_content$error$message %||% "unknown") |> stop()
    }
    response_content$data |>
      purrr::map("embedding") |>
      purrr::map(unlist)
  }

  perform_embedding_request(
    .request               = request,
    .timeout               = .timeout,
    .max_tries             = .max_tries,
    .input_texts           = input_texts,
    .fn_extract_embeddings = extract_embeddings_fn
  )
}


#' List Models Loaded in the llama.cpp Server
#'
#' @description
#' Calls the `/v1/models` endpoint of a running llama.cpp server and returns
#' the currently loaded model(s) as a tibble. In normal operation this is one
#' row; two rows appear when speculative decoding is active (main + draft model).
#'
#' @param .server Base URL of the llama.cpp server. Defaults to `LLAMACPP_SERVER`
#'   env var, falling back to `"http://localhost:8080"`.
#' @param .api_key API key for the server (default: `LLAMACPP_API_KEY` env var).
#' @param .timeout Request timeout in seconds (default: 30).
#' @param .max_tries Maximum retries (default: 3).
#'
#' @return A tibble with columns `id`, `object`, and `created`.
#'
#' @export
llamacpp_list_models <- function(.server  = Sys.getenv("LLAMACPP_SERVER", "http://localhost:8080"),
                                 .api_key = Sys.getenv("LLAMACPP_API_KEY", ""),
                                 .timeout = 30,
                                 .max_tries = 3) {
  auth_header <- if (nzchar(.api_key)) {
    list(Authorization = sprintf("Bearer %s", .api_key))
  } else {
    list()
  }

  response <- httr2::request(.server) |>
    httr2::req_url_path("/v1/models") |>
    httr2::req_headers(!!!auth_header) |>
    httr2::req_timeout(.timeout) |>
    httr2::req_retry(max_tries = .max_tries) |>
    httr2::req_perform() |>
    httr2::resp_body_json()

  if (is.null(response$data) || length(response$data) == 0) {
    return(tibble::tibble(id = character(), object = character(), created = integer()))
  }

  tibble::tibble(
    id      = vapply(response$data, function(m) m$id %||% NA_character_, character(1)),
    object  = vapply(response$data, function(m) m$object %||% NA_character_, character(1)),
    created = vapply(response$data, function(m) {
      v <- m$created; if (is.null(v)) NA_integer_ else as.integer(v)
    }, integer(1))
  )
}


#' Check Health of the llama.cpp Server
#'
#' @description
#' Calls the `/health` endpoint of a running llama.cpp server.
#' Returns the status string (`"ok"`, `"loading model"`, `"no model loaded"`,
#' or `"error"`) along with the full parsed response body as a named list.
#'
#' @param .server Base URL of the llama.cpp server. Defaults to `LLAMACPP_SERVER`
#'   env var, falling back to `"http://localhost:8080"`.
#' @param .timeout Request timeout in seconds (default: 10).
#'
#' @return A named list with at least a `status` element.
#'
#' @export
llamacpp_health <- function(.server  = Sys.getenv("LLAMACPP_SERVER", "http://localhost:8080"),
                            .timeout = 10) {
  resp <- tryCatch(
    httr2::request(.server) |>
      httr2::req_url_path("/health") |>
      httr2::req_timeout(.timeout) |>
      httr2::req_error(is_error = \(r) FALSE) |>
      httr2::req_perform() |>
      httr2::resp_body_json(),
    error = function(e) list(status = paste0("connection error: ", e$message))
  )
  resp
}


#' List Local GGUF Model Files
#'
#' @description
#' Scans a directory for `.gguf` files and returns a tibble with their names,
#' sizes, and modification times. No server needed.
#'
#' @param .path Directory to scan (default: `LLAMACPP_MODEL_DIR` env var, then
#'   `"~/models"`). The path is expanded with `path.expand()`.
#'
#' @return A tibble with columns `filename`, `size_gb`, `modified`, and `path`.
#'
#' @export
llamacpp_list_local_models <- function(.path = Sys.getenv("LLAMACPP_MODEL_DIR", "~/models")) {
  .path <- path.expand(.path)

  c(
    "Path must be an existing directory" = dir.exists(.path)
  ) |> validate_inputs()

  files <- list.files(.path, pattern = "\\.gguf$", full.names = TRUE, ignore.case = TRUE)

  if (length(files) == 0) {
    return(tibble::tibble(
      filename = character(),
      size_gb  = numeric(),
      modified = as.POSIXct(character()),
      path     = character()
    ))
  }

  info <- file.info(files)
  tibble::tibble(
    filename = basename(files),
    size_gb  = round(info$size / 1e9, 2),
    modified = info$mtime,
    path     = files
  )
}


#' List GGUF Files Available in a Hugging Face Repository
#'
#' @description
#' Calls the Hugging Face Hub API to list all `.gguf` files in a repository,
#' along with their sizes. Useful for choosing a quantization level before
#' calling `llamacpp_download_model()`.
#'
#' @param .repo Hugging Face repository ID in `"owner/model"` format
#'   (e.g. `"Qwen/Qwen3-8B-GGUF"`).
#' @param .timeout Request timeout in seconds (default: 30).
#'
#' @return A tibble with columns `filename`, `size_gb`, and `url`.
#'
#' @export
list_hf_gguf_files <- function(.repo, .timeout = 30) {
  c(
    "Input .repo must be a non-empty string in 'owner/model' format" = is.character(.repo) && nzchar(.repo) && grepl("/", .repo)
  ) |> validate_inputs()

  response <- httr2::request(glue::glue("https://huggingface.co/api/models/{.repo}/tree/main")) |>
    httr2::req_timeout(.timeout) |>
    httr2::req_retry(max_tries = 3) |>
    httr2::req_perform() |>
    httr2::resp_body_json()

  gguf_files <- purrr::keep(response, ~ grepl("\\.gguf$", .x$path %||% "", ignore.case = TRUE))

  if (length(gguf_files) == 0) {
    return(tibble::tibble(filename = character(), size_gb = numeric(), url = character()))
  }

  tibble::tibble(
    filename = vapply(gguf_files, function(f) basename(f$path %||% ""), character(1)),
    size_gb  = vapply(gguf_files, function(f) {
      v <- f$size; if (is.null(v)) NA_real_ else round(as.numeric(v) / 1e9, 2)
    }, numeric(1)),
    url = vapply(gguf_files, function(f) {
      glue::glue("https://huggingface.co/{.repo}/resolve/main/{f$path}")
    }, character(1))
  )
}


#' Download a GGUF Model from Hugging Face
#'
#' @description
#' Downloads a single GGUF file from a Hugging Face repository with a streaming
#' progress bar. Use `list_hf_gguf_files()` first to browse available quantizations.
#'
#' @param .repo Hugging Face repository ID (e.g. `"Qwen/Qwen3-8B-GGUF"`).
#' @param .filename The exact filename to download (e.g. `"Qwen3-8B-Q4_K_M.gguf"`).
#'   Use `list_hf_gguf_files(.repo)` to see what is available.
#' @param .dir Destination directory. Defaults to `LLAMACPP_MODEL_DIR` env var,
#'   then `"~/models"`. Created if it does not exist.
#' @param .timeout Download timeout in seconds (default: 3600).
#'
#' @return Invisibly returns the full path of the downloaded file.
#'
#' @export
llamacpp_download_model <- function(.repo,
                                    .filename,
                                    .dir     = Sys.getenv("LLAMACPP_MODEL_DIR", "~/models"),
                                    .timeout = 3600) {
  c(
    "Input .repo must be a non-empty string in 'owner/model' format" = is.character(.repo) && nzchar(.repo) && grepl("/", .repo),
    "Input .filename must be a non-empty string" = is.character(.filename) && nzchar(.filename)
  ) |> validate_inputs()

  .dir <- path.expand(.dir)
  if (!dir.exists(.dir)) dir.create(.dir, recursive = TRUE)

  dest <- file.path(.dir, .filename)
  url  <- glue::glue("https://huggingface.co/{.repo}/resolve/main/{.filename}")

  progress_bar <- cli::cli_progress_bar(
    format      = "{cli::pb_name} {cli::pb_bar} {cli::pb_percent} | {cli::pb_current_bytes}/{cli::pb_total_bytes}",
    format_done = "{cli::pb_name} {cli::pb_bar} downloaded in {cli::pb_elapsed}",
    total       = NA,
    auto_terminate = FALSE,
    type        = "download"
  )

  con <- file(dest, "wb")
  on.exit({ close(con); cli::cli_progress_done(id = progress_bar) }, add = TRUE)

  httr2::request(url) |>
    httr2::req_timeout(.timeout) |>
    httr2::req_perform_stream(
      callback = function(chunk) {
        writeBin(chunk, con)
        TRUE
      },
      buffer_kb = 512
    )

  invisible(dest)
}


#' Delete a Local GGUF Model File
#'
#' @description
#' Deletes a GGUF file from disk. Issues a warning confirming the deletion.
#' Mirrors the `ollama_delete_model()` pattern.
#'
#' @param .path Full path to the `.gguf` file to delete.
#'
#' @return Invisibly returns `TRUE` on success.
#'
#' @export
llamacpp_delete_model <- function(.path) {
  .path <- path.expand(.path)
  c(
    "Input .path must point to an existing file" = file.exists(.path)
  ) |> validate_inputs()

  filename <- basename(.path)
  unlink(.path)
  warning(paste0("Model ", filename, " deleted"))
  invisible(TRUE)
}


#' Rerank Documents Using a llama.cpp Server
#'
#' @description
#' Calls the `/v1/reranking` endpoint of a running llama.cpp server to score
#' documents by relevance to a query. Useful for building fully-local RAG
#' pipelines (embed → cosine search → rerank → chat, all with `llamacpp()`).
#'
#' @param .query A single query string.
#' @param .documents A character vector of documents to rerank.
#' @param .model The model name (default: `"local-model"`).
#' @param .server Base URL of the llama.cpp server. Defaults to `LLAMACPP_SERVER`
#'   env var, falling back to `"http://localhost:8080"`.
#' @param .api_key API key for the server (default: `LLAMACPP_API_KEY` env var).
#' @param .timeout Request timeout in seconds (default: 60).
#' @param .max_tries Maximum retries (default: 3).
#' @param .dry_run If TRUE, returns the request object without executing it (default: FALSE).
#'
#' @return A tibble with columns `index` (original position), `document` (text),
#'   and `relevance_score`, sorted by descending score.
#'
#' @export
llamacpp_rerank <- function(.query,
                            .documents,
                            .model   = "local-model",
                            .server  = Sys.getenv("LLAMACPP_SERVER", "http://localhost:8080"),
                            .api_key = Sys.getenv("LLAMACPP_API_KEY", ""),
                            .timeout = 60,
                            .max_tries = 3,
                            .dry_run = FALSE) {
  c(
    "Input .query must be a non-empty string" = is.character(.query) && length(.query) == 1 && nzchar(.query),
    "Input .documents must be a non-empty character vector" = is.character(.documents) && length(.documents) >= 1,
    "Input .dry_run must be logical" = is.logical(.dry_run)
  ) |> validate_inputs()

  auth_header <- if (nzchar(.api_key)) {
    list(Authorization = sprintf("Bearer %s", .api_key))
  } else {
    list()
  }

  request <- httr2::request(.server) |>
    httr2::req_url_path("/v1/reranking") |>
    httr2::req_headers(!!!auth_header, `Content-Type` = "application/json") |>
    httr2::req_body_json(list(
      model     = .model,
      query     = .query,
      documents = as.list(.documents)
    ))

  if (.dry_run) {
    return(request)
  }

  response <- request |>
    httr2::req_timeout(.timeout) |>
    httr2::req_retry(max_tries = .max_tries) |>
    httr2::req_perform() |>
    httr2::resp_body_json()

  results <- response$results %||% response$data %||% list()

  if (length(results) == 0) {
    return(tibble::tibble(index = integer(), document = character(), relevance_score = numeric()))
  }

  tbl <- tibble::tibble(
    index           = vapply(results, function(r) as.integer(r$index), integer(1)),
    document        = .documents[vapply(results, function(r) as.integer(r$index) + 1L, integer(1))],
    relevance_score = vapply(results, function(r) {
      v <- r$relevance_score %||% r$score
      if (is.null(v)) NA_real_ else as.numeric(v)
    }, numeric(1))
  )

  tbl[order(-tbl$relevance_score), ]
}


#' llama.cpp Provider Function
#'
#' @description
#' The `llamacpp()` provider connects tidyllm to a locally running
#' [llama.cpp](https://github.com/ggml-org/llama.cpp) server. It exposes
#' the same verb/provider pattern as every other tidyllm provider while
#' also offering llama.cpp-specific features: BNF grammar constraints
#' (`.grammar`), token logprobs (`.logprobs`), and model management helpers.
#'
#' The server must be started separately before calling `llamacpp()`. See
#' `llamacpp_health()` to verify the server is running, and
#' `llamacpp_download_model()` / `list_hf_gguf_files()` to obtain models.
#'
#' @param ... Parameters passed to the appropriate llama.cpp-specific function.
#' @param .called_from An internal argument specifying which verb invoked this
#'   function. Managed automatically by tidyllm verbs; do not set manually.
#'
#' @return The result of the requested action (e.g., an updated `LLMMessage`
#'   for `chat()`, a tibble for `embed()` or `list_models()`).
#'
#' @export
llamacpp <- create_provider_function(
  .name       = "llamacpp",
  chat        = llamacpp_chat,
  embed       = llamacpp_embedding,
  list_models = llamacpp_list_models
)

structure(list(method = "POST", url = "https://api.anthropic.com/v1/messages", 
    status_code = 200L, headers = structure(list(date = "Wed, 16 Oct 2024 00:52:01 GMT", 
        `content-type` = "application/json", `anthropic-ratelimit-requests-limit` = "50", 
        `anthropic-ratelimit-requests-remaining` = "49", `anthropic-ratelimit-requests-reset` = "2024-10-16T00:52:10Z", 
        `anthropic-ratelimit-tokens-limit` = "40000", `anthropic-ratelimit-tokens-remaining` = "40000", 
        `anthropic-ratelimit-tokens-reset` = "2024-10-16T00:52:01Z", 
        `request-id` = "req_01SoZo63WzaC9eP6vs7y8Sfb", `x-cloud-trace-context` = "90045c15ee7b45a1fced4842cea42ce4", 
        via = "1.1 google", `cf-cache-status` = "DYNAMIC", `x-robots-tag` = "none", 
        server = "cloudflare", `cf-ray` = "8d3410048e1e1cc9-FRA", 
        `content-encoding` = "gzip"), class = "httr2_headers"), 
    body = charToRaw("{\"id\":\"msg_01EeoWV2RPMZARvAgDyAdx7Q\",\"type\":\"message\",\"role\":\"assistant\",\"model\":\"claude-3-5-sonnet-20240620\",\"content\":[{\"type\":\"text\",\"text\":\"This plot shows a scatter plot with a linear regression line fitted to the data points. Here's an analysis of the plot and the provided data summary:\\n\\n1. Correlation: There's a strong positive linear relationship between x and y variables, as evidenced by the upward-sloping regression line and the clustering of points around it.\\n\\n2. Distribution:\\n   - X-axis: Values range from approximately -3 to 3\\n   - Y-axis: Values range from about -7 to 7\\n\\n3. Regression line: The blue line represents the best-fit linear regression line. The shaded area around it likely represents the confidence interval for the regression.\\n\\n4. Data summary:\\n   - X variable:\\n     * Range: -2.77 to 2.52\\n     * Mean: -0.06141 (close to 0)\\n     * Median: -0.17643 (slightly below 0)\\n   - Y variable:\\n     * Range: -7.07 to 6.75\\n     * Mean: 0.01294 (very close to 0)\\n     * Median: 0.20385 (slightly above 0)\\n\\n5. Spread: The interquartile range (IQR) for x is from -0.83 to 0.73, and for y is from -1.59 to 1.35, indicating a moderate spread of data.\\n\\n6. Symmetry: Both variables appear to be roughly symmetrically distributed around their means, which are close to zero. This suggests the possibility of a normal distribution for both x and y.\\n\\n7. Outliers: There are a few potential outliers, particularly at the extremes of both x and y axes, but they don't seem to strongly deviate from the overall trend.\\n\\n8. Model fit: The regression line appears to fit the data well, with most points falling close to the line and evenly distributed on both sides.\\n\\nIn conclusion, this plot demonstrates a strong positive linear relationship between x and y variables, with data that appears to be normally distributed and well-represented by the linear regression model.\"}],\"stop_reason\":\"end_turn\",\"stop_sequence\":null,\"usage\":{\"input_tokens\":494,\"output_tokens\":469}}"), 
    cache = new.env(parent = emptyenv())), class = "httr2_response")

structure(list(method = "POST", url = "https://api.mistral.ai/v1/chat/completions", 
    status_code = 200L, headers = structure(list(date = "Thu, 21 Aug 2025 12:41:10 GMT", 
        `content-type` = "application/json", `content-encoding` = "gzip", 
        `mistral-correlation-id` = "0198cca5-ae6f-7c35-9c52-d3340710d226", 
        `x-kong-request-id` = "0198cca5-ae6f-7c35-9c52-d3340710d226", 
        `x-envoy-upstream-service-time` = "3080", `x-ratelimit-limit-tokens-minute` = "2000000", 
        `x-ratelimit-remaining-tokens-minute` = "1999932", `x-ratelimit-limit-tokens-month` = "10000000000", 
        `x-ratelimit-remaining-tokens-month` = "9999999932", 
        `x-ratelimit-tokens-query-cost` = "68", `x-ratelimit-limit-req-10-second` = "60", 
        `x-ratelimit-remaining-req-10-second` = "59", `access-control-allow-origin` = "*", 
        `x-kong-upstream-latency` = "3081", `x-kong-proxy-latency` = "5", 
        `cf-cache-status` = "DYNAMIC", `set-cookie` = "REDACTED", 
        `set-cookie` = "REDACTED", `strict-transport-security` = "max-age=15552000; includeSubDomains", 
        server = "cloudflare", `cf-ray` = "972a33ea0e9b691b-FRA", 
        `alt-svc` = "h3=\":443\"; ma=86400"), redact = character(0), class = "httr2_headers"), 
    body = charToRaw("{\"id\":\"d0a4a06db45446809c07ca9d1cc4158d\",\"created\":1755780066,\"model\":\"mistral-large-latest\",\"usage\":{\"prompt_tokens\":7,\"total_tokens\":68,\"completion_tokens\":61},\"object\":\"chat.completion\",\"choices\":[{\"index\":0,\"finish_reason\":\"stop\",\"message\":{\"role\":\"assistant\",\"tool_calls\":null,\"content\":\"Hello, world! 🌍✨ How can I brighten your day today? Whether you're coding, exploring, or just saying hi, I'm here to help! 😊\\n\\n(And if you're learning programming—great start with that classic greeting! 🚀)\"}}]}"), 
    cache = new.env(parent = emptyenv())), class = "httr2_response")

structure(list(method = "POST", url = "https://api.anthropic.com/v1/messages", 
    status_code = 200L, headers = structure(list(date = "Thu, 21 Aug 2025 12:41:00 GMT", 
        `content-type` = "application/json", `content-encoding` = "gzip", 
        `anthropic-ratelimit-input-tokens-limit` = "80000", `anthropic-ratelimit-input-tokens-remaining` = "80000", 
        `anthropic-ratelimit-input-tokens-reset` = "2025-08-21T12:40:59Z", 
        `anthropic-ratelimit-output-tokens-limit` = "16000", 
        `anthropic-ratelimit-output-tokens-remaining` = "16000", 
        `anthropic-ratelimit-output-tokens-reset` = "2025-08-21T12:41:00Z", 
        `anthropic-ratelimit-requests-limit` = "1000", `anthropic-ratelimit-requests-remaining` = "999", 
        `anthropic-ratelimit-requests-reset` = "2025-08-21T12:40:59Z", 
        `anthropic-ratelimit-tokens-limit` = "96000", `anthropic-ratelimit-tokens-remaining` = "96000", 
        `anthropic-ratelimit-tokens-reset` = "2025-08-21T12:40:59Z", 
        `request-id` = "req_011CSLsW7mFqvHHnzrKdKjAE", `strict-transport-security` = "max-age=31536000; includeSubDomains; preload", 
        `anthropic-organization-id` = "36562039-1ae4-48e6-ae28-1e517e3d2d12", 
        `x-envoy-upstream-service-time` = "690", via = "1.1 google", 
        `cf-cache-status` = "DYNAMIC", `x-robots-tag` = "none", 
        server = "cloudflare", `cf-ray` = "972a33bb3977d246-FRA"), redact = character(0), class = "httr2_headers"), 
    body = charToRaw("{\"id\":\"msg_01QgNtCXZKCJgpWHW3NEwmdP\",\"type\":\"message\",\"role\":\"assistant\",\"model\":\"claude-3-5-sonnet-20240620\",\"content\":[{\"type\":\"text\",\"text\":\"Hello! How can I assist you today? Is there anything specific you'd like to know or discuss?\"}],\"stop_reason\":\"end_turn\",\"stop_sequence\":null,\"usage\":{\"input_tokens\":16,\"cache_creation_input_tokens\":0,\"cache_read_input_tokens\":0,\"cache_creation\":{\"ephemeral_5m_input_tokens\":0,\"ephemeral_1h_input_tokens\":0},\"output_tokens\":24,\"service_tier\":\"standard\"}}"), 
    cache = new.env(parent = emptyenv())), class = "httr2_response")

tidyllm_field <- new_class("tidyllm_field", properties = list(
  type = class_character,
  description = class_character,
  enum = class_character,   # For factor (enum) fields only.
  vector = class_logical,
  schema = class_list       # New: holds a nested schema (if any)
))

#' Define Field Descriptors for JSON Schema
#'
#' These functions create field descriptors used in `tidyllm_schema()` or `field_object()` to define JSON schema fields. They support character, factor, numeric, and logical types.
#'
#' @param .description A character string describing the field (optional).
#' @param .vector A logical value indicating if the field is a vector (default: FALSE).
#' @param .levels A character vector specifying allowable values (for `field_fct()` only).
#'
#' @return An S7 `tidyllm_field` object representing the field descriptor.
#'
#' @examples
#' field_chr("A common street name")
#' field_fct("State abbreviation", .levels = c("CA", "TX", "Other"))
#' field_dbl("House number")
#' field_lgl("Is residential")
#' field_dbl("A list of appartment numbers at the address",.vector=TRUE )
#' @export
field_chr <- function(.description = character(0), .vector = FALSE) {
  tidyllm_field(
    type = "string",
    description = .description,
    enum = character(0),
    vector = .vector,
    schema = list()
  )
}

#' @rdname field_chr
#' @export
field_fct <- function(.description = character(0), .levels, .vector = FALSE) {
  tidyllm_field(
    type = "string",
    description = .description,
    enum = .levels,
    vector = .vector,
    schema = list()
  )
}

#' @rdname field_chr
#' @export
field_dbl <- function(.description = character(0), .vector = FALSE) {
  tidyllm_field(
    type = "number",
    description = .description,
    enum = character(0),
    vector = .vector,
    schema = list()
  )
}

#' @rdname field_chr
#' @export
field_lgl <- function(.description = character(0), .vector = FALSE) {
  tidyllm_field(
    type = "boolean",
    description = .description,
    enum = character(0),
    vector = .vector,
    schema = list()
  )
}

#' Define a nested object field
#' @param .description A character string describing the field (optional).
#' @param ... Named fields to include in the object definition (required).
#' @param .vector A logical value indicating if the field is a list of objects (default: FALSE).
#' @return An S7 `tidyllm_field` object of type "object" containing nested fields.
#' @examples
#' # Define an address object with nested fields
#' address <- field_object("A mailing address",
#'   street = field_chr("Street name"),
#'   city = field_chr("City name"),
#'   zipcode = field_chr("Postal code")
#' )
#' 
#' # Create a vector of objects
#' addresses <- field_object("List of addresses", 
#'   street = field_chr("Street name"),
#'   city = field_chr("City name"),
#'   .vector = TRUE
#' )
#' @export
field_object <- function(.description = character(0), ..., .vector = FALSE) {
  nested_fields <- list(...)
  if (length(nested_fields) == 0) {
    stop("field_object must have at least one nested field.")
  }
  nested_schema <- build_schema(nested_fields)
  
  tidyllm_field(
    type = "object",
    description = .description,
    enum = character(0),
    vector = .vector,
    schema = nested_schema
  )
}


#'parse_field converts a tidyllm_field (or a shorthand string) to its JSON schema fragment.
#'  
#'@noRd
parse_field <- function(field) {
  # If ellmer is available and the field is an ellmer type, use ellmer's to_schema().
  if (requireNamespace("ellmer", quietly = TRUE) &&
      any(class(field) %in% c("ellmer::TypeBasic", "ellmer::TypeEnum", "ellmer::TypeObject", "ellmer::TypeArray"))) {
    json_type <- to_schema(field)
    return(json_type)
  }
  
  # If field is an S7 tidyllm_field, process it.
  if (S7_inherits(field, tidyllm_field)) {
    # If the field is an object with a nested schema.
    if (field@type == "object" && length(field@schema)>=1) {
      json_type <- list(
        type = "object",
        properties = field@schema$properties,
        required = field@schema$required
      )
      if (length(field@description) > 0) {
        json_type$description <- field@description
      }
      if (isTRUE(field@vector)) {
        json_type <- list(
          type = "array",
          items = json_type
        )
      }
      return(json_type)
    }
    
    # For non-object fields.
    json_type <- list(type = field@type)
    if (length(field@description) > 0) {
      json_type$description <- field@description
    }
    if (length(field@enum) > 0) {
      json_type$enum <- field@enum
    }
    if (isTRUE(field@vector)) {
      json_type <- list(
        type = "array",
        items = json_type
      )
    }
    return(json_type)
  }
  
  # Fallback for character shorthand specifications.
  is_array <- grepl("\\[\\]$", field)
  base_field <- sub("\\[\\]$", "", field)
  
  if (grepl("^factor\\(.*\\)$", base_field)) {
    enums <- base_field |>
      stringr::str_replace("^factor\\((.*)\\)$", "\\1") |>
      stringr::str_split(",") |>
      (\(x) x[[1]])()
    enums <- stringr::str_trim(enums)
    json_type <- list(type = "string", enum = enums)
  } else if (base_field %in% c("string", "character", "logical", "numeric", "factor")) {
    type_map <- list(
      string    = "string",
      character = "string",
      factor    = "string",
      logical   = "boolean",
      numeric   = "number"
    )
    json_type <- list(type = type_map[[base_field]])
  } else {
    stop("Unsupported field type: ", field)
  }
  
  if (is_array) {
    json_type <- list(
      type = "array",
      items = json_type
    )
  }
  return(json_type)
}

#' build_schema constructs a JSON schema list from a list of named field definitions.
#'  
#' @noRd
build_schema <- function(fields, name = NULL) {
  if (!all(nzchar(names(fields)))) {
    stop("Field names must be non-empty character strings")
  }
  properties <- lapply(fields, parse_field)
  schema <- list(
    type = "object",
    properties = properties,
    required = I(names(fields))
  )
  if (!is.null(name)) {
    attr(schema, "name") <- name
  }
  schema
}

#' Create a JSON Schema for Structured Outputs
#'
#' This function creates a JSON schema for structured outputs, supporting both character-based shorthand and S7 `tidyllm_field` objects.
#' It also integrates with `ellmer` types like `ellmer::type_string()` if ellmer is in your namespace
#'
#' @param name A character string specifying the schema name (default: "tidyllm_schema").
#' @param ... Named arguments where each name represents a field, and each value is either a character string, a `tidyllm_field`, or an `ellmer` type.
#'
#' Supported character shorthand types:
#' - "character" or "string" for character fields
#' - "logical" for boolean fields
#' - "numeric" for number fields
#' - "factor(...)" for enumerations
#' - Use `[]` to indicate vectors, e.g., "character[]"
#'
#' @return A list representing the JSON schema, suitable for use with `.json_schema` in LLM API calls.
#'
#' @examples
#' \dontrun{
#' # Example using different field types
#' address_schema <- tidyllm_schema(
#'   name = "AddressSchema",
#'   Street = field_chr("A common street name"),
#'   house_number = field_dbl(),
#'   City = field_chr("Name of a city"),
#'   State = field_fct("State abbreviation", .levels = c("CA", "TX", "Other")),
#'   Country = "string",
#'   PostalCode = "string"
#' )
#'
#' llm_message("Imagine an address") |> chat(openai, .json_schema = address_schema)
#'
#' # Example with vector field
#' tidyllm_schema(
#'   plz = field_dbl(.vector = TRUE)
#' )
#'}
#' @export
tidyllm_schema <- function(name = "tidyllm_schema", ...) {
  fields <- list(...)
  # Validate inputs: each field must be either a supported character string,
  # an S7 tidyllm_field, or (if ellmer is available) an ellmer type.
  validate_inputs(c(
    "Schema name must be a non-empty character string" = is.character(name) && nzchar(name),
    "Field names must be non-empty character strings" = all(nzchar(names(fields))),
    "Field types must be either a supported character string, an S7 tidyllm_field, or an ellmer type" =
      all(sapply(fields, function(field) {
        (is.character(field) && grepl("^(string|character|logical|numeric|factor\\(.*\\))(\\[\\])?$", field)) ||
          S7_inherits(field, tidyllm_field)  ||
          is_ellmer_type(field)
      }))
  ))
  build_schema(fields, name = name)
}


to_schema <- new_generic("to_schema", "x")
if (requireNamespace("ellmer", quietly = TRUE)) {
  method(to_schema, ellmer::TypeBasic) <- function(x) {
    list(type = x@type, description = x@description %||% "")
  }
  
  method(to_schema, ellmer::TypeEnum) <- function(x) {
    list(
      type = "string",
      description = x@description %||% "",
      enum = as.list(x@values)
    )
  }
  
  method(to_schema, ellmer::TypeObject) <- function(x) {
    names <- rlang::names2(x@properties)
    required <- purrr::map_lgl(x@properties, function(prop) prop@required)
    
    properties <- to_schema(x@properties)
    names(properties) <- names
    
    list(
      type = "object",
      description = x@description %||% "",
      properties = properties,
      required = as.list(names[required]),
      additionalProperties = x@additional_properties
    )
  }
  
  method(to_schema, ellmer::TypeArray) <- function(x) {
    list(
      type = "array",
      description = x@description %||% "",
      items = to_schema(x@items)
    )
  }
  
  method(to_schema, class_list) <- function(x) {
    lapply(x, to_schema)
  }
}

test_that("tidyllm_tool works with an explicit function", {
  explicit_tool <- tidyllm_tool(
    .f = runif,
    .description = "Generates n uniformly distributed random numbers on a range.",
    n = field_dbl("An integer specifying how many random numbers to generate"),
    min = field_dbl("The smallest floating point number on the range"),
    max = field_dbl("The largest floating point number on the range")
  )
  
  expect_true(S7::S7_inherits(explicit_tool, TOOL))  # Check S7 class
  expect_equal(explicit_tool@description, "Generates n uniformly distributed random numbers on a range.")
  expect_named(explicit_tool@input_schema, c("n", "min", "max"))
  expect_true(is.function(explicit_tool@func))
  expect_type(explicit_tool@func(n = 1, min = 0, max = 1), "double")
})

test_that("tidyllm_tool works with an anonymous function", {
  anon_tool <- tidyllm_tool(
    .f = function(tz = "UTC", format = "%Y-%m-%d %H:%M:%S") {
      format(Sys.time(), format, tz = tz)
    },
    .description = "Returns the current time in a specified timezone. Use this to determine the current time in any location.",
    tz = field_chr("The time zone identifier (e.g., 'Europe/Berlin', 'America/New_York', 'Asia/Tokyo', 'UTC'). Required."),
    format = field_chr("Format string for the time output. Default is '%Y-%m-%d %H:%M:%S'.")
  )
  
  expect_true(S7::S7_inherits(anon_tool, TOOL))  # Check S7 class
  expect_equal(anon_tool@description, "Returns the current time in a specified timezone. Use this to determine the current time in any location.")
  expect_named(anon_tool@input_schema, c("tz", "format"))
  expect_true(is.function(anon_tool@func))
  expect_type(anon_tool@func(), "character")
})

test_that("tidyllm_tool works with formula-based anonymous functions", {
  fml_tool <- tidyllm_tool(
    ~{.x * 12345.6789},
    .description = "Multiplies a number by 12345.6789",
    .x = field_dbl("A number")
  )
  
  expect_true(S7::S7_inherits(fml_tool, TOOL))  # Check S7 class
  expect_equal(fml_tool@description, "Multiplies a number by 12345.6789")
  expect_named(fml_tool@input_schema, c(".x"))
  expect_true(is.function(fml_tool@func))
  expect_equal(fml_tool@func(2), 2 * 12345.6789)
})

test_that("tidyllm_tool assigns unique names to anonymous functions", {
  tool1 <- tidyllm_tool(~ .x * 2, "Doubles a number", .x = field_dbl("A number"))
  tool2 <- tidyllm_tool(~ .x + 3, "Adds 3", .x = field_dbl("A number"))
  
  expect_match(tool1@name, "anonymous_")
  expect_match(tool2@name, "anonymous_")
  expect_false(tool1@name == tool2@name)  # Should be unique
})

test_that("tidyllm_tool validates missing schema definitions", {
  expect_error(
    tidyllm_tool(
      function(x, y) x + y,
      "Adds two numbers",
      x = field_dbl("First number") # Missing `y`
    ),
    "Missing schema definitions for required arguments: y"
  )
})

test_that("tidyllm_tool does not allow invalid schema arguments", {
  expect_error(
    tidyllm_tool(
      function(x) x + 1,
      "Adds 1 to a number",
      x = field_dbl("A number"),
      y = field_dbl("A number")  # `y` does not exist in the function
    ),
    "Schema defines arguments not present in function: y"
  )
})

test_that("tidyllm_schema handles basic field types correctly", {
  schema <- tidyllm_schema(
    name = "TestSchema",
    field1 = field_chr("A test character field"),
    field2 = field_dbl("A test numeric field"),
    field3 = field_lgl("A test logical field"),
    field4 = field_fct("A test factor field", .levels = c("A", "B", "C"))
  )
  
  expect_equal(schema$type, "object")
  expect_named(schema$properties, c("field1", "field2", "field3", "field4"))
  
  expect_equal(schema$properties$field1$type, "string")
  expect_equal(schema$properties$field1$description, "A test character field")
  
  expect_equal(schema$properties$field2$type, "number")
  expect_equal(schema$properties$field2$description, "A test numeric field")
  
  expect_equal(schema$properties$field3$type, "boolean")
  expect_equal(schema$properties$field3$description, "A test logical field")
  
  expect_equal(schema$properties$field4$type, "string")
  expect_equal(schema$properties$field4$enum, c("A", "B", "C"))
  expect_equal(schema$properties$field4$description, "A test factor field")
})


test_that("tidyllm_schema correctly handles vector fields", {
  schema <- tidyllm_schema(
    name = "TestVectorSchema",
    field_vec = field_dbl("A numeric vector field", .vector = TRUE)
  )
  
  expect_named(schema$properties, "field_vec")
  expect_equal(schema$properties$field_vec$type, "array")
  expect_equal(schema$properties$field_vec$items$type, "number")
  expect_equal(schema$properties$field_vec$items$description, "A numeric vector field")
})


test_that("tidyllm_schema handles character-based type specifications", {
  schema <- tidyllm_schema(
    name = "TestShorthandSchema",
    field1 = "string",
    field2 = "logical",
    field3 = "numeric",
    field4 = "factor(A, B, C)",
    field5 = "character[]"
  )
  
  expect_named(schema$properties, c("field1", "field2", "field3", "field4", "field5"))
  
  expect_equal(schema$properties$field1$type, "string")
  expect_equal(schema$properties$field2$type, "boolean")
  expect_equal(schema$properties$field3$type, "number")
  expect_equal(schema$properties$field4$type, "string")
  expect_equal(schema$properties$field4$enum, c("A", "B", "C"))
  expect_equal(schema$properties$field5$type, "array")
  expect_equal(schema$properties$field5$items$type, "string")
})


test_that("tidyllm_schema validates incorrect field types", {
  expect_error(
    tidyllm_schema(
      name = "InvalidSchema",
      field1 = 42  # Invalid type
    ),
    "Field types must be either a supported character string, an S7 tidyllm_field, or an ellmer type"
  )
})

#Uncommented (causes issues in devtools::check)
#test_that("tidyllm_schema integrates with ellmer types if available", {
#  skip_if_not(requireNamespace("ellmer", quietly = TRUE))
#  library(ellmer)
#  
#  field_ellmer <- ellmer::type_string("An ellmer string field")
#  schema <- tidyllm_schema(
#    name = "EllmerSchema",
#    field1 = field_ellmer
#  )
#  
#  expect_named(schema$properties, "field1")
#  expect_equal(schema$properties$field1$type, "string")
#  expect_equal(schema$properties$field1$description, "An ellmer string field")
#})

#' The base class for API-provider specific classes
#'
#' @noRd
APIProvider <- new_class("APIProvider",properties = list(
  short_name = class_character,
  long_name  = class_character,
  api_key_env_var = class_character
))

parse_chat_response        <- new_generic("parse_chat_response",c(".api",".content"))
handle_stream              <- new_generic("handle_stream",c(".api",".stream_response"))
generate_callback_function <- new_generic("generate_callback_function",".api")
ratelimit_from_header      <- new_generic("ratelimit_from_header",c(".api", ".headers"))
get_api_key                <- new_generic("get_api_key",".api")
prepare_llms_for_batch     <- new_generic("prepare_llms_for_batch",".api")
extract_metadata           <- new_generic("extract_metadata",c(".api", ".response"))
extract_metadata_stream    <- new_generic("extract_metadata_stream",c(".api", ".stream_raw_data"))
parse_logprobs             <- new_generic("parse_logprobs", c(".api", ".input"))

#Default method for the streaming callback function
#'
#' @noRd
method(generate_callback_function,APIProvider) <- function(.api) {
  # Default testing callback implementation
  function(.data) {
    if (is.null(.tidyllm_stream_env$testing_chunck_list)) {
      .tidyllm_stream_env$testing_chunck_list <- list()
    }
    
    # Append new data to the buffer
    new_data <- rawToChar(.data, multiple = FALSE)
    .tidyllm_stream_env$testing_chunck_list <- append(.tidyllm_stream_env$testing_chunck_list, new_data)
    length(.tidyllm_stream_env$testing_chunck_list) |> cat("\n")
    utils::flush.console()
    TRUE
  }
}


#Default method for metadata extraction
#'
#' @noRd
method(extract_metadata, list(APIProvider,class_list))<- function(.api,.response) {
  list(
    model             = NA_character_,
    timestamp         = lubridate::as_datetime(lubridate::now()),
    prompt_tokens     = NA_integer_,
    completion_tokens = NA_integer_,
    total_tokens      = NA_integer_,
    stream            = FALSE,
    specific_metadata = list() 
  )
}  

#Default method for metadata extraction
#'
#' @noRd
method(extract_metadata_stream, list(APIProvider,class_list))<- function(.api,.stream_raw_data) {
  list(
    model             = NA_character_,
    timestamp         = lubridate::as_datetime(lubridate::now()),
    prompt_tokens     = NA_integer_,
    completion_tokens = NA_integer_,
    total_tokens      = NA_integer_,
    stream            = TRUE,
    specific_metadata = list() 
  )
}  


#Default method for the API key checks
#'
#' @noRd
method(get_api_key, APIProvider) <- function(.api,.dry_run=FALSE) {
  api_key <- Sys.getenv(.api@api_key_env_var)
  if (api_key == "" & .dry_run==FALSE) {
    paste0("API key is not set. Please set it with: Sys.setenv(",api@api_key_env_var," = \"YOUR-KEY-GOES-HERE\")." ) |>
      rlang::abort()
  }
  return(api_key)
}


#Prepare a list of LLMs for batch requests 
#'
#' @noRd
method(prepare_llms_for_batch, APIProvider) <- function(.api, .llms, .id_prefix, .overwrite = FALSE) {
  # Check for unique non-missing names
  non_missing_names <- names(.llms)[!(is.na(names(.llms)) | names(.llms) == "")]
  if (anyDuplicated(non_missing_names)) {
    rlang::abort("Each specified name in .llms must be unique. Please ensure that all non-missing names are unique.")
  }
  
  # Check for existing batch_id
  if (!is.null(attr(.llms, "batch_id"))) {
    if (.overwrite) {
      rlang::warn("Batch ID is already set in the provided list. Overwriting with a new batch ID.")
    } else {
      rlang::abort("Batch ID is already set in the provided list. Set .overwrite = TRUE to overwrite.")
    }
  }
  
  # Generate custom IDs for missing names
  names(.llms) <- lapply(seq_along(.llms), function(i) {
    current_name <- names(.llms)[i]
    if (is.null(current_name) || current_name == "" || is.na(current_name)) {
      paste0(.id_prefix, i)
    } else {
      current_name
    }
  })
  
  return(.llms)
}


#api_fart <- APIProvider(short_name = "Fart",long_name = "fart.ai",api_key_env_var = "FOPENAI_API_KEY" )
#get_api_key(api_fart,TRUE)

structure(list(method = "POST", url = "https://api.anthropic.com/v1/messages", 
    status_code = 200L, headers = structure(list(date = "Wed, 16 Oct 2024 00:51:40 GMT", 
        `content-type` = "application/json", `anthropic-ratelimit-requests-limit` = "50", 
        `anthropic-ratelimit-requests-remaining` = "49", `anthropic-ratelimit-requests-reset` = "2024-10-16T00:52:10Z", 
        `anthropic-ratelimit-tokens-limit` = "40000", `anthropic-ratelimit-tokens-remaining` = "40000", 
        `anthropic-ratelimit-tokens-reset` = "2024-10-16T00:51:40Z", 
        `request-id` = "req_01AaWumjjQfjdMDvTRrnp868", `x-cloud-trace-context` = "a6fa53423ab69ee7833cecf18d85df41", 
        via = "1.1 google", `cf-cache-status` = "DYNAMIC", `x-robots-tag` = "none", 
        server = "cloudflare", `cf-ray` = "8d340fcc795b1cc9-FRA", 
        `content-encoding` = "gzip"), class = "httr2_headers"), 
    body = charToRaw("{\"id\":\"msg_01Lm6GLp6jUyEEN991EjirVr\",\"type\":\"message\",\"role\":\"assistant\",\"model\":\"claude-3-5-sonnet-20240620\",\"content\":[{\"type\":\"text\",\"text\":\"The capital of France is Paris.\"}],\"stop_reason\":\"end_turn\",\"stop_sequence\":null,\"usage\":{\"input_tokens\":20,\"output_tokens\":10}}"), 
    cache = new.env(parent = emptyenv())), class = "httr2_response")

#' Create or Update Large Language Model Message Object
#'
#' This function creates a new `LLMMessage` object or updates an existing one.
#' It supports adding text prompts and various media types, such as images, PDFs, text files, or plots.
#'
#' @param .llm An existing LLMMessage object or an initial text prompt.
#' @param .prompt Text prompt to add to the message history.
#' @param .role The role of the message sender, typically "user" or "assistant".
#' @param .system_prompt Default system prompt if a new LLMMessage needs to be created.
#' @param .media An inline media object or a list of them. Accepted types: `img()`, `audio_file()`, `video_file()`, `pdf_file()`.
#' @param .files A `tidyllm_file` object or list of them returned by `upload_file()`. These are remote file references stored on a provider's server.
#' @param .imagefile Path to an image file to be attached (optional). Deprecated; use `.media = img(path)` instead.
#' @param .pdf Path to a PDF file to be attached (optional). Deprecated; use `.media = pdf_file(path)` instead.
#' @param .textfile Path to a text file to be read and attached (optional).
#' @param .capture_plot Boolean to indicate whether a plot should be captured and attached as an image (optional).
#' @param .f An R function or an object coercible to a function via `rlang::as_function`, whose output should be captured and attached (optional).
#' @return Returns an updated or new LLMMessage object.
#' @importFrom jsonlite toJSON
#' @importFrom base64enc base64encode
#' @family Message Creation Utilities
#' @seealso [df_llm_message()]
#' @export
llm_message <- function(.llm = NULL,
                        .prompt = NULL,
                        .role = "user",
                        .system_prompt = "You are a helpful assistant",
                        .media = NULL,
                        .files = NULL,
                        .imagefile = NULL,
                        .pdf = NULL,
                        .textfile = NULL,
                        .capture_plot = FALSE,
                        .f = NULL) {

  media_list <- list()
  files_list <- NULL

  validate_inputs(c(
    ".llm must be NULL, a character vector, or an LLMMessage object" = is.null(.llm) || is.character(.llm) || S7_inherits(.llm, LLMMessage),
    ".prompt must be a non-empty string if provided" = is.null(.prompt) || (is.character(.prompt) && length(.prompt) == 1 && nchar(.prompt) > 0),
    ".role must be a non-empty string" = is.character(.role) && length(.role) == 1 && nchar(.role) > 0,
    ".system_prompt must be a string" = is.character(.system_prompt),
    ".imagefile must be NULL or a valid file path" = is.null(.imagefile) || (is.character(.imagefile) && length(.imagefile) == 1 && file.exists(.imagefile)),
    ".pdf must be NULL, a file path, or a list with filename, start_page, and end_page" = is.null(.pdf) ||
      (is.character(.pdf) && length(.pdf) == 1 && file.exists(.pdf)) ||
      (is.list(.pdf) && all(c("filename", "start_page", "end_page") %in% names(.pdf)) &&
         is.character(.pdf$filename) && length(.pdf$filename) == 1 && file.exists(.pdf$filename) &&
         is.numeric(.pdf$start_page) && is.numeric(.pdf$end_page) && .pdf$start_page >= 1 && .pdf$end_page >= .pdf$start_page),
    ".textfile must be NULL or a valid file path" = is.null(.textfile) || (is.character(.textfile) && length(.textfile) == 1 && file.exists(.textfile)),
    ".capture_plot must be logical" = is.logical(.capture_plot) && length(.capture_plot) == 1,
    ".f must be NULL or coercible to a function via rlang::as_function" = is.null(.f) || (tryCatch({rlang::as_function(.f); TRUE}, error = function(e) FALSE))
  ))

  pre_existing_object <- S7_inherits(.llm, LLMMessage)

  # Handle .media — accept a single S7 media object or a list
  if (!is.null(.media)) {
    media_items <- if (is.list(.media) && !S7_inherits(.media, tidyllm_image) &&
                       !S7_inherits(.media, tidyllm_audio) && !S7_inherits(.media, tidyllm_video) &&
                       !S7_inherits(.media, tidyllm_pdf)) {
      .media
    } else {
      list(.media)
    }
    media_list <- c(media_list, media_items)
  }

  # Handle .files — accept a tidyllm_file or list of tidyllm_file
  if (!is.null(.files)) {
    files_list <- if (S7_inherits(.files, tidyllm_file)) list(.files) else .files
  }

  # Handle .imagefile (deprecated)
  if (!is.null(.imagefile) || .capture_plot) {
    if (.capture_plot) {
      plot_file <- tempfile(fileext = ".png")
      dev.copy(png, filename = plot_file)
      dev.off()
      .imagefile <- plot_file
    }
    lifecycle::deprecate_warn("0.5.0", "llm_message(.imagefile=)", "llm_message(.media=)",
      details = "Replace with .media = img(path).")
    media_list <- c(media_list, list(img(.imagefile)))
  }

  # Handle .pdf (deprecated)
  if (!is.null(.pdf)) {
    lifecycle::deprecate_warn("0.5.0", "llm_message(.pdf=)", "llm_message(.media=)",
      details = "Replace with .media = pdf_file(path, .text_extract = TRUE) to preserve identical behaviour.")
    if (!requireNamespace("pdftools", quietly = TRUE)) {
      stop("The 'pdftools' package is required to read PDF files. Please install it.")
    }
    if (is.character(.pdf) && length(.pdf) == 1) {
      media_list <- c(media_list, list(pdf_file(.pdf, .text_extract = TRUE)))
    } else if (is.list(.pdf)) {
      pdf_path   <- .pdf$filename
      start_page <- .pdf$start_page
      end_page   <- .pdf$end_page
      pages      <- seq(start_page, end_page)
      media_list <- c(media_list, list(pdf_file(pdf_path, pages = pages, .text_extract = TRUE)))
    } else {
      stop(".pdf must be either a file path or a list with filename, start_page, and end_page.")
    }
  }

  # Handle .textfile
  if (!is.null(.textfile)) {
    text_content <- readLines(.textfile) |>
      stringr::str_c(collapse = "\n")
    media_list <- c(media_list, list(list(type = "TextFile", content = text_content, filename = basename(.textfile))))
  }

  # Handle .f
  if (!is.null(.f)) {
    output <- utils::capture.output(rlang::as_function(.f)(), file = NULL) |>
      stringr::str_c(collapse = "\n")
    media_list <- c(media_list, list(list(type = "RConsole", content = output, filename = "RConsole.txt")))
  }

  media_arg <- if (length(media_list) > 0) media_list else NULL

  if (is.character(.llm)) {
    initial_prompt <- .llm
    .llm <- LLMMessage(system_prompt = .system_prompt,
                       message_history = list(list(role = "system", content = .system_prompt)))
    if (is.null(.prompt)) {
      return(add_message(.llm, .role, initial_prompt, .media = media_arg, .files = files_list))
    }
  }

  if (is.null(.llm)) {
    .llm <- LLMMessage(system_prompt = .system_prompt,
                       message_history = list(list(role = "system", content = .system_prompt)))
  }

  if (!is.null(.prompt)) {
    return(add_message(.llm, .role, .prompt, .media = media_arg, .files = files_list))
  }
}

#' Convert a Data Frame to an LLMMessage Object
#'
#' This function converts a data frame into an `LLMMessage` object representing a conversation history.
#' The data frame must have specific columns (`role` and `content`), with each row representing a message.
#'
#' @param .df A data frame with at least two rows and columns `role` and `content`.
#'   The `role` column should contain "user", "assistant", or "system". The `content` column should contain the corresponding message text.
#' @return An `LLMMessage` object representing the structured conversation.
#' @family Message Creation Utilities
#' @seealso [llm_message()]
#' @export
df_llm_message <- function(.df){
  #Validate the inputs
  c(
    "Input .df must be a dataframe" = is.data.frame(.df),
    "Input .df must contain the columns role and content" = (sum(c("role","content") %in% names(.df))==2),
    "Input .df column content must be character" = is.character(.df$content),
    "Input .df column role  must be character" = is.character(.df$role),
    "Input .df must have at least two rows" = nrow(.df)>=2,
    "Only user, system and assistant are allowed as values of .df column role" =all(.df$role %in% c("user","assistant","system"))
  ) |>
    validate_inputs()
  
  # Splitting the dataframe into a list of single-row dataframes
  single_rows <- split(.df, seq(nrow(.df)))
  #Get the system prompt and the first message into a message history
  if(single_rows[[1]]$role=="system"){
    initial_message <- llm_message(.llm = single_rows[[2]]$content,.system_prompt = single_rows[[1]]$content)
  } else {
    initial_message <- llm_message(.llm = single_rows[[1]]$content)
  }
  
  #We are done if it is just a two-row data.frame
  if(nrow(.df)<=2){final_message <- initial_message}
  
  #For more than 3 rows we iteratively apply the messages to a message history
  if(nrow(.df)>2){
    last_output <- initial_message
    for (i in 3:nrow(.df)) {
      current_row <- single_rows[[i]]
      last_output <- add_message(llm = last_output,
                                 role = current_row$role,
                                 content=current_row$content)
    }
    final_message <- last_output
  }
  return(final_message)
}

testthat::skip_if_not_installed("httptest2")
library(httptest2)


test_that("gemini_chat function constructs a correct request and dry runs it", {
  # Create a mock LLMMessage object
  llm <- llm_message("Write a poem about an evil online search monopolist")
  
  # Call chatgpt with .dry_run = TRUE and perform the dry run
  request <- llm |>
    chat(gemini,.dry_run = TRUE) 
  
  dry_run <- request |>
    httr2::req_dry_run(redact_headers = TRUE, quiet = TRUE)
  
  # Check that the method is POST
  expect_equal(dry_run$method, "POST")
  
  # Check that the URL path is correct
  expect_true(grepl("/v1beta/models/gemini-2.5-flash:generateContent", dry_run$path))
  
  # Inspect headers
  headers <- dry_run$headers
  expect_type(headers, "list")
  
  expect_true("content-type" %in% names(headers))
  
  # Check that the content-type is JSON
  expect_equal(headers$`content-type`, "application/json")
  
  # Now check the body content to ensure the JSON is constructed as expected
  body_json <- request$body |> jsonlite::toJSON() |> as.character()
  
  expected_json <- "{\"data\":{\"model\":[\"gemini-2.5-flash\"],\"system_instruction\":{\"parts\":{\"text\":[\"You are a helpful assistant\"]}},\"contents\":[{\"role\":[\"user\"],\"parts\":{\"text\":[\"Write a poem about an evil online search monopolist \"]}}]},\"type\":[\"json\"],\"content_type\":[\"application/json\"],\"params\":{\"auto_unbox\":[true],\"digits\":[22],\"null\":[\"null\"]}}"
  # Check if the JSON matches the expected JSON
  expect_equal(body_json, expected_json)
  
})

test_that("llm_message creates a new LLMMessage with initial prompt", {
  llm <- llm_message("Hello, assistant!")
  expect_true(S7_inherits(llm, LLMMessage))
  expect_equal(length(llm@message_history), 2)
  expect_equal(llm@message_history[[2]]$role, "user")
  expect_equal(llm@message_history[[2]]$content, "Hello, assistant!")
})

test_that("llm_message adds a prompt to an existing LLMMessage", {
  llm <- llm_message("Hello, assistant!")
  llm <- llm_message(.llm = llm, .prompt = "Another message")
  expect_equal(length(llm@message_history), 3)
  expect_equal(llm@message_history[[3]]$content, "Another message")
})

test_that("llm_message handles image files", {
  # Create a temporary image file
  temp_img <- tempfile(fileext = ".png")
  png(temp_img)
  plot(1:10)
  dev.off()
  
  llm <- llm_message("Please analyze the attached image.", .imagefile = temp_img)
  expect_equal(length(llm@message_history), 2)
  media <- llm@message_history[[2]]$media
  expect_equal(length(media), 1)
  expect_equal(media[[1]]$type, "Image")
  expect_equal(media[[1]]$filename, basename(temp_img))
  
  # Clean up
  unlink(temp_img)
})

test_that("llm_message handles PDFs with file path", {
  pdf_file <- test_path("pdf-sample_150kB.pdf")
  
  llm <- llm_message("Please summarize the attached PDF.", .pdf = pdf_file)
  expect_equal(length(llm@message_history), 2)
  media <- llm@message_history[[2]]$media
  expect_equal(length(media), 1)
  expect_equal(media[[1]]$type, "PDF")
  expect_equal(media[[1]]$filename, basename(pdf_file))
})

test_that("llm_message handles PDFs with specified pages", {
  pdf_file <- test_path("pdf-sample_150kB.pdf")
  
  llm <- llm_message("Please summarize pages 2 to 3.",
    .pdf = list(filename = pdf_file, start_page = 2, end_page = 3)
  )
  expect_equal(length(llm@message_history), 2)
  media <- llm@message_history[[2]]$media
  expect_equal(length(media), 1)
  expect_equal(media[[1]]$type, "PDF")
  expect_equal(media[[1]]$filename, basename(pdf_file))
  
  # Check that the content corresponds to pages 2 and 3
  pdf_text <- pdftools::pdf_text(pdf_file)[2:3] |> stringr::str_c(collapse = "\n")
  expect_equal(media[[1]]$content, pdf_text)
})

test_that("llm_message adjusts end_page if it exceeds total pages", {
  # Create a sample PDF with 2 pages
  pdf_file <- test_path("pdf-sample_150kB.pdf")
  
  # Expect a warning about end_page adjustment
  expect_warning(
    llm_message("Please summarize pages 1 to 5.",
      .pdf = list(filename = pdf_file, start_page = 1, end_page = 100)
    ),
    "end_page is greater than total pages. Setting end_page to total pages."
  )
  
})

test_that("llm_message handles text files", {
  # Create a temporary text file
  temp_txt <- tempfile(fileext = ".txt")
  writeLines(c("Line 1", "Line 2"), temp_txt)
  
  llm <- llm_message("Please analyze the attached text file.", .textfile = temp_txt)
  expect_equal(length(llm@message_history), 2)
  media <- llm@message_history[[2]]$media
  expect_equal(length(media), 1)
  expect_equal(media[[1]]$type, "TextFile")
  expect_equal(media[[1]]$filename, basename(temp_txt))
  expect_equal(media[[1]]$content, "Line 1\nLine 2")
  
  # Clean up
  unlink(temp_txt)
})

test_that("llm_message handles function outputs with .f", {
  llm <- llm_message("Please analyze the function output.",
    .f = function() {
      print("Test output")
    }
  )
  expect_equal(length(llm@message_history), 2)
  media <- llm@message_history[[2]]$media
  expect_equal(length(media), 1)
  expect_equal(media[[1]]$type, "RConsole")
  expect_equal(media[[1]]$filename, "RConsole.txt")
  expect_equal(media[[1]]$content, "[1] \"Test output\"")
})

test_that("llm_message validates inputs and throws errors for invalid inputs", {
  expect_error(llm_message(.llm = 123), "must be NULL, a character vector, or an LLMMessage object")
  expect_error(llm_message(.prompt = ""), "must be a non-empty string if provided")
  expect_error(llm_message(.role = ""), "must be a non-empty string")
  expect_error(llm_message(.imagefile = "nonexistent.png"), "must be NULL or a valid file path")
  expect_error(llm_message(.pdf = "nonexistent.pdf"), ".pdf must be NULL, a file path, or a list with filename, start_page, and end_page")
  expect_error(
    llm_message(.pdf = list(filename = "nonexistent.pdf", start_page = 1, end_page = 2)),
    ".pdf must be NULL, a file path, or a list with filename, start_page, and end_page"
  )
  expect_error(llm_message(.textfile = "nonexistent.txt"), ".textfile must be NULL or a valid file path")
  expect_error(llm_message(.capture_plot = "yes"), "must be logical")
  expect_error(llm_message(.f = 123), "must be NULL or coercible to a function via rlang::as_function")
})

#strange discrepancy between devtools::check() and devtools::test()
#test_that("llm_message handles .capture_plot correctly", {
#  # Create a plot
#  plot(1:10)
#  
#  llm <- llm_message("Please analyze the attached plot.", .capture_plot = TRUE)
#  expect_equal(length(llm$message_history), 2)
#  media <- llm$message_history[[2]]$media
#  expect_equal(length(media), 1)
#  expect_equal(media[[1]]$type, "Image")
#  expect_true(grepl("\\.png$", media[[1]]$filename))
#})

test_that("llm_message handles initial prompt when .llm is character", {
  llm <- llm_message(.llm = "Initial prompt")
  expect_true(S7_inherits(llm, LLMMessage))
  expect_equal(length(llm@message_history), 2)
  expect_equal(llm@message_history[[2]]$role, "user")
  expect_equal(llm@message_history[[2]]$content, "Initial prompt")
})

test_that("llm_message handles anonymous functions with .f", {
  llm <- llm_message("Please analyze the function output.",
    .f = ~ print("Anonymous function output")
  )
  expect_equal(length(llm@message_history), 2)
  media <- llm@message_history[[2]]$media
  expect_equal(media[[1]]$content, "[1] \"Anonymous function output\"")
})

structure(list(method = "POST", url = "https://api.openai.com/v1/chat/completions", 
    status_code = 200L, headers = structure(list(date = "Wed, 16 Oct 2024 00:51:46 GMT", 
        `content-type` = "application/json", `access-control-expose-headers` = "X-Request-ID", 
        `openai-organization` = "user-j2dlcmauhfkzofrqjehceoc7", 
        `openai-processing-ms` = "4211", `openai-version` = "2020-10-01", 
        `x-ratelimit-limit-requests` = "5000", `x-ratelimit-limit-tokens` = "800000", 
        `x-ratelimit-remaining-requests` = "4999", `x-ratelimit-remaining-tokens` = "798188", 
        `x-ratelimit-reset-requests` = "12ms", `x-ratelimit-reset-tokens` = "135ms", 
        `x-request-id` = "req_f9be3b0cb133408bbbc7b441f3bcbc4e", 
        `strict-transport-security` = "max-age=31536000; includeSubDomains; preload", 
        `cf-cache-status` = "DYNAMIC", `set-cookie` = "REDACTED", 
        `x-content-type-options` = "nosniff", `set-cookie` = "REDACTED", 
        server = "cloudflare", `cf-ray` = "8d340fd98bdadc68-FRA", 
        `content-encoding` = "gzip", `alt-svc` = "h3=\":443\"; ma=86400"), class = "httr2_headers"), 
    body = charToRaw("{\n  \"id\": \"chatcmpl-AImYKyWVCkGgSaGcWhc02zgv693AK\",\n  \"object\": \"chat.completion\",\n  \"created\": 1729039904,\n  \"model\": \"gpt-4o-2024-08-06\",\n  \"choices\": [\n    {\n      \"index\": 0,\n      \"message\": {\n        \"role\": \"assistant\",\n        \"content\": \"The picture shows a picturesque landscape featuring a large body of water bordered by rugged mountains. Below, there's a town surrounded by lush greenery, including fields or vineyards. The sun is shining brightly, creating a serene and idyllic atmosphere. \\n\\nThis scene likely depicts a location similar to Lake Garda in Italy, known for its beautiful lakeside towns and mountainous surroundings.\",\n        \"refusal\": null\n      },\n      \"logprobs\": null,\n      \"finish_reason\": \"stop\"\n    }\n  ],\n  \"usage\": {\n    \"prompt_tokens\": 795,\n    \"completion_tokens\": 70,\n    \"total_tokens\": 865,\n    \"prompt_tokens_details\": {\n      \"cached_tokens\": 0\n    },\n    \"completion_tokens_details\": {\n      \"reasoning_tokens\": 0\n    }\n  },\n  \"system_fingerprint\": \"fp_a20a4ee344\"\n}\n"), 
    cache = new.env(parent = emptyenv())), class = "httr2_response")

structure(list(method = "GET", url = "https://api.openai.com/v1/batches?limit=20", 
    status_code = 200L, headers = structure(list(date = "Thu, 21 Aug 2025 12:41:49 GMT", 
        `content-type` = "application/json", `openai-version` = "2020-10-01", 
        `openai-organization` = "user-j2dlcmauhfkzofrqjehceoc7", 
        `openai-project` = "proj_wN2FU9W9VUckqc7nz44xte9H", `x-request-id` = "req_e2c3d7a1e53f39735069740eff16fe0e", 
        `openai-processing-ms` = "215", `x-envoy-upstream-service-time` = "217", 
        `strict-transport-security` = "max-age=31536000; includeSubDomains; preload", 
        `cf-cache-status` = "DYNAMIC", `set-cookie` = "REDACTED", 
        `x-content-type-options` = "nosniff", `set-cookie` = "REDACTED", 
        server = "cloudflare", `cf-ray` = "972a34efca31bbc2-FRA", 
        `content-encoding` = "gzip", `alt-svc` = "h3=\":443\"; ma=86400"), redact = character(0), class = "httr2_headers"), 
    body = charToRaw("{\n  \"object\": \"list\",\n  \"data\": [\n    {\n      \"id\": \"batch_680b8e3d92208190a7c2b25ef89cd8b5\",\n      \"object\": \"batch\",\n      \"endpoint\": \"/v1/chat/completions\",\n      \"errors\": null,\n      \"input_file_id\": \"file-8LjgrmGyywoGv8vrDb2mWb\",\n      \"completion_window\": \"24h\",\n      \"status\": \"completed\",\n      \"output_file_id\": \"file-D1Tn2JDcYztLWs1P59XUtx\",\n      \"error_file_id\": null,\n      \"created_at\": 1745587773,\n      \"in_progress_at\": 1745587775,\n      \"expires_at\": 1745674173,\n      \"finalizing_at\": 1745598803,\n      \"completed_at\": 1745598858,\n      \"failed_at\": null,\n      \"expired_at\": null,\n      \"cancelling_at\": null,\n      \"cancelled_at\": null,\n      \"request_counts\": {\n        \"total\": 575,\n        \"completed\": 575,\n        \"failed\": 0\n      },\n      \"metadata\": null\n    },\n    {\n      \"id\": \"batch_67df363bc4508190a0fb4160b4117949\",\n      \"object\": \"batch\",\n      \"endpoint\": \"/v1/chat/completions\",\n      \"errors\": null,\n      \"input_file_id\": \"file-J763TdywMfH4xVtc8gb1v5\",\n      \"completion_window\": \"24h\",\n      \"status\": \"completed\",\n      \"output_file_id\": \"file-5yu9BV9MPdneMkmHBN7Sfd\",\n      \"error_file_id\": null,\n      \"created_at\": 1742681659,\n      \"in_progress_at\": 1742681660,\n      \"expires_at\": 1742768059,\n      \"finalizing_at\": 1742681727,\n      \"completed_at\": 1742681730,\n      \"failed_at\": null,\n      \"expired_at\": null,\n      \"cancelling_at\": null,\n      \"cancelled_at\": null,\n      \"request_counts\": {\n        \"total\": 24,\n        \"completed\": 24,\n        \"failed\": 0\n      },\n      \"metadata\": null\n    },\n    {\n      \"id\": \"batch_67daa551c13c8190ab7b855832f423ca\",\n      \"object\": \"batch\",\n      \"endpoint\": \"/v1/chat/completions\",\n      \"errors\": null,\n      \"input_file_id\": \"file-Y1gaustRLWZW31bmgGPJHP\",\n      \"completion_window\": \"24h\",\n      \"status\": \"completed\",\n      \"output_file_id\": \"file-G8zouWYRcyxEvAi75yBE1u\",\n      \"error_file_id\": null,\n      \"created_at\": 1742382417,\n      \"in_progress_at\": 1742382418,\n      \"expires_at\": 1742468817,\n      \"finalizing_at\": 1742384829,\n      \"completed_at\": 1742384830,\n      \"failed_at\": null,\n      \"expired_at\": null,\n      \"cancelling_at\": null,\n      \"cancelled_at\": null,\n      \"request_counts\": {\n        \"total\": 4,\n        \"completed\": 4,\n        \"failed\": 0\n      },\n      \"metadata\": null\n    },\n    {\n      \"id\": \"batch_67daa459f6388190a343393776061bbe\",\n      \"object\": \"batch\",\n      \"endpoint\": \"/v1/chat/completions\",\n      \"errors\": null,\n      \"input_file_id\": \"file-A4amUoNcQz4aiB3MSuwCtx\",\n      \"completion_window\": \"24h\",\n      \"status\": \"completed\",\n      \"output_file_id\": \"file-JRMPb3f1v7wZWeMko7b5BZ\",\n      \"error_file_id\": null,\n      \"created_at\": 1742382170,\n      \"in_progress_at\": 1742382170,\n      \"expires_at\": 1742468570,\n      \"finalizing_at\": 1742382232,\n      \"completed_at\": 1742382232,\n      \"failed_at\": null,\n      \"expired_at\": null,\n      \"cancelling_at\": null,\n      \"cancelled_at\": null,\n      \"request_counts\": {\n        \"total\": 3,\n        \"completed\": 3,\n        \"failed\": 0\n      },\n      \"metadata\": null\n    },\n    {\n      \"id\": \"batch_67b9268b399c8190b134c840c184d160\",\n      \"object\": \"batch\",\n      \"endpoint\": \"/v1/chat/completions\",\n      \"errors\": null,\n      \"input_file_id\": \"file-2Z5UABm7zAUYPzBZP28SG1\",\n      \"completion_window\": \"24h\",\n      \"status\": \"completed\",\n      \"output_file_id\": \"file-RmWfqciD6d44ydhv1SUVrp\",\n      \"error_file_id\": null,\n      \"created_at\": 1740187275,\n      \"in_progress_at\": 1740187276,\n      \"expires_at\": 1740273675,\n      \"finalizing_at\": 1740187380,\n      \"completed_at\": 1740187382,\n      \"failed_at\": null,\n      \"expired_at\": null,\n      \"cancelling_at\": null,\n      \"cancelled_at\": null,\n      \"request_counts\": {\n        \"total\": 3,\n        \"completed\": 3,\n        \"failed\": 0\n      },\n      \"metadata\": null\n    },\n    {\n      \"id\": \"batch_67b66cd8678481909ce499abded372c6\",\n      \"object\": \"batch\",\n      \"endpoint\": \"/v1/chat/completions\",\n      \"errors\": null,\n      \"input_file_id\": \"file-Rr9bwuBaxKFSs5uVj9eZtT\",\n      \"completion_window\": \"24h\",\n      \"status\": \"completed\",\n      \"output_file_id\": \"file-KiBzbjkA1rhopE8cQh9FtY\",\n      \"error_file_id\": null,\n      \"created_at\": 1740008664,\n      \"in_progress_at\": 1740008664,\n      \"expires_at\": 1740095064,\n      \"finalizing_at\": 1740009030,\n      \"completed_at\": 1740009031,\n      \"failed_at\": null,\n      \"expired_at\": null,\n      \"cancelling_at\": null,\n      \"cancelled_at\": null,\n      \"request_counts\": {\n        \"total\": 4,\n        \"completed\": 4,\n        \"failed\": 0\n      },\n      \"metadata\": null\n    },\n    {\n      \"id\": \"batch_674c7bf16aa4819080dfa5e9b8efbdee\",\n      \"object\": \"batch\",\n      \"endpoint\": \"/v1/chat/completions\",\n      \"errors\": null,\n      \"input_file_id\": \"file-HXCQ23DJLvooBEeV6RCpxk\",\n      \"completion_window\": \"24h\",\n      \"status\": \"completed\",\n      \"output_file_id\": \"file-W2HSoUBnPfZZWdovVut7wC\",\n      \"error_file_id\": null,\n      \"created_at\": 1733065713,\n      \"in_progress_at\": 1733065713,\n      \"expires_at\": 1733152113,\n      \"finalizing_at\": 1733065794,\n      \"completed_at\": 1733065796,\n      \"failed_at\": null,\n      \"expired_at\": null,\n      \"cancelling_at\": null,\n      \"cancelled_at\": null,\n      \"request_counts\": {\n        \"total\": 31,\n        \"completed\": 31,\n        \"failed\": 0\n      },\n      \"metadata\": null\n    },\n    {\n      \"id\": \"batch_674b399f2e288190a9aad4d6d2b13587\",\n      \"object\": \"batch\",\n      \"endpoint\": \"/v1/chat/completions\",\n      \"errors\": null,\n      \"input_file_id\": \"file-548TC1XrHkJH6SPyNd3RHk\",\n      \"completion_window\": \"24h\",\n      \"status\": \"completed\",\n      \"output_file_id\": \"file-JcZ57aJ1SMQPdwZwjonQTT\",\n      \"error_file_id\": null,\n      \"created_at\": 1732983199,\n      \"in_progress_at\": 1732983199,\n      \"expires_at\": 1733069599,\n      \"finalizing_at\": 1732983234,\n      \"completed_at\": 1732983234,\n      \"failed_at\": null,\n      \"expired_at\": null,\n      \"cancelling_at\": null,\n      \"cancelled_at\": null,\n      \"request_counts\": {\n        \"total\": 3,\n        \"completed\": 3,\n        \"failed\": 0\n      },\n      \"metadata\": null\n    },\n    {\n      \"id\": \"batch_674b186550e8819087b2407dbc200ef7\",\n      \"object\": \"batch\",\n      \"endpoint\": \"/v1/chat/completions\",\n      \"errors\": null,\n      \"input_file_id\": \"file-2k454PBUtkhmg1XT9FtzA1\",\n      \"completion_window\": \"24h\",\n      \"status\": \"completed\",\n      \"output_file_id\": \"file-CzwinJRHXjg3Cbec8KMDKd\",\n      \"error_file_id\": null,\n      \"created_at\": 1732974693,\n      \"in_progress_at\": 1732974694,\n      \"expires_at\": 1733061093,\n      \"finalizing_at\": 1732974706,\n      \"completed_at\": 1732974707,\n      \"failed_at\": null,\n      \"expired_at\": null,\n      \"cancelling_at\": null,\n      \"cancelled_at\": null,\n      \"request_counts\": {\n        \"total\": 3,\n        \"completed\": 3,\n        \"failed\": 0\n      },\n      \"metadata\": null\n    },\n    {\n      \"id\": \"batch_674b17e882f481909f438bec763b0375\",\n      \"object\": \"batch\",\n      \"endpoint\": \"/v1/chat/completions\",\n      \"errors\": null,\n      \"input_file_id\": \"file-HXuoMmZTYCDcB19bTDYcNC\",\n      \"completion_window\": \"24h\",\n      \"status\": \"completed\",\n      \"output_file_id\": \"file-X3j7WdUh5YyFxJ13WtobqB\",\n      \"error_file_id\": null,\n      \"created_at\": 1732974568,\n      \"in_progress_at\": 1732974569,\n      \"expires_at\": 1733060968,\n      \"finalizing_at\": 1732974578,\n      \"completed_at\": 1732974579,\n      \"failed_at\": null,\n      \"expired_at\": null,\n      \"cancelling_at\": null,\n      \"cancelled_at\": null,\n      \"request_counts\": {\n        \"total\": 3,\n        \"completed\": 3,\n        \"failed\": 0\n      },\n      \"metadata\": null\n    },\n    {\n      \"id\": \"batch_673f45f04ed88190b54425d1ad02097d\",\n      \"object\": \"batch\",\n      \"endpoint\": \"/v1/chat/completions\",\n      \"errors\": null,\n      \"input_file_id\": \"file-lbYKD1lAVtqELKXv57YeyQIx\",\n      \"completion_window\": \"24h\",\n      \"status\": \"completed\",\n      \"output_file_id\": \"file-Vf06YkflOrr09vd1fbaSg8l9\",\n      \"error_file_id\": null,\n      \"created_at\": 1732199920,\n      \"in_progress_at\": 1732199921,\n      \"expires_at\": 1732286320,\n      \"finalizing_at\": 1732201331,\n      \"completed_at\": 1732201335,\n      \"failed_at\": null,\n      \"expired_at\": null,\n      \"cancelling_at\": null,\n      \"cancelled_at\": null,\n      \"request_counts\": {\n        \"total\": 24,\n        \"completed\": 24,\n        \"failed\": 0\n      },\n      \"metadata\": null\n    },\n    {\n      \"id\": \"batch_673df61b7fa881909a3636f92a8edbba\",\n      \"object\": \"batch\",\n      \"endpoint\": \"/v1/chat/completions\",\n      \"errors\": null,\n      \"input_file_id\": \"file-x6a5otNxlsHZJae5N5UOd0lW\",\n      \"completion_window\": \"24h\",\n      \"status\": \"completed\",\n      \"output_file_id\": \"file-SNR0fMQAFJlT5CTCz1VoM3fQ\",\n      \"error_file_id\": null,\n      \"created_at\": 1732113947,\n      \"in_progress_at\": 1732113948,\n      \"expires_at\": 1732200347,\n      \"finalizing_at\": 1732113975,\n      \"completed_at\": 1732113976,\n      \"failed_at\": null,\n      \"expired_at\": null,\n      \"cancelling_at\": null,\n      \"cancelled_at\": null,\n      \"request_counts\": {\n        \"total\": 3,\n        \"completed\": 3,\n        \"failed\": 0\n      },\n      \"metadata\": null\n    },\n    {\n      \"id\": \"batch_67336038ab588190a73c21909032d51a\",\n      \"object\": \"batch\",\n      \"endpoint\": \"/v1/chat/completions\",\n      \"errors\": null,\n      \"input_file_id\": \"file-T4kerjYXuZNyHaiFauZt5xBF\",\n      \"completion_window\": \"24h\",\n      \"status\": \"completed\",\n      \"output_file_id\": \"file-DbEjAkYOZjADNKBytBroXtNL\",\n      \"error_file_id\": null,\n      \"created_at\": 1731420216,\n      \"in_progress_at\": 1731420217,\n      \"expires_at\": 1731506616,\n      \"finalizing_at\": 1731420433,\n      \"completed_at\": 1731420434,\n      \"failed_at\": null,\n      \"expired_at\": null,\n      \"cancelling_at\": null,\n      \"cancelled_at\": null,\n      \"request_counts\": {\n        \"total\": 2,\n        \"completed\": 2,\n        \"failed\": 0\n      },\n      \"metadata\": null\n    },\n    {\n      \"id\": \"batch_67335c97baa881908914084a447a75a8\",\n      \"object\": \"batch\",\n      \"endpoint\": \"/v1/chat/completions\",\n      \"errors\": null,\n      \"input_file_id\": \"file-JnzHT7Dg4upZpstY2diCxkbq\",\n      \"completion_window\": \"24h\",\n      \"status\": \"completed\",\n      \"output_file_id\": \"file-5D1wTdE8Z5WUAhNCVjPu25dg\",\n      \"error_file_id\": null,\n      \"created_at\": 1731419287,\n      \"in_progress_at\": 1731419288,\n      \"expires_at\": 1731505687,\n      \"finalizing_at\": 1731453625,\n      \"completed_at\": 1731453626,\n      \"failed_at\": null,\n      \"expired_at\": null,\n      \"cancelling_at\": null,\n      \"cancelled_at\": null,\n      \"request_counts\": {\n        \"total\": 2,\n        \"completed\": 2,\n        \"failed\": 0\n      },\n      \"metadata\": null\n    },\n    {\n      \"id\": \"batch_672e77b2dd588190b2f751653989e98e\",\n      \"object\": \"batch\",\n      \"endpoint\": \"/v1/chat/completions\",\n      \"errors\": null,\n      \"input_file_id\": \"file-REqzkDJPSzOD8iVlztZdjnjH\",\n      \"completion_window\": \"24h\",\n      \"status\": \"completed\",\n      \"output_file_id\": \"file-nK0YxSWaAlxd36JqMc98rOvn\",\n      \"error_file_id\": null,\n      \"created_at\": 1731098547,\n      \"in_progress_at\": 1731098547,\n      \"expires_at\": 1731184947,\n      \"finalizing_at\": 1731098552,\n      \"completed_at\": 1731098553,\n      \"failed_at\": null,\n      \"expired_at\": null,\n      \"cancelling_at\": null,\n      \"cancelled_at\": null,\n      \"request_counts\": {\n        \"total\": 2,\n        \"completed\": 2,\n        \"failed\": 0\n      },\n      \"metadata\": null\n    },\n    {\n      \"id\": \"batch_672e7751c0148190bd8a56a9e1a97859\",\n      \"object\": \"batch\",\n      \"endpoint\": \"/v1/chat/completions\",\n      \"errors\": null,\n      \"input_file_id\": \"file-lU6SesGKEgH13pAivF30orc6\",\n      \"completion_window\": \"24h\",\n      \"status\": \"completed\",\n      \"output_file_id\": \"file-yD2R2GbVBEEYIR3CRb1EajXk\",\n      \"error_file_id\": null,\n      \"created_at\": 1731098449,\n      \"in_progress_at\": 1731098450,\n      \"expires_at\": 1731184849,\n      \"finalizing_at\": 1731098455,\n      \"completed_at\": 1731098456,\n      \"failed_at\": null,\n      \"expired_at\": null,\n      \"cancelling_at\": null,\n      \"cancelled_at\": null,\n      \"request_counts\": {\n        \"total\": 2,\n        \"completed\": 2,\n        \"failed\": 0\n      },\n      \"metadata\": null\n    },\n    {\n      \"id\": \"batch_6726c748bd588190a506a1c1882a3741\",\n      \"object\": \"batch\",\n      \"endpoint\": \"/v1/chat/completions\",\n      \"errors\": null,\n      \"input_file_id\": \"file-3JEyZLqIX0qj9o2pY96MpHNv\",\n      \"completion_window\": \"24h\",\n      \"status\": \"completed\",\n      \"output_file_id\": \"file-bbCv9zivkbf1ulnjf15MjVHx\",\n      \"error_file_id\": null,\n      \"created_at\": 1730594632,\n      \"in_progress_at\": 1730594634,\n      \"expires_at\": 1730681032,\n      \"finalizing_at\": 1730594681,\n      \"completed_at\": 1730594683,\n      \"failed_at\": null,\n      \"expired_at\": null,\n      \"cancelling_at\": null,\n      \"cancelled_at\": null,\n      \"request_counts\": {\n        \"total\": 24,\n        \"completed\": 24,\n        \"failed\": 0\n      },\n      \"metadata\": null\n    },\n    {\n      \"id\": \"batch_672671ea162c8190993e5871d9226fa0\",\n      \"object\": \"batch\",\n      \"endpoint\": \"/v1/chat/completions\",\n      \"errors\": null,\n      \"input_file_id\": \"file-6ECf6nJBp1NzglmoFQ1NXet5\",\n      \"completion_window\": \"24h\",\n      \"status\": \"completed\",\n      \"output_file_id\": \"file-zARt37HQsMaPOqeGytw6G24M\",\n      \"error_file_id\": null,\n      \"created_at\": 1730572778,\n      \"in_progress_at\": 1730572839,\n      \"expires_at\": 1730659178,\n      \"finalizing_at\": 1730573447,\n      \"completed_at\": 1730573448,\n      \"failed_at\": null,\n      \"expired_at\": null,\n      \"cancelling_at\": null,\n      \"cancelled_at\": null,\n      \"request_counts\": {\n        \"total\": 4,\n        \"completed\": 4,\n        \"failed\": 0\n      },\n      \"metadata\": null\n    },\n    {\n      \"id\": \"batch_67265e33b0148190b48d9d0c11f04ae1\",\n      \"object\": \"batch\",\n      \"endpoint\": \"/v1/chat/completions\",\n      \"errors\": null,\n      \"input_file_id\": \"file-bVodIkZMrEUuw8FOK4NmHQkn\",\n      \"completion_window\": \"24h\",\n      \"status\": \"completed\",\n      \"output_file_id\": null,\n      \"error_file_id\": \"file-NwRZbUZtQHlaEC81WkTulhc2\",\n      \"created_at\": 1730567731,\n      \"in_progress_at\": 1730567732,\n      \"expires_at\": 1730654131,\n      \"finalizing_at\": 1730567736,\n      \"completed_at\": 1730567737,\n      \"failed_at\": null,\n      \"expired_at\": null,\n      \"cancelling_at\": null,\n      \"cancelled_at\": null,\n      \"request_counts\": {\n        \"total\": 4,\n        \"completed\": 0,\n        \"failed\": 4\n      },\n      \"metadata\": null\n    },\n    {\n      \"id\": \"batch_67265d847bd48190afc0006531008ae4\",\n      \"object\": \"batch\",\n      \"endpoint\": \"/v1/chat/completions\",\n      \"errors\": null,\n      \"input_file_id\": \"file-io9wSihisrt4jqY6eWdnBWXs\",\n      \"completion_window\": \"24h\",\n      \"status\": \"completed\",\n      \"output_file_id\": \"file-AOOihugly3xB4S79wEZ0jTzw\",\n      \"error_file_id\": null,\n      \"created_at\": 1730567556,\n      \"in_progress_at\": 1730567557,\n      \"expires_at\": 1730653956,\n      \"finalizing_at\": 1730567560,\n      \"completed_at\": 1730567561,\n      \"failed_at\": null,\n      \"expired_at\": null,\n      \"cancelling_at\": null,\n      \"cancelled_at\": null,\n      \"request_counts\": {\n        \"total\": 2,\n        \"completed\": 2,\n        \"failed\": 0\n      },\n      \"metadata\": null\n    }\n  ],\n  \"first_id\": \"batch_680b8e3d92208190a7c2b25ef89cd8b5\",\n  \"last_id\": \"batch_67265d847bd48190afc0006531008ae4\",\n  \"has_more\": true\n}"), 
    cache = new.env(parent = emptyenv())), class = "httr2_response")

#' Create a Provider Function for Routing LLM Actions
#'
#' This function generates a provider-specific function that dynamically routes 
#' different LLM-related actions (e.g., chat, generate embeddings, batch operations) 
#' based on the `.called_from` argument or the presence of an `LLMMessage` object.
#'
#' @param .name A string representing the name of the provider (e.g., "openai").
#' @param ... Named functions corresponding to the various actions the provider supports 
#'   (e.g., `chat`, `embed`, `send_batch`).
#'
#' @return A function that dynamically routes to the appropriate action based on its 
#'   inputs.
#' @noRd 
dispatch_to_provider <- function(provider_expr, verb_name, common_args,
                                  validate = TRUE, env = rlang::caller_env()) {
  if (validate) {
    meta      <- rlang::call_modify(provider_expr, .called_from = "metadata") |> rlang::eval_tidy()
    supported <- meta$supported_args[[verb_name]]
    unsupported <- setdiff(names(common_args), supported)
    if (length(unsupported) > 0) {
      stop(glue::glue(
        "The following arguments are not supported by the provider's `{verb_name}()` function: {paste(unsupported, collapse = ', ')}."
      ))
    }
    valid_args <- common_args[names(common_args) %in% supported]
  } else {
    valid_args <- common_args
  }
  valid_args$.called_from <- verb_name
  rlang::eval_tidy(rlang::call_modify(provider_expr, !!!valid_args), env = env)
}

create_provider_function <- function(.name, ...) {
  function_map <- list(...)
  
  if ("metadata" %in% names(function_map)) {
    stop(glue::glue("'metadata' is a reserved keyword and cannot be used as a function name in provider '{.name}'."))
  }
  
  # Capture the arguments used in the provider function
  supported_args <- purrr::map(function_map, formals) |> 
    purrr::map(names)
  
  # Create the provider function
  provider_function <- function(..., .called_from = NULL) {
    args <- list(...)
    
    if (!is.null(.called_from)) {
      if (.called_from == "metadata") {
        # Return provider metadata
        return(list(
          provider_name = .name,
          supported_args = supported_args
        ))
      } else if (.called_from %in% names(function_map)) {
        # Call the appropriate function
        return(do.call(function_map[[.called_from]], args))
      } else {
        stop(glue::glue("'{.called_from}' is not available for API-provider '{.name}'."), call. = FALSE)
      }
    }
    
    # Check if any argument is an LLMMessage
    if (any(purrr::map_lgl(args, ~ S7_inherits(.x, LLMMessage)))) {
      lifecycle::deprecate_warn(
        when = "0.2.3",
        what = glue::glue("{.name}(.llm)"),
        details = glue::glue(
          "Passing an `LLMMessage` directly to `{.name}()` is deprecated.\n",
          "Please use either `chat({.name}(...))` for a verb-based approach, or `{.name}_chat()`."
        ))
      return(do.call(function_map[["chat"]], args))
    }
    
    # Default: Construct and return an unevaluated call to the provider function
    return(rlang::call2(.name, !!!args))
  }
  
  # Return the provider function with metadata accessible via "metadata" call
  provider_function
}



#' Chat with a Language Model
#'
#' The `chat()` function sends a message to a language model via a specified provider and returns the response. 
#' It routes the provided `LLMMessage` object to the appropriate provider-specific chat function, 
#' while allowing for the specification of common arguments applicable across different providers.
#'
#' @param .llm An `LLMMessage` object containing the message or conversation history to send to the language model.
#' @param .provider A function or function call specifying the language model provider and any additional parameters.
#'   This should be a call to a provider function like `openai()`, `claude()`, etc. 
#'   You can also set a default provider function via the `tidyllm_chat_default` option.
#' @param .dry_run Logical; if `TRUE`, simulates the request without sending it to the provider. Useful for testing.
#' @param .stream Logical; if `TRUE`, streams the response from the provider in real-time.
#' @param .temperature Numeric; controls the randomness of the model's output (0 = deterministic).
#' @param .timeout Numeric; the maximum time (in seconds) to wait for a response.
#' @param .top_p Numeric; nucleus sampling parameter, which limits the sampling to the top cumulative probability `p`.
#' @param .max_tries Integer; the maximum number of retries for failed requests.
#' @param .model Character; the model identifier to use (e.g., `"gpt-4"`).
#' @param .verbose Logical; if `TRUE`, prints additional information about the request and response.
#' @param .json_schema List; A JSON schema object as R list to enforce the output structure 
#' @param .tools Either a single TOOL object or a list of TOOL objects representing the available functions for tool calls.
#' @param .max_tool_rounds Integer; the maximum number of tool use iterations for multi-turn tool calling (default varies by provider).
#' @param .seed Integer; sets a random seed for reproducibility.
#' @param .stop Character vector; specifies sequences where the model should stop generating further tokens.
#' @param .frequency_penalty Numeric; adjusts the likelihood of repeating tokens (positive values decrease repetition).
#' @param .presence_penalty Numeric; adjusts the likelihood of introducing new tokens (positive values encourage novelty).
#'
#' @return An updated `LLMMessage` object containing the response from the language model.
#'
#' @examples
#' \dontrun{
#' # Basic usage with OpenAI provider
#' llm_message("Hello World") |>
#'    chat(ollama(.ollama_server = "https://my-ollama-server.de"),.model="mixtral")
#'    
# llm_message("Hello World") |>
#'    chat(mistral,.model="mixtral")
#'
#' # Use streaming with Claude provider
#' llm_message("Tell me a story") |>
#'    chat(claude(),.stream=TRUE)
#' }
#'
#' @details
#' The `chat()` function provides a unified interface for interacting with different language model providers.
#' Common arguments such as `.temperature`, `.model`, and `.stream` are supported by most providers and can be
#' passed directly to `chat()`. If a provider does not support a particular argument, an error will be raised.
#'
#' Advanced provider-specific configurations can be accessed via the provider functions. 
#'
#' @export
chat <- function(
    .llm,
    .provider = getOption("tidyllm_chat_default"),
    .dry_run = NULL,
    .stream = NULL,
    .temperature = NULL,
    .timeout = NULL,
    .top_p = NULL,
    .max_tries = NULL,
    .model = NULL,
    .verbose = NULL,
    .json_schema = NULL,
    .tools = NULL,
    .max_tool_rounds = NULL,
    .seed = NULL,
    .stop = NULL,
    .frequency_penalty = NULL,
    .presence_penalty = NULL) {
  
  if (!S7_inherits(.llm, LLMMessage)) {
    stop("Input .llm must be an LLMMessage object.")
  }
  
  if (is.null(.provider)) {
    stop("You need to specify a .provider function in chat().")
  }
  
  # Evaluate or capture the provider
  if (rlang::is_function(.provider)) {
    .provider <- .provider()
  }
  
  # Capture the provider call expression
  if(!rlang::is_call(.provider)){
    provider_expr <- rlang::quo_get_expr(rlang::enquo(.provider))
  } else {
    provider_expr <- .provider
  }
  
  common_args <- list(
    .llm = .llm,
    .model = .model,
    .verbose = .verbose,
    .max_tries = .max_tries,
    .stream = .stream,
    .timeout = .timeout,
    .temperature = .temperature,
    .dry_run = .dry_run,
    .top_p = .top_p,
    .json_schema = .json_schema,
    .seed = .seed,
    .stop = .stop,
    .frequency_penalty = .frequency_penalty,
    .presence_penalty = .presence_penalty,
    .tools = .tools,
    .max_tool_rounds = .max_tool_rounds
  )
  common_args <- common_args[!sapply(common_args, is.null)]

  return(dispatch_to_provider(provider_expr, "chat", common_args))
}


#' Generate text embeddings
#'
#' The `embed()` function allows you to embed a text via a specified provider.
#' It routes the input to the appropriate provider-specific embedding function.
#'
#' @param .input  A character vector of texts v, a list of texts and image objects,  or an `LLMMessage` object
#' @param .provider A function or function call specifying the language model provider and any additional parameters.
#'   This should be a call to a provider function like `openai()`, `ollama()`, etc. 
#'   You can also set a default provider function via the `tidyllm_embed_default` option.
#' @param .model The embedding model to use
#' @param .truncate Whether to truncate inputs to fit the model's context length
#' @param .timeout Timeout for the API request in seconds
#' @param .dry_run If TRUE, perform a dry run and return the request object.
#' @param .max_tries Maximum retry attempts for requests
#' @return A tibble with two columns: `input` and `embeddings`. 
#' The `input` column contains the texts sent to embed, and the `embeddings` column 
#' is a list column where each row contains an embedding vector of the sent input.
#' @examples
#' \dontrun{
#'c("What is the meaning of life, the universe and everything?",
#'  "How much wood would a woodchuck chuck?",
#'  "How does the brain work?") |>
#'  embed(gemini)
#'  }
#' @export
embed <- function(.input, 
                  .provider = getOption("tidyllm_embed_default"),
                  .model=NULL,
                  .truncate=NULL,
                  .timeout=NULL,
                  .dry_run=NULL,
                  .max_tries=NULL) {

  # Validate the inputs
  c(
    "Input .input must be a character vector, a list or an LLMMessage object" = S7_inherits(.input, LLMMessage) | is.character(.input) | is.list(.input),
    "You need to specify a .provider function in embed()" = !is.null(.provider)
  ) |> validate_inputs()
  
  #Accept an unevaluated function
  if(rlang::is_function(.provider)){
    .provider <- .provider()
  }
  
  # Capture the provider call expression
  if(!rlang::is_call(.provider)){
    provider_expr <- rlang::quo_get_expr(rlang::enquo(.provider))
  } else {
    provider_expr <- .provider
  }
  
  
  common_args <- list(
    .input = .input,
    .model = .model,
    .truncate = .truncate,
    .max_tries = .max_tries,
    .timeout = .timeout,
    .dry_run = .dry_run
  )
  common_args <- common_args[!sapply(common_args, is.null)]

  return(dispatch_to_provider(provider_expr, "embed", common_args))
}


#' Send a batch of messages to a batch API
#'
#' The `send_batch()` function allows you to send a list of `LLMMessage` objects
#' to an API.
#' It routes the input to the appropriate provider-specific batch API function.
#'
#' @param .llms A list of `LLMMessage` objects containing conversation histories.
#' @param .provider A function or function call specifying the language model provider and any additional parameters.
#'   This should be a call to a provider function like `openai()`, `claude()`, etc. 
#'   You can also set a default provider function via the `tidyllm_sbatch_default` option.
#' @param .model Character; the model identifier to use (e.g., `"gpt-4"`).
#' @param .temperature Numeric; controls the randomness of the model's output (0 = deterministic).
#' @param .timeout Numeric; the maximum time (in seconds) to wait for a response.
#' @param .top_p Numeric; nucleus sampling parameter, which limits the sampling to the top cumulative probability `p`.
#' @param .max_tries Integer; the maximum number of retries for failed requests.
#' @param .verbose Logical; if `TRUE`, prints additional information about the request and response.
#' @param .json_schema List; A JSON schema object as R list to enforce the output structure 
#' @param .seed Integer; sets a random seed for reproducibility.
#' @param .stop Character vector; specifies sequences where the model should stop generating further tokens.
#' @param .frequency_penalty Numeric; adjusts the likelihood of repeating tokens (positive values decrease repetition).
#' @param .presence_penalty Numeric; adjusts the likelihood of introducing new tokens (positive values encourage novelty).
#' @param .dry_run Logical; if `TRUE`, simulates the request without sending it to the provider. Useful for testing.
#' @param .id_prefix Character string to specify a prefix for generating custom IDs when names in `.llms` are missing 
#' @return An updated and named list of `.llms` with identifiers that align with batch responses, including a `batch_id` attribute.
#' @export
send_batch <- function(.llms, 
                       .provider = getOption("tidyllm_sbatch_default"),
                       .dry_run = NULL,
                       .temperature = NULL,
                       .timeout = NULL,
                       .top_p = NULL,
                       .max_tries = NULL,
                       .model = NULL,
                       .verbose = NULL,
                       .json_schema = NULL,
                       .seed = NULL,
                       .stop = NULL,
                       .frequency_penalty = NULL,
                       .presence_penalty = NULL,
                       .id_prefix = NULL){

  # Validate the inputs
  c(
    ".llms must be a list of LLMMessage objects" = is.list(.llms) && all(sapply(.llms, S7_inherits, LLMMessage)),
    "You need to specify a .provider function in send_batch()" = !is.null(.provider)
  ) |> validate_inputs()
  
  #Accept an unevaluated function
  if(rlang::is_function(.provider)){
    .provider <- .provider()
  }
  
  # Capture the provider call expression
  if(!rlang::is_call(.provider)){
    provider_expr <- rlang::quo_get_expr(rlang::enquo(.provider))
  } else {
    provider_expr <- .provider
  }
  
  common_args <- list(
    .llms = .llms,
    .model = .model,
    .verbose = .verbose,
    .max_tries = .max_tries,
    .timeout = .timeout,
    .temperature = .temperature,
    .dry_run = .dry_run,
    .top_p = .top_p,
    .json_schema = .json_schema,
    .seed = .seed,
    .stop = .stop,
    .frequency_penalty = .frequency_penalty,
    .presence_penalty = .presence_penalty,
    .id_prefix = .id_prefix
  )
  common_args <- common_args[!sapply(common_args, is.null)]

  return(dispatch_to_provider(provider_expr, "send_batch", common_args))
}

#' Check Batch Processing Status 
#'
#' This function retrieves the processing status and other details of a specified 
#' batchid or a list of `LLMMessage` objects with batch attribute.
#' It routes the input to the appropriate provider-specific batch API function.
#'
#' @param .llms A list of `LLMMessage` objects or a character vector with a batch ID.
#' @param .provider A function or function call specifying the language model provider and any additional parameters.
#'   This should be a call to a provider function like `openai()`, `claude()`, etc. 
#'   You can also set a default provider function via the `tidyllm_cbatch_default` option.
#' @param .dry_run Logical; if TRUE, returns the prepared request object without executing it 
#' @param .max_tries Maximum retries to perform the request
#' @param .timeout Integer specifying the request timeout in seconds
#' @return A tibble with information about the status of batch processing.
#' @export
check_batch <- function(.llms,
                        .provider = getOption("tidyllm_cbatch_default"),
                        .dry_run  = NULL,
                        .max_tries = NULL,
                        .timeout = NULL) {

  # Validate the inputs
  c(
    ".llms must be a list of LLMMessage objects or a character vector with a Batch ID" = (is.list(.llms) && all(sapply(.llms, S7_inherits, LLMMessage))) | is.character(.llms),
    "You need to specify a .provider function in check_batch()" = !is.null(.provider)
  ) |> validate_inputs()
  
  
  # Extract batch_id
  if((is.list(.llms) && all(sapply(.llms, S7_inherits, LLMMessage)))){
    batch_id <- attr(.llms, "batch_id")
  }
  if(is.character(.llms)){
    batch_id <- .llms
  }
  
  #Accept an unevaluated function
  if(rlang::is_function(.provider)){
    .provider <- .provider()
  }
  
  # Capture the provider call expression
  if(!rlang::is_call(.provider)){
    provider_expr <- rlang::quo_get_expr(rlang::enquo(.provider))
  } else {
    provider_expr <- .provider
  }
  
  
  common_args <- list(
    .batch_id = batch_id,
    .max_tries = .max_tries,
    .timeout = .timeout,
    .dry_run = .dry_run
  )
  common_args <- common_args[!sapply(common_args, is.null)]

  return(dispatch_to_provider(provider_expr, "check_batch", common_args, validate = FALSE))
}

#' List all Batch Requests on a Batch API
#'
#' @param .provider A function or function call specifying the language model provider and any additional parameters.
#'   This should be a call to a provider function like `openai()`, `claude()`, etc. 
#'   You can also set a default provider function via the `tidyllm_lbatch_default` option.
#' @return A tibble with information about the status of batch processing.
#' @export
list_batches <- function(.provider = getOption("tidyllm_lbatch_default")) {
  
  if (is.null(.provider)) {
    stop("You need to specify a .provider function in list_batches().")
  }
  
  #Accept an unevaluated function
  if(rlang::is_function(.provider)){
    .provider <- .provider()
  }
  
  # Capture the provider call expression
  if(!rlang::is_call(.provider)){
    provider_expr <- rlang::quo_get_expr(rlang::enquo(.provider))
  } else {
    provider_expr <- .provider
  }
  
  return(dispatch_to_provider(provider_expr, "list_batches", list(), validate = FALSE))
}


#' Fetch Results from a Batch API
#'
#' This function retrieves the results of a completed  batch and updates
#' the provided list of `LLMMessage` objects with the responses. It aligns each
#' response with the original request using the `custom_id`s generated in `send_batch()`.
#'
#' The function routes the input to the appropriate provider-specific batch API function.
#'
#' @param .llms A list of `LLMMessage` objects containing conversation histories.
#' @param .provider A function or function call specifying the language model provider and any additional parameters.
#'   This should be a call to a provider function like `openai()`, `claude()`, etc. 
#'   You can also set a default provider function via the `tidyllm_fbatch_default` option.
#' @param .dry_run Logical; if `TRUE`, returns the constructed request without executing it
#' @param .max_tries Integer; maximum number of retries if the request fails
#' @param .timeout Integer; request timeout in seconds
#' @return A list of updated `LLMMessage` objects, each with the assistant's response added if successful.
#' @export
fetch_batch <- function(.llms, 
                        .provider = getOption("tidyllm_fbatch_default"),
                        .dry_run=NULL,
                        .max_tries=NULL,
                        .timeout=NULL) {

  # Validate the inputs
  c(
    ".llms must be a list of LLMMessage objects with names as custom IDs" = is.list(.llms) && all(sapply(.llms, S7_inherits, LLMMessage)),
    "You need to specify a .provider function in fetch_batch()" = !is.null(.provider)
  ) |> validate_inputs()
  
  #Accept an unevaluated function
  if(rlang::is_function(.provider)){
    .provider <- .provider()
  }
  
  # Capture the provider call expression
  if(!rlang::is_call(.provider)){
    provider_expr <- rlang::quo_get_expr(rlang::enquo(.provider))
  } else {
    provider_expr <- .provider
  }
  
  common_args <- list(
    .llms = .llms,
    .max_tries = .max_tries,
    .timeout = .timeout,
    .dry_run = .dry_run
  )
  common_args <- common_args[!sapply(common_args, is.null)]

  return(dispatch_to_provider(provider_expr, "fetch_batch", common_args, validate = FALSE))
}

#' Run Deep Research via a Provider
#'
#' The `deep_research()` function sends a message to a provider's deep research endpoint.
#' Currently supported: Perplexity (`sonar-deep-research` via async API).
#'
#' @param .llm An `LLMMessage` object containing the research question.
#' @param .provider A function or function call specifying the provider (e.g., `perplexity()`).
#' @param .background Logical; if TRUE, returns a `tidyllm_research_job` immediately (default: FALSE).
#' @param ... Additional arguments passed to the provider's deep research function.
#'
#' @return If `.background = FALSE`, an `LLMMessage` with the research reply.
#'   If `.background = TRUE`, a `tidyllm_research_job` for use with `check_job()`/`fetch_job()`.
#' @export
deep_research <- function(.llm, .provider, .background = FALSE, ...) {
  if (!S7_inherits(.llm, LLMMessage)) {
    stop("Input .llm must be an LLMMessage object.")
  }
  if (is.null(.provider)) {
    stop("You need to specify a .provider function in deep_research().")
  }
  if (rlang::is_function(.provider)) {
    .provider <- .provider()
  }
  if (!rlang::is_call(.provider)) {
    provider_expr <- rlang::quo_get_expr(rlang::enquo(.provider))
  } else {
    provider_expr <- .provider
  }
  common_args <- list(.llm = .llm, .background = .background, ...)
  return(dispatch_to_provider(provider_expr, "deep_research", common_args))
}


#' Check the Status of a Batch or Research Job
#'
#' `check_job()` dispatches to `check_batch()` for batch objects or
#' `perplexity_check_research()` for `tidyllm_research_job` objects.
#'
#' @param .job An object with a `batch_id` attribute (from `send_batch()`) or
#'   a `tidyllm_research_job` (from `deep_research(.background = TRUE)`).
#' @param ... Additional arguments passed to the underlying function.
#' @return Status information; type depends on `.job` class.
#' @export
check_job <- function(.job, ...) {
  if (!is.null(attr(.job, "batch_id"))) {
    check_batch(.job, ...)
  } else if (inherits(.job, "tidyllm_research_job")) {
    perplexity_check_research(.job, ...)
  } else {
    stop("check_job() expects an object with a 'batch_id' attribute or a tidyllm_research_job.")
  }
}


#' Fetch Results from a Batch or Research Job
#'
#' `fetch_job()` dispatches to `fetch_batch()` for batch objects or
#' `perplexity_fetch_research()` for `tidyllm_research_job` objects.
#'
#' @param .job An object with a `batch_id` attribute (from `send_batch()`) or
#'   a `tidyllm_research_job` (from `deep_research(.background = TRUE)`).
#' @param .provider A provider function (required for batch jobs, ignored for research jobs).
#' @param ... Additional arguments passed to the underlying function.
#' @return Fetched results; type depends on `.job` class.
#' @export
fetch_job <- function(.job, .provider = NULL, ...) {
  if (!is.null(attr(.job, "batch_id"))) {
    fetch_batch(.job, .provider, ...)
  } else if (inherits(.job, "tidyllm_research_job")) {
    perplexity_fetch_research(.job, ...)
  } else {
    stop("fetch_job() expects an object with a 'batch_id' attribute or a tidyllm_research_job.")
  }
}


#' List Available Models for a Provider
#'
#' The `list_models()` function retrieves available models from the specified provider.
#'
#' @param .provider A function or function call specifying the provider and any additional parameters.
#'   You can also set a default provider via the `tidyllm_lmodels_default` option.
#' @param ... Additional arguments to be passed to the provider-specific list_models function.
#'
#' @return A tibble containing model information.
#' @export
list_models <- function(.provider = getOption("tidyllm_lmodels_default"), ...) {
  if (is.null(.provider)) {
    stop("You need to specify a .provider function in list_models().")
  }
  
  # Accept an unevaluated function
  if (rlang::is_function(.provider)) {
    .provider <- .provider()
  }
  
  # Capture the provider call expression
  if (!rlang::is_call(.provider)) {
    provider_expr <- rlang::quo_get_expr(rlang::enquo(.provider))
  } else {
    provider_expr <- .provider
  }
  
  extra_args <- list(...)
  return(dispatch_to_provider(provider_expr, "list_models", extra_args, validate = FALSE))
}

testthat::skip_if_not_installed("httptest2")
library(httptest2)

test_that("openai function constructs a correct request and dry runs it", {
  # Create a mock LLMMessage object
  llm <- llm_message("Write a poem about a (stochastic) parrot")
  
  # Call chatgpt with .dry_run = TRUE and perform the dry run
  request <- llm |>
    chat(openai,.dry_run = TRUE) 
  
  dry_run <- request |>
    httr2::req_dry_run(redact_headers = TRUE, quiet = TRUE)
  
  # Check that the method is POST
  expect_equal(dry_run$method, "POST")
  
  # Check that the URL path is correct
  expect_true(grepl("/v1/chat/completions", dry_run$path))
  
  # Inspect headers
  headers <- dry_run$headers
  expect_type(headers, "list")
  
  # Check that the required headers are present
  expect_true("authorization" %in% names(headers))
  expect_true("content-type" %in% names(headers))
  
  # Check that the content-type is JSON
  expect_equal(headers$`content-type`, "application/json")
  
  # Now check the body content to ensure the JSON is constructed as expected
  body_json <- request$body |> jsonlite::toJSON() |> as.character()
  
  expected_json <- "{\"data\":{\"model\":[\"gpt-5.1-chat-latest\"],\"messages\":[{\"role\":[\"system\"],\"content\":[\"You are a helpful assistant \"]},{\"role\":[\"user\"],\"content\":[\"Write a poem about a (stochastic) parrot \"]}]},\"type\":[\"json\"],\"content_type\":[\"application/json\"],\"params\":{\"auto_unbox\":[true],\"digits\":[22],\"null\":[\"null\"]}}"
  # Check if the JSON matches the expected JSON
  expect_equal(body_json, expected_json)
  
})


test_that("openai returns expected response", {
  with_mock_dir("openai",expr = {
    
    # Make sure the environment starts clean
    if (exists("openai", envir = .tidyllm_rate_limit_env)) {
      .tidyllm_rate_limit_env[["openai"]] <- NULL
    }
    
    # Store the current API key and set a dummy key if none exists
    if (Sys.getenv("OPENAI_API_KEY") == "") {
      Sys.setenv(OPENAI_API_KEY = "DUMMY_KEY_FOR_TESTING")
    }
    
    llm <- llm_message("Hello, world")
    
    result <- openai_chat(
      .llm = llm,
      .stream = FALSE
    )
    result_tbl <- as_tibble(result)

    if (Sys.getenv("OPENAI_API_KEY") == "DUMMY_KEY_FOR_TESTING") {
      Sys.setenv(OPENAI_API_KEY = "")
    }
    
    ## Assertions based on the message in the captured mock response
    expect_true(S7_inherits(result, LLMMessage))
    expect_equal(
      result_tbl$content[3],
      "Hello! How can I help you today?"
    )
    expect_equal(result_tbl$role[3], "assistant")
    
    # Now, check that the rate limit environment has been populated with correct values
    expect_true(exists("openai", envir = .tidyllm_rate_limit_env))
    
    ## Retrieve rate limit info for chatgpt
    rl_info <- rate_limit_info("openai")
    #
    ## Assertions for rate limit values based on the mocked response
    expect_equal(rl_info$requests_remaining, 4999)
    expect_equal(rl_info$tokens_remaining, 799986.0)
  },simplify = FALSE)
})

test_that("openai_embedding returns expected response", {
  with_mock_dir("openai_embedding",expr = {
    
    # Set a dummy key if none exists
    if (Sys.getenv("OPENAI_API_KEY") == "") {
      Sys.setenv(OPENAI_API_KEY = "DUMMY_KEY_FOR_TESTING")
    }
    
    result <- c("It is not that I am mad, it is only that my head is different from yours",
                "A man can do as he wills, but not will as he wills",
                "Whereof one cannot speak, thereof one must be silent",
                "The limits of my language mean the limits of my world") |>
      openai_embedding() 
    
    if (Sys.getenv("OPENAI_API_KEY") == "DUMMY_KEY_FOR_TESTING") {
      Sys.setenv(OPENAI_API_KEY = "")
    }
    
    # Test that the result is a tibble
    expect_s3_class(result, "tbl_df")
    
    # Test that the tibble has two columns: input and embeddings
    expect_named(result, c("input", "embeddings"))
    
    # Test that the input column contains the original input texts
    expect_equal(result$input, c("It is not that I am mad, it is only that my head is different from yours",
                                 "A man can do as he wills, but not will as he wills",
                                 "Whereof one cannot speak, thereof one must be silent",
                                 "The limits of my language mean the limits of my world"))
    
    
    purrr::walk(result$embeddings, function(embedding) {
      expect_equal(length(embedding), 1536)
    })
    
  },simplify = FALSE)
})

test_that("tidyllm_schema() handles single element correctly", {
  with_mock_dir("openai_schema_single", {
    
    # Set a dummy key if none exists
    if (Sys.getenv("OPENAI_API_KEY") == "") {
      Sys.setenv(OPENAI_API_KEY = "DUMMY_KEY_FOR_TESTING")
    }
    
    if (exists("openai", envir = .tidyllm_rate_limit_env)) {
      .tidyllm_rate_limit_env[["openai"]] <- NULL
    }
    
    schema_single <- tidyllm_schema(
      name = "AreaSchema",
      area = "numeric"
    )
    message_single <- llm_message("Imagine an area in JSON format that matches the schema.") |>
      chat(openai(), .json_schema = schema_single) %>%
      get_metadata()
    
    expect_true("completion_tokens" %in% colnames(message_single))
    expect_gt(message_single$total_tokens, 0)
  }, simplify = FALSE)
})

test_that("tidyllm_schema() handles multiple elements correctly", {
  with_mock_dir("openai_schema_multiple", {
    
    # Set a dummy key if none exists
    if (Sys.getenv("OPENAI_API_KEY") == "") {
      Sys.setenv(OPENAI_API_KEY = "DUMMY_KEY_FOR_TESTING")
    }
    
    if (exists("openai", envir = .tidyllm_rate_limit_env)) {
      .tidyllm_rate_limit_env[["openai"]] <- NULL
    }
    
    schema_multiple <- tidyllm_schema(
      name = "AreaSchema",
      area = "numeric",
      population = "numeric"
    )
    message_multiple <- llm_message("Imagine an area and population in JSON format that matches the schema.") |>
      chat(openai(), .json_schema = schema_multiple) %>%
      get_metadata()
    
    expect_true("completion_tokens" %in% colnames(message_multiple))
    expect_gt(message_multiple$total_tokens, 0)
    
    if (Sys.getenv("OPENAI_API_KEY") == "DUMMY_KEY_FOR_TESTING") {
      Sys.setenv(OPENAI_API_KEY = "")
    }
  }, simplify = FALSE)
})


test_that("send_batch creates correct JSONL for batch requests", {
  # Generate batch of messages
  messages <- glue::glue("Write a haiku about {x}",
                         x = c("Mannheim", "Stuttgart", "Heidelberg")) |>
    purrr::map(llm_message)
  
  jsonl_lines <- send_batch(messages, openai, .model = "gpt-4o-mini", .dry_run = TRUE)
  
  
  # Check that we have 3 lines (one for each request)
  expect_equal(length(jsonl_lines), 3)
  
  # Parse each line as JSON
  parsed_lines <- lapply(jsonl_lines, jsonlite::fromJSON)
  
  # Verify structure and content of each line
  purrr::iwalk(parsed_lines, function(x,y){
    expect_equal(x$custom_id, paste0("tidyllm_openai_req_", y))
    expect_equal(x$method, "POST")
    expect_equal(x$url, "/v1/chat/completions")
    expect_equal(x$body$model, "gpt-4o-mini")
    expect_equal(x$body$messages$role, c("system","user"))
    expect_equal(x$body$messages$content[1], "You are a helpful assistant ")
    expect_equal(x$body$messages$role[2], "user")
  })
  
  content_lines <- parsed_lines  |>
    purrr::map_chr(~.x$body$messages$content[2])
  
  expect_true(stringr::str_detect(content_lines[1],"Mannheim"))
  expect_true(stringr::str_detect(content_lines[2],"Stuttgart"))
  expect_true(stringr::str_detect(content_lines[3],"Heidelberg"))
})


test_that("list_openai_batches functions correctly", {
  with_mock_dir("list_openai_batches", {
    # Set a dummy key if none exists
    if (Sys.getenv("OPENAI_API_KEY") == "") {
      Sys.setenv(OPENAI_API_KEY = "DUMMY_KEY_FOR_TESTING")
    }
    
    # Test the function
    result <- list_openai_batches()
    
    # Clean up environment
    if (Sys.getenv("OPENAI_API_KEY") == "DUMMY_KEY_FOR_TESTING") {
      Sys.setenv(OPENAI_API_KEY = "")
    }
    
    # Check that the result is a tibble
    expect_s3_class(result, "tbl_df")
    
    # Check the structure of the returned tibble
    expected_columns <- c("batch_id", "status", "created_at", "expires_at", 
                          "request_total", "request_completed", "request_failed")
    expect_named(result, expected_columns)
    
    # Check column types
    expect_s3_class(result$created_at, "POSIXct")
    expect_s3_class(result$expires_at, "POSIXct")
    expect_type(result$batch_id, "character")
    expect_type(result$status, "character")
    expect_type(result$request_total, "integer")
    expect_type(result$request_completed, "integer")
    expect_type(result$request_failed, "integer")
  }, simplify = FALSE)
})


#test_that("fetch_batch retrieves and processes batch responses correctly", {
#  with_mock_dir("fetch_openai_batch", {
#    # Set a dummy key if none exists
#    if (Sys.getenv("OPENAI_API_KEY") == "") {
#      Sys.setenv(OPENAI_API_KEY = "DUMMY_KEY_FOR_TESTING")
#    }
#    
#    # Create test batch with necessary attributes
#    test_batch <- glue::glue("Write a haiku about {x}", 
#                             x = c("Mannheim", "Stuttgart", "Heidelberg", "Konstanz")) |>
#      purrr::map(llm_message) |>
#      purrr::set_names(paste0("tidyllm_openai_req_", 1:4))
#    
#    attr(test_batch, "batch_id") <- "batch_67daa551c13c8190ab7b855832f423ca"
#    attr(test_batch, "json") <- "FALSE"
#    
#    # Fetch the batch results
#    fetched_batch <- fetch_batch(test_batch, openai)
#    
#    # Clean up environment
#    if (Sys.getenv("OPENAI_API_KEY") == "DUMMY_KEY_FOR_TESTING") {
#      Sys.setenv(OPENAI_API_KEY = "")
#    }
#    
#    # Test that the result is a list of LLMMessage objects
#    expect_type(fetched_batch, "list")
#    expect_length(fetched_batch, 4)
#
#    # Test that each message has an assistant response
#    haikus <- purrr::map_chr(fetched_batch, get_reply)
#    expect_length(haikus, 4)
#    expect_true(all(nchar(haikus) > 0))
#    
#    # Check specific content for each city
#    expect_true(stringr::str_detect(haikus[1], "Mannheim"))
#    expect_true(stringr::str_detect(haikus[2], "Stuttgart"))
#    expect_true(stringr::str_detect(haikus[3], "Heidelberg"))
#    expect_true(stringr::str_detect(haikus[4], "Konstanz"))
#    
#    # Check that metadata was properly included in responses
#    expect_true(all(purrr::map_lgl(fetched_batch, function(msg) {
#      meta <- get_metadata(msg)
#      all(c("prompt_tokens", "completion_tokens", "total_tokens") %in% names(meta))
#    })))
#  }, simplify = FALSE)
#})

test_that("openai_list_models returns expected data", {
  with_mock_dir("openai_models", {
    # Set a dummy key if none exists
    if (Sys.getenv("OPENAI_API_KEY") == "") {
      Sys.setenv(OPENAI_API_KEY = "DUMMY_KEY_FOR_TESTING")
    }
    
    # Test the function
    result <- openai_list_models()
    
    # Clean up environment
    if (Sys.getenv("OPENAI_API_KEY") == "DUMMY_KEY_FOR_TESTING") {
      Sys.setenv(OPENAI_API_KEY = "")
    }
    
    # Check that the result is a tibble
    expect_s3_class(result, "tbl_df")
    
    # Check the structure of the returned tibble
    expect_named(result, c("id", "created", "owned_by"))
    
    # Check that columns have the right types
    expect_type(result$id, "character")
    expect_type(result$created, "character")
    expect_type(result$owned_by, "character")
    
    # Check that we have some models in the result
    expect_gt(nrow(result), 0)
    
    # Verify that common models are included
    common_models <- c("gpt-4o", "gpt-4o-mini")
    expect_true(any(stringr::str_detect(result$id, paste(common_models, collapse = "|"))))
  }, simplify = FALSE)
})

testthat::skip_if_not_installed("httptest2")
library(httptest2)


test_that("claude function constructs a correct request and dry runs it", {
  # Call claude with .dry_run = TRUE and perform the dry run
  request <- llm_message("Make a poem about anteaters!") |>
    chat(claude,.dry_run = TRUE,.model="claude-3-5-sonnet-20240620") 
  
  dry_run <- request |>
    httr2::req_dry_run(redact_headers = TRUE, quiet = TRUE)
  
  # Check that the method is POST
  expect_equal(dry_run$method, "POST")
  
  # Check that the path is correct
  expect_equal(dry_run$path, "/v1/messages")
  
  # Inspect headers
  headers <- dry_run$headers
  expect_type(headers, "list")
  
  # Check that the required headers are present
  expect_true("accept" %in% names(headers))
  expect_true("anthropic-version" %in% names(headers))
  expect_true("content-type" %in% names(headers))
  #expect_true("x-api-key" %in% names(headers))
  
  # Check that the content-type is JSON
  expect_equal(headers$`content-type`, "application/json; charset=utf-8")
  
  # Now check the body content to ensure the JSON is constructed as expected
  body_json <- request$body |> jsonlite::toJSON()  |> as.character()
  
  expected_json <- "{\"data\":{\"model\":[\"claude-3-5-sonnet-20240620\"],\"max_tokens\":[2048],\"messages\":[{\"role\":[\"user\"],\"content\":[{\"type\":[\"text\"],\"text\":[\"Make a poem about anteaters! \"]}]}],\"system\":[\"You are a helpful assistant\"],\"stream\":[false]},\"type\":[\"json\"],\"content_type\":[\"application/json\"],\"params\":{\"auto_unbox\":[true],\"digits\":[22],\"null\":[\"null\"]}}"

  # Check if the JSON  matches the expected JSON
  expect_equal(body_json, expected_json)
  
})


test_that("claude returns expected response", {
  with_mock_dir("claude",expr = {
    
    # Make sure the environment starts clean
    if (exists("claude", envir = .tidyllm_rate_limit_env)) {
      .tidyllm_rate_limit_env[["claude"]] <- NULL
    }
    
    messages <- llm_message("Hello, world")
    
    # Store the current API key and set a dummy key if none exists
    if (Sys.getenv("ANTHROPIC_API_KEY") == "") {
      Sys.setenv(ANTHROPIC_API_KEY = "DUMMY_KEY_FOR_TESTING")
    }
    
    result <- claude_chat(
      .llm = messages,
      .model = "claude-3-5-sonnet-20240620",
      .max_tokens = 1024,
      .temperature = 0,
      .stream = FALSE
    )
    
    # Store the current API key and set a dummy key if none exists
    if (Sys.getenv("ANTHROPIC_API_KEY") == "DUMMY_KEY_FOR_TESTING") {
      Sys.setenv(ANTHROPIC_API_KEY = "")
    }
    
    # Assertions based on the message in the captured mock response
    expect_true(S7_inherits(result, LLMMessage))
    expect_equal(
      result@message_history[[length(result@message_history)]]$content,
      "Hello! How can I assist you today? Is there anything specific you'd like to know or discuss?")
    expect_equal(result@message_history[[length(result@message_history)]]$role, "assistant")
    
    # Now, check that the rate limit environment has been populated with correct values
    expect_true(exists("claude", envir = .tidyllm_rate_limit_env))
    
    # Get the rate limit info for "claude"
    rl_info <- rate_limit_info("claude")
    
    # Assertions for rate limit values based on the mocked response
    expect_equal(rl_info$api, "claude")
    expect_equal(rl_info$requests_remaining, 999)
    expect_equal(rl_info$tokens_remaining, 96000)
  },simplify = FALSE)
})

test_that("claude_chat input validation works correctly", {
  llm <- llm_message("Test message")
  
  # Test temperature validation
  expect_error(
    claude_chat(.llm = llm, .temperature = 1.5),
    ".temperature must be numeric between 0 and 1 if provided"
  )
  expect_error(
    claude_chat(.llm = llm, .temperature = -0.1),
    ".temperature must be numeric between 0 and 1 if provided"
  )
  
  # Test top_p validation
  expect_error(
    claude_chat(.llm = llm, .top_p = 1.5),
    ".top_p must be numeric between 0 and 1 if provided"
  )
  
  # Test top_k validation
  expect_error(
    claude_chat(.llm = llm, .top_k = 0.5),
    ".top_k must be a positive integer if provided"
  )
  expect_error(
    claude_chat(.llm = llm, .top_k = -1),
    ".top_k must be a positive integer if provided"
  )
  
  # Test temperature and top_p mutual exclusivity
  expect_error(
    claude_chat(.llm = llm, .temperature = 0.7, .top_p = 0.9),
    "Only one of .temperature or .top_p should be specified"
  )
  
  # Test stop_sequences validation
  expect_error(
    claude_chat(.llm = llm, .stop_sequences = 123),
    ".stop_sequences must be a character vector"
  )
})

test_that("send_claude_batch creates correct request format", {
  # Generate batch of messages
  messages <- glue::glue("Write a haiku about {x}",
                         x = c("Mannheim", "Stuttgart", "Heidelberg")) |>
    purrr::map(llm_message)
  
  # Get request object with dry run
  request <- send_batch(messages, claude, 
                        .model = "claude-3-5-sonnet-20241022", 
                        .dry_run = TRUE)
  
  # Parse the request to check its structure
  dry_run <- request |>
    httr2::req_dry_run(redact_headers = TRUE, quiet = TRUE)
  
  # Check basic request structure
  expect_equal(dry_run$method, "POST")
  expect_equal(dry_run$path, "/v1/messages/batches")
  

  request_body <- request$body
  # Verify requests structure
  expect_true("data" %in% names(request_body))
  expect_equal(length(request_body$data$requests), 3)
  
  # Check each request has the right structure
  purrr::walk(request_body$data$requests, function(req) {
    expect_true("custom_id" %in% names(req))
    expect_true("params" %in% names(req))
    expect_true("model" %in% names(req$params))
    expect_true("messages" %in% names(req$params))
    expect_equal(req$params$model, "claude-3-5-sonnet-20241022")
  })
  
  # Verify that the requests contain the correct prompts
  messages_content <- purrr::map(request_body$data$requests, function(req) {
    # Extract text content from the first message
    req$params$messages[[1]]$content[[1]]$text
  })
  
  expect_true(any(stringr::str_detect(unlist(messages_content), "Mannheim")))
  expect_true(any(stringr::str_detect(unlist(messages_content), "Stuttgart")))
  expect_true(any(stringr::str_detect(unlist(messages_content), "Heidelberg")))
})


test_that("fetch_claude_batch constructs correct request", {
  # Create test batch with necessary attributes
  test_batch <- glue::glue("Write a haiku about {x}", 
                           x = c("Mannheim", "Stuttgart")) |>
    purrr::map(llm_message) |>
    purrr::set_names(paste0("tidyllm_claude_req_", 1:2))
  
  attr(test_batch, "batch_id") <- "test-batch-id-67890"
  
  # Get request object with dry run
  request <- fetch_batch(test_batch, claude, .dry_run = TRUE)
  
  # Parse the request to check its structure
  dry_run <- request |>
    httr2::req_dry_run(redact_headers = TRUE, quiet = TRUE)
  
  # Check request structure
  expect_equal(dry_run$method, "GET")
  expect_equal(dry_run$path, "/v1/messages/batches/test-batch-id-67890")
  
})

testthat::skip_if_not_installed("httptest2")
library(httptest2)

test_that("perplexity function constructs a correct request and dry runs it", {
  # Create a mock LLMMessage object
  llm <- llm_message("Write a haiku about search indices")
  
  # Call perplexity_chat with .dry_run = TRUE
  request <- llm |> chat(perplexity, .dry_run = TRUE) 
  
  dry_run <- request |> httr2::req_dry_run(redact_headers = TRUE, quiet = TRUE)
  
  # Check that the method is POST
  expect_equal(dry_run$method, "POST")
  
  # Check that the URL path is correct
  expect_true(grepl("/chat/completions", dry_run$path))
  
  # Inspect headers
  headers <- dry_run$headers
  expect_type(headers, "list")
  
  # Check that the required headers are present
  expect_true("authorization" %in% names(headers))
  expect_true("content-type" %in% names(headers))
  
  # Check that the content-type is JSON
  expect_equal(headers$`content-type`, "application/json")
  
  # Now check the body content to ensure the JSON is constructed as expected
  body_json <- request$body |> jsonlite::toJSON() |> as.character()
  
  expected_json <- "{\"data\":{\"model\":[\"sonar\"],\"messages\":[{\"role\":[\"user\"],\"content\":[\"Write a haiku about search indices \"]}],\"max_tokens\":[1024],\"return_images\":[false],\"search_mode\":[\"web\"],\"return_related_questions\":[false],\"stream\":[false]},\"type\":[\"json\"],\"content_type\":[\"application/json\"],\"params\":{\"auto_unbox\":[true],\"digits\":[22],\"null\":[\"null\"]}}"
  # Check if the JSON matches the expected JSON
  expect_equal(body_json, expected_json)
})

test_that("perplexity_chat dry run includes search_domain_filter in body", {
  request <- llm_message("Any news?") |>
    chat(perplexity(.search_domain_filter = c("arxiv.org", "-reddit.com")), .dry_run = TRUE)
  body_json <- request$body |> jsonlite::toJSON() |> as.character()
  expect_true(grepl("arxiv.org", body_json))
  expect_true(grepl("-reddit.com", body_json))
})

test_that("perplexity_chat dry run includes search_recency_filter in body", {
  request <- llm_message("What happened today?") |>
    chat(perplexity(.search_recency_filter = "day"), .dry_run = TRUE)
  body_json <- request$body |> jsonlite::toJSON() |> as.character()
  expect_true(grepl("\"day\"", body_json))
})

test_that("check_job errors for non-job object", {
  expect_error(check_job("not a job"), "check_job\\(\\) expects")
  expect_error(check_job(list(x = 1)), "check_job\\(\\) expects")
})

test_that("fetch_job errors for non-job object", {
  expect_error(fetch_job("not a job"), "fetch_job\\(\\) expects")
  expect_error(fetch_job(42), "fetch_job\\(\\) expects")
})

test_that("perplexity_deep_research validates reasoning_effort", {
  old_key <- Sys.getenv("PERPLEXITY_API_KEY")
  Sys.setenv(PERPLEXITY_API_KEY = "DUMMY_KEY_FOR_TESTING")
  on.exit(Sys.setenv(PERPLEXITY_API_KEY = old_key))
  expect_error(
    perplexity_deep_research(llm_message("test"), .reasoning_effort = "extreme"),
    ".reasoning_effort must be"
  )
})

test_that("perplexity_deep_research validates search_domain_filter max length", {
  old_key <- Sys.getenv("PERPLEXITY_API_KEY")
  Sys.setenv(PERPLEXITY_API_KEY = "DUMMY_KEY_FOR_TESTING")
  on.exit(Sys.setenv(PERPLEXITY_API_KEY = old_key))
  expect_error(
    perplexity_deep_research(
      llm_message("test"),
      .search_domain_filter = paste0("domain", 1:11, ".com")
    ),
    "must be <= 10 domains"
  )
})

test_that("perplexity returns expected response", {
  with_mock_dir("perplexity", expr = {
    
    # Store the current API key and set a dummy key if none exists
    if (Sys.getenv("PERPLEXITY_API_KEY") == "") {
      Sys.setenv(PERPLEXITY_API_KEY = "DUMMY_KEY_FOR_TESTING")
    }
    
    llm <- llm_message("Write a haiku about Perplexity")
    
    result <- perplexity_chat(llm)
    
    reply <- result |>
      get_reply() 
    
    if (Sys.getenv("PERPLEXITY_API_KEY") == "DUMMY_KEY_FOR_TESTING") {
      Sys.setenv(PERPLEXITY_API_KEY = "")
    }
    
    ## Assertions based on the message in the captured mock response
    expect_true(S7::S7_inherits(result, LLMMessage))
    expect_equal(
      reply,
      "Silent web unfolds,  \nPerplexity finds the truth—  \nAnswers bloom like spring."
    )

  }, simplify = FALSE)
})

testthat::skip_if_not_installed("httptest2")
library(httptest2)

test_that("mistral function constructs a correct request and dry runs it", {
  # Call mistral with .dry_run = TRUE and perform the dry run
  request <- llm_message("Write a poem about the Gallic Rooster") |> 
    chat(mistral,.dry_run = TRUE)
  
  dry_run <- request |>
    httr2::req_dry_run(redact_headers = TRUE, quiet = TRUE)
  
  # Check that the method is POST
  expect_equal(dry_run$method, "POST")
  
  # Check that the path is correct
  expect_equal(dry_run$path, "/v1/chat/completions")
  
  # Inspect headers
  headers <- dry_run$headers
  expect_type(headers, "list")
  
  # Check that the required headers are present
  expect_true("authorization" %in% names(headers))
  expect_true("content-type" %in% names(headers))
  
  # Check that the content-type is JSON
  expect_equal(headers$`content-type`, "application/json")
  
  # Now check the body content to ensure the JSON is constructed as expected
  body_json <- request$body |> jsonlite::toJSON() |> as.character()
  
  expected_json <- "{\"data\":{\"model\":[\"mistral-large-latest\"],\"messages\":[{\"role\":[\"user\"],\"content\":[\"Write a poem about the Gallic Rooster \"]}],\"safe_prompt\":[false],\"temperature\":[0.7],\"top_p\":[1]},\"type\":[\"json\"],\"content_type\":[\"application/json\"],\"params\":{\"auto_unbox\":[true],\"digits\":[22],\"null\":[\"null\"]}}"
  # Check if the JSON matches the expected JSON
  expect_equal(body_json, expected_json)
})

test_that("mistral returns expected response",{ 
  with_mock_dir("mistral",expr = {
    
    # Store the current API key and set a dummy key if none exists
    if (Sys.getenv("MISTRAL_API_KEY") == "") {
      Sys.setenv(MISTRAL_API_KEY = "DUMMY_KEY_FOR_TESTING")
    }
    
    # Make sure the environment starts clean
    if (exists("mistral", envir = .tidyllm_rate_limit_env)) {
      .tidyllm_rate_limit_env[["mistral"]] <- NULL
    }
    
    llm <- llm_message("Hello, world")
    
    result <- mistral_chat(
      .llm = llm,
      .max_tokens = 1024,
      .temperature = 0,
    )
    

    if (Sys.getenv("MISTRAL_API_KEY") == "DUMMY_KEY_FOR_TESTING") {
      Sys.setenv(MISTRAL_API_KEY = "")
    }
    
    # Assertions based on the message in the captured mock response
    expect_true(S7_inherits(result, LLMMessage))
    result_tbl <- as_tibble(result) 
    

    expect_equal(result_tbl$role[3], "assistant")
    
    # Now, check that the rate limit environment has been populated with correct values
    expect_true(exists("mistral", envir = .tidyllm_rate_limit_env))
    

  },simplify = FALSE)
  
})
 

test_that("mistral_embedding returns expected response", {
  with_mock_dir("mistral_embedding",expr = {
    
    # Store the current API key and set a dummy key if none exists
    if (Sys.getenv("MISTRAL_API_KEY") == "") {
      Sys.setenv(MISTRAL_API_KEY = "DUMMY_KEY_FOR_TESTING")
    }
    
    
    result <- c("It is not that I am mad, it is only that my head is different from yours",
                "A man can do as he wills, but not will as he wills",
                "Whereof one cannot speak, thereof one must be silent",
                "The limits of my language mean the limits of my world") |>
      mistral_embedding() 
    
    if (Sys.getenv("MISTRAL_API_KEY") == "DUMMY_KEY_FOR_TESTING") {
      Sys.setenv(MISTRAL_API_KEY = "")
    }
    
    # Test that the result is a tibble
    expect_s3_class(result, "tbl_df")
    
    # Test that the tibble has two columns: input and embeddings
    expect_named(result, c("input", "embeddings"))
    
    # Test that the input column contains the original input texts
    expect_equal(result$input, c("It is not that I am mad, it is only that my head is different from yours",
                                 "A man can do as he wills, but not will as he wills",
                                 "Whereof one cannot speak, thereof one must be silent",
                                 "The limits of my language mean the limits of my world"))
    
    
    purrr::walk(result$embeddings, function(embedding) {
      expect_equal(length(embedding), 1024)
    })
    
  },simplify = FALSE)
})

test_that("send_batch creates correct JSONL for batch requests", {
  # Generate batch of messages
  messages <- glue::glue("Write a haiku about {x}",
                         x = c("Mannheim", "Stuttgart", "Heidelberg")) |>
    purrr::map(llm_message)
  
  jsonl_lines <- send_batch(messages, mistral, .model = "mistral_large", .dry_run = TRUE)
  
  
  # Check that we have 3 lines (one for each request)
  expect_equal(length(jsonl_lines), 3)
  
  # Parse each line as JSON
  parsed_lines <- lapply(jsonl_lines, jsonlite::fromJSON)
  
  # Verify structure and content of each line
  purrr::iwalk(parsed_lines, function(x,y){
    expect_equal(x$custom_id, paste0("tidyllm_mistral_req_", y))
    expect_equal(x$method, "POST")
    expect_equal(x$url, "/v1/chat/completions")
    expect_equal(x$body$model, "mistral_large")
    expect_equal(x$body$messages$role, c("user"))
    expect_equal(x$body$messages$role[1], "user")
  })
  
  content_lines <- parsed_lines  |>
    purrr::map_chr(~.x$body$messages$content[1])
  
  expect_true(stringr::str_detect(content_lines[1],"Mannheim"))
  expect_true(stringr::str_detect(content_lines[2],"Stuttgart"))
  expect_true(stringr::str_detect(content_lines[3],"Heidelberg"))
})

testthat::skip_if_not_installed("httptest2")
library(httptest2)

test_that("voyage_embedding returns expected response for text input", {
  with_mock_dir("voyage_embedding_text", {
    
    # Store the current API key and set a dummy key if none exists
    original_key <- Sys.getenv("VOYAGE_API_KEY")
    if (original_key == "") {
      Sys.setenv(VOYAGE_API_KEY = "DUMMY_KEY_FOR_TESTING")
    }
    
    input_texts <- c(
      "How does photosynthesis work?",
      "The quick brown fox jumps over the lazy dog"
    )
    
    result <- input_texts |>
      embed(voyage)
    
    # Reset the API key if we used a dummy
    if (original_key == "") {
      Sys.setenv(VOYAGE_API_KEY = "")
    }
    
    # Check that the result is a tibble with the correct columns and values
    expect_s3_class(result, "tbl_df")
    expect_named(result, c("input", "embeddings"))
    expect_equal(result$input, input_texts)
    
    # Assuming embeddings are 1024-dimensional vectors
    purrr::walk(result$embeddings, function(embedding) {
      expect_equal(length(embedding), 1024)
    })
  })
})

structure(list(method = "POST", url = "https://api.deepseek.com/chat/completions", 
    status_code = 200L, headers = structure(list(date = "Thu, 21 Aug 2025 12:41:01 GMT", 
        `content-type` = "application/json", `content-encoding` = "gzip", 
        vary = "origin, access-control-request-method, access-control-request-headers", 
        `access-control-allow-credentials` = "true", `x-ds-trace-id` = "754bfdec3ed289904df8314d9c221d25", 
        `strict-transport-security` = "max-age=31536000; includeSubDomains; preload", 
        `x-content-type-options` = "nosniff", `cf-cache-status` = "DYNAMIC", 
        `set-cookie` = "REDACTED", server = "cloudflare", `cf-ray` = "972a33c67ac19736-FRA"), redact = character(0), class = "httr2_headers"), 
    body = charToRaw("{\"id\":\"b55e6172-d379-4806-b4e0-d21dd4b243b7\",\"object\":\"chat.completion\",\"created\":1755780061,\"model\":\"deepseek-chat\",\"choices\":[{\"index\":0,\"message\":{\"role\":\"assistant\",\"content\":\"Hello! How can I assist you today?\"},\"logprobs\":null,\"finish_reason\":\"stop\"}],\"usage\":{\"prompt_tokens\":14,\"completion_tokens\":9,\"total_tokens\":23,\"prompt_tokens_details\":{\"cached_tokens\":0},\"prompt_cache_hit_tokens\":0,\"prompt_cache_miss_tokens\":14},\"system_fingerprint\":\"fp_feb633d1f5_prod0820_fp8_kvcache\"}"), 
    cache = new.env(parent = emptyenv())), class = "httr2_response")

structure(list(method = "POST", url = "https://api.groq.com/openai/v1/chat/completions", 
    status_code = 200L, headers = structure(list(date = "Sun, 16 Nov 2025 13:02:44 GMT", 
        `content-type` = "application/json", `cache-control` = "private, max-age=0, no-store, no-cache, must-revalidate", 
        server = "cloudflare", vary = "Origin", `x-groq-region` = "fra", 
        `x-ratelimit-limit-requests` = "500000", `x-ratelimit-limit-tokens` = "250000", 
        `x-ratelimit-remaining-requests` = "499999", `x-ratelimit-remaining-tokens` = "249969", 
        `x-ratelimit-reset-requests` = "172.799999ms", `x-ratelimit-reset-tokens` = "7.44ms", 
        `x-request-id` = "req_01ka6c5fgseaksg4wxp79d43hm", via = "1.1 google", 
        `cf-cache-status` = "DYNAMIC", `content-encoding` = "gzip", 
        `set-cookie` = "REDACTED", `cf-ray` = "99f72f3679d791fc-FRA", 
        `alt-svc` = "h3=\":443\"; ma=86400"), redact = character(0), class = "httr2_headers"), 
    body = charToRaw("{\"id\":\"chatcmpl-59364eff-df3b-4826-b4d6-1562b9cdf2be\",\"object\":\"chat.completion\",\"created\":1763298164,\"model\":\"moonshotai/kimi-k2-instruct-0905\",\"choices\":[{\"index\":0,\"message\":{\"role\":\"assistant\",\"content\":\"Hello! How can I help you today?\"},\"logprobs\":null,\"finish_reason\":\"stop\"}],\"usage\":{\"queue_time\":0.190203705,\"prompt_tokens\":30,\"prompt_time\":0.010721307,\"completion_tokens\":10,\"completion_time\":0.012596456,\"total_tokens\":40,\"total_time\":0.023317763},\"usage_breakdown\":null,\"system_fingerprint\":\"fp_05df423bab\",\"x_groq\":{\"id\":\"req_01ka6c5fgseaksg4wxp79d43hm\"},\"service_tier\":\"on_demand\"}\n"), 
    cache = new.env(parent = emptyenv())), class = "httr2_response")

#' The Google Gemini API provider stub
#'
#' At the moment this is just a stub but is needed for methods dispatch
#'
#' @noRd
api_gemini <- new_class("Google Gemini", APIProvider)

#' Convert LLMMessage to Gemini API-Compatible Format
#'
#' Converts the `message_history` of an `LLMMessage` object into the
#' one needed for the Google Gemini API.
#'
#' @noRd
method(to_api_format, list(LLMMessage, api_gemini)) <- function(.llm, 
                                                                .api) {
  
  # Filter to only include user and assistant messages
  gemini_history <- filter_roles(.llm@message_history, c("user", "assistant"))
  
  # Map each message to the expected Gemini format
  lapply(gemini_history, function(m) {
    
    formatted_message <- format_message(m)
    if (!is.null(formatted_message$image)) {
      list(
        role = ifelse(m$role == "user", "user", "model"), 
        parts = list(
          list(text = formatted_message$content),
          list(
            inline_data = list(
              mime_type = formatted_message$image$media_type,
              data      = formatted_message$image$data
            )
          )
        )
      )
    } else {
      list(
        role = ifelse(m$role == "user", "user", "model"), 
        parts = list(text = formatted_message$content)
      )
    }
  })
}


#' A chat response parsing method for Gemini to extract the assistant response 
#'
#' @noRd
method(parse_chat_response, list(api_gemini,class_list)) <- function(.api,.content) {
  api_label <- .api@long_name 
  if("error" %in% names(.content)){
    sprintf("%s returned an Error:\nCode: %s\nMessage: %s",
            api_label,
            .content$error$code,
            .content$error$message) |>
      stop()
  }
  
  if (!"candidates" %in% names(.content) || length(.content$candidates) == 0) {
    paste0("Received empty response from ", api_label) |>
      stop()
  }
  
  .content$candidates[[1]]$content$parts[[1]]$text
}


#' A method to handle streaming requests for Gemini
#'
#' @noRd
method(handle_stream,list(api_gemini,new_S3_class("httr2_response"))) <- function(.api,.stream_response) {
  stream_data <- ""
  current_buffer <-  ""
  repeat {
    stream_chunk   <- httr2::resp_stream_lines(.stream_response)
    stream_data    <- paste0(stream_data,stream_chunk)
    current_buffer <- paste0(current_buffer,stream_chunk) 
    
    parts_section <- stringr::str_extract(current_buffer, 
                                          '"parts":\\s*\\[\\s*\\{[^\\}]*\\}\\s*\\]')
    
    if(!is.na(parts_section)){
      current_part <- jsonlite::fromJSON(paste0("{",parts_section,"}"))
      stream_response <- current_part$parts$text
      cat(stream_response)
      utils::flush.console()
      current_buffer <-  ""
    }
    
    # Skip empty chunks
    if (httr2::resp_stream_is_complete(.stream_response)) {
      close(.stream_response)
      message("\n---------\nStream finished\n---------\n")
      break
    }
  }
  stream_data_parsed <- jsonlite::fromJSON(stream_data)
  stream_text <- stream_data_parsed$candidates |> 
    purrr::map_chr(~.x$content$parts[[1]]$text) |>
    stringr::str_c(collapse = "")
  
  list(
    reply = stream_text,
    raw_data = list(parsed=stream_data_parsed)
  )
}
    


#' A function to get metadata from Openai responses
#'
#' @noRd
method(extract_metadata, list(api_gemini,class_list))<- function(.api,.response) {
  list(
    model             = .response$modelVersion,
    timestamp         = lubridate::as_datetime(lubridate::now()),
    prompt_tokens     = .response$usageMetadata$promptTokenCount,
    completion_tokens = .response$usageMetadata$candidatesTokenCount,
    total_tokens      = .response$usageMetadata$totalTokenCount,
    specific_metadata = list(
      finishReason      = .response$candidates[[1]]$finishReason,
      avgLogprobs       = .response$candidates[[1]]$avgLogprobs,
      groundingMetadata = .response$candidates[[1]]$groundingMetadata,
      thinking_tokens   = .response$usageMetadata$thoughtsTokenCount
    )
  )
}  

#' A function to get metadata from Openai streaming responses
#'
#' @noRd
method(extract_metadata_stream, list(api_gemini,class_list))<- function(.api,.stream_raw_data) {
  parsed_stream_data <- .stream_raw_data$parsed
  final_stream_chunk <- parsed_stream_data[nrow(parsed_stream_data),]

  list(
    model             = final_stream_chunk$modelVersion,
    timestamp         = lubridate::as_datetime(lubridate::now()),
    prompt_tokens     = final_stream_chunk$usageMetadata$promptTokenCount,
    completion_tokens = final_stream_chunk$usageMetadata$candidatesTokenCount,
    total_tokens      = final_stream_chunk$usageMetadata$totalTokenCount,
    stream            = TRUE,
    specific_metadata = list(
      warning    = "Gemini outputs different metadata for streaming and non-streaming responses",
      token_details = final_stream_chunk$usageMetadata
    ) 
  )
}  


#' Method to convert a tidyllm TOOL definition to the expected input for Gemini
#'
#' @noRd
method(tools_to_api, list(api_gemini, class_list)) <- function(.api, .tools) {
  list(
    function_declarations = purrr::map(.tools, function(tool) {
      if (length(tool@builtin) > 0) {
        tool@builtin[[1]]
      } else {
        tool_def <- list(
          name = tool@name,
          description = tool@description
        )
        if (length(tool@input_schema) > 0) {
          tool_def$parameters <- list(
            type = "object",
            properties = purrr::map(tool@input_schema, field_to_param_schema),
            required = names(tool@input_schema)
          )
        }
        tool_def
      }
    })
  )
}

method(has_tool_calls, list(api_gemini, class_any)) <- function(.api, .response) {
  parts <- .response$raw$content$candidates[[1]]$content$parts
  if (is.null(parts)) return(FALSE)
  any(purrr::map_lgl(parts, ~!is.null(.x$functionCall)))
}

method(extract_tool_calls, list(api_gemini, class_any)) <- function(.api, .response) {
  parts <- .response$raw$content$candidates[[1]]$content$parts
  purrr::keep(parts, ~!is.null(.x$functionCall)) |>
    purrr::map(~.x$functionCall)
}

method(append_tool_messages, list(api_gemini, class_any, class_any, class_any)) <-
  function(.api, .request_body, .response, .tool_results) {
    tool_calls <- extract_tool_calls(.api, .response)
    assistant_message <- list(
      role = "model",
      parts = purrr::map(tool_calls, ~list(functionCall = .x))
    )
    .request_body$contents <- c(
      .request_body$contents,
      list(assistant_message),
      list(.tool_results)
    )
    .request_body
  }

#' A method to run tool calls on Gemini and create the expected response
#'
#' @noRd
method(run_tool_calls, list(api_gemini, class_list, class_list)) <- function(.api, .tool_calls, .tools) {
  # Iterate over each tool call returned by Gemini and build a list of parts.
  tool_parts <- purrr::map(.tool_calls, function(tool_call) {
    # Gemini returns the tool call information in a structure like:
    # list(name = "<tool_name>", args = list(...))
    tool_name <- tool_call$name
    tool_args <- tool_call$args
    
    # Find the corresponding tool in the provided tools list.
    matching_tool <- purrr::keep(.tools, ~ .x@name == tool_name)
    if (length(matching_tool) == 0) {
      warning(sprintf("No matching tool found for: %s", tool_name))
      return(NULL)
    }
    
    tool_function <- matching_tool[[1]]@func
    
    # Execute the tool function with the provided arguments.
    tool_result <- utils::capture.output(
      do.call(tool_function, as.list(tool_args)),
      file = NULL
    ) |> 
      stringr::str_c(collapse = "\n")
    
    # Format the tool response as a part for a single user message.
    list(
      functionResponse = list(
        name = tool_name,
        response = list(
          name = tool_name,
          content = tool_result
        )
      )
    )
  })
  
  tool_parts <- purrr::compact(tool_parts)
  
  # Combine all parts into one user message.
  list(
    role = "user",
    parts = tool_parts
  )
}





#' Inject files into Gemini message contents
#'
#' @param .gemini_contents The existing gemini contents list
#' @param .file_ids A vector or list of file IDs to inject
#' @return Updated gemini_contents with files injected
#' @noRd
gemini_inject_files <- function(.gemini_contents, 
                                .file_ids) {
  # If file_ids is NULL or empty, just return gemini_contents
  if (is.null(.file_ids) || length(.file_ids) == 0) {
    return(.gemini_contents)
  }
  
  # For each file ID, get the metadata and create file_data
  file_data_list <- lapply(.file_ids, function(file_id) {
    file_info <- gemini_file_metadata(file_id)
    list(
      fileData = list(
        fileUri = file_info$uri,
        mimeType = file_info$mime_type
      )
    )
  })
  
  # Now, inject the file_data into the parts of the last user message
  # First, get the last message
  last_msg <- .gemini_contents[[length(.gemini_contents)]]
  
  # Ensure the last message is a user message
  if (last_msg$role != "user") {
    stop("The last message must be a user message to inject files")
  }
  
  # Ensure that last_msg$parts is a list of parts
  if (!is.list(last_msg$parts)) {
    stop("The 'parts' of the last message is not a list")
  }
  
  # Check if last_msg$parts[[1]] is a list; if not, wrap it
  if (!is.list(last_msg$parts[[1]])) {
    last_msg$parts <- list(last_msg$parts)
  }
  
  # Append the file_data_list to the parts
  last_msg$parts <- c(
    last_msg$parts,
    file_data_list
  )
  
  .gemini_contents[[length(.gemini_contents)]] <- last_msg
}





#' Send LLMMessage to Gemini API
#'
#' @param .llm An existing LLMMessage object or an initial text prompt.
#' @param .model The model identifier (default: "gemini-1.5-flash").
#' @param .fileid Optional vector of file IDs uploaded via `gemini_upload_file()` (default: NULL).
#' @param .temperature Controls randomness in generation (default: NULL, range: 0.0-2.0).
#' @param .max_output_tokens Maximum tokens in the response (default: NULL).
#' @param .top_p Controls nucleus sampling (default: NULL, range: 0.0-1.0).
#' @param .top_k Controls diversity in token selection (default: NULL, range: 0 or more).
#' @param .presence_penalty Penalizes new tokens (default: NULL, range: -2.0 to 2.0).
#' @param .frequency_penalty Penalizes frequent tokens (default: NULL, range: -2.0 to 2.0).
#' @param .stop_sequences Optional character sequences to stop generation (default: NULL, up to 5).
#' @param .safety_settings A list of safety settings (default: NULL).
#' @param .json_schema A schema to enforce an output structure
#' @param .grounding_threshold A grounding threshold between 0 and 1. With lower 
#' grounding thresholds  Gemini will use Google to search for relevant information 
#' before answering.  (default: NULL).
#' @param .tools Either a single TOOL object or a list of TOOL objects representing the available functions for tool calls.
#' @param .timeout When should our connection time out (default: 120 seconds).
#' @param .dry_run If TRUE, perform a dry run and return the request object.
#' @param .max_tries Maximum retries to perform request (default: 3).
#' @param .verbose Should additional information be shown after the API call.
#' @param .stream Should the response be streamed (default: FALSE).
#' @param .max_tool_rounds Integer specifying the maximum number of tool use iterations (default: 10).
#'   Set to 1 for single-round tool use, or higher for multi-turn agentic loops.
#' @param .thinking_budget Token budget for internal reasoning (default: NULL). Works with `gemini-2.5-flash` and `gemini-2.5-pro`.
#'
#' @return A new `LLMMessage` object containing the original messages plus the assistant's response.
#'
#' @export
gemini_chat <- function(.llm,
                   .model = "gemini-2.5-flash",
                   .fileid = NULL,
                   .temperature = NULL,
                   .max_output_tokens = NULL,
                   .top_p = NULL,
                   .top_k = NULL,
                   .grounding_threshold = NULL,
                   .presence_penalty = NULL,
                   .frequency_penalty = NULL,
                   .stop_sequences = NULL,
                   .safety_settings = NULL,
                   .json_schema = NULL,
                   .tools = NULL,
                   .thinking_budget = NULL,
                   .timeout = 120,
                   .dry_run = FALSE,
                   .max_tries = 3,
                   .verbose = FALSE,
                   .stream = FALSE,
                   .max_tool_rounds = 10) {

  # Validate inputs
  c(
    "Input .llm must be an LLMMessage object" = S7_inherits(.llm, LLMMessage),
    "Input .model must be a string" = is.character(.model) && length(.model) == 1,
    "Input .fileid must be NULL or a charcater vector of file IDs" = is.null(.fileid) | is.character(.fileid) ,
    "Input .temperature must be NULL or in [0.0, 2.0]" = is.null(.temperature) | (.temperature >= 0.0 & .temperature <= 2.0),
    "Input .max_output_tokens must be NULL or an integer-valued numeric greater than 1" = is.null(.max_output_tokens) | (.max_output_tokens >= 1 & is_integer_valued(.max_output_tokens)),
    "Input .top_p must be NULL or in [0.0, 1.0]" = is.null(.top_p) | (.top_p >= 0.0 & .top_p <= 1.0),
    "Input .top_k must be NULL or non-negative" = is.null(.top_k) | (.top_k >= 0),
    "Input .grounding_threshold must be NULL or in [0.0, 1.0]" = is.null(.grounding_threshold) | (.grounding_threshold >= 0.0 & .grounding_threshold <= 1.0),
    "Input .presence_penalty must be NULL or in [-2.0, 2.0]" = is.null(.presence_penalty) | (.presence_penalty >= -2.0 & .presence_penalty <= 2.0),
    "Input .frequency_penalty must be NULL or in [-2.0, 2.0]" = is.null(.frequency_penalty) | (.frequency_penalty >= -2.0 & .frequency_penalty <= 2.0),
    "Input .stop_sequences must be NULL or a list of up to 5 strings" = is.null(.stop_sequences) | (is.list(.stop_sequences) & length(.stop_sequences) <= 5),
    "Input .safety_settings must be NULL or a list" = is.null(.safety_settings) | is.list(.safety_settings),
    "Input .json_schema must be NULL or a list or an ellmer type object" = is.null(.json_schema) | is.list(.json_schema) | is_ellmer_type(.json_schema),
    "Input .timeout must be an integer-valued numeric and positive" = is_integer_valued(.timeout) & .timeout > 0,
    "Input .max_tries must be integer-valued numeric and positive" = is_integer_valued(.max_tries) & .max_tries > 0,
    "Input .dry_run must be logical" = is.logical(.dry_run),
    "Input .verbose must be logical" = is.logical(.verbose),
    "Input .stream must be logical" = is.logical(.verbose),
    "Input .tools must be NULL, a TOOL object, or a list of TOOL objects" = is.null(.tools) || S7_inherits(.tools, TOOL) || (is.list(.tools) && all(purrr::map_lgl(.tools, ~ S7_inherits(.x, TOOL)))),
    "Streaming is not supported for requests with tool calls" = is.null(.tools) || !isTRUE(.stream),
    ".max_tool_rounds must be a positive integer" = is_integer_valued(.max_tool_rounds) && .max_tool_rounds >= 1,
    ".thinking_budget must be NULL or a non-negative integer" = is.null(.thinking_budget) || (is_integer_valued(.thinking_budget) && .thinking_budget >= 0)
  ) |>
    validate_inputs()
  
  api_obj <- api_gemini(short_name = "gemini",
                        long_name  = "Google Gemini",
                        api_key_env_var = "GOOGLE_API_KEY")
  
  api_key <- get_api_key(api_obj,.dry_run)
  
  # Prepare message contents (inject files if available)
  gemini_contents <- to_api_format(.llm,api_obj) |>
    gemini_inject_files(.fileid)
  
  # Handle JSON schema
  response_format <- NULL
  json=FALSE
  if (requireNamespace("ellmer", quietly = TRUE)) {
    #Handle ellmer json schemata Objects
    if(S7_inherits(.json_schema,ellmer::TypeObject)){
      .json_schema = to_schema(.json_schema)
      # Remove additionalProperties if it exists
      if (!is.null(.json_schema) && "additionalProperties" %in% names(.json_schema)) {
        .json_schema <- .json_schema[setdiff(names(.json_schema), "additionalProperties")]
      }
    } 
  }
  if (!is.null(.json_schema)) {
    json=TRUE
    response_format <- list(
      response_mime_type = "application/json",
      response_schema = .json_schema
    )
  } 
  
  
  #Put a single tool into a list if only one is provided.
  raw_tools_def <- if (!is.null(.tools)) {
    if (S7_inherits(.tools, TOOL)) list(.tools) else .tools
  } else {
    NULL
  }
  tools_def <- if (!is.null(raw_tools_def)) {
    tools_to_api(api_obj, raw_tools_def)
  } else {
    NULL
  }
  
  # Add grounding tool configuration if grounding_threshold is set
  if (!is.null(.grounding_threshold)) {
    grounding_tool <- list(
      google_search_retrieval = list(
        dynamic_retrieval_config = list(
          mode = "MODE_DYNAMIC",
          dynamic_threshold = .grounding_threshold
        )
      )
    )
    if(!is.null(tools_def)){
      tools_def <- tools_def |>
        append(list(grounding_tool))
    } else {
      tools_def <- list(grounding_tool)
    }
  }
  
  #Handle system prompt
  system_prompt <- list(parts = list(
    text = filter_roles(.llm@message_history, c("system"))[[1]]$content
    ))
  
  # Build generationConfig
  generation_config <- list(
    temperature = .temperature,
    maxOutputTokens = .max_output_tokens,
    topP = .top_p,
    topK = .top_k,
    presencePenalty = .presence_penalty,
    frequencyPenalty = .frequency_penalty,
    stopSequences = .stop_sequences,
    thinkingConfig = if (!is.null(.thinking_budget)) list(thinkingBudget = .thinking_budget) else NULL
  ) |>
    append(response_format) |>
    purrr::compact()
  
  # Construct the request body
  request_body <- list(
    model = .model,
    system_instruction = system_prompt,
    contents = gemini_contents,
    generationConfig = generation_config,
    safetySettings = .safety_settings,
    tools = tools_def
  ) |>
    purrr::compact()


  
  if(.stream==FALSE) request_type <- ":generateContent"
  if(.stream==TRUE)  request_type <- ":streamGenerateContent"
  
  # Build the request
  request <- httr2::request("https://generativelanguage.googleapis.com") |>
    httr2::req_url_path(paste0("/v1beta/models/", .model, request_type)) |>
    httr2::req_url_query(key = api_key) |>
    httr2::req_headers(`Content-Type` = "application/json") |>
    httr2::req_body_json(request_body)
  
  if (.dry_run) return(request)
  
  # Perform the API request
  response <- perform_chat_request(request,api_obj,.stream,.timeout,.max_tries)
  
  if (.stream == FALSE && !is.null(raw_tools_def)) {
    response <- process_tool_loop(
      .api = api_obj,
      .response = response,
      .tools_def = raw_tools_def,
      .request_body = request_body,
      .request = request,
      .timeout = .timeout,
      .max_tries = .max_tries,
      .max_tool_rounds = .max_tool_rounds
    )
  }
  
  add_message(.llm     = .llm,
              .role    = "assistant", 
              .content = response$assistant_reply, 
              .json    = json,
              .meta    = response$meta)
}



#' Upload a File to Gemini API
#'
#' Uploads a file to the Gemini API and returns its metadata as a tibble.
#'
#' @param .file_path The local file path of the file to upload.
#' @return A tibble containing metadata about the uploaded file, including its name, URI, and MIME type.
#' @export
gemini_upload_file <- function(.file_path) {
  mime_type <- guess_mime_type(.file_path)
  num_bytes <- file.info(.file_path)$size
  display_name <- basename(.file_path)
  
  # Retrieve API key
  api_key <- Sys.getenv("GOOGLE_API_KEY")
  if ((api_key == "")) {
    stop("API key is not set. Please set it with: Sys.setenv(GOOGLE_API_KEY = 'YOUR-KEY-GOES-HERE')")
  }
  
  # Step 1: Initiate the upload and get the resumable upload URL
  init_response <- httr2::request("https://generativelanguage.googleapis.com/upload/v1beta/files") |>
    httr2::req_url_query(key = api_key) |>
    httr2::req_headers(
      `X-Goog-Upload-Protocol` = "resumable",
      `X-Goog-Upload-Command` = "start",
      `X-Goog-Upload-Header-Content-Length` = as.character(num_bytes),
      `X-Goog-Upload-Header-Content-Type` = mime_type,
      `Content-Type` = "application/json"
    ) |>
    httr2::req_body_json(list(file = list(display_name = display_name))) |>
    httr2::req_perform()
  
  upload_url <- httr2::resp_header(init_response, "x-goog-upload-url")
  
  if (is.null(upload_url)) {
    stop("Failed to get upload URL")
  }
  
  # Step 2: Upload the file bytes
  response <- httr2::request(upload_url) |>
    httr2::req_headers(
      `Content-Length` = as.character(num_bytes),
      `X-Goog-Upload-Offset` = "0",
      `X-Goog-Upload-Command` = "upload, finalize"
    ) |>
    httr2::req_body_raw(readBin(.file_path, "raw", num_bytes)) |>
    httr2::req_progress(type = "up") |>
    httr2::req_perform() |>
    httr2::resp_body_json()
  
  tibble::tibble(
    name = response$file$name,
    display_name = response$file$displayName,
    mime_type = response$file$mimeType,
    size_bytes = as.numeric(response$file$sizeBytes),
    create_time = response$file$createTime,
    uri = response$file$uri,
    state = response$file$state
  )
}

#' Retrieve Metadata for a File from Gemini API
#'
#' Retrieves metadata for a specific file uploaded to the Gemini API.
#'
#' @param .file_name The file ID (e.g., "files/abc-123") to retrieve metadata for.
#' @return A tibble containing metadata fields such as name, display name, MIME type, size, and URI.
#' @export
gemini_file_metadata <- function(.file_name) {
  # Retrieve API key
  api_key <- Sys.getenv("GOOGLE_API_KEY")
  if (api_key == "") {
    stop("API key is not set. Please set it with: Sys.setenv(GOOGLE_API_KEY = 'YOUR-KEY-GOES-HERE')")
  }
  
  # Request to get file metadata
  response <- httr2::request(paste0("https://generativelanguage.googleapis.com/v1beta/", .file_name)) |>
    httr2::req_url_query(key = api_key) |>
    httr2::req_perform() |>
    httr2::resp_body_json()
  
  # Convert the response to a tibble
  tibble::tibble(
    name = response$name,
    display_name = response$displayName,
    mime_type = response$mimeType,
    size_bytes = as.numeric(response$sizeBytes),
    create_time = response$createTime,
    update_time = response$updateTime,
    expiration_time = response$expirationTime,
    sha256_hash = response$sha256Hash,
    uri = response$uri,
    state = response$state
  )
}


#' List Files in Gemini API
#'
#' Lists metadata for files uploaded to the Gemini API, supporting pagination.
#'
#' @param .page_size The maximum number of files to return per page (default: 10, maximum: 100).
#' @param .page_token A token for fetching the next page of results (default: NULL).
#' @return A tibble containing metadata for each file, including fields such as name, display name, MIME type, and URI.
#' @export
gemini_list_files <- function(.page_size = 10, 
                              .page_token = NULL) {
  # Retrieve API key
  api_key <- Sys.getenv("GOOGLE_API_KEY")
  if (api_key == "") {
    stop("API key is not set. Please set it with: Sys.setenv(GOOGLE_API_KEY = 'YOUR-KEY-GOES-HERE')")
  }
  
  # Request to list files
  response <- httr2::request("https://generativelanguage.googleapis.com/v1beta/files") |>
    httr2::req_url_query(key = api_key, pageSize = .page_size, pageToken = .page_token) |>
    httr2::req_perform() |>
    httr2::resp_body_json()
  
  files <- response$files
  tibble::tibble(
    name = sapply(files, function(x) x$name),
    display_name = sapply(files, function(x) x$displayName),
    mime_type = sapply(files, function(x) x$mimeType),
    size_bytes = as.numeric(sapply(files, function(x) x$sizeBytes)),
    create_time = sapply(files, function(x) x$createTime),
    update_time = sapply(files, function(x) x$updateTime),
    expiration_time = sapply(files, function(x) x$expirationTime),
    sha256_hash = sapply(files, function(x) x$sha256Hash),
    uri = sapply(files, function(x) x$uri),
    state = sapply(files, function(x) x$state)
  )
}

#' Delete a File from Gemini API
#'
#' Deletes a specific file from the Gemini API using its file ID.
#'
#' @param .file_name The file ID (e.g., "files/abc-123") to delete.
#' @return Invisibly returns `NULL`. Prints a confirmation message upon successful deletion.
#' @export
gemini_delete_file <- function(.file_name) {
  # Retrieve API key
  api_key <- Sys.getenv("GOOGLE_API_KEY")
  if (api_key == "") {
    stop("API key is not set. Please set it with: Sys.setenv(GOOGLE_API_KEY = 'YOUR-KEY-GOES-HERE')")
  }
  
  # Request to delete the file
  httr2::request(paste0("https://generativelanguage.googleapis.com/v1beta/", .file_name)) |>
    httr2::req_url_query(key = api_key) |>
    httr2::req_method("DELETE") |>
    httr2::req_perform()
  
  message("File ", .file_name, " has been successfully deleted.")
}


#' Generate Embeddings Using the Google Gemini API
#'
#' @param .input  A character vector of texts to embed or an `LLMMessage` object
#' @param .model The embedding model identifier (default: "text-embedding-3-small").
#' @param .truncate Whether to truncate inputs to fit the model's context length (default: TRUE).
#' @param .timeout Timeout for the API request in seconds (default: 120).
#' @param .dry_run If TRUE, perform a dry run and return the request object.
#' @param .max_tries Maximum retry attempts for requests (default: 3).
#' @return A matrix where each column corresponds to the embedding of a message in the message history.
#' @export
gemini_embedding <- function(.input,
                             .model = "gemini-embedding-2-preview",
                             .truncate = TRUE,
                             .timeout = 120,
                             .dry_run = FALSE,
                             .max_tries = 3) {

  # Get the API key
  api_key <- Sys.getenv("GOOGLE_API_KEY")
  if ((api_key == "") & .dry_run == FALSE) {
    stop("API key is not set. Please set it with: Sys.setenv(GOOGLE_API_KEY = 'YOUR-KEY-GOES-HERE')")
  }
  
  # Validate inputs
  c(
    "Input .input must be a character vector or an LLMMessage object" = S7_inherits(.input, LLMMessage) | is.character(.input),
    "Input .model must be a string" = is.character(.model),
    "Input .truncate must be logical" = is.logical(.truncate),
    "Input .timeout must be a positive numeric value" = is.numeric(.timeout) && .timeout > 0,
    ".dry_run must be logical" = is.logical(.dry_run)
  ) |> validate_inputs()
  
  # Prepare message texts
  input_texts <- parse_embedding_input(.input)
  
  # Prepare batch request
  request_body <- list(
    requests = lapply(input_texts, function(text) {
      list(
        model = paste0("models/",.model),
        content = list(
          parts = list(
            list(text = text)
          )
        )
      )
    })
  )
  
  # Build the request
  request <-httr2::request("https://generativelanguage.googleapis.com") |>
    httr2::req_url_path(paste0("/v1beta/models/", .model, ":batchEmbedContents")) |>
    httr2::req_url_query(key = api_key) |>
    httr2::req_headers(`Content-Type` = "application/json") |>
    httr2::req_body_json(request_body)
  
  # Dry run
  if (.dry_run) {
    return(request)
  }
  
  extract_embeddings_fn <- function(response_content,error,headers){
    if(error){
      paste0("API error response - ", response_content$error$message) |>
        stop()
    }
    response_content$embeddings |>
      purrr::map(unlist)
  }
  
  # Perform a standard embedding API request
  perform_embedding_request(.request = request,
                            .timeout = .timeout,
                            .max_tries = 3,
                            .input_texts = input_texts, 
                            .fn_extract_embeddings = extract_embeddings_fn)
}

#' Submit a list of LLMMessage objects to Gemini's batch API
#'
#' Returns a named list (same as input) with batch_id and json attributes.
#' @param .llms List of LLMMessage objects (named or unnamed).
#' @param .model The model identifier (default: "gemini-1.5-flash").
#' @param .temperature Controls randomness (default: NULL, range: 0-2).
#' @param .max_output_tokens Maximum tokens in the response (default: NULL).
#' @param .top_p Nucleus sampling (default: NULL, range: 0-1).
#' @param .top_k Diversity in token selection (default: NULL).
#' @param .presence_penalty Penalizes new tokens (default: NULL, -2 to 2).
#' @param .frequency_penalty Penalizes frequent tokens (default: NULL, -2 to 2).
#' @param .stop_sequences Character vector or NULL of up to 5.
#' @param .safety_settings Optional list of safety settings (default: NULL).
#' @param .json_schema Optional schema to enforce output structure.
#' @param .grounding_threshold Optional grounding threshold (0-1) to enable Google Search.
#' @param .timeout Timeout in seconds (default: 120).
#' @param .dry_run If TRUE, returns the constructed request (default: FALSE).
#' @param .max_tries Maximum retry attempts (default: 3).
#' @param .display Display name for this batch (default: "tidyllm_batch").
#' @param .id_prefix Prefix for message IDs (default: "tidyllm_gemini_req_").
#' @return Named list of LLMMessage objects with attributes `batch_id` and `json`
#' @export
send_gemini_batch <- function(.llms,
                              .model = "gemini-2.5-flash",
                              .temperature = NULL,
                              .max_output_tokens = NULL,
                              .top_p = NULL,
                              .top_k = NULL,
                              .presence_penalty = NULL,
                              .frequency_penalty = NULL,
                              .stop_sequences = NULL,
                              .safety_settings = NULL,
                              .json_schema = NULL,
                              .grounding_threshold = NULL,
                              .timeout = 120,
                              .dry_run = FALSE,
                              .max_tries = 3,
                              .display = "tidyllm_batch",
                              .id_prefix = "tidyllm_gemini_req_") {
  
  # --- Input validation ---
  c(
    "Input .llms must be a list of LLMMessage objects" =
      is.list(.llms) && all(vapply(.llms, S7_inherits, logical(1), LLMMessage)),
    "Input .model must be a string" = is.character(.model) && length(.model) == 1,
    "Input .temperature must be NULL or in [0.0, 2.0]" =
      is.null(.temperature) || (.temperature >= 0.0 && .temperature <= 2.0),
    "Input .max_output_tokens must be NULL or integer >= 1" =
      is.null(.max_output_tokens) || (is_integer_valued(.max_output_tokens) && .max_output_tokens >= 1),
    "Input .top_p must be NULL or in [0.0, 1.0]" =
      is.null(.top_p) || (.top_p >= 0.0 && .top_p <= 1.0),
    "Input .top_k must be NULL or non-negative" =
      is.null(.top_k) || .top_k >= 0,
    "Input .presence_penalty must be NULL or in [-2.0, 2.0]" =
      is.null(.presence_penalty) || (.presence_penalty >= -2.0 && .presence_penalty <= 2.0),
    "Input .frequency_penalty must be NULL or in [-2.0, 2.0]" =
      is.null(.frequency_penalty) || (.frequency_penalty >= -2.0 && .frequency_penalty <= 2.0),
    "Input .stop_sequences must be NULL or a character vector of less or equal 5" =
      is.null(.stop_sequences) || (is.character(.stop_sequences) && length(.stop_sequences) <= 5),
    "Input .safety_settings must be NULL or a list" =
      is.null(.safety_settings) || is.list(.safety_settings),
    "Input .json_schema must be NULL, a list, or an ellmer type" =
      is.null(.json_schema) || is.list(.json_schema) || is_ellmer_type(.json_schema),
    "Input .grounding_threshold must be NULL or in [0.0, 1.0]" =
      is.null(.grounding_threshold) || (.grounding_threshold >= 0.0 && .grounding_threshold <= 1.0),
    "Input .timeout must be a positive integer" =
      is_integer_valued(.timeout) && .timeout > 0,
    "Input .max_tries must be a positive integer" =
      is_integer_valued(.max_tries) && .max_tries > 0,
    "Input .dry_run must be logical" = is.logical(.dry_run)
  ) |> validate_inputs()
  
  # --- Auto-name messages if needed ---
  if (is.null(names(.llms)) || anyNA(names(.llms)) || any(names(.llms) == "")) {
    names(.llms) <- sprintf("%s%s", .id_prefix, seq_along(.llms))
  }
  ids <- names(.llms)
  
  # --- Schema handling ---
  json <- FALSE
  if (requireNamespace("ellmer", quietly = TRUE)) {
    if (S7_inherits(.json_schema, ellmer::TypeObject)) {
      .json_schema <- to_schema(.json_schema)
    }
  }
  if (!is.null(.json_schema)) {
    json <- TRUE
    if ("additionalProperties" %in% names(.json_schema)) {
      .json_schema <- .json_schema[setdiff(names(.json_schema), "additionalProperties")]
    }
  }
  
  # --- Generation config ---
  gen_config <- list(
    temperature = .temperature,
    topP = .top_p,
    topK = .top_k,
    presencePenalty = .presence_penalty,
    frequencyPenalty = .frequency_penalty,
    maxOutputTokens = .max_output_tokens,
    stopSequences = .stop_sequences
  )
  if (json) {
    gen_config$response_mime_type <- "application/json"
    gen_config$response_schema <- .json_schema
  }
  gen_config <- purrr::compact(gen_config)
  
  # --- Only Google search grounding tool if specified ---
  grounding_tool <- if (!is.null(.grounding_threshold)) {
    list(
      google_search_retrieval = list(
        dynamic_retrieval_config = list(
          mode = "MODE_DYNAMIC",
          dynamic_threshold = .grounding_threshold
        )
      )
    )
  } else NULL
  
  # --- API key and API object ---
  api_obj <- api_gemini(short_name = "gemini",
                        long_name = "Google Gemini",
                        api_key_env_var = "GOOGLE_API_KEY")
  api_key <- get_api_key(api_obj, .dry_run)
  
  # --- Build batch requests, tagged with names as keys ---
  requests <- purrr::imap(.llms, function(msg, id) {
    list(
      request = purrr::compact(list(
        contents = to_api_format(msg, api_obj),
        system_instruction = list(parts = list(text = msg@system_prompt)),
        generationConfig = gen_config,
        safetySettings = .safety_settings,
        tools = if (!is.null(grounding_tool)) list(grounding_tool)
      )),
      metadata = list(key = id)
    )
  })
  
  # --- Final body for Gemini batch ---
  body <- list(
    batch = list(
      display_name = .display,
      input_config = list(
        requests = list(requests = unname(requests))
      )
    )
  )
  
  req <- httr2::request("https://generativelanguage.googleapis.com") |>
    httr2::req_url_path(paste0("/v1beta/models/", .model, ":batchGenerateContent")) |>
    httr2::req_url_query(key = api_key) |>
    httr2::req_body_json(body)
  
  if (.dry_run) return(req)
  
  resp <- perform_generic_request(req, .timeout, .max_tries)
  
  # --- Surface Gemini API errors as R errors ---
  if (!is.null(resp$content$error)) {
    err <- resp$content$error
    stop(sprintf("Gemini batch request failed [%s]: %s\n%s",
                 err$status, err$message,
                 if (!is.null(err$details[[1]]$links[[1]]$url)) err$details[[1]]$links[[1]]$url else ""
    ),
    call. = FALSE
    )
  }
  
  batch_id <- resp$content$name
  
  # --- Attach attributes, return named list (same as Claude) ---
  attr(.llms, "batch_id") <- batch_id
  attr(.llms, "json") <- json
  .llms
}



#' Check the Status of a Gemini Batch Operation
#'
#' Retrieves processing status and metadata for a Gemini batch operation.
#'
#' You can supply either the `.batch_id` string (e.g. `"batches/xyz..."`) **or**
#' a list of LLMMessage objects (`.llms`) with a `"batch_id"` attribute as returned by `send_gemini_batch()`.
#'
#' @param .llms (Optional) List of LLMMessage objects with a `"batch_id"` attribute (as returned by `send_gemini_batch()`).
#' @param .batch_id (Optional) Character string: full batch operation name, e.g. `"batches/xyz123"`.
#'   If both `.llms` and `.batch_id` are provided, `.batch_id` is used.
#' @param .timeout Integer. Request timeout in seconds. Default: 60.
#' @param .max_tries Integer. Maximum retry attempts. Default: 3.
#' @param .dry_run Logical. If TRUE, return the request object instead of making the request (for debugging). Default: FALSE.
#' @return A tibble with the operation's metadata, including name, state, creation time, completion time, and done status.
#' @export
check_gemini_batch <- function(.llms = NULL,
                               .batch_id = NULL,
                               .timeout = 60,
                               .max_tries = 3,
                               .dry_run = FALSE) {
  # If .batch_id missing, try to get from .llms 
  if (is.null(.batch_id)) {
    if (!is.null(.llms)) {
      .batch_id <- attr(.llms, "batch_id")
      if (is.null(.batch_id)) {
        stop("No batch_id attribute found in provided .llms object. Use send_gemini_batch() to generate batch_id.")
      }
    } else {
      stop("Must provide either .batch_id or .llms with a batch_id attribute.")
    }
  }

  
  api_key <- Sys.getenv("GOOGLE_API_KEY")
  if (api_key == "") stop("Google API key is not set (GOOGLE_API_KEY).")
  op_path <- if (!startsWith(.batch_id, "/v1beta/")) paste0("/v1beta/", .batch_id) else .batch_id
  req <- httr2::request("https://generativelanguage.googleapis.com") |>
    httr2::req_url_path(op_path) |>
    httr2::req_url_query(key = api_key)
  if (.dry_run) return(req)
  resp <- perform_generic_request(req, .timeout, .max_tries)
  content <- resp$content
  if (!is.null(content$error)) {
    stop(sprintf("Gemini API Error: %s - %s", content$error$code, content$error$message))
  }
  tibble::tibble(
    name = content$name,
    state = purrr::pluck(content, "metadata", "state", .default=NA_character_),
    done = purrr::pluck(content, "done", .default=NA),
    create_time = lubridate::ymd_hms(purrr::pluck(content, "metadata", "createTime", .default=NA_character_), tz="UTC"),
    complete_time = lubridate::ymd_hms(purrr::pluck(content, "metadata", "completeTime", .default=NA_character_), tz="UTC")
  )
}
#' List Recent Gemini Batch Operations
#'
#' Returns a tibble with recent Gemini batch operations and their metadata.
#'
#' @param .filter Optional filter expression for batch listing (see Gemini API docs).
#' @param .page_size Integer. Maximum number of results to return. Default: 20.
#' @param .timeout Integer. Request timeout in seconds. Default: 60.
#' @param .max_tries Integer. Maximum retry attempts. Default: 3.
#' @param .dry_run Logical. If TRUE, returns the request object (for debugging). Default: FALSE.
#' @return A tibble with columns: name, state, done, create_time, complete_time.
#' @export
list_gemini_batches <- function(.filter    = NULL,
                                .page_size = 20,
                                .timeout   = 60,
                                .max_tries = 3,
                                .dry_run   = FALSE) {
  api_key <- Sys.getenv("GOOGLE_API_KEY")
  if (api_key == "") stop("Google API key is not set (GOOGLE_API_KEY).")
  req <- httr2::request("https://generativelanguage.googleapis.com") |>
    httr2::req_url_path("/v1beta/batches") |>
    httr2::req_url_query(
      key      = api_key,
      filter   = .filter,
      pageSize = .page_size
    )
  if (.dry_run) return(req)
  resp <- perform_generic_request(req, .timeout, .max_tries)
  content <- resp$content
  if (!is.null(content$error)) {
    stop(sprintf("Gemini API Error: %s - %s", content$error$code, content$error$message))
  }
  batches <- purrr::pluck(content, "operations", .default = list())
  purrr::map_dfr(batches, function(x) tibble::tibble(
    name = x$name,
    state = purrr::pluck(x, "metadata", "state", .default=NA_character_),
    done = purrr::pluck(x, "done", .default=NA),
    create_time = lubridate::ymd_hms(purrr::pluck(x, "metadata", "createTime", .default=NA_character_), tz="UTC"),
    complete_time = lubridate::ymd_hms(purrr::pluck(x, "metadata", "completeTime", .default=NA_character_), tz="UTC")
  ))
}

#' Fetch Results for a Gemini Batch
#'
#' Retrieves the results of a completed Gemini batch and updates
#' the provided list of LLMMessage objects with the assistant's responses,
#' matching by original list order.
#'
#' @param .llms List of `LLMMessage` objects (as from `send_gemini_batch()`), must have a `batch_id` attribute if `.batch_name` is not given.
#' @param .batch_name (Optional) Character; batch operation name (e.g. "batches/xyz123"). If not provided, is taken from `attr(.llms, "batch_id")`.
#' @param .timeout Integer; request timeout in seconds (default: 60).
#' @param .max_tries Integer; maximum retry attempts (default: 3).
#' @param .dry_run Logical; if `TRUE`, returns the GET request object (default: FALSE).
#'
#' @return A list of updated LLMMessage objects with the assistant response appended to each, in the same order.
#' @export
fetch_gemini_batch <- function(.llms,
                               .batch_name = NULL,
                               .timeout = 60,
                               .max_tries = 3,
                               .dry_run = FALSE){
  
  # Preserve original names
  original_names <- names(.llms)
  if(!is.null(attr(.llms,"json"))) json <- attr(.llms,"json") else json <- FALSE
  
  # Validate inputs
  stopifnot(is.list(.llms), !is.null(.llms[[1]]))
  
  # Retrieve batch name
  if (is.null(.batch_name)) {
    .batch_name <- attr(.llms, "batch_id")
    if (is.null(.batch_name)) stop("No batch_name provided and no batch_id attribute found in .llms.")
  }
  
  api_obj <- api_gemini(short_name = "gemini",
                        long_name  = "Google Gemini",
                        api_key_env_var = "GOOGLE_API_KEY")
  
  api_key <- get_api_key(api_obj,.dry_run)
  
  op_path <- if (!startsWith(.batch_name, "/v1beta/")) paste0("/v1beta/", .batch_name) else .batch_name
  
  # Build GET request for batch status
  req <- httr2::request("https://generativelanguage.googleapis.com") |>
    httr2::req_url_path(op_path) |>
    httr2::req_url_query(key = api_key) |>
    httr2::req_headers(`Content-Type` = "application/json")
  
  if (.dry_run) return(req)
  
  resp <- perform_generic_request(req, .timeout, .max_tries)
  content <- resp$content
  
  # Handle batch not done or error state
  batch_state <- purrr::pluck(content, "metadata", "state", .default = "UNKNOWN")
  is_done <- purrr::pluck(content, "done", .default = FALSE)
  
  if (!is_done) stop(sprintf("Batch not finished processing (state: %s)", batch_state))
  
  if (batch_state == "JOB_STATE_FAILED") {
    stop(sprintf("Batch failed: %s", jsonlite::toJSON(content$error, auto_unbox=TRUE)))
  }
  if (batch_state == "JOB_STATE_CANCELLED") {
    stop("Batch was cancelled by the user.")
  }
  
  # Get responses: either inline or downloadable file
  response <- content$response
  responses <- NULL
  
  if (!is.null(response$inlinedResponses)) {
    # Inline responses: take as-is
    responses <- response$inlinedResponses$inlinedResponses
  } else if (!is.null(response$responsesFile)) {
    # Download responses file (JSONL format)
    file_path <- paste0("/download/v1beta/", response$responsesFile, ":download")
    download_req <- httr2::request("https://generativelanguage.googleapis.com") |>
      httr2::req_url_path(file_path) |>
      httr2::req_url_query(key = api_key, alt = "media")
    download_resp <- perform_generic_request(download_req, .timeout, .max_tries)
    responses_lines <- strsplit(httr2::resp_body_string(download_resp$raw_response), "\n")[[1]]
    responses <- lapply(responses_lines, function(line) if (nzchar(line)) jsonlite::fromJSON(line) else NULL)
    responses <- Filter(Negate(is.null), responses)
  } else {
    stop("No batch responses found in Gemini batch response.")
  }
  
  # Gemini batches preserve order; map responses to .llms in order
  if (length(responses) != length(.llms)) {
    stop(sprintf("Number of responses (%d) does not match number of LLM messages (%d).", length(responses), length(.llms)))
  }
  
  
  # Update LLMMessage objects with responses
  # Map results back to the original .llms list using names as custom IDs
  updated_llms <- purrr::imap(names(.llms),function(x,y){
    chat_response <- parse_chat_response(api_obj,responses[[y]]$response) 
    
    llm <- add_message(.llm = .llms[[x]],
                .role = "assistant", 
                .content = chat_response,
                .json = json,
                .meta = extract_metadata(api_obj,responses[[y]]$response) )
    llm
  }) 
  
  
  # Return updated list, with batch_id and json attributes removed
  attr(updated_llms, "batch_id") <- NULL
  attr(updated_llms, "json") <- NULL
  names(updated_llms) <- original_names
  updated_llms
}


#' List Available Models from the Google Gemini API
#'
#' @param .timeout Request timeout in seconds (default: 60).
#' @param .max_tries Maximum number of retries for the API request (default: 3).
#' @param .dry_run Logical; if TRUE, returns the prepared request object without executing it.
#'
#' @return A tibble containing model information with columns including `name`, `base_model_id`, 
#'   `version`, `display_name`, `description`, `input_token_limit`, `output_token_limit`, 
#'   `supported_generation_methods`, `thinking`, `temperature`, `max_temperature`, `top_p`, and `top_k`,
#'   or NULL if no models are found.
#'
#' @export
gemini_list_models <- function(.timeout = 60,
                               .max_tries = 3,
                               .dry_run = FALSE) {
  
  api_obj <- api_gemini(short_name = "gemini",
                        long_name  = "Google Gemini",
                        api_key_env_var = "GOOGLE_API_KEY")
  
  api_key <- get_api_key(api_obj, .dry_run)
  
  request <- httr2::request("https://generativelanguage.googleapis.com") |>
    httr2::req_url_path("/v1beta/models") |>
    httr2::req_url_query(key = api_key)
  
  if (.dry_run) {
    return(request)
  }
  
  response <- request |>
    httr2::req_timeout(.timeout) |>
    httr2::req_retry(max_tries = .max_tries) |>
    httr2::req_perform() |>
    httr2::resp_body_json()
  
  if (!is.null(response$models)) {
    models <- response$models
    
    model_info <- purrr::map_dfr(models, function(m) {
      tibble::tibble(
        name = m$name,
        display_name = purrr::pluck(m, "displayName", .default = NA_character_),
        version = purrr::pluck(m, "version", .default = NA_character_),
        description = purrr::pluck(m, "description", .default = NA_character_),
        input_token_limit = purrr::pluck(m, "inputTokenLimit", .default = NA_integer_),
        output_token_limit = purrr::pluck(m, "outputTokenLimit", .default = NA_integer_),
        supported_generation_methods = list(purrr::pluck(m, "supportedGenerationMethods", .default = list())),
        thinking = purrr::pluck(m, "thinking", .default = NA),
        temperature = purrr::pluck(m, "temperature", .default = NA_real_),
        max_temperature = purrr::pluck(m, "maxTemperature", .default = NA_real_),
        top_p = purrr::pluck(m, "topP", .default = NA_real_),
        top_k = purrr::pluck(m, "topK", .default = NA_integer_)
      )
    })
    
    return(model_info)
  } else {
    return(NULL)
  }
}




#' Google Gemini Provider Function
#'
#' The `gemini()` function acts as a provider interface for interacting with the Google Gemini API 
#' through `tidyllm`'s main verbs such as `chat()` and `embed()`. 
#' It dynamically routes requests to Gemini-specific functions 
#' like `gemini_chat()` and `gemini_embedding()` based on the context of the call.
#'
#' Some functions, such as `gemini_upload_file()` and `gemini_delete_file()`, 
#' are specific to Gemini and do not have general verb counterparts.
#'
#' @param ... Parameters to be passed to the appropriate Gemini-specific function, 
#'   such as model configuration, input text, or API-specific options.
#' @param .called_from An internal argument specifying which action (e.g., 
#'   `chat`, `embed`) the function is invoked from. 
#'   This argument is automatically managed by the `tidyllm` verbs and should not be modified by the user.
#'
#' @return The result of the requested action, depending on the specific function invoked 
#'   (e.g., an updated `LLMMessage` object for `chat()`).
#'
#' @export
gemini <- create_provider_function(
  .name = "gemini",
  chat = gemini_chat,
  embed = gemini_embedding,
  send_batch = send_gemini_batch,
  check_batch = check_gemini_batch,
  list_batches = list_gemini_batches,
  fetch_batch = fetch_gemini_batch,
  list_models = gemini_list_models
)

#' The Mistral API provider class
#'
#'Inherit most of the functionality from vanilla OpenAI API
#'
#' @noRd
api_mistral <- new_class("Mistral", api_openai)


#' Extract rate limit info from  Mistral API-Headers
#'
#' @noRd
method(ratelimit_from_header, list(api_mistral,new_S3_class("httr2_headers"))) <- function(.api,.headers){
  request_time <- strptime(.headers[["date"]], format = "%a, %d %b %Y %H:%M:%S", tz = "GMT")
  
  # Extract tokens remaining and reset time
  ratelimit_tokens_remaining <- as.integer(.headers[["x-ratelimitbysize-remaining-minute"]])
  ratelimit_tokens_limit <- as.integer(.headers[["x-ratelimitbysize-limit-minute"]])
  ratelimit_tokens_reset_dt <- as.integer(.headers[["ratelimitbysize-reset"]])
  ratelimit_tokens_reset_time <- request_time + ratelimit_tokens_reset_dt
  
  # For requests per second limit
  ratelimit_requests_remaining <- 0
  ratelimit_requests_reset_time <- request_time + 1  # Mistral allows 1 request per second
  
  list(
    this_request_time = request_time,
    ratelimit_requests_remaining = ratelimit_requests_remaining,
    ratelimit_requests_reset_time = ratelimit_requests_reset_time,
    ratelimit_tokens_remaining = ratelimit_tokens_remaining,
    ratelimit_tokens_reset_time = ratelimit_tokens_reset_time
  )
}

#' Prepare Mistral API request parameters
#'
#' @description
#' Helper function to prepare common request parameters for Mistral API calls,
#' centralizing duplicated logic between chat and batch operations
#'
#' @param .llm An LLMMessage object containing conversation history
#' @param .model The model identifier to use
#' @param .max_tokens Maximum tokens for completion
#' @param .frequency_penalty Frequency penalty parameter
#' @param .logit_bias Logit bias parameters
#' @param .presence_penalty Presence penalty parameter
#' @param .seed Random seed for reproducibility
#' @param .stop Sequence to stop generation
#' @param .temperature Temperature parameter
#' @param .top_p Top p parameter
#' @param .json_schema JSON schema for structured output
#'
#' @return A list containing prepared request parameters
#' @noRd
prepare_mistral_request <- function(
    .llm,
    .api,
    .model = "mistral-large-latest",
    .min_tokens = NULL,
    .max_tokens = NULL,
    .frequency_penalty = NULL,
    .logit_bias = NULL,
    .presence_penalty = NULL,
    .seed = NULL,
    .stop = NULL,
    .safe_prompt =NULL,
    .temperature = NULL,
    .top_p = NULL,
    .json_schema = NULL
) {
  # Format messages using the API format
  messages <- to_api_format(.llm, .api, TRUE)
  
  # Handle JSON schema and response format
  response_format <- NULL
  json <- FALSE
  if (!is.null(.json_schema)) {
    json <- TRUE
    schema_name <- "empty"
    if (requireNamespace("ellmer", quietly = TRUE)) {
      # Handle ellmer json schema objects
      if (S7_inherits(.json_schema, ellmer::TypeObject)) {
        .json_schema <- to_schema(.json_schema)
        schema_name <- "ellmer_schema"
      }
    }
    if (schema_name != "ellmer_schema") {
      schema_name <- attr(.json_schema, "name")
    }
    
  
    # Ensure additionalProperties is set to FALSE if not already provided
    if (is.null(.json_schema$additionalProperties)) {
      .json_schema$additionalProperties <- FALSE
    }
    
    response_format <- list(
      type = "json_schema",
      json_schema = list(
        schema = .json_schema,
        name = schema_name,
        strict = TRUE  # Mistral API expects this field
      )
    )
  }
  
  # Build common request body parameters
  request_body <- list(
    model = .model,
    messages = messages,
    frequency_penalty = .frequency_penalty,
    logit_bias = .logit_bias,
    max_tokens = .max_tokens,  # Note: This might also need to be max_tokens not max_completion_tokens
    presence_penalty = .presence_penalty,
    response_format = response_format,
    random_seed = .seed,  # Note: Mistral uses random_seed not seed
    safe_prompt = .safe_prompt,
    stop = .stop,
    temperature = .temperature,
    top_p = .top_p
  ) |> purrr::compact()
  
  # Return all prepared elements
  list(
    messages = messages,
    request_body = request_body,
    json = json,
    no_system_prompt = TRUE
  )
}

#' Send LLMMessage to Mistral API
#'
#' @param .llm An `LLMMessage` object.
#' @param .model The model identifier to use (default: `"mistral-large-latest"`). 
#' @param .frequency_penalty Number between -2.0 and 2.0. Positive values penalize new tokens based on their existing frequency.
#' @param .logit_bias A named list modifying the likelihood of specified tokens appearing in the completion.
#' @param .presence_penalty Number between -2.0 and 2.0. Positive values penalize new tokens based on whether they appear in the text so far.
#' @param .seed If specified, the system will make a best effort to sample deterministically.
#' @param .stop Up to 4 sequences where the API will stop generating further tokens.
#' @param .stream If set to TRUE, the answer will be streamed to console as it comes (default: FALSE).
#' @param .temperature What sampling temperature to use, between 0 and 2. Higher values make the output more random.
#' @param .top_p An alternative to sampling with temperature, called nucleus sampling.
#' @param .min_tokens The minimum number of tokens to generate in the completion. Must be `>= 0` (optional).
#' @param .max_tokens An upper bound for the number of tokens that can be generated for a completion.
#' @param .tools Either a single TOOL object or a list of TOOL objects representing the available functions for tool calls.
#' @param .tool_choice A character string specifying the tool-calling behavior; valid values are "none", "auto", or "required".
#' @param .json_schema A JSON schema object provided by tidyllm schema or ellmer schemata.
#' @param .safe_prompt Whether to inject a safety prompt before all conversations (default: `FALSE`).
#' @param .timeout When should our connection time out in seconds (default: `120`).
#' @param .verbose Should additional information be shown after the API call? (default: `FALSE`)
#' @param .dry_run If `TRUE`, perform a dry run and return the request object (default: `FALSE`).
#' @param .max_tries Maximum retries to peform request
#' @param .max_tool_rounds Integer specifying the maximum number of tool use iterations (default: 10). 
#'   Set to 1 for single-round tool use, or higher for multi-turn agentic loops.
#' @return Returns an updated `LLMMessage` object.
#' @export
mistral_chat <- function(.llm,
                         .model = "mistral-large-latest",
                         .frequency_penalty = NULL,
                         .logit_bias = NULL,
                         .presence_penalty = NULL,
                         .seed = NULL,
                         .stop = NULL,
                         .stream = FALSE,
                         .temperature = 0.7,
                         .top_p = 1,
                         .min_tokens = NULL,
                         .max_tokens = NULL,
                         .json_schema = NULL,
                         .safe_prompt = FALSE,
                         .timeout = 120,
                         .max_tries = 3,
                         .dry_run = FALSE,
                         .verbose = FALSE,
                         .tools = NULL,
                         .tool_choice = NULL,
                         .max_tool_rounds = 10) {
  
  # Validate the inputs
  c(
    "Input .llm must be an LLMMessage object" = S7_inherits(.llm, LLMMessage),
    "Input .model must be a string or NULL" = is.null(.model) || (is.character(.model) && length(.model) == 1),
    "Input .stream must be logical" = is.logical(.stream) && length(.stream) == 1,
    "Input .temperature must be numeric between 0 and 1.5" = is.null(.temperature) || (is.numeric(.temperature) && .temperature >= 0 && .temperature <= 1.5),
    "Input .top_p must be numeric between 0 and 1" = is.null(.top_p) || (is.numeric(.top_p) && .top_p >= 0 && .top_p <= 1),
    "Input .max_tokens must be NULL or a positive integer" = is.null(.max_tokens) || (is_integer_valued(.max_tokens) && .max_tokens > 0),
    "Input .min_tokens must be integer >= 0 if provided" = is.null(.min_tokens) || (is_integer_valued(.min_tokens) && .min_tokens >= 0),
    "Input .seed must be integer >= 0 if provided" = is.null(.seed) || (is_integer_valued(.seed) && .seed >= 0),
    "Input .stop must be a string or vector of strings if provided" = is.null(.stop) || (is.character(.stop)),
    "Input .frequency_penalty must be numeric or NULL" = is.null(.frequency_penalty) || is.numeric(.frequency_penalty),
    "Input .logit_bias must be a list or NULL" = is.null(.logit_bias) || is.list(.logit_bias),
    "Input .presence_penalty must be numeric or NULL" = is.null(.presence_penalty) || is.numeric(.presence_penalty),
    "Input .safe_prompt must be logical" = is.logical(.safe_prompt) && length(.safe_prompt) == 1,
    "Input .max_tries must be integer-valued numeric" = is_integer_valued(.max_tries),
    "Input .timeout must be integer-valued numeric (seconds till timeout)" = is_integer_valued(.timeout),
    "Input .json_schema must be NULL or a list or an ellmer type object" = is.null(.json_schema) || is.list(.json_schema) || is_ellmer_type(.json_schema),
    "Input .tools must be NULL, a TOOL object, or a list of TOOL objects" = is.null(.tools) || S7_inherits(.tools, TOOL) || (is.list(.tools) && all(purrr::map_lgl(.tools, ~ S7_inherits(.x, TOOL)))),
    "Input .tool_choice must be NULL or a character (one of 'none', 'auto', 'required')" = is.null(.tool_choice) || (is.character(.tool_choice) && .tool_choice %in% c("none", "auto", "required")),
    "Input .dry_run must be logical" = is.logical(.dry_run) && length(.dry_run) == 1,
    "Input .verbose must be logical" = is.logical(.verbose) && length(.verbose) == 1,
    ".max_tool_rounds must be a positive integer" = is_integer_valued(.max_tool_rounds) && .max_tool_rounds >= 1,
    "Streaming is not supported for requests with tool calls" = is.null(.tools) || !isTRUE(.stream)
  ) |>
    validate_inputs()
  
  # Create API object
  api_obj <- api_mistral(
    short_name = "mistral",
    long_name = "Mistral",
    api_key_env_var = "MISTRAL_API_KEY"
  )
  
  # Get API key
  api_key <- get_api_key(api_obj, .dry_run)
  
  # Use the prepare_openai_request helper to set up common parameters
  request_data <- prepare_mistral_request(
    .llm = .llm,
    .api = api_obj,
    .model = .model,
    .max_tokens = .max_tokens,
    .frequency_penalty = .frequency_penalty,
    .logit_bias = .logit_bias,
    .presence_penalty = .presence_penalty,
    .seed = .seed,
    .stop = .stop,
    .safe_prompt = .safe_prompt,
    .min_tokens = .min_tokens,
    .temperature = .temperature,
    .top_p = .top_p,
    .json_schema = .json_schema
  )
  
  # Get components from the request data
  request_body <- request_data$request_body
  json <- request_data$json
  
  # Add streaming if requested
  if (.stream) {
    request_body$stream <- TRUE
  }
  
  # Handle tools
  tools_def <- if (!is.null(.tools)) {
    if (S7_inherits(.tools, TOOL)) list(.tools) else .tools
  } else {
    NULL
  }
  
  # Add tools to request body if provided
  if (!is.null(tools_def)) {
    request_body$tools <- tools_to_api(api_obj, tools_def)
    request_body$tool_choice <- .tool_choice
  }
  

  # Build the request
  request <- httr2::request("https://api.mistral.ai") |>
    httr2::req_url_path("/v1/chat/completions") |>
    httr2::req_headers(
      `Authorization` = sprintf("Bearer %s", api_key),
      `Content-Type` = "application/json",
      .redact = "Authorization"
    ) |>
    httr2::req_body_json(request_body)
  
  # Return only the request object in a dry run
  if (.dry_run) {
    return(request)  
  }
  
  # Perform the request
  response <- perform_chat_request(request, api_obj, .stream, .timeout, .max_tries)
  
  # Handle tool calls with multi-turn support
  if (.stream == FALSE && !is.null(tools_def)) {
    response <- process_tool_loop(
      .api = api_obj,
      .response = response,
      .tools_def = tools_def,
      .request_body = request_body,
      .request = request,
      .timeout = .timeout,
      .max_tries = .max_tries,
      .max_tool_rounds = .max_tool_rounds
    )
  }
  
  # Extract assistant reply
  assistant_reply <- response$assistant_reply
  
  # Track rate limit
  track_rate_limit(api_obj, response$headers, .verbose)
  
  # Add model's message to the history of the LLMMessage object
  add_message(
    .llm = .llm,
    .role = "assistant", 
    .content = assistant_reply,
    .json = json,
    .meta = response$meta,
    .logprobs = NULL
  )
}


#' Generate Embeddings Using Mistral API
#'
#' @param .input A character vector of texts to embed or an `LLMMessage` object
#' @param .model The embedding model identifier (default: "mistral-embed").
#' @param .timeout Timeout for the API request in seconds (default: 120).
#' @param .dry_run If TRUE, perform a dry run and return the request object.
#' @param .max_tries Maximum retries to peform request
#' @return A matrix where each column corresponds to the embedding of a message in the message history.
#' @export
mistral_embedding <- function(.input,
                              .model = "mistral-embed",
                              .timeout = 120,
                              .max_tries = 3,
                              .dry_run = FALSE) {

  # Retrieve API key from environment variables
  api_key <- Sys.getenv("MISTRAL_API_KEY")
  if ((api_key == "")& .dry_run==FALSE){
    stop("API key is not set. Please set it with: Sys.setenv(MISTRAL_API_KEY = \"YOUR-KEY-GOES-HERE\")")
  }
  # Validate the inputs
  c(
    "Input .llm must character vector or an LLMMessage object" = S7_inherits(.input, LLMMessage) | is.character(.input),
    "Input .model must be a string" = is.character(.model),
    "Input .timeout must be an integer-valued numeric (seconds till timeout)" = is.numeric(.timeout) && .timeout > 0,
    ".dry_run must be logical" = is.logical(.dry_run)
  ) |> validate_inputs()
  
  input_texts <- parse_embedding_input(.input)
  # Prepare the request body
  request_body <- list(
    model = .model,
    input = input_texts
  )
  
  # Build the request
  request <- httr2::request("https://api.mistral.ai/v1/embeddings") |>
    httr2::req_headers(
      "Content-Type" = "application/json",
      "Authorization" = paste("Bearer", api_key),
      .redact = "Authorization"
    ) |>
    httr2::req_body_json(request_body)
  
  # Dry run
  if (.dry_run) {
    return(request)
  }
  
  extract_embeddings_fn <- function(response_content,error,headers){
    if(error){
      paste0("API error response (Code ", response_content$code ,") ", response_content$message)|>
        stop()
    }
    response_content$data |> 
      purrr::map("embedding")  |>
      purrr::map(unlist)
  }
  
  # Perform a standard embedding API request
  perform_embedding_request(.request = request,
                            .timeout = .timeout,
                            .max_tries = .max_tries,
                            .input_texts = input_texts, 
                            .fn_extract_embeddings = extract_embeddings_fn)
}

#' Send a Batch of Requests to the Mistral API
#'
#' This function creates and submits a batch of messages to the Mistral API for asynchronous processing.
#'
#' @param .llms A list of LLMMessage objects containing conversation histories.
#' @param .model The Mistral model version (default: "mistral-small-latest").
#' @param .endpoint The API endpoint (default: "/v1/chat/completions").
#' @param .metadata Optional metadata for the batch.
#' @param .temperature Sampling temperature to use, between `0.0` and `1.5` (default: `0.7`).
#' @param .top_p Nucleus sampling parameter, between `0.0` and `1.0` (default: `1`).
#' @param .max_tokens The maximum number of tokens to generate in the completion (default: `1024`).
#' @param .min_tokens The minimum number of tokens to generate (optional).
#' @param .frequency_penalty Numeric value (or NULL) for frequency penalty.
#' @param .logit_bias A named list modifying the likelihood of specific tokens (or NULL).
#' @param .presence_penalty Numeric value (or NULL) for presence penalty.
#' @param .seed Random seed for deterministic outputs (optional).
#' @param .stop Sequence(s) at which to stop generation (optional).
#' @param .safe_prompt Logical; if TRUE, injects a safety prompt (default: FALSE).
#' @param .json_schema A JSON schema object for structured output (optional).
#' @param .dry_run Logical; if TRUE, returns the prepared request without executing it (default: FALSE).
#' @param .overwrite Logical; if TRUE, allows overwriting existing custom IDs (default: FALSE).
#' @param .max_tries Maximum retry attempts for requests (default: 3).
#' @param .timeout Request timeout in seconds (default: 60).
#' @param .id_prefix Prefix for generating custom IDs (default: "tidyllm_mistral_req_").
#'
#' @return The prepared LLMMessage list with a batch_id attribute.
#' @export
send_mistral_batch <- function(.llms,
                               .model = "mistral-small-latest",
                               .endpoint = "/v1/chat/completions",
                               .metadata = NULL,
                               .temperature = 0.7,
                               .top_p = 1,
                               .max_tokens = 1024,
                               .min_tokens = NULL,
                               .frequency_penalty = NULL,
                               .logit_bias = NULL,
                               .presence_penalty = NULL,
                               .seed = NULL,
                               .stop = NULL,
                               .safe_prompt = FALSE,
                               .json_schema = NULL,
                               .dry_run = FALSE,
                               .overwrite = FALSE,
                               .max_tries = 3,
                               .timeout = 60,
                               .id_prefix = "tidyllm_mistral_req_") {
  
  # Input validation
  c(
    ".llms must be a list of LLMMessage objects" = is.list(.llms) && all(sapply(.llms, S7_inherits, LLMMessage)),
    ".model must be a string" = is.character(.model),
    ".endpoint must be a string" = is.character(.endpoint),
    ".metadata must be NULL or a list" = is.null(.metadata) || is.list(.metadata),
    ".temperature must be numeric between 0.0 and 1.5" = is.numeric(.temperature) && .temperature >= 0 && .temperature <= 1.5,
    ".top_p must be numeric between 0.0 and 1.0" = is.numeric(.top_p) && .top_p >= 0 && .top_p <= 1.0,
    ".max_tokens must be integer >= 0" = is_integer_valued(.max_tokens) && .max_tokens >= 0,
    ".min_tokens must be integer >= 0 if provided" = is.null(.min_tokens) || (is_integer_valued(.min_tokens) && .min_tokens >= 0),
    ".frequency_penalty must be numeric or NULL" = is.null(.frequency_penalty) || is.numeric(.frequency_penalty),
    ".logit_bias must be a list or NULL" = is.null(.logit_bias) || is.list(.logit_bias),
    ".presence_penalty must be numeric or NULL" = is.null(.presence_penalty) || is.numeric(.presence_penalty),
    ".seed must be integer >= 0 if provided" = is.null(.seed) || (is_integer_valued(.seed) && .seed >= 0),
    ".stop must be NULL or character" = is.null(.stop) || is.character(.stop),
    ".safe_prompt must be logical" = is.logical(.safe_prompt) && length(.safe_prompt) == 1,
    ".json_schema must be NULL or a list or an ellmer type object" = is.null(.json_schema) || is.list(.json_schema) || is_ellmer_type(.json_schema),
    ".dry_run must be logical" = is.logical(.dry_run),
    ".overwrite must be logical" = is.logical(.overwrite),
    ".id_prefix must be a string" = is.character(.id_prefix)
  ) |> validate_inputs()
  
  # Initialize API object and get API key
  api_obj <- api_mistral(short_name = "mistral",
                         long_name = "Mistral",
                         api_key_env_var = "MISTRAL_API_KEY")
  api_key <- get_api_key(api_obj, .dry_run)
  
  # Prepare LLMs for the batch
  prepared_llms <- prepare_llms_for_batch(api_obj, 
                                          .llms = .llms, 
                                          .id_prefix = .id_prefix, 
                                          .overwrite = .overwrite)
  
  # Generate request lines for the batch using the common prepare function
  request_lines <- lapply(seq_along(prepared_llms), function(i) {
    custom_id <- names(prepared_llms)[i]
    
    # Prepare common request parameters using prepare_mistral_request
    request_data <- prepare_mistral_request(
      .llm = prepared_llms[[i]],
      .api = api_obj,
      .model = .model,
      .min_tokens = .min_tokens,
      .max_tokens = .max_tokens,
      .frequency_penalty = .frequency_penalty,
      .logit_bias = .logit_bias,
      .presence_penalty = .presence_penalty,
      .seed = .seed,
      .stop = .stop,
      .safe_prompt = .safe_prompt,
      .temperature = .temperature,
      .top_p = .top_p,
      .json_schema = .json_schema
    )
    
    list(
      custom_id = custom_id,
      method = "POST",
      url = .endpoint,
      body = request_data$request_body
    ) |> jsonlite::toJSON(auto_unbox = TRUE)
  })
  
  # Save the batch requests to a temporary .jsonl file
  temp_file <- tempfile(fileext = ".jsonl")
  writeLines(unlist(request_lines), con = temp_file)
  
  if (.dry_run) {
    return(readLines(temp_file))
  }
  
  # Upload the batch file
  upload_request <- httr2::request("https://api.mistral.ai/v1/files") |>
    httr2::req_headers(Authorization = sprintf("Bearer %s", api_key)) |>
    httr2::req_body_multipart(purpose = "batch", file = curl::form_file(temp_file))
  
  upload_response <- upload_request |>
    perform_generic_request(.timeout = .timeout, .max_tries = .max_tries)
  
  input_file_id <- upload_response$content$id
  
  # Create the batch job
  batch_request_body <- list(
    input_files = list(input_file_id),
    model = .model,
    endpoint = .endpoint,
    metadata = .metadata
  )
  
  batch_request <- httr2::request("https://api.mistral.ai/v1/batch/jobs") |>
    httr2::req_headers(
      Authorization = sprintf("Bearer %s", api_key),
      `Content-Type` = "application/json"
    ) |>
    httr2::req_body_json(batch_request_body)
  
  batch_response <- batch_request |>
    perform_generic_request(.timeout = .timeout, .max_tries = .max_tries)
  
  if(batch_response$status == 403) {
    rlang::abort(sprintf("Mistral Batch API error: %s", batch_response$content$detail))
  }
  
  # Attach batch_id as an attribute to the prepared LLMs
  batch_id <- batch_response$content$id
  attr(prepared_llms, "batch_id") <- batch_id
  attr(prepared_llms, "json") <- !is.null(.json_schema)

  # Optionally remove the temporary file
  unlink(temp_file)

  return(prepared_llms)
}


#' Check Batch Processing Status for Mistral Batch API
#'
#' This function retrieves the processing status and other details of a specified Mistral batch ID
#' from the Mistral Batch API.
#'
#' @param .llms A list of LLMMessage objects.
#' @param .batch_id A manually set batch ID.
#' @param .dry_run Logical; if TRUE, returns the prepared request object without executing it (default: FALSE).
#' @param .max_tries Maximum retries to perform the request (default: 3).
#' @param .timeout Integer specifying the request timeout in seconds (default: 60).
#' @return A tibble with information about the status of batch processing.
#' @export
check_mistral_batch <- function(.llms = NULL,
                               .batch_id = NULL,
                               .dry_run = FALSE,
                               .max_tries = 3,
                               .timeout = 60) {
  # Extract batch_id
  if (is.null(.batch_id)) {
    if (!is.null(.llms)) {
      .batch_id <- attr(.llms, "batch_id")
      if (is.null(.batch_id)) {
        stop("No batch_id attribute found in the provided list.")
      }
    } else {
      stop("Either .llms or .batch_id must be provided.")
    }
  }
  
  # Retrieve API key
  api_key <- Sys.getenv("MISTRAL_API_KEY")
  if ((api_key == "") & !.dry_run){
    stop("API key is not set.")
  }
  
  # Build request
  request_url <- paste0("https://api.mistral.ai/v1/batch/jobs/", .batch_id)
  request <- httr2::request(request_url) |>
    httr2::req_headers(
      Authorization = sprintf("Bearer %s", api_key),
      `Content-Type` = "application/json"
    )
  
  # If .dry_run is TRUE, return the request object for inspection
  if (.dry_run) {
    return(request)
  }
  
  # Perform request with retries and error handling
  response <- request |>
    perform_generic_request(.timeout=.timeout,
                            .max_tries = .max_tries)
  
  # Parse response
  response_body <- response$content
  if("error" %in% names(response_body)){
    sprintf("Mistral API returned an Error:\nType: %s\nMessage: %s",
            response_body$error$type,
            response_body$error$message) |>
      stop()
  }
  
  # Extract relevant fields and handle timestamps
  tibble::tibble(
    batch_id = response_body$id,
    status = response_body$status,
    model = response_body$model,
    endpoint = response_body$endpoint,
    created_at = lubridate::as_datetime(response_body$created_at),
    started_at = if (!is.null(response_body$started_at)) lubridate::as_datetime(response_body$started_at) else NA,
    completed_at = if (!is.null(response_body$completed_at)) lubridate::as_datetime(response_body$completed_at) else NA,
    total_requests = response_body$total_requests,
    completed_requests = response_body$completed_requests,
    succeeded_requests = response_body$succeeded_requests,
    failed_requests = response_body$failed_requests,
    output_file = response_body$output_file,
    error_file = response_body$error_file
  )
}



#' List Mistral Batch Requests
#'
#' Retrieves batch request details from the OpenAI Batch API.
#'
#' @param .limit Maximum number of batches to retrieve (default: 20).
#' @param .max_tries Maximum retry attempts for requests (default: 3).
#' @param .timeout Request timeout in seconds (default: 60).
#' @param .status Filter by status. (default: NULL)
#' @param .created_after created after a string specifiying a date-time  (default: NULL)
#'
#' @return A tibble with batch details for all batches fitting the request
#'
#' @export
list_mistral_batches <- function(.limit = 100,
                                 .max_tries = 3,
                                 .timeout = 60,
                                 .status = NULL,
                                 .created_after = NULL
                                 ) {
  # Retrieve API key
  api_key <- Sys.getenv("MISTRAL_API_KEY")
  if (api_key == "") {
    stop("API key is not set. Please set it with: Sys.setenv(MISTRAL_API_KEY = \"YOUR-KEY-GOES-HERE\").")
  }
  
  query_params = list(
    page_size = .limit,
    status    = .status,
    created_after = .created_after
  ) |>
    purrr::compact()
  
  # Set up request URL with query parameters
  request <- httr2::request("https://api.mistral.ai/v1/batch/jobs") |>
    httr2::req_headers(
      Authorization = sprintf("Bearer %s", api_key),
      `Content-Type` = "application/json"
    ) |>
    httr2::req_url_query(!!!query_params)
  
  
  # Perform the request with retries and error handling
  response <- request |>
    perform_generic_request(.timeout=.timeout,
                            .max_tries = .max_tries)
  
  # Parse response
  response_body <- response$content
  if ("error" %in% names(response_body)) {
    sprintf("Mistral API returned an Error:\nType: %s\nMessage: %s",
            response_body$error$type,
            response_body$error$message) |>
      stop()  
  }
  
  # Extract batch data and format as tibble
  batch_data <- response_body$data
  batch_tibble <- purrr::map_dfr(batch_data, ~tibble::tibble(
    batch_id = .x$id,
    status   = .x$status,
    created_at = as.POSIXct(.x$created_at, origin = "1970-01-01", tz = "UTC"),
    started_at = if (!is.null(.x$started_at)) as.POSIXct(.x$started_at, origin = "1970-01-01", tz = "UTC") else NA,
    completed_at = if (!is.null(.x$completed_at)) as.POSIXct(.x$completed_at, origin = "1970-01-01", tz = "UTC") else NA,
    total_requests = .x$total_requests,
    completed_requests = .x$completed_requests,
    succeeded_requests = .x$succeeded_requests,
    failed_requests = .x$failed_requests,
    input_files = .x$input_files, 
    model = .x$model,
    endpoint = .x$endpoint,
    output_file = .x$output_file,
    error_file = .x$error_file
  ))
  
  return(batch_tibble)
}


#' Fetch Results for an Mistral Batch
#'
#' This function retrieves the results of a completed Mistral batch and updates
#' the provided list of `LLMMessage` objects with the responses. It aligns each
#' response with the original request using the `custom_id`s generated in `send_mistral_batch()`.
#'
#' @param .llms A list of `LLMMessage` objects that were part of the batch.
#' @param .batch_id Character; the unique identifier for the batch. By default this is NULL
#'                  and the function will attempt to use the `batch_id` attribute from `.llms`.
#' @param .dry_run Logical; if `TRUE`, returns the constructed request without executing it (default: `FALSE`).
#' @param .max_tries Integer; maximum number of retries if the request fails (default: `3`).
#' @param .timeout Integer; request timeout in seconds (default: `60`).
#'
#' @return A list of updated `LLMMessage` objects, each with the assistant's response added if successful.
#' @export
fetch_mistral_batch <- function(.llms,
                               .batch_id = NULL,
                               .dry_run = FALSE,
                               .max_tries = 3,
                               .timeout = 60) {

  c(
    ".llms must be a list of LLMMessage objects with names as custom IDs" = is.list(.llms) && all(sapply(.llms, S7_inherits, LLMMessage)),
    ".batch_id must be a non-empty character string or NULL" = is.null(.batch_id) || (is.character(.batch_id) && nzchar(.batch_id)),
    ".dry_run must be logical" = is.logical(.dry_run),
    ".max_tries must be integer-valued numeric" = is_integer_valued(.max_tries),
    ".timeout must be integer-valued numeric" = is_integer_valued(.timeout)
  ) |> validate_inputs()
  
  .json <- attr(.llms, "json")
  if (is.null(.json)) {.json <- FALSE}
  
  # Preserve original names
  original_names <- names(.llms)
  
  # Retrieve batch_id from .llms if not provided
  if (is.null(.batch_id)) {
    .batch_id <- attr(.llms, "batch_id")
    if (is.null(.batch_id)) {
      stop("No batch_id provided and no batch_id attribute found in the provided list.")
    }
  }
  
  #Get batch infos
  batch_info <- check_mistral_batch(.batch_id = .batch_id)
  # Check if batch has completed processing
  if (batch_info$status != "SUCCESS") {
    stop("Batch processing has not completed yet. Please check again later.")
  }
  

  api_obj <- api_mistral(short_name = "mistral",
                         long_name  = "Mistral",
                         api_key_env_var = "MISTRAL_API_KEY")
  
  api_key <- get_api_key(api_obj,.dry_run)

  # Download the output file
  results_url <- paste0("https://api.mistral.ai/v1/files/",  batch_info$output_file, "/content")
  results_request <- httr2::request(results_url) |>
    httr2::req_headers(
      Authorization = sprintf("Bearer %s", api_key)
    )
  
  results_response <- results_request |>
    httr2::req_timeout(.timeout) |>
    httr2::req_error(is_error = function(resp) FALSE) |>
    httr2::req_retry(
      max_tries = .max_tries,
      retry_on_failure = TRUE,
      is_transient = function(resp) httr2::resp_status(resp) %in% c(429, 503)
    ) |>
    httr2::req_perform()
  
  # Parse JSONL response and map results by custom_id
  results_lines <- strsplit(httr2::resp_body_string(results_response), "\n")[[1]]
  results_list <- lapply(results_lines, function(line) {
    if (nzchar(line)) jsonlite::fromJSON(line) else NULL
  })
  results_list <- Filter(Negate(is.null), results_list)
  
  results_by_custom_id <- purrr::set_names(results_list, sapply(results_list, function(x) x$custom_id))
  
  # Map results back to the original .llms list using names as custom IDs
  updated_llms <- lapply(names(.llms), function(custom_id) {
    result <- results_by_custom_id[[custom_id]]
    
    if (!is.null(result) && is.null(result$error) && result$response$status_code == 200) {
      assistant_reply <- result$response$body$choices$message$content
      meta_data <- extract_metadata(api_obj,result$response$body)
      llm <- add_message(.llm = .llms[[custom_id]],
                         .role = "assistant", 
                         .content =  assistant_reply,
                         .json = .json,
                         .meta = meta_data)
      return(llm)
    } else {
      warning(sprintf("Result for custom_id %s was unsuccessful or not found", custom_id))
      return(.llms[[custom_id]])
    }
  })
  
  # Restore original names
  names(updated_llms) <- original_names
  
  # Remove batch_id attribute before returning to avoid reuse conflicts
  attr(updated_llms, "batch_id") <- NULL
  attr(updated_llms, "json") <- NULL
  
  return(updated_llms)
}


#' List Available Models from the Mistral API
#'
#' @param .api_url Base URL for the API (default: "https://api.mistral.ai").
#' @param .timeout Request timeout in seconds (default: 60).
#' @param .max_tries Maximum number of retries for the API request (default: 3).
#' @param .dry_run Logical; if TRUE, returns the prepared request object without executing it.
#' @param .verbose Logical; if TRUE, prints additional information about the request.
#'
#' @return A tibble containing model information (columns include `id` and `created`),
#'   or NULL if no models are found.
#'
#' @export
mistral_list_models <- function(.api_url = "https://api.mistral.ai",
                                .timeout = 60,
                                .max_tries = 3,
                                .dry_run = FALSE,
                                .verbose = FALSE) {
  # Create an API object for Mistral using the tidyllm helper
  api_obj <- api_mistral(short_name = "mistral",
                         long_name  = "Mistral",
                         api_key_env_var = "MISTRAL_API_KEY")
  
  # Retrieve the API key (will error if not set, unless in dry run mode)
  api_key <- get_api_key(api_obj, .dry_run)
  
  # Build the request to the /v1/models endpoint
  request <- httr2::request(.api_url) |>
    httr2::req_url_path("/v1/models") |>
    httr2::req_headers(
      Authorization = sprintf("Bearer %s", api_key),
      `Content-Type` = "application/json"
    )
  
  # If dry run is requested, return the constructed request object
  if (.dry_run) {
    return(request)
  }
  
  # Perform the request with specified timeout and retry logic
  response <- request |>
    httr2::req_timeout(.timeout) |>
    httr2::req_retry(max_tries = .max_tries) |>
    httr2::req_perform() |>
    httr2::resp_body_json()
  
  if (.verbose) {
    message("Retrieved response from Mistral: ", response$object)
  }
  
  # Check if the "data" field exists and contains models
  if (!is.null(response$data)) {
    models <- response$data
    
    # Create a tibble with selected model information
    model_info <- tibble::tibble(
      id = vapply(models, function(model) model$id, character(1)),
      created = vapply(models, function(model) {
        as.character(strptime(
          format(as.POSIXct(model$created, origin = "1970-01-01", tz = "GMT"),
                 format = "%a, %d %b %Y %H:%M:%S"),
          format = "%a, %d %b %Y %H:%M:%S", tz = "GMT"))
      }, character(1))
    )
    
    return(model_info)
  } else {
    return(NULL)
  }
}


#' Mistral Provider Function
#'
#' The `mistral()` function acts as an interface for interacting with the Mistral API 
#' through main `tidyllm` verbs such as `chat()` and `embed()`. 
#' It dynamically routes requests to Mistral-specific functions 
#' like `mistral_chat()` and `mistral_embedding()` based on the context of the call.
#'
#' @param ... Parameters to be passed to the appropriate Mistral-specific function, 
#'   such as model configuration, input text, or API-specific options.
#' @param .called_from An internal argument that specifies which action (e.g., 
#'   `chat`, `embed`, `send_batch`) the function is being invoked from. 
#'   This argument is automatically managed and should not be modified by the user.
#'
#' @return The result of the requested action, depending on the specific function invoked 
#'   (e.g., an updated `LLMMessage` object for `chat()`, or a matrix for `embed()`).
#' 
#' @export
mistral <- create_provider_function(
  .name = "mistral",
  chat  = mistral_chat,
  embed = mistral_embedding,
  send_batch = send_mistral_batch,
  check_batch = check_mistral_batch,
  list_batches = list_mistral_batches,
  fetch_batch = fetch_mistral_batch,
  list_models = mistral_list_models
)

#' Large Language Model Message Class
#'
#' @description
#' `LLMMessage` is an S7 class for managing a conversation history intended for use with large language models (LLMs). Please use 
#' `llm_message()`to create or modify `LLMMessage` objects. 
#'
#' @param message_history A list containing messages. Each message is a named list with keys like `role`, `content`, `media`, etc.
#' @param system_prompt A character string representing the default system prompt used for the conversation.
#' @details
#' The `LLMMessage` class includes the following features:
#' - Stores message history in a structured format.
#' - Supports attaching media and metadata to messages.
#' - Provides generics like `add_message()`, `has_image()`, and `remove_message()` for interaction.
#' - Enables API-specific formatting through the `to_api_format()` generic.
#' - `message_history`: A list containing messages. Each message is a named list with keys like `role`, `content`, `media`, etc.
#' - `system_prompt`: A character string representing the default system prompt used for the conversation.
#' @export
LLMMessage <- new_class("LLMMessage", 
                        properties = list(
                          message_history = class_list,
                          system_prompt = class_character
                        )
)

#' An API provider
#'
#' At the moment this is just a stub but is needed for methods dispatch
#' @noRd
api_provider <- new_class("api_provider")

#LLMMessage generics
add_message <- new_generic("add_message", ".llm")
has_image   <- new_generic("has_image", ".llm")
remove_message <- new_generic("remove_message", ".llm")
to_api_format <- new_generic("to_api_format", c(".llm", ".api"))
print.LLMMessage     <- new_external_generic("base", "print", "x")
as_tibble.LLMMessage <- new_external_generic("tibble", "as_tibble", "x")

#' Add a New Message to LLMMessage
#'
#' Adds a new message to the `message_history` property of an `LLMMessage` object.
#' Supports roles such as "user", "assistant", and "system" along with optional
#' media, metadata, and JSON flags.
#'
#' @noRd
method(add_message, LLMMessage) <- function(.llm,
                                            .role,
                                            .content,
                                            .media = NULL,
                                            .json = FALSE,
                                            .meta = NULL,
                                            .logprobs = NULL) {
  
  message_details <- list(role = .role, content = .content, json = .json)
  if (!is.null(.media)) {
    message_details$media <- .media
  }
  if (!is.null(.meta)) {
    message_details$meta <- .meta
  }
  if (!is.null(.logprobs)) {  
    message_details$logprobs <- .logprobs
  }
  .llm@message_history <- c(.llm@message_history, list(message_details))
  .llm
}


#' Check if Any Messages Contain Images
#'
#' Determines whether any messages in the `message_history` property of an `LLMMessage` object
#' include attached images.
#'
#' @noRd
method(has_image, LLMMessage) <- function(.llm) {
  .llm@message_history |>
    purrr::map_lgl(function(m) {
      length(extract_media(m$media, "image")) > 0
    }) |>
    any()
}

#' Prints the current message history in a structured format.
#'
#' By default, this function respects the `tidyllm_print_metadata` option. 
#' If the option is set to `TRUE`, metadata for each message (if available) will be printed.
#' 
#' Users can override the default behavior by explicitly passing a value to the `.meta` parameter.
#'
#' @param .meta Logical; if `TRUE`, prints metadata for each message. 
#' Defaults to the value of `getOption("tidyllm_print_metadata", FALSE)`.
#' @noRd
method(print.LLMMessage,LLMMessage) <- function(x,...,.meta = getOption("tidyllm_print_metadata", FALSE)) {
  cat("Message History:\n")
  
  purrr::walk(seq_along(x@message_history), function(i) {
    message <- x@message_history[[i]]
    
    # Print role and content
    content_lines <- unlist(strsplit(message$content, "\n"))
    wrapped_lines <- purrr::map(content_lines, function(line) {
      stringr::str_wrap(line, width = 60)
    })
    wrapped_content <- paste(unlist(wrapped_lines), collapse = "\n")
    cat(sprintf("%s:\n%s\n", message$role, wrapped_content))
    
    # Print media details if available
    if (!is.null(message$media)) {
      media_files <- sapply(message$media, function(media) media$filename)
      if (length(media_files) != 0) {
        cat(" -> Attached Media Files: ", paste(media_files, collapse = ", "), "\n")
      }
    }
    
    # Print metadata if .meta is TRUE and metadata is available
    if (.meta && !is.null(message$meta)) {
      cat("\n  Metadata:\n")
      meta_fields <- names(message$meta)
      lapply(meta_fields[meta_fields!="specific_metadata"], function(key) {
        meta_value <- message$meta[[key]]
        # Ensure proper formatting of timestamp
        if (inherits(meta_value, "POSIXct")) {
          meta_value <- format(meta_value, "%Y-%m-%d %H:%M:%S")
        } else if (key == "timestamp" && is.numeric(meta_value)) {
          meta_value <- format(as.POSIXct(meta_value, origin = "1970-01-01", tz = "UTC"), "%Y-%m-%d %H:%M:%S")
        }
        
        cat(sprintf("    %s: %s\n", key, ifelse(is.null(meta_value), "NA", meta_value)))
      })
    }
    
    # Print message separator
    cat("--------------------------------------------------------------\n")
  })
  invisible(x)
}

method(remove_message, LLMMessage) <- function(.llm, .index) {
  # Validate index
  c(
    "Index must be a positive integer" = is_integer_valued(.index) && .index > 0,
    "Index is out of bounds" = .index <= length(.llm@message_history)) |> 
    validate_inputs()
  
  # Remove the message
  .llm@message_history <- .llm@message_history[-.index]
  .llm
}


#' Convert a LLMMessage Object to a Tibble
#'
#' Converts the `message_history` of an `LLMMessage` object into a tibble.
#'
#' @param x A `LLMMessage` object.
#' @param ... Additional arguments (ignored).
#' @param .rows Number of rows (ignored).
#' @param .name_repair Name repair strategy (ignored).
#' @param rownames Row names (ignored).
#' 
#' @return A tibble with two columns:
#'   - `role`: The role of the message sender (e.g., "user", "assistant", "system").
#'   - `content`: The message content.
#'   
#' @noRd 
method(as_tibble.LLMMessage, LLMMessage) <- function(x, ..., .rows = NULL, .name_repair = "check_unique", rownames = NULL) {
  purrr::map_dfr(x@message_history, ~{
    tibble::tibble(
      role = .x$role,
      content = .x$content
    )
  })
}

#' Pre-format Message for API format methods
#'
#' A helper function to process and format a message object for use with  `to_api_format` methods.
#'
#' @param message A list containing:
#'   - `content`: The base textual content of the message.
#'   - `media`: A list of media objects, where each object has `type` (e.g., "text", "image"), `filename` (for images), and `content` (the base64-encoded content or text content).
#'
#' @return A list formatted for APIs format. If `message$media` includes an image, the returned list contains:
#'   - `content`: The combined text content.
#'   - `image`: A sublist with:
#'     - `type`: Always `"base64"`.
#'     - `media_type`: The MIME type of the image (e.g., "image/png").
#'     - `data`: The base64-encoded image content.
#' If no image is present, the list only contains `content`.
#'
#' @keywords internal
#' @noRd
format_message <- function(message) {
  base_content <- message$content
  media_list <- message$media
  text_media <- extract_media(media_list, "text")
  image_media <- extract_media(media_list, "image")
  
  combined_text <- paste(base_content, text_media, sep = " ")
  
  if (length(image_media) > 0) {
    image_file_type <- guess_mime_type(image_media[[1]]$filename)
    base64_image <- image_media[[1]]$content
    
    list(
      content = combined_text,
      image = list(type = "base64", media_type = image_file_type, data = base64_image)
    )
  } else {
    list(content = combined_text)
  }
}

  

structure(list(method = "POST", url = "http://localhost:11434/api/chat", 
    status_code = 200L, headers = structure(list(`Content-Type` = "application/json; charset=utf-8", 
        Date = "Thu, 21 Aug 2025 12:41:21 GMT", `Content-Length` = "330"), redact = character(0), class = "httr2_headers"), 
    body = charToRaw("{\"model\":\"gemma2\",\"created_at\":\"2025-08-21T12:41:21.294827Z\",\"message\":{\"role\":\"assistant\",\"content\":\"Hello! 👋  How can I help you today?\"},\"done_reason\":\"stop\",\"done\":true,\"total_duration\":9819624917,\"load_duration\":4533008625,\"prompt_eval_count\":26,\"prompt_eval_duration\":4819809375,\"eval_count\":12,\"eval_duration\":462770333}"), 
    cache = new.env(parent = emptyenv())), class = "httr2_response")

#' Batch Process PDF into LLM Messages
#'
#' This function processes a PDF file page by page. For each page, it extracts the text
#' and converts the page into an image. It creates a list of LLMMessage objects with
#' the text and the image for multimodal processing. Users can specify a range of pages
#' to process and provide a custom function to generate prompts for each page.
#'
#' @param .pdf Path to the PDF file.
#' @param .general_prompt A default prompt that is applied to each page if `.prompt_fn` is not provided.
#' @param .system_prompt Optional system prompt to initialize the LLMMessage (default is "You are a helpful assistant").
#' @param .page_range A vector of two integers specifying the start and end pages to process. If NULL, all pages are processed.
#' @param .prompt_fn An optional custom function that generates a prompt for each page. The function takes the page text as input 
#' and returns a string. If NULL, `.general_prompt` is used for all pages.
#' 
#' @return A list of LLMMessage objects, each containing the text and image for a page.
#' @import pdftools
#' @import base64enc
#' @export
pdf_page_batch <- function(.pdf, 
                           .general_prompt, 
                           .system_prompt = "You are a helpful assistant",
                           .page_range = NULL,
                           .prompt_fn = NULL) {
  
  # Validate the inputs using validate_inputs
  validate_inputs(c(
    "The 'pdftools' package is required to read PDF files. Please install it." = requireNamespace("pdftools", quietly = TRUE),
    "The .pdf file must exist" = file.exists(.pdf),
    "The .general_prompt must be a non-empty string" = is.character(.general_prompt) && nchar(.general_prompt) > 0,
    "The .system_prompt must be a non-empty string" = is.character(.system_prompt) && nchar(.system_prompt) > 0,
    "If provided, .page_range must be a vector of two integer-valued numerics" = is.null(.page_range) || (
      is.numeric(.page_range) && length(.page_range) == 2 && 
        is_integer_valued(.page_range[1]) && is_integer_valued(.page_range[2])
    ),
    "The page range must be positive and start before the end" = is.null(.page_range) || (.page_range[1] > 0 && .page_range[1] <= .page_range[2])
  ))
  
  # Allow for anonymous prompt functions
  if (!is.null(.prompt_fn)) {.prompt_fn <- rlang::as_function(.prompt_fn)}
  
  # Extract text from PDF
  pdf_text <- pdftools::pdf_text(.pdf)
  num_pages <- length(pdf_text)
  
  # Limit to specified page range
  if (!is.null(.page_range)) {
    pdf_text <- pdf_text[.page_range[1]:min(.page_range[2], num_pages)]
  }
  
  message_list <- purrr::imap(pdf_text, function(text, i) {
    # Render the page as an image
    rendered_page <- pdftools::pdf_render_page(.pdf, page = i)
    
    # Convert the image to PNG format and then encode it as base64
    image_file <- tempfile(fileext = ".png")
    png::writePNG(rendered_page, image_file)
    
    # Generate the prompt for the current page
    llm_prompt <- if (is.null(.prompt_fn)) {
      paste0("Here are a picture of a page in a pdf and its text in <pdf></pdf> delimiters. 
           Please answer the following prompt:\n", .general_prompt, "\n<pdf>", text, "</pdf>")
    } else {
      .prompt_fn(text)
    }
    
    # Create an LLMMessage object for the page
    llm_message(
      .llm = llm_prompt, 
      .imagefile = image_file,
      .role = "user", 
      .system_prompt = .system_prompt
    )
  })
  
  # Return the list of LLMMessage objects
  return(message_list)
}

  
  

#' The Ollama API provider stub
#'
#' At the moment this is just a stub but is needed for methods dispatch
#'
#' @noRd
api_ollama <- new_class("Ollama", APIProvider)


#' Convert LLMMessage to Ollama API-Compatible Format
#'
#' Converts the `message_history` of an `LLMMessage` object into the
#' one needed for the Ollama API.
#'
#' @noRd
method(to_api_format, list(LLMMessage, api_ollama)) <- function(.llm, 
                                                                .api) {
  
  ollama_history <- filter_roles(.llm@message_history, c("user", "assistant"))
  lapply(ollama_history, function(m) {
    formatted_message <- format_message(m)
    if (!is.null(formatted_message$image)) {
      list(
        role = m$role,
        content = formatted_message$content,
        images = list(glue::glue("{formatted_message$image$data}"))
      )
    } else {
      list(role = m$role, content = formatted_message$content)
    }
  })
}

#' Format messages for Ollama tool calling API
#'
#' Converts assistant tool_calls messages and tool result messages
#' into the format expected by the Ollama API.
#'
#' @noRd
format_ollama_tool_messages <- function(.assistant_message, .tool_results) {
  assistant_msg <- list(
    role = "assistant",
    content = .assistant_message$content %||% "",
    tool_calls = .assistant_message$tool_calls
  )
  
  tool_msgs <- purrr::map(.tool_results, function(result) {
    list(
      role = "tool",
      content = result$content,
      tool_name = result$tool_name
    )
  })
  
  c(list(assistant_msg), tool_msgs)
}

#' A function to get metadata from Ollama responses
#'
#' @noRd
method(extract_metadata, list(api_ollama,class_list))<- function(.api,.response) {
  list(
    model             = .response$model,
    timestamp         = lubridate::as_datetime(.response$created_at),
    prompt_tokens     = .response$prompt_eval_count,
    completion_tokens = .response$eval_count,
    total_tokens      = .response$prompt_eval_count + .response$eval_count,
    stream            = FALSE,
    specific_metadata = list(
      done_reason     = .response$done_reason,
      total_duration_ns  = .response$total_duration,
      load_duration_ns   = .response$load_duration,
      eval_duration_ns   = .response$eval_duration
    ) 
  )
}  

#' A function to get metadata from Openai streaming responses
#'
#' @noRd
method(extract_metadata_stream, list(api_ollama,class_list))<- function(.api,.stream_raw_data) {
  final_stream_chunk <- .stream_raw_data[[length(.stream_raw_data)]]
  
  list(
    model             = final_stream_chunk$model,
    timestamp         = lubridate::as_datetime(final_stream_chunk$created),
    prompt_tokens     = final_stream_chunk$prompt_eval_count,
    completion_tokens = final_stream_chunk$eval_count,
    total_tokens      = final_stream_chunk$prompt_eval_count + final_stream_chunk$eval_count,
    stream            = TRUE,
    specific_metadata = list(
      done_reason     = final_stream_chunk$done_reason,
      total_duration_ns  = final_stream_chunk$total_duration,
      load_duration_ns   = final_stream_chunk$load_duration,
      eval_duration_ns   = final_stream_chunk$eval_duration
    ) 
  )
}  

#' A method to handle streaming requests for ollama
#'
#' @noRd
method(handle_stream,list(api_ollama,new_S3_class("httr2_response"))) <- function(.api,.stream_response) {
  stream_text <- ""
  stream_data <- list()
  repeat {
    #Sleep for a tiny bit because Ollama often hangs on sending complete lines
    Sys.sleep(0.25)
    stream_chunk   <- httr2::resp_stream_lines(.stream_response)
  
    stream_content <- stream_chunk |> 
      jsonlite::fromJSON()
    
    stream_data <- append(stream_data,list(stream_content))
    if (stream_content$done==TRUE) {
      close(.stream_response)
      message("\n---------\nStream finished\n---------\n")
      break
    }

    chunk_text  <- stream_content$message$content
    stream_text <- paste0(stream_text, chunk_text)
    cat(chunk_text)
    utils::flush.console()
    next
  }
  
  list(
    reply = stream_text,
    raw_data = stream_data
  )
}



#' A chat parsing method for Ollama to extract the assitant response 
#'
#' @noRd
method(parse_chat_response, list(api_ollama,class_list)) <- function(.api,.content) {
  api_label <- .api@long_name 
  if("error" %in% names(.content)){
    sprintf("%s returned an Error:\nMessage: %s",
            api_label,
            .content$error) |>
      stop()
  }
  
  .content$message$content
}



#' A method to run tool calls on Ollama and create the expected response
#'
#' @noRd
method(run_tool_calls, list(api_ollama, class_list, class_list)) <- function(.api, .tool_calls, .tools) {

  tool_results <- purrr::map(.tool_calls, function(tool_call) {
    tool_name <- tool_call$`function`$name
    tool_args <- tool_call$`function`$arguments
    
    if (is.null(tool_args)) {
      tool_args <- list()
    }
    
    matching_tool <- purrr::keep(.tools, ~ .x@name == tool_name)
    
    if (length(matching_tool) == 0) {
      warning(sprintf("No matching tool found for: %s", tool_name))
      return(NULL)
    }
    
    tool_function <- matching_tool[[1]]@func
    
    tool_result <- tryCatch({
      result <- do.call(tool_function, as.list(tool_args))
      if (is.character(result)) {
        result
      } else {
        jsonlite::toJSON(result, auto_unbox = TRUE)
      }
    }, error = function(e) {
      paste("Error executing tool:", e$message)
    })
    
    list(
      tool_name = tool_name,
      content = as.character(tool_result)
    )
  })
  
  purrr::compact(tool_results)
}

method(has_tool_calls, list(api_ollama, class_any)) <- function(.api, .response) {
  !is.null(.response$raw$content$message$tool_calls) &&
    length(.response$raw$content$message$tool_calls) > 0
}

method(extract_tool_calls, list(api_ollama, class_any)) <- function(.api, .response)
  .response$raw$content$message$tool_calls

method(append_tool_messages, list(api_ollama, class_any, class_any, class_any)) <-
  function(.api, .request_body, .response, .tool_results) {
    assistant_message <- .response$raw$content$message
    tool_messages <- format_ollama_tool_messages(assistant_message, .tool_results)
    .request_body$messages <- .request_body$messages |> append(tool_messages)
    .request_body
  }


#' Interact with local AI models via the Ollama API
#'
#'
#' @param .llm An LLMMessage object containing the conversation history and system prompt.
#' @param .model Character string specifying the Ollama model to use (default: "qwen3-vl")
#' @param .stream Logical; whether to stream the response (default: FALSE)
#' @param .seed Integer; seed for reproducible generation (default: NULL)
#' @param .json_schema A JSON schema object as R list to enforce the output structure (default: NULL)
#' @param .temperature Float between 0-2; controls randomness in responses (default: NULL)
#' @param .num_ctx Integer; sets the context window size (default: 2048)
#' @param .num_predict Integer; maximum number of tokens to predict (default: NULL)
#' @param .top_k Integer; controls diversity by limiting top tokens considered (default: NULL)
#' @param .top_p Float between 0-1; nucleus sampling threshold (default: NULL)
#' @param .min_p Float between 0-1; minimum probability threshold (default: NULL)
#' @param .mirostat Integer (0,1,2); enables Mirostat sampling algorithm (default: NULL)
#' @param .mirostat_eta Float; Mirostat learning rate (default: NULL)
#' @param .mirostat_tau Float; Mirostat target entropy (default: NULL)
#' @param .repeat_last_n Integer; tokens to look back for repetition (default: NULL)
#' @param .repeat_penalty Float; penalty for repeated tokens (default: NULL)
#' @param .tools Either a single TOOL object or a list of TOOL objects representing the available functions for tool calls.
#' @param .max_tool_rounds Integer; maximum number of tool use iterations for multi-turn tool calling (default: 10).
#'   Set to 1 for single-round tool use, or higher for multi-turn agentic loops.
#' @param .tfs_z Float; tail free sampling parameter (default: NULL)
#' @param .stop Character; custom stop sequence(s) (default: NULL)
#' @param .think Logical or character; controls thinking mode for supported models like Qwen3. Use FALSE to disable, TRUE to enable, or "high"/"medium"/"low" to set effort level (default: NULL - model default)
#' @param .keep_alive Character; How long should the ollama model be kept in memory after request (default: NULL - 5 Minutes)
#' @param .ollama_server String; Ollama API endpoint (default: "http://localhost:11434")
#' @param .timeout Integer; API request timeout in seconds (default: 120)
#' @param .dry_run Logical; if TRUE, returns request object without execution (default: FALSE)
#'
#' @return A new LLMMessage object containing the original messages plus the model's response
#'
#' @details
#' The function provides extensive control over the generation process through various parameters:
#' - Temperature (0-2): Higher values increase creativity, lower values make responses more focused
#' - Top-k/Top-p: Control diversity of generated text
#' - Mirostat: Advanced sampling algorithm for maintaining consistent complexity
#' - Repeat penalties: Prevent repetitive text
#' - Context window: Control how much previous conversation is considered
#'
#' @examples
#' \dontrun{
#' llm_message("user", "Hello, how are you?")
#' response <- ollama_chat(llm, .model = "gemma2", .temperature = 0.7)
#' 
#' # With custom parameters
#' response <- ollama_chat(
#'   llm,
#'   .model = "llama2",
#'   .temperature = 0.8,
#'   .top_p = 0.9,
#'   .num_ctx = 4096
#' )
#' }
#'
#' @export
ollama_chat <- function(.llm,
                   .model = "qwen3.5:4b",
                   .stream = FALSE,
                   .seed = NULL,
                   .json_schema = NULL,
                   .temperature = NULL,
                   .num_ctx = 2048,
                   .num_predict = NULL,
                   .top_k = NULL,
                   .top_p = NULL,
                   .min_p = NULL,
                   .mirostat = NULL,
                   .mirostat_eta = NULL,
                   .mirostat_tau = NULL,
                   .repeat_last_n = NULL,
                   .repeat_penalty = NULL,
                   .tools = NULL,
                   .max_tool_rounds = 10,
                   .tfs_z = NULL,
                   .stop = NULL,
                   .think = NULL,
                   .ollama_server = "http://localhost:11434",
                   .timeout = 120,
                   .keep_alive = NULL,
                   .dry_run = FALSE) {

  # Validate the inputs
  c(
    "Input .llm must be an LLMMessage object" = S7_inherits(.llm, LLMMessage),
    "Input .model must be a string" = is.character(.model),
    "Input .stream must be logical if provided" = is.logical(.stream),
    "Input .json_schema must be NULL or a list or an ellmer type object" = is.null(.json_schema) | is.list(.json_schema) | is_ellmer_type(.json_schema),
    "Input .temperature must be numeric between 0 and 2 if provided" = is.null(.temperature) || (is.numeric(.temperature) && .temperature >= 0 && .temperature <= 2),
    "Input .seed must be an integer-valued numeric if provided" = is.null(.seed) || is_integer_valued(.seed),
    "Input .num_ctx must be a positive integer if provided" = is.null(.num_ctx) || (is_integer_valued(.num_ctx) && .num_ctx > 0),
    "Input .num_predict must be an integer if provided" = is.null(.num_predict) || is_integer_valued(.num_predict),
    "Input .top_k must be a positive integer if provided" = is.null(.top_k) || (is_integer_valued(.top_k) && .top_k > 0),
    "Input .top_p must be numeric between 0 and 1 if provided" = is.null(.top_p) || (is.numeric(.top_p) && .top_p >= 0 && .top_p <= 1),
    "Input .min_p must be numeric between 0 and 1 if provided" = is.null(.min_p) || (is.numeric(.min_p) && .min_p >= 0 && .min_p <= 1),
    "Input .mirostat must be 0, 1, or 2 if provided" = is.null(.mirostat) || (.mirostat %in% c(0, 1, 2)),
    "Input .mirostat_eta must be positive numeric if provided" = is.null(.mirostat_eta) || (is.numeric(.mirostat_eta) && .mirostat_eta > 0),
    "Input .mirostat_tau must be positive numeric if provided" = is.null(.mirostat_tau) || (is.numeric(.mirostat_tau) && .mirostat_tau > 0),
    "Input .repeat_last_n must be integer if provided" = is.null(.repeat_last_n) || is_integer_valued(.repeat_last_n),
    "Input .repeat_penalty must be positive numeric if provided" = is.null(.repeat_penalty) || (is.numeric(.repeat_penalty) && .repeat_penalty > 0),
    "Input .tfs_z must be positive numeric if provided" = is.null(.tfs_z) || (is.numeric(.tfs_z) && .tfs_z > 0),
    "Input .stop must be character if provided" = is.null(.stop) || is.character(.stop),
    "Input .timeout must be a positive integer (seconds)" = is_integer_valued(.timeout) && .timeout > 0,
    "Input .keep_alive must be character" =  is.null(.keep_alive) ||  is.character(.keep_alive),
    "Input .tools must be NULL, a TOOL object, or a list of TOOL objects" = is.null(.tools) || S7_inherits(.tools, TOOL) || (is.list(.tools) && all(purrr::map_lgl(.tools, ~ S7_inherits(.x, TOOL)))),
    "Input .think must be logical or one of 'high', 'medium', 'low' if provided" = is.null(.think) || is.logical(.think) || (.think %in% c("high", "medium", "low")),
    "Input .dry_run must be logical" = is.logical(.dry_run),
    "Streaming is not supported for requests with tool calls" = is.null(.tools) || !isTRUE(.stream),
    ".max_tool_rounds must be a positive integer" = is_integer_valued(.max_tool_rounds) && .max_tool_rounds >= 1
  ) |>
    validate_inputs()

  api_obj <- api_ollama(short_name = "ollama",long_name = "Ollama")
  # Get formatted message list for ollama models
  ollama_messages <-  to_api_format(.llm,api_obj)
  
  # Handle JSON schema
  json=FALSE
  if (!is.null(.json_schema)) {
    json=TRUE
  } 
  if (requireNamespace("ellmer", quietly = TRUE)) {
    #Handle ellmer json schemata Objects
    if(S7_inherits(.json_schema,ellmer::TypeObject)){
      .json_schema = to_schema(.json_schema)
    } 
  }
  
  #Put a single tool into a list if only one is provided. 
  tools_def <- if (!is.null(.tools)) {
    if (S7_inherits(.tools, TOOL))  list(.tools) else .tools
  } else {
    NULL
  }
  
  ollama_options <- list(
    temperature = .temperature,
    seed = .seed,
    num_ctx = .num_ctx,
    num_predict = .num_predict,
    top_k = .top_k,
    top_p = .top_p,
    min_p = .min_p,
    mirostat = .mirostat,
    mirostat_eta = .mirostat_eta,
    mirostat_tau = .mirostat_tau,
    repeat_last_n = .repeat_last_n,
    repeat_penalty = .repeat_penalty,
    tfs_z = .tfs_z,
    stop = .stop
  ) |> purrr::compact()
  
  ollama_request_body <- list(
    model = .model,
    messages = ollama_messages,
    options = ollama_options,
    stream = .stream,
    think = .think,
    format = .json_schema,
    tools = if(!is.null(tools_def)) tools_to_api(api_obj,tools_def) else NULL
  )  |> purrr::compact()
  
  # Add keep_alive to request body only if it's provided
  if (!is.null(.keep_alive)) {
    ollama_request_body$keep_alive <- .keep_alive
  }

  # Build the request
  request <- httr2::request(.ollama_server) |>
    httr2::req_url_path("/api/chat") |>
    httr2::req_body_json(ollama_request_body)
  
  # Return only the request object in a dry run.
  if (.dry_run) {
    return(request)  
  }
  
  # Perform the API request
  response <- perform_chat_request(request, api_obj, .stream, .timeout, 3)
  
  if (.stream == FALSE && !is.null(tools_def)) {
    response <- process_tool_loop(
      .api = api_obj,
      .response = response,
      .tools_def = tools_def,
      .request_body = ollama_request_body,
      .request = request,
      .timeout = .timeout,
      .max_tries = 3,
      .max_tool_rounds = .max_tool_rounds
    )
  }
  
  add_message(.llm     = .llm,
              .role    = "assistant", 
              .content = response$assistant_reply, 
              .json    = json,
              .meta    = response$meta)
}


#' Generate Embeddings Using Ollama API
#'
#' @param .input Aa charachter vector of texts to embed or an `LLMMessage` object
#' @param .model The embedding model identifier (default: "all-minilm").
#' @param .truncate Whether to truncate inputs to fit the model's context length (default: TRUE).
#' @param .ollama_server The URL of the Ollama server to be used (default: "http://localhost:11434").
#' @param .timeout Timeout for the API request in seconds (default: 120).
#' @param .dry_run If TRUE, perform a dry run and return the request object.
#' @return A matrix where each column corresponds to the embedding of a message in the message history.
#' @export
ollama_embedding <- function(.input,
                             .model = "qwen3-embedding:0.6b",
                             .truncate = TRUE,
                             .ollama_server = "http://localhost:11434",
                             .timeout = 120,
                             .dry_run=FALSE) {

  # Validate the inputs
  c(
    "Input .input must be a character vector or an LLMMessage object" = S7_inherits(.input, LLMMessage) | is.character(.input),
    "Input .model must be a string" = is.character(.model),
    "Input .truncate must be logical" = is.logical(.truncate),
    "Input .timeout must be an integer-valued numeric (seconds till timeout)" = is.numeric(.timeout) && .timeout > 0,
    ".dry_run must be logical"                   = is.logical(.dry_run)
  ) |> validate_inputs()
  
  input_texts <- parse_embedding_input(.input)
  
  # Prepare the request body
  request_body <- list(
    model = .model,
    input = input_texts,
    truncate = .truncate
  )
  
  # Build the request
  request <- httr2::request(.ollama_server) |>
    httr2::req_url_path("/api/embed") |>
    httr2::req_body_json(request_body)
  
  # Return only the request object in a dry run.
  if (.dry_run) {
    return(request)  
  }
  
  extract_embeddings_fn <- function(response_content,error,headers){
    if(error){
      paste0("API error response - ", response_content$error) |>
        stop()
    }
    response_content$embeddings |>
      purrr::map(unlist)
  }
  
  # Perform a standard embedding API request
 perform_embedding_request(.request = request,
                           .timeout = .timeout,
                           .max_tries = 3,
                           .input_texts = input_texts, 
                           .fn_extract_embeddings = extract_embeddings_fn)
}


#' Send a Batch of Messages to Ollama API
#'
#' This function creates and submits a batch of messages to the Ollama API
#' Contrary to other batch functions, this functions waits for the batch to finish and receives requests.
#' The advantage compared to sending single messages via `chat()` is that Ollama handles large parallel
#' requests quicker than many individual chat requests.
#'
#' @param .llms A list of LLMMessage objects containing conversation histories.
#' @param .model Character string specifying the Ollama model to use (default: "qwen3-vl")
#' @param .stream Logical; whether to stream the response (default: FALSE)
#' @param .seed Integer; seed for reproducible generation (default: NULL)
#' @param .json_schema A JSON schema object as R list to enforce the output structure (default: NULL)
#' @param .temperature Float between 0-2; controls randomness in responses (default: NULL)
#' @param .num_ctx Integer; sets the context window size (default: 2048)
#' @param .num_predict Integer; maximum number of tokens to predict (default: NULL)
#' @param .top_k Integer; controls diversity by limiting top tokens considered (default: NULL)
#' @param .top_p Float between 0-1; nucleus sampling threshold (default: NULL)
#' @param .min_p Float between 0-1; minimum probability threshold (default: NULL)
#' @param .mirostat Integer (0,1,2); enables Mirostat sampling algorithm (default: NULL)
#' @param .mirostat_eta Float; Mirostat learning rate (default: NULL)
#' @param .mirostat_tau Float; Mirostat target entropy (default: NULL)
#' @param .repeat_last_n Integer; tokens to look back for repetition (default: NULL)
#' @param .repeat_penalty Float; penalty for repeated tokens (default: NULL)
#' @param .tfs_z Float; tail free sampling parameter (default: NULL)
#' @param .stop Character; custom stop sequence(s) (default: NULL)
#' @param .keep_alive Character; How long should the ollama model be kept in memory after request (default: NULL - 5 Minutes)
#' @param .ollama_server String; Ollama API endpoint (default: "http://localhost:11434")
#' @param .timeout Integer; API request timeout in seconds (default: 120)
#' @param .dry_run Logical; if TRUE, returns request object without execution (default: FALSE)
#'
#' @return A list of updated `LLMMessage` objects, each with the assistant's response added if successful.
#'
#' @details
#' The function provides extensive control over the generation process through various parameters:
#' - Temperature (0-2): Higher values increase creativity, lower values make responses more focused
#' - Top-k/Top-p: Control diversity of generated text
#' - Mirostat: Advanced sampling algorithm for maintaining consistent complexity
#' - Repeat penalties: Prevent repetitive text
#' - Context window: Control how much previous conversation is considered
#'
#' @export
send_ollama_batch <- function(.llms,
                        .model = "qwen3.5:4b",
                        .stream = FALSE,
                        .seed = NULL,
                        .json_schema = NULL,
                        .temperature = NULL,
                        .num_ctx = 2048,
                        .num_predict = NULL,
                        .top_k = NULL,
                        .top_p = NULL,
                        .min_p = NULL,
                        .mirostat = NULL,
                        .mirostat_eta = NULL,
                        .mirostat_tau = NULL,
                        .repeat_last_n = NULL,
                        .repeat_penalty = NULL,
                        .tfs_z = NULL,
                        .stop = NULL,
                        .ollama_server = "http://localhost:11434",
                        .timeout = 120,
                        .keep_alive = NULL,
                        .dry_run = FALSE) {

  # Validate the inputs
  c(
    ".llms must be a list of LLMMessage objects" = is.list(.llms) && all(sapply(.llms, S7_inherits, LLMMessage)),
    "Input .model must be a string" = is.character(.model),
    "Input .stream must be logical if provided" = is.logical(.stream),
    "Input .json_schema must be NULL or a list" = is.null(.json_schema) | is.list(.json_schema),
    "Input .temperature must be numeric between 0 and 2 if provided" = is.null(.temperature) || (is.numeric(.temperature) && .temperature >= 0 && .temperature <= 2),
    "Input .seed must be an integer-valued numeric if provided" = is.null(.seed) || is_integer_valued(.seed),
    "Input .num_ctx must be a positive integer if provided" = is.null(.num_ctx) || (is_integer_valued(.num_ctx) && .num_ctx > 0),
    "Input .num_predict must be an integer if provided" = is.null(.num_predict) || is_integer_valued(.num_predict),
    "Input .top_k must be a positive integer if provided" = is.null(.top_k) || (is_integer_valued(.top_k) && .top_k > 0),
    "Input .top_p must be numeric between 0 and 1 if provided" = is.null(.top_p) || (is.numeric(.top_p) && .top_p >= 0 && .top_p <= 1),
    "Input .min_p must be numeric between 0 and 1 if provided" = is.null(.min_p) || (is.numeric(.min_p) && .min_p >= 0 && .min_p <= 1),
    "Input .mirostat must be 0, 1, or 2 if provided" = is.null(.mirostat) || (.mirostat %in% c(0, 1, 2)),
    "Input .mirostat_eta must be positive numeric if provided" = is.null(.mirostat_eta) || (is.numeric(.mirostat_eta) && .mirostat_eta > 0),
    "Input .mirostat_tau must be positive numeric if provided" = is.null(.mirostat_tau) || (is.numeric(.mirostat_tau) && .mirostat_tau > 0),
    "Input .repeat_last_n must be integer if provided" = is.null(.repeat_last_n) || is_integer_valued(.repeat_last_n),
    "Input .repeat_penalty must be positive numeric if provided" = is.null(.repeat_penalty) || (is.numeric(.repeat_penalty) && .repeat_penalty > 0),
    "Input .tfs_z must be positive numeric if provided" = is.null(.tfs_z) || (is.numeric(.tfs_z) && .tfs_z > 0),
    "Input .stop must be character if provided" = is.null(.stop) || is.character(.stop),
    "Input .timeout must be a positive integer (seconds)" = is_integer_valued(.timeout) && .timeout > 0,
    "Input .keep_alive must be character" =  is.null(.keep_alive) ||  is.character(.keep_alive),
    "Input .dry_run must be logical" = is.logical(.dry_run)
  ) |>
    validate_inputs()
  
  api_obj <- api_ollama(short_name = "ollama",long_name = "Ollama")
  
  # Handle JSON schema
  json=FALSE
  if (!is.null(.json_schema)) {
    json=TRUE
  } 
  if (requireNamespace("ellmer", quietly = TRUE)) {
    #Handle ellmer json schemata Objects
    if(S7_inherits(.json_schema,ellmer::TypeObject)){
      .json_schema = to_schema(.json_schema)
    } 
  }
  
  # Prepare the request lines
  ollama_requests <- lapply(seq_along(.llms), function(i) { 

    # Get messages from each LLMMessage object
    ollama_messages <- to_api_format(.llm = .llms[[i]],
                              .api=api_obj)
    

    ollama_options <- list(
      temperature = .temperature,
      seed = .seed,
      num_ctx = .num_ctx,
      num_predict = .num_predict,
      top_k = .top_k,
      top_p = .top_p,
      min_p = .min_p,
      mirostat = .mirostat,
      mirostat_eta = .mirostat_eta,
      mirostat_tau = .mirostat_tau,
      repeat_last_n = .repeat_last_n,
      repeat_penalty = .repeat_penalty,
      tfs_z = .tfs_z,
      stop = .stop
    ) |> purrr::compact()
    
    ollama_request_body <- list(
      model = .model,
      messages = ollama_messages,
      options = ollama_options,
      stream = .stream,
      format = .json_schema
    )  |> purrr::compact()
    
    # Add keep_alive to request body only if it's provided
    if (!is.null(.keep_alive)) {
      ollama_request_body$keep_alive <- .keep_alive
    }
    
    # Build the request
    request <- httr2::request(.ollama_server) |>
      httr2::req_url_path("/api/chat") |>
      httr2::req_body_json(ollama_request_body)
  })
  

  # Return only the request object in a dry run.
  if (.dry_run) {
    return(ollama_requests)  
  }
  
  ollama_responses <- ollama_requests |>
    httr2::req_perform_parallel()
  
  updated_llms <- lapply(seq_along(.llms), function(i) {
    response_content <- httr2::resp_body_json(ollama_responses[[i]])
    
    meta <- extract_metadata(api_obj,response_content)
    chat_content <- parse_chat_response(api_obj,response_content)
    add_message(.llm     = .llms[[i]],
                .role    = "assistant", 
                .content = chat_content, 
                .json    = json,
                .meta    = meta)
  })
  
  return(updated_llms)
}


#' Retrieve and return model information from the Ollama API
#' 
#' This function connects to the Ollama API and retrieves information 
#' about available models, returning it as a tibble.
#' @return A tibble containing model information, or NULL if no models are found.
#' @param .ollama_server The URL of the ollama server to be used
#' @export
ollama_list_models <- function(.ollama_server = "http://localhost:11434") {
  
  # Perform the request and save the response
  response <- httr2::request(.ollama_server) |>
    httr2::req_url_path("/api/tags") |>
    httr2::req_perform() |>
    httr2::resp_body_json()
  
  # Check if 'models' key exists in the response
  if (!is.null(response$models)) {
    models <- response$models
    
    # Create a tibble with model information
    model_info <- tibble::tibble(
      name = sapply(models, function(model) model$name),
      modified_at = sapply(models, function(model) model$modified_at),
      size = sapply(models, function(model) model$size),
      format = sapply(models, function(model) model$details$format),
      family = sapply(models, function(model) model$details$family),
      parameter_size = sapply(models, function(model) model$details$parameter_size),
      quantization_level = sapply(models, function(model) model$details$quantization_level)
    )
    
    return(model_info)
    
  } else {
    # Return NULL if no models are found
    return(NULL)
  }
}


#' Download a model from the Ollama API
#'
#' This function sends a request to the Ollama API to download a specified model 
#' from Ollama's large online library of models. 
#' 
#' @param .model The name of the model to download.
#' @param .ollama_server The base URL of the Ollama API (default is "http://localhost:11434").
#' @return NULL
#' @export
ollama_download_model <- function(.model, .ollama_server = "http://localhost:11434") {
  # Initialize a progress bar
  progress_bar <- cli::cli_progress_bar(auto_terminate = FALSE, type = "download")
  
  # Prepare the callback function for handling stream data
  callback_modeldownload_stream <- function(.stream, progress_bar) {
    line <- rawToChar(.stream)
    line <- trimws(line)
    
    # Ignore empty lines
    if (nchar(line) == 0) {
      return(TRUE)
    }
    
    # Parse the JSON with error handling
    stream_content <- tryCatch(
      jsonlite::fromJSON(line),
      error = function(e) NULL
    )
    
    if (is.null(stream_content)) {
      return(TRUE)
    }
    
    # Show a progress bar for the download 
    if (stringr::str_detect(stream_content$status, "pulling")) {
      if (!is.null(stream_content$total) && !is.null(stream_content$completed) &&
          stream_content$total > 0 && stream_content$completed >= 0) {
        # Update the progress bar
        cli::cli_progress_update(set = stream_content$completed,
                                 total = stream_content$total,
                                 id = progress_bar)
      }
    }
    TRUE
  }
  
  # Build and perform the request
  request <- httr2::request(.ollama_server) |>
    httr2::req_url_path("/api/pull") |>
    httr2::req_body_json(list(name = .model)) |>
    httr2::req_perform_stream(function(.stream) callback_modeldownload_stream(.stream, progress_bar),
                              buffer_kb = 0.05, round = "line")
  
  cli::cli_progress_done(id = progress_bar)
  invisible(NULL)
}

#' Delete a model from the Ollama API
#'
#' This function sends a DELETE request to the Ollama API to remove a specified model.
#'
#' @param .model The name of the model to delete.
#' @param .ollama_server The base URL of the Ollama API (default is "http://localhost:11434").
#' @return NULL
#' @export
ollama_delete_model <- function(.model, .ollama_server = "http://localhost:11434") {
  

  # Perform DELETE request to remove the model
  httr2::request(.ollama_server) |>
    httr2::req_url_path_append("/api/delete") |>
    httr2::req_method("DELETE") |>
    httr2::req_body_json(list(name = .model)) |>
    httr2::req_error(body = function(resp) httr2::resp_body_json(resp)$error) |>
    httr2::req_perform()
  
  # Run ollama_list_models() to show available models after deletion
  warning(paste0("Model ", .model," deleted"))
  
  ollama_list_models(.ollama_server)
}



#' Ollama API Provider Function
#'
#' The `ollama()` function acts as an interface for interacting with local AI models via the Ollama API.
#' It integrates seamlessly with the main `tidyllm` verbs such as `chat()` and `embed()`.
#' 
#' Some functionalities, like `ollama_download_model()` or `ollama_list_models()`
#' are unique to the Ollama API and do not have a general verb counterpart. 
#' These functions can be only accessed directly.
#'
#' @param ... Parameters to be passed to the appropriate Ollama-specific function, 
#'   such as model configuration, input text, or API-specific options.
#' @param .called_from An internal argument specifying the verb (e.g., `chat`, `embed`) 
#'   the function is invoked from. This argument is automatically managed by `tidyllm` and 
#'   should not be set by the user.
#'
#' @return The result of the requested action:
#'   - For `chat()`: An updated `LLMMessage` object containing the model's response.
#'   - For `embed()`: A matrix where each column corresponds to an embedding.
#'
#' @details
#' Supported Verbs:
#' - **`chat()`**: Sends a message to an Ollama model and retrieves the model's response.
#' - **`embed()`**: Generates embeddings for input texts using an Ollama model.
#' - **`send_batch()`**: Behaves different than the other `send_batch()` verbs since it immediately processes the answers
#' @export
ollama <- create_provider_function(
  .name = "ollama",
  chat = ollama_chat,
  embed = ollama_embedding,
  send_batch = send_ollama_batch,
  list_models = ollama_list_models   
)

testthat::skip_if_not_installed("httptest2")
library(httptest2)

test_that("ollama function constructs a correct request and dry runs it", {
  # Call groq with .dry_run = TRUE and perform the dry run
  request <- llm_message("Write a poem about porcupines") |> 
    chat(ollama,.model="gemma2",.temperature = 0,.dry_run = TRUE)
  
  dry_run <- request |>
    httr2::req_dry_run(redact_headers = TRUE, quiet = TRUE)
  
  # Check that the method is POST
  expect_equal(dry_run$method, "POST")
  
  #Dry run has a reasonable path 
  expect_equal(dry_run$path, "/api/chat")
  
  # Inspect headers
  headers <- dry_run$headers
  expect_type(headers, "list")
  
  # Check that the required headers are present
  expect_true("content-length" %in% names(headers))
  expect_true("content-type" %in% names(headers))
  
  #Content length is identical in header
  expect_equal(headers$`content-length`,"148")
  
  # Check that the content-type is JSON
  expect_equal(headers$`content-type`, "application/json")
  
  # Now check the body content to ensure the JSON is constructed as expected
  body_json <- request$body |> jsonlite::toJSON() |> as.character()
  
  expected_json <- "{\"data\":{\"model\":[\"gemma2\"],\"messages\":[{\"role\":[\"user\"],\"content\":[\"Write a poem about porcupines \"]}],\"options\":{\"temperature\":[0],\"num_ctx\":[2048]},\"stream\":[false]},\"type\":[\"json\"],\"content_type\":[\"application/json\"],\"params\":{\"auto_unbox\":[true],\"digits\":[22],\"null\":[\"null\"]}}"
  # Check if the JSON matches the expected JSON
  expect_equal(body_json, expected_json)
  
})

test_that("ollama_embedding function constructs a correct request and dry runs it", {
  
  #Character vector request
  request <- c("It is not that I am mad, it is only that my head is different from yours",
    "A man can do as he wills, but not will as he wills",
    "Whereof one cannot speak, thereof one must be silent",
    "The limits of my language mean the limits of my world") |>
    ollama_embedding(.dry_run = TRUE)
  
  dry_run <- request |>
    httr2::req_dry_run(redact_headers = TRUE, quiet = TRUE) 
  
  # Check that the method is POST
  expect_equal(dry_run$method, "POST")
  
  #Dry run has a reasonable path 
  expect_equal(dry_run$path, "/api/embed")
  
  # Inspect headers
  headers <- dry_run$headers
  expect_type(headers, "list")
  
  # Check that the required headers are present
  expect_true("content-length" %in% names(headers))
  expect_true("content-type" %in% names(headers))
  
  # Check that the content-type is JSON
  expect_equal(headers$`content-type`, "application/json")
  
  # Now check the body content to ensure the JSON is constructed as expected
  body_json <- request$body |> jsonlite::toJSON() |> as.character()
  expected_json <- "{\"data\":{\"model\":[\"qwen3-embedding:0.6b\"],\"input\":[\"It is not that I am mad, it is only that my head is different from yours\",\"A man can do as he wills, but not will as he wills\",\"Whereof one cannot speak, thereof one must be silent\",\"The limits of my language mean the limits of my world\"],\"truncate\":[true]},\"type\":[\"json\"],\"content_type\":[\"application/json\"],\"params\":{\"auto_unbox\":[true],\"digits\":[22],\"null\":[\"null\"]}}"
  expect_equal(body_json, expected_json)
})

test_that("ollama returns expected response", {
  with_mock_dir("ollama",expr = {
    

    llm <- llm_message("Hello, world")
    
    result <- ollama_chat(
      .llm = llm,
      .temperature = 0,
      .model = "gemma2"
    )
    result_tbl <- as_tibble(result)
    
    # Assertions based on the message in the captured mock response
    expect_true(S7_inherits(result, LLMMessage))
    expect_equal(
      result_tbl$content[3],
      "Hello! 👋  How can I help you today?"
    )
    expect_equal(result_tbl$role[3], "assistant")
    
  },simplify = FALSE)
})


test_that("ollama_embedding returns expected response", {
  with_mock_dir("ollama_embedding",expr = {
    
    result <- c("It is not that I am mad, it is only that my head is different from yours",
                "A man can do as he wills, but not will as he wills",
                "Whereof one cannot speak, thereof one must be silent",
                "The limits of my language mean the limits of my world") |>
      ollama_embedding(.model = "all-minilm")
    
    
    # Test that the result is a tibble
    expect_s3_class(result, "tbl_df")
    
    # Test that the tibble has two columns: input and embeddings
    expect_named(result, c("input", "embeddings"))
    
    # Test that the input column contains the original input texts
    expect_equal(result$input, c("It is not that I am mad, it is only that my head is different from yours",
                                 "A man can do as he wills, but not will as he wills",
                                 "Whereof one cannot speak, thereof one must be silent",
                                 "The limits of my language mean the limits of my world"))
    
    purrr::walk(result$embeddings, function(embedding) {
      expect_equal(length(embedding), 384)
    })
    
  },simplify = FALSE)
})

#' Generate Embeddings Using Voyage AI API
#'
#' This function creates embedding vectors from text or multimodal inputs (text and images)
#' using the Voyage AI API. It supports three types of input:
#' 
#' 1. Character vector: Embeds each text string separately
#' 2. LLMMessage object: Extracts and embeds text content from messages
#' 3. List of mixed content: Processes a combination of text strings and image objects created with `img()`
#'
#' For multimodal inputs, the function automatically switches to Voyage's multimodal API
#' and formats the response with appropriate labels (e.g., \code{"[IMG] image.png"}) for images.
#'
#' @param .input Input to embed. Can be:
#'   - A character vector of texts
#'   - An `LLMMessage` object (all textual components will be embedded)
#'   - A list containing a mix of character strings and `tidyllm_image` objects created with `img()`
#' @param .model The embedding model identifier. For text-only: "voyage-3.5-lite" (default).
#'               For multimodal inputs: "voyage-multimodal-3" is used automatically.
#' @param .timeout Timeout for the API request in seconds (default: 120).
#' @param .dry_run If TRUE, perform a dry run and return the request object without sending.
#' @param .max_tries Maximum retry attempts for requests (default: 3).
#' @param .verbose Should information about current rate limits be printed? (default: FALSE).
#' @param .output_dimension Optional integer to control output vector size (default: NULL, uses model default).
#'
#' @return A tibble with two columns: `input` and `embeddings`.
#'   - The `input` column contains the input texts or image labels
#'   - The `embeddings` column is a list column where each row contains an embedding vector
#'
#' @examples
#' \dontrun{
#' # Text embeddings
#' voyage_embedding("How does photosynthesis work?")
#'
#' # Multimodal embeddings
#' list("A banana", img("banana.jpg"), "Yellow fruit") |>
#'   voyage_embedding()
#' }
#' @export
voyage_embedding <- function(.input,
                             .model = "voyage-4",
                             .output_dimension = NULL,
                             .timeout = 120,
                             .dry_run = FALSE,
                             .max_tries = 3,
                             .verbose   = FALSE) {

  # Get the Voyage API key
  api_key <- Sys.getenv("VOYAGE_API_KEY")
  if ((api_key == "") & .dry_run == FALSE) {
    stop("API key is not set. Please set it with: Sys.setenv(VOYAGE_API_KEY = \"YOUR-KEY-GOES-HERE\")")
  }
  
  # Determine input type
  is_message <- S7::S7_inherits(.input, LLMMessage)
  is_character <- is.character(.input)
  is_list <- is.list(.input) && !is_message
  
  # Validate the inputs
  c(
    "Input .input must be an LLMMessage object, a character vector, or a list with texts and images" =
      is_message | is_character | is_list,
    "Input .model must be a string" = is.character(.model),
    "Input .output_dimension must be NULL or a positive integer" = is.null(.output_dimension) || (is_integer_valued(.output_dimension) && .output_dimension > 0),
    "Input .timeout must be an integer-valued numeric (seconds till timeout)" = is.numeric(.timeout) && .timeout > 0,
    ".dry_run must be logical" = is.logical(.dry_run)
  ) |> validate_inputs()
  
  # Process inputs based on type
  if (is_list) {
    # Handle multimodal list input
    has_images <- any(sapply(.input, function(x) S7::S7_inherits(x, tidyllm_image)))
    
    if (has_images) {
      # Multimodal API
      model_to_use <- if (grepl("multimodal", .model)) .model else "voyage-multimodal-3"
      
      # Prepare the multimodal content: ONE content wrapper per item
      inputs_list <- purrr::map(.input, function(item) {
        
        if (S7::S7_inherits(item, tidyllm_image)) {
          list(content = list(
            list(
              type = "image_base64",
              image_base64 = item@imagebase64
            )
          ))
          
        } else if (is.character(item)) {
          list(content = list(
            list(
              type = "text",
              text = item
            )
          ))
          
        } else {
          stop("Unsupported item type in list input")
        }
        
      })
      
      # Create input labels
      input_labels <- purrr::map_chr(.input, function(item) {
        if (S7::S7_inherits(item, tidyllm_image)) {
          paste0("[IMG] ", item@imagename)
        } else if (is.character(item)) {
          item
        } else {
          "Unknown item"
        }
      })
      
      # Prepare multimodal request body
      request_body <- list(
        inputs = inputs_list,
        model  = model_to_use,
        output_dimension = .output_dimension
      ) |> purrr::compact()
      
      # Build multimodal request
      request <- httr2::request("https://api.voyageai.com/v1/multimodalembeddings") |>
        httr2::req_headers(
          "Content-Type" = "application/json",
          "Authorization" = paste("Bearer", api_key)
        ) |>
        httr2::req_body_json(request_body)
      
      # Return the request object if it's a dry run
      if (.dry_run) {
        return(request)
      }
      
      # Extract embeddings function for multimodal
      extract_embeddings_fn <- function(response_content, error, response_headers) {
        if (error) {
          paste0("API error response - ", response_content$error$message) |> stop()
        }
        
        response_content$data |> 
          purrr::map("embedding") |>
          purrr::map(unlist)
      }
      
      # Perform request
      return(perform_embedding_request(
        .request = request,
        .timeout = .timeout,
        .max_tries = .max_tries,
        .input_texts = input_labels,
        .fn_extract_embeddings = extract_embeddings_fn
      ))
    } else {
      # No images, treat as regular text input
      .input <- unlist(.input)
      is_character <- TRUE
    }
  }
  
  # Handle standard text or LLMMessage input
  if (is_character || is_message) {
    input_texts <- parse_embedding_input(.input)
    
    # Prepare the request body
    request_body <- list(
      model = .model,
      input = input_texts,
      output_dimension = .output_dimension
    ) |> purrr::compact()
    
    # Build the request
    request <- httr2::request("https://api.voyageai.com/v1/embeddings") |>
      httr2::req_headers(
        "Content-Type" = "application/json",
        "Authorization" = paste("Bearer", api_key)
      ) |>
      httr2::req_body_json(request_body)
    
    # Return the request object if it's a dry run
    if (.dry_run) {
      return(request)
    }
    
    extract_embeddings_fn <- function(response_content, error, response_headers) {
      if (error) {
        paste0("API error response - ", response_content$error$message) |> stop()
      }
      
      response_content$data |> 
        purrr::map("embedding") |>
        purrr::map(unlist)
    }
    
    # Perform a standard embedding API request
    perform_embedding_request(
      .request = request,
      .timeout = .timeout,
      .max_tries = .max_tries,
      .input_texts = input_texts, 
      .fn_extract_embeddings = extract_embeddings_fn
    )
  }
}

#' Rerank Documents Using Voyage AI API
#'
#' @param .query A single character string representing the search query.
#' @param .documents A character vector of documents to rerank.
#' @param .model The reranking model identifier (default: "rerank-2").
#' @param .top_k Integer; return only the top-k results (default: NULL, returns all).
#' @param .api_key Character; Voyage API key (default: from environment).
#' @param .timeout Integer; request timeout in seconds (default: 60).
#' @param .max_tries Integer; maximum retries (default: 3).
#'
#' @return A tibble with columns `index`, `document`, and `relevance_score`, sorted by score descending.
#' @export
voyage_rerank <- function(.query,
                          .documents,
                          .model = "rerank-2",
                          .top_k = NULL,
                          .api_key = Sys.getenv("VOYAGE_API_KEY"),
                          .timeout = 60,
                          .max_tries = 3) {
  c(
    "Input .query must be a non-empty string" = is.character(.query) && length(.query) == 1 && nzchar(.query),
    "Input .documents must be a non-empty character vector" = is.character(.documents) && length(.documents) > 0,
    "Input .model must be a non-empty string" = is.character(.model) && nzchar(.model),
    "Input .top_k must be NULL or a positive integer" = is.null(.top_k) || (is_integer_valued(.top_k) && .top_k > 0),
    "Input .api_key must be non-empty" = nzchar(.api_key),
    "Input .timeout must be a positive number" = is.numeric(.timeout) && .timeout > 0
  ) |> validate_inputs()

  request_body <- list(
    query     = .query,
    documents = as.list(.documents),
    model     = .model,
    top_k     = .top_k
  ) |> purrr::compact()

  response <- httr2::request("https://api.voyageai.com/v1/rerank") |>
    httr2::req_headers(
      "Content-Type"  = "application/json",
      "Authorization" = paste("Bearer", .api_key)
    ) |>
    httr2::req_body_json(request_body) |>
    httr2::req_retry(max_tries = .max_tries) |>
    httr2::req_timeout(.timeout) |>
    httr2::req_perform() |>
    httr2::resp_body_json()

  results <- response$data
  tibble::tibble(
    index           = purrr::map_int(results, ~ .x$index),
    document        = purrr::map_chr(results, ~ .documents[.x$index + 1]),
    relevance_score = purrr::map_dbl(results, ~ .x$relevance_score)
  ) |>
    (\(x) x[order(-x$relevance_score), ])()
}


#' Voyage Provider Function
#'
#' The `voyage()` function acts as a provider interface for interacting with the Voyage.ai API 
#' through `tidyllm`'s verbs.
#' It dynamically routes requests to voyage-specific functions. At the moment this is only
#' `voyage_embed()`
#'
#' @param ... Parameters to be passed to the appropriate Voyage-specific function, 
#'   such as model configuration, input text, or API-specific options.
#' @param .called_from An internal argument specifying which action (e.g., 
#'  `embed`) the function is invoked from. 
#'   This argument is automatically managed by the `tidyllm` verbs and should not be modified by the user.
#'
#' @return The result of the requested action, depending on the specific function invoked 
#'  
#'
#' @export
voyage <- create_provider_function(
  .name = "voyage",
  embed = voyage_embedding
)  

---
title: "Get Started"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Get Started}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

# Introduction to tidyllm

**tidyllm** is an R package providing a unified interface for interacting with various large language model APIs. This vignette guides you through the basic setup and usage of **tidyllm**.

## Installation

To install **tidyllm** from CRAN:

```{r, eval=FALSE}
install.packages("tidyllm")
```

Or install the development version from GitHub:

```{r, eval=FALSE}
devtools::install_github("edubruell/tidyllm")
```

### Setting up API Keys

Set API keys as environment variables. The easiest way is to add them to your `.Renviron` file (run `usethis::edit_r_environ()`) and restart R:

```
ANTHROPIC_API_KEY="your-key-here"
OPENAI_API_KEY="your-key-here"
```

Or set them temporarily in your session:

```{r, eval=FALSE}
Sys.setenv(OPENAI_API_KEY = "your-key-here")
```

| Provider | Environment Variable | Where to get a key |
|----------|---------------------|--------------------|
| **Claude** (Anthropic) | `ANTHROPIC_API_KEY` | [Anthropic Console](https://console.anthropic.com/settings/keys) |
| **OpenAI** | `OPENAI_API_KEY` | [OpenAI API Keys](https://platform.openai.com/account/api-keys) |
| **Google Gemini** | `GOOGLE_API_KEY` | [Google AI Studio](https://aistudio.google.com/app/apikey) |
| **Mistral** | `MISTRAL_API_KEY` | [Mistral Console](https://console.mistral.ai/api-keys/) |
| **Groq** | `GROQ_API_KEY` | [Groq Console](https://console.groq.com/playground) |
| **Perplexity** | `PERPLEXITY_API_KEY` | [Perplexity API Settings](https://www.perplexity.ai/settings/api) |
| **DeepSeek** | `DEEPSEEK_API_KEY` | [DeepSeek Platform](https://platform.deepseek.com/api_keys) |
| **Voyage AI** | `VOYAGE_API_KEY` | [Voyage AI Dashboard](https://dashboard.voyageai.com/api-keys) |
| **OpenRouter** | `OPENROUTER_API_KEY` | [OpenRouter Dashboard](https://openrouter.ai/keys) |
| **Azure OpenAI** | `AZURE_OPENAI_API_KEY` | [Azure Portal](https://portal.azure.com/) |

### Running Local Models

**tidyllm** supports local inference via [Ollama](https://ollama.com/) and [llama.cpp](https://github.com/ggml-org/llama.cpp). Both run entirely on your machine; no API key required.

For Ollama, install from [ollama.com](https://ollama.com/) and pull a model:

```{r, eval=FALSE}
ollama_download_model("qwen3.5:4b")
```

For llama.cpp, see the Local Models article on the tidyllm website for a step-by-step setup guide.

## Basic Usage

**tidyllm** is built around a **message-centric** interface. Interactions work through a message history created by `llm_message()` and passed to verbs like `chat()`:

```{r convo1, eval=FALSE}
library(tidyllm)

conversation <- llm_message("What is the capital of France?") |>
  chat(claude())

conversation
```
```{r convo1_out, eval=TRUE, echo=FALSE, message=FALSE}
library(tidyllm)

conversation <- llm_message("What is the capital of France?") |>
  llm_message("The capital of France is Paris.", .role = "assistant")

conversation
```

```{r convo2, eval=FALSE}
# Continue the conversation with a different provider
conversation <- conversation |>
  llm_message("What's a famous landmark in this city?") |>
  chat(openai())

get_reply(conversation)
```
```{r convo2_out, eval=TRUE, echo=FALSE}
c("A famous landmark in Paris is the Eiffel Tower.")
```

All API interactions follow the **verb + provider** pattern:

- **Verbs** define the action: `chat()`, `embed()`, `send_batch()`, `check_batch()`, `fetch_batch()`, `list_batches()`, `list_models()`, `deep_research()`, `check_job()`, `fetch_job()`, `upload_file()`, `list_files()`, `file_info()`, `delete_file()`
- **Providers** specify the API: `openai()`, `claude()`, `gemini()`, `ollama()`, `mistral()`, `groq()`, `perplexity()`, `deepseek()`, `voyage()`, `openrouter()`, `llamacpp()`, `azure_openai()`

Provider-specific functions like `openai_chat()` or `claude_chat()` also work directly and expose the full range of parameters for each API.

### Sending Images and Media

Attach images, PDFs, audio, or video to a message using the `.media` argument of `llm_message()`. Pass a single media object or a list of them:

```{r, echo=FALSE, out.width="70%", fig.alt="A photograph showing lake Garda and the scenery near Torbole, Italy."}
knitr::include_graphics("picture.jpeg")
```

```{r images, eval=FALSE}
# Single image
image_description <- llm_message("Describe this picture. Can you guess where it was taken?",
                                  .media = img("picture.jpeg")) |>
  chat(openai(.model = "gpt-5.4"))

get_reply(image_description)
```
```{r images_out, eval=TRUE, echo=FALSE}
c("The picture shows a beautiful landscape with a lake, mountains, and a town nestled below. The area appears lush and green, with agricultural fields visible. This scenery is reminiscent of northern Italy, particularly around Lake Garda.")
```

Send multiple images in one message by passing a list. The list can also mix different media types; for example, an image alongside a PDF or an audio file alongside a chart. Gemini handles mixed-type lists most flexibly:

```{r, eval=FALSE}
llm_message("Compare these two charts.",
            .media = list(img("chart_a.png"), img("chart_b.png"))) |>
  chat(claude())

# Mixed types in one message
llm_message("Does this figure match the results described in the paper?",
            .media = list(img("figure.png"), pdf_file("paper.pdf", pages = 1:5))) |>
  chat(gemini())
```

For audio and video, use `audio_file()` and `video_file()`. Gemini supports both natively; llama.cpp supports audio with compatible models (Ultravox, Qwen2.5-Omni, Qwen3-Omni); OpenRouter routes audio and video to underlying models that accept them:

```{r, eval=FALSE}
# Audio transcription or analysis
llm_message("Transcribe and summarise this recording.",
            .media = audio_file("interview.mp3")) |>
  chat(gemini())

# Video understanding
llm_message("Describe what happens in this clip.",
            .media = video_file("demo.mp4")) |>
  chat(gemini())
```

### Adding PDFs to Messages

Pass a PDF to `pdf_file()` and attach it with `.media`. By default the file is sent as binary to providers that support native PDF rendering (Claude, Gemini), and the text is automatically extracted as a fallback for others:

```{r pdf, eval=FALSE}
llm_message("Summarize the key points from this document.",
            .media = pdf_file("die_verwandlung.pdf")) |>
  chat(claude())
```
```{r pdf_out, eval=TRUE, echo=FALSE}
llm_message("Summarize the key points from this document.",
            .media = pdf_file("die_verwandlung.pdf")) |>
  llm_message("The story centres on Gregor Samsa, who wakes up transformed into a giant insect. Unable to work, he becomes isolated while his family struggles. Eventually Gregor dies, and his relieved family looks ahead to a better future.",
              .role = "assistant")
```

Restrict to specific pages with the `pages` argument:

```{r, eval=FALSE}
pdf_file("report.pdf", pages = 1:5)
```

For large or frequently reused documents, upload them once to the provider's Files API and reference the handle in any message:

```{r, eval=FALSE}
report <- upload_file(claude(), .path = "annual_report.pdf")

llm_message("What are the key risks mentioned?", .files = report) |>
  chat(claude())

# Manage uploaded files
list_files(claude())
file_info(claude(), report)
delete_file(claude(), report)
```

File management verbs are currently only supported by Claude, Gemini, and OpenAI. Files are uploaded to and stored on the specific provider's servers; a file uploaded to Claude cannot be used with Gemini or OpenAI. Supported file types vary by provider: Gemini accepts a wide range including video, audio, images, PDFs, and various document formats. OpenAI accepts 80+ formats including Office documents (PPTX, DOCX, XLSX), source code, ZIP archives, and more. Claude focuses on PDFs and images. Always check the provider's current documentation for the full list.

### Sending R Outputs to Models

Use `.f` to capture console output from a function and `.capture_plot` to include the last plot:

```{r routputs_base, eval=TRUE, echo=TRUE, message=FALSE, fig.alt="Scatter plot of car weight (x-axis) versus miles per gallon (y-axis) from the mtcars dataset, with a fitted linear regression line showing a negative relationship between weight and fuel efficiency."}
library(tidyverse)

ggplot(mtcars, aes(wt, mpg)) +
  geom_point() +
  geom_smooth(method = "lm", formula = "y ~ x") +
  labs(x = "Weight", y = "Miles per gallon")
```

```{r routputs_base_out, eval=FALSE}
llm_message("Analyze this plot and data summary:",
            .capture_plot = TRUE,
            .f = ~{summary(mtcars)}) |>
  chat(claude())
```
```{r routputs_fake, eval=TRUE, echo=FALSE}
llm_message("Analyze this plot and data summary:",
            .imagefile = "file1568f6c1b4565.png",
            .f = ~{summary(mtcars)}) |>
  llm_message("The scatter plot shows a clear negative correlation between weight and fuel efficiency. Heavier cars consistently achieve lower mpg, with the linear trend confirming this relationship. Variability around the line suggests other factors, engine size and transmission, also play a role.", .role = "assistant")
```

## Getting Replies from the API

`get_reply()` retrieves assistant text from a message history. Use an index to access a specific reply, or omit it to get the last:

```{r eval=FALSE}
conversation <- llm_message("Imagine a German address.") |>
  chat(groq()) |>
  llm_message("Imagine another address.") |>
  chat(claude())

conversation
```
```{r eval=TRUE, echo=FALSE}
conversation <- llm_message("Imagine a German address.") |>
  llm_message("Herr Müller\nMusterstraße 12\n53111 Bonn", .role = "assistant") |>
  llm_message("Imagine another address.") |>
  llm_message("Frau Schmidt\nFichtenweg 78\n42103 Wuppertal", .role = "assistant")
conversation
```

```{r standard_last_reply, eval=TRUE}
conversation |> get_reply(1)   # first reply
conversation |> get_reply()    # last reply (default)
```

Convert the message history (without attachments) to a tibble with `as_tibble()`:

```{r as_tibble, eval=TRUE}
conversation |> as_tibble()
```

`get_metadata()` returns token counts, model names, and API-specific metadata for all assistant replies:

```{r get_metadata, eval=FALSE}
conversation |> get_metadata()
```
```{r get_metadata_out, eval=TRUE, echo=FALSE}
tibble::tibble(
  model             = c("groq-model", "claude-sonnet-4-6-20251001"),
  timestamp         = as.POSIXct(c("2025-11-08 14:25:43", "2025-11-08 14:26:02")),
  prompt_tokens     = c(20L, 80L),
  completion_tokens = c(45L, 40L),
  total_tokens      = c(65L, 120L),
  api_specific      = list(list(), list())
) |> print(width = 60)
```

The `api_specific` list column holds provider-only metadata such as citations (Perplexity), thinking traces (Claude, Gemini, DeepSeek), or grounding information (Gemini).

### Listing Available Models

Use `list_models()` to query which models a provider offers:

```{r list_models, eval=FALSE}
list_models(openai())
list_models(openrouter())   # 300+ models across providers
list_models(ollama())       # models installed locally
```

## Working with Structured Outputs

**tidyllm** supports JSON schema enforcement so models return structured, machine-readable data. Use `tidyllm_schema()` with field helpers to define your schema:

- `field_chr()`: text
- `field_dbl()`: numeric
- `field_lgl()`: boolean
- `field_fct(.levels = ...)`: enumeration
- `field_object(...)`: nested object (supports `.vector = TRUE` for arrays)

```{r schema, eval=FALSE}
person_schema <- tidyllm_schema(
  first_name = field_chr("A male first name"),
  last_name  = field_chr("A common last name"),
  occupation = field_chr("A quirky occupation"),
  address    = field_object(
    "The person's home address",
    street  = field_chr("Street name"),
    number  = field_dbl("House number"),
    city    = field_chr("A large city"),
    zip     = field_dbl("Postal code"),
    country = field_fct("Country", .levels = c("Germany", "France"))
  )
)

profile <- llm_message("Imagine a person profile matching the schema.") |>
  chat(openai(), .json_schema = person_schema)

profile
```
```{r schema_out, eval=TRUE, echo=FALSE}
profile <- llm_message("Imagine a person profile matching the schema.") |>
  llm_message("{\"first_name\":\"Julien\",\"last_name\":\"Martin\",\"occupation\":\"Gondola Repair Specialist\",\"address\":{\"street\":\"Rue de Rivoli\",\"number\":112,\"city\":\"Paris\",\"zip\":75001,\"country\":\"France\"}}",
              .role = "assistant")
profile@message_history[[3]]$json <- TRUE
profile
```

Extract the parsed result as an R list with `get_reply_data()`:

```{r get_reply_data}
profile |> get_reply_data() |> str()
```

Set `.vector = TRUE` on `field_object()` to get outputs that unpack directly into a data frame:

```{r fobj, eval=FALSE}
llm_message("Imagine five people.") |>
  chat(openai(),
       .json_schema = tidyllm_schema(
         person = field_object(
           first_name = field_chr(),
           last_name  = field_chr(),
           occupation = field_chr(),
           .vector    = TRUE
         )
       )) |>
  get_reply_data() |>
  bind_rows()
```
```{r fobjout, echo=FALSE, eval=TRUE}
data.frame(
  first_name = c("Alice", "Robert", "Maria", "Liam", "Sophia"),
  last_name  = c("Johnson", "Anderson", "Gonzalez", "O'Connor", "Lee"),
  occupation = c("Software Developer", "Graphic Designer", "Data Scientist",
                 "Mechanical Engineer", "Content Writer")
)
```

If the **ellmer** package is installed, you can use ellmer type objects (`ellmer::type_string()`, `ellmer::type_object()`, etc.) directly as field definitions in `tidyllm_schema()` or as the `.json_schema` argument.

## API Parameters

Common arguments like `.model`, `.temperature`, and `.json_schema` can be set directly in verbs like `chat()`. Provider-specific arguments go in the provider function:

```{r temperature, eval=FALSE}
# Common args in the verb
llm_message("Write a haiku about tidyllm.") |>
  chat(ollama(), .temperature = 0)

# Provider-specific args in the provider
llm_message("Hello") |>
  chat(ollama(.ollama_server = "http://my-server:11434"), .temperature = 0)
```

When an argument appears in both `chat()` and the provider function, `chat()` takes precedence. If a common argument is not supported by the chosen provider, `chat()` raises an error rather than silently ignoring it.

## Tool Use

Define R functions that the model can call during a conversation. Wrap them with `tidyllm_tool()`:

```{r, eval=FALSE}
get_current_time <- function(tz, format = "%Y-%m-%d %H:%M:%S") {
  format(Sys.time(), tz = tz, format = format, usetz = TRUE)
}

time_tool <- tidyllm_tool(
  .f          = get_current_time,
  .description = "Returns the current time in a specified timezone.",
  tz     = field_chr("The timezone identifier, e.g. 'Europe/Berlin'."),
  format = field_chr("strftime format string. Default: '%Y-%m-%d %H:%M:%S'.")
)

llm_message("What time is it in Stuttgart right now?") |>
  chat(openai(), .tools = time_tool)
```
```{r, eval=TRUE, echo=FALSE}
llm_message("What time is it in Stuttgart right now?") |>
  llm_message("The current time in Stuttgart (Europe/Berlin) is 2025-03-03 09:51:22 CET.",
              .role = "assistant")
```

When a tool is provided, the model can request its execution, tidyllm runs it in your R session, and the result is fed back automatically. Multi-turn tool loops (where the model makes several tool calls) are handled transparently.

If you use **ellmer**, convert existing ellmer tool definitions with `ellmer_tool()`:

```{r, eval=FALSE}
library(ellmer)
btw_tool <- ellmer_tool(btw::btw_tool_files_list_files)
llm_message("List files in the R/ folder.") |>
  chat(claude(), .tools = btw_tool)
```

For a detailed guide, see the Tool Use article on the tidyllm website.

## Embeddings

Generate vector representations of text with `embed()`. The output is a tibble with one row per input:

```{r embed, eval=FALSE}
c("What is the meaning of life?",
  "How much wood would a woodchuck chuck?",
  "How does the brain work?") |>
  embed(ollama())
```
```{r embed_output, eval=TRUE, echo=FALSE}
tibble::tibble(
  input      = c("What is the meaning of life?",
                 "How much wood would a woodchuck chuck?",
                 "How does the brain work?"),
  embeddings = purrr::map(1:3, ~runif(min = -1, max = 1, n = 384))
)
```

Voyage AI supports **multimodal embeddings**, meaning text and images share the same vector space:

```{r, eval=FALSE}
list("tidyllm logo", img("docs/logo.png")) |>
  embed(voyage())
```

`voyage_rerank()` re-orders documents by relevance to a query:

```{r, eval=FALSE}
voyage_rerank(
  "best R package for LLMs",
  c("tidyllm", "ellmer", "httr2", "rvest")
)
```

## Batch Requests

Batch APIs (Claude, OpenAI, Mistral, Groq, Gemini) process large numbers of requests overnight at roughly half the cost of standard calls. Use `send_batch()` to submit, `check_batch()` to poll status, and `fetch_batch()` to retrieve results:

```{r, eval=FALSE}
# Submit a batch and save the job handle to disk
glue::glue("Write a poem about {x}", x = c("cats", "dogs", "hamsters")) |>
  purrr::map(llm_message) |>
  send_batch(claude()) |>
  saveRDS("claude_batch.rds")
```

```{r, eval=FALSE}
# Check status
readRDS("claude_batch.rds") |>
  check_batch(claude())
```
```{r, echo=FALSE}
tibble::tribble(
  ~batch_id,                          ~status, ~created_at,                      ~expires_at,                      ~req_succeeded, ~req_errored, ~req_expired, ~req_canceled,
  "msgbatch_02A1B2C3D4E5F6G7H8I9J",   "ended", as.POSIXct("2025-11-01 10:30:00"), as.POSIXct("2025-11-02 10:30:00"), 3L,             0L,           0L,           0L
)
```

```{r, eval=FALSE}
# Fetch completed results
conversations <- readRDS("claude_batch.rds") |>
  fetch_batch(claude())

poems <- purrr::map_chr(conversations, get_reply)
```

`check_job()` and `fetch_job()` are type-dispatching aliases; they work on both batch objects and Perplexity research jobs (see `deep_research()` below).

## Deep Research

For long-horizon research tasks, `deep_research()` runs an extended web search and synthesis. Perplexity's `sonar-deep-research` model is currently supported:

```{r, eval=FALSE}
# Blocking: waits for completion, returns an LLMMessage
result <- llm_message("Compare Rust and Go for systems programming in 2025.") |>
  deep_research(perplexity())

get_reply(result)
get_metadata(result)$api_specific[[1]]$citations
```

```{r, eval=FALSE}
# Background mode: returns a job handle immediately
job <- llm_message("Summarise recent EU AI Act developments.") |>
  deep_research(perplexity(), .background = TRUE)

check_job(job)           # poll status
result <- fetch_job(job) # retrieve when complete
```

## Streaming

Stream reply tokens to the console as they are generated with `.stream = TRUE`:

```{r, eval=FALSE}
llm_message("Write a short poem about R.") |>
  chat(claude(), .stream = TRUE)
```

Streaming is useful for monitoring long responses interactively. For production data-analysis workflows the non-streaming mode is preferred because it provides complete metadata and is more reliable.

## Choosing the Right Provider

| Provider | Strengths and tidyllm-specific features |
|----------|-----------------------------------------|
| `openai()` | Top benchmark performance across coding, math, and reasoning; `o3`/`o4` reasoning models; built-in web search via `openai_websearch()`; Files API |
| `claude()` | Best-in-class coding (SWE-bench leader); `.thinking = TRUE` for extended reasoning; Files API for document workflows; batch |
| `gemini()` | 1M-token context window; native image, audio, and video via `.media`; search grounding; `.thinking_budget` for reasoning; Files API; batch |
| `mistral()` | EU-hosted, GDPR-friendly; Magistral reasoning models; Voxtral audio models (`.media = audio_file()`); embeddings; batch |
| `groq()` | Fastest available inference (300-1200 tokens/s); `groq_transcribe()` for Whisper audio; `.json_schema`; batch |
| `perplexity()` | Real-time web search with citations in metadata; `deep_research()` for long-horizon research; search domain filter |
| `deepseek()` | Top math and reasoning benchmarks at very low cost; `.thinking = TRUE` auto-switches to `deepseek-reasoner` |
| `voyage()` | State-of-the-art retrieval embeddings; `voyage_rerank()`; multimodal embeddings (text + images); `.output_dimension` |
| `openrouter()` | 500+ models via one API key; audio and video support on Gemini, Qwen 3.6, and other multimodal routes; automatic fallback routing |
| `ollama()` | Local models with full data privacy; no API costs; `ollama_download_model()` for model management |
| `llamacpp()` | Often the most performant local inference stack; audio input with Ultravox, Qwen2.5-Omni, Qwen3-Omni, and Gemma 4 models; BNF grammar constraints; logprobs; `llamacpp_rerank()` |
| `azure_openai()` | Enterprise Azure deployments of OpenAI models; batch support |

For getting started with local models (Ollama and llama.cpp), see the Local Models article on the tidyllm website.

## Setting a Default Provider

Avoid specifying a provider on every call by setting options:

```{r, eval=FALSE}
options(tidyllm_chat_default   = openai(.model = "gpt-5.4"))
options(tidyllm_embed_default  = ollama())
options(tidyllm_sbatch_default = claude(.temperature = 0))
options(tidyllm_cbatch_default = claude())
options(tidyllm_fbatch_default = claude())
options(tidyllm_lbatch_default = claude())

# Now the provider argument can be omitted
llm_message("Hello!") |> chat()

c("text one", "text two") |> embed()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tools.R
\name{tidyllm_tool}
\alias{tidyllm_tool}
\title{Create a Tool Definition for tidyllm}
\usage{
tidyllm_tool(.f, .description = character(0), ...)
}
\arguments{
\item{.f}{The function to wrap as a tool}

\item{.description}{Character string describing what the tool does}

\item{...}{Named arguments providing schema definitions for each function parameter using tidyllm_fields}
}
\value{
A \code{TOOL} class object that can be used with tidyllm \code{chat()} functions
}
\description{
Creates a tool definition for use with Language Model API calls that support function calling.
This function wraps an existing R function with schema information for LLM interaction.
}
\details{
Each parameter schema in \code{...} should correspond to a parameter in the wrapped function.
All required function parameters must have corresponding schema definitions.
}
\examples{
get_weather <- function(location){}
weather_tool <- tidyllm_tool(
  get_weather,
  "Get the current weather in a given location",
  location = field_chr("The city and state, e.g., San Francisco, CA")
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/message_retrieval.R
\name{get_metadata}
\alias{get_metadata}
\alias{last_metadata}
\title{Retrieve Metadata from Assistant Replies}
\usage{
get_metadata(.llm, .index = NULL)

last_metadata(.llm)
}
\arguments{
\item{.llm}{An \code{LLMMessage} object containing the message history.}

\item{.index}{A positive integer specifying which assistant reply's metadata to extract.
If \code{NULL} (default), metadata for all replies is returned.}
}
\value{
A tibble containing metadata for the specified assistant reply or all replies.
}
\description{
Retrieves metadata from assistant replies within an \code{LLMMessage} object.
It returns the metadata as a tibble.
}
\details{
Metadata columns may include:
\itemize{
\item \code{model}: The model used for generating the reply.
\item \code{timestamp}: The time when the reply was generated.
\item \code{prompt_tokens}: The number of tokens in the input prompt.
\item \code{completion_tokens}: The number of tokens in the assistant's reply.
\item \code{total_tokens}: The total number of tokens (prompt + completion).
\item \code{api_specific}: A list column with API-specific metadata.
}

For convenience, \code{\link[=last_metadata]{last_metadata()}} is provided to retrieve the metadata for the last message.
}
\seealso{
\code{\link[=last_metadata]{last_metadata()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/message_retrieval.R
\name{get_user_message}
\alias{get_user_message}
\alias{last_user_message}
\title{Retrieve a User Message by Index}
\usage{
get_user_message(.llm, .index = NULL)

last_user_message(.llm)
}
\arguments{
\item{.llm}{An \code{LLMMessage} object.}

\item{.index}{A positive integer indicating which user message to retrieve. Defaults to \code{NULL}, which retrieves the last message.}
}
\value{
Returns the content of the user's message at the specified index. If no messages are found, returns \code{NA_character_}.
}
\description{
Extracts the content of a user's message from an \code{LLMMessage} object at a specific index.
}
\details{
For convenience, \code{\link[=last_user_message]{last_user_message()}} is provided as a wrapper to retrieve the
latest user message without specifying an index.
}
\seealso{
\code{\link[=last_user_message]{last_user_message()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_perplexity.R
\name{perplexity_chat}
\alias{perplexity_chat}
\title{Send LLM Messages to the Perplexity Chat API}
\usage{
perplexity_chat(
  .llm,
  .model = "sonar",
  .max_tokens = 1024,
  .temperature = NULL,
  .top_p = NULL,
  .frequency_penalty = NULL,
  .presence_penalty = NULL,
  .stop = NULL,
  .search_domain_filter = NULL,
  .search_language_filter = NULL,
  .language_preference = NULL,
  .return_images = FALSE,
  .image_domain_filter = NULL,
  .image_format_filter = NULL,
  .search_recency_filter = NULL,
  .search_mode = "web",
  .search_after_date_filter = NULL,
  .search_before_date_filter = NULL,
  .last_updated_after_filter = NULL,
  .last_updated_before_filter = NULL,
  .disable_search = FALSE,
  .enable_search_classifier = FALSE,
  .reasoning_effort = NULL,
  .return_related_questions = FALSE,
  .user_location = NULL,
  .search_context_size = NULL,
  .search_type = NULL,
  .stream_mode = NULL,
  .json_schema = NULL,
  .top_k = NULL,
  .web_search_options = NULL,
  .api_url = "https://api.perplexity.ai/",
  .timeout = 60,
  .stream = FALSE,
  .verbose = FALSE,
  .max_tries = 3,
  .dry_run = FALSE
)
}
\arguments{
\item{.llm}{An \code{LLMMessage} object containing the conversation history.}

\item{.model}{Model name to use (default: "sonar").}

\item{.max_tokens}{Max completion tokens (default: 1024).}

\item{.temperature}{Controls response randomness (0 < x < 2).}

\item{.top_p}{Nucleus sampling threshold (0 < x < 1).}

\item{.frequency_penalty}{Number > 0. Penalizes frequent tokens.}

\item{.presence_penalty}{Numeric between -2 and 2. Penalizes present tokens.}

\item{.stop}{Stop sequence(s), string or character vector/list.}

\item{.search_domain_filter}{Character vector of domains to allowlist/denylist (max 10; prefix with "-" to denylist).}

\item{.search_language_filter}{ISO 639-1 language code to restrict search results (e.g. "en", "de", "fr").}

\item{.language_preference}{ISO 639-1 code for preferred response language (e.g. "en").}

\item{.return_images}{Logical; if TRUE, returns images from search.}

\item{.image_domain_filter}{Character vector of domains to restrict image results to.}

\item{.image_format_filter}{Character vector of image formats to include (e.g. c("png", "jpg")).}

\item{.search_recency_filter}{Restrict search to recent results: "hour", "day", "week", "month", or "year".}

\item{.search_mode}{Search index to use: "web" (default), "academic", or "sec".}

\item{.search_after_date_filter}{Only include content published after this date (MM/DD/YYYY).}

\item{.search_before_date_filter}{Only include content published before this date (MM/DD/YYYY).}

\item{.last_updated_after_filter}{Only include content last updated after this date (MM/DD/YYYY).}

\item{.last_updated_before_filter}{Only include content last updated before this date (MM/DD/YYYY).}

\item{.disable_search}{Logical; if TRUE, disables web search entirely (default: FALSE).}

\item{.enable_search_classifier}{Logical; if TRUE, lets the model decide whether to search (default: FALSE).}

\item{.reasoning_effort}{Reasoning level: "low", "medium", or "high".}

\item{.return_related_questions}{Logical; if TRUE, returns related questions.}

\item{.user_location}{Named list for geographic search personalisation. Accepted fields:
\code{country} (ISO 3166-1 alpha-2), \code{city}, \code{region}, \code{latitude}, \code{longitude}.
Example: \code{list(country = "DE", city = "Berlin")}.}

\item{.search_context_size}{Amount of search context to include: "low", "medium" (default), or "high".}

\item{.search_type}{Search quality preference inside \code{web_search_options}: "fast", "pro", or "auto".}

\item{.stream_mode}{Response format: "full" (default) or "concise".}

\item{.json_schema}{A tidyllm schema created with \code{tidyllm_schema()} for structured JSON output (optional).}

\item{.top_k}{Top-k token sampling (integer, 0 disables).}

\item{.web_search_options}{Named list with raw web_search_options overrides. Values set here take
precedence over the dedicated parameters above.}

\item{.api_url}{API endpoint (default: "https://api.perplexity.ai/").}

\item{.timeout}{Timeout in seconds (default: 60).}

\item{.stream}{If TRUE, streams response.}

\item{.verbose}{If TRUE, prints additional info.}

\item{.max_tries}{Max request retries (default: 3).}

\item{.dry_run}{If TRUE, returns constructed request instead of sending.}
}
\value{
An updated \code{LLMMessage} object with the assistant's reply and metadata, including citations and search_results.
}
\description{
Sends a chat message history to the Perplexity Chat API, supporting all documented API parameters.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_llamacpp.R
\name{llamacpp_delete_model}
\alias{llamacpp_delete_model}
\title{Delete a Local GGUF Model File}
\usage{
llamacpp_delete_model(.path)
}
\arguments{
\item{.path}{Full path to the \code{.gguf} file to delete.}
}
\value{
Invisibly returns \code{TRUE} on success.
}
\description{
Deletes a GGUF file from disk. Issues a warning confirming the deletion.
Mirrors the \code{ollama_delete_model()} pattern.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_gemini.R
\name{fetch_gemini_batch}
\alias{fetch_gemini_batch}
\title{Fetch Results for a Gemini Batch}
\usage{
fetch_gemini_batch(
  .llms,
  .batch_name = NULL,
  .timeout = 60,
  .max_tries = 3,
  .dry_run = FALSE
)
}
\arguments{
\item{.llms}{List of \code{LLMMessage} objects (as from \code{send_gemini_batch()}), must have a \code{batch_id} attribute if \code{.batch_name} is not given.}

\item{.batch_name}{(Optional) Character; batch operation name (e.g. "batches/xyz123"). If not provided, is taken from \code{attr(.llms, "batch_id")}.}

\item{.timeout}{Integer; request timeout in seconds (default: 60).}

\item{.max_tries}{Integer; maximum retry attempts (default: 3).}

\item{.dry_run}{Logical; if \code{TRUE}, returns the GET request object (default: FALSE).}
}
\value{
A list of updated LLMMessage objects with the assistant response appended to each, in the same order.
}
\description{
Retrieves the results of a completed Gemini batch and updates
the provided list of LLMMessage objects with the assistant's responses,
matching by original list order.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tools.R
\name{ellmer_tool}
\alias{ellmer_tool}
\title{Convert an ellmer Tool to a tidyllm TOOL}
\usage{
ellmer_tool(.ellmer_tool)
}
\arguments{
\item{.ellmer_tool}{An ellmer \code{ToolDef} object created via \code{ellmer::tool()},
or a builtin tool like \code{ellmer::claude_tool_web_search()}}
}
\value{
A \code{TOOL} class object that can be used with tidyllm \code{chat()} functions
}
\description{
Converts an ellmer \code{ToolDef} or \code{ToolBuiltIn} object to a tidyllm \code{TOOL} object,
allowing seamless integration of ellmer-defined tools and builtin provider tools
with tidyllm workflows.
}
\details{
This function supports two types of ellmer tools:

\strong{Custom ToolDef objects}: Extracts the function, description, and argument
schemas from an ellmer tool and converts them to tidyllm's internal representation.
Ellmer type objects are automatically converted to tidyllm field descriptors.

\strong{Builtin ToolBuiltIn objects}: Converts provider-native builtin tools (like
Claude's web search) to tidyllm format. For builtin tools, the tool definition
is passed through as-is to the provider's API, which handles the tool execution
natively.
}
\examples{
\dontrun{
library(ellmer)

# Example 1: Custom tool
tool_rnorm <- ellmer::tool(
  rnorm,
  description = "Draw numbers from a random normal distribution",
  arguments = list(
    n = ellmer::type_integer("The number of observations"),
    mean = ellmer::type_number("The mean value"),
    sd = ellmer::type_number("The standard deviation")
  )
)

tidyllm_tool_rnorm <- ellmer_tool(tool_rnorm)

llm_message("Generate 100 random numbers") |> 
  chat(openai(), .tools = tidyllm_tool_rnorm)

# Example 2: Builtin tool
web_search <- ellmer_tool(ellmer::claude_tool_web_search())
llm_message("What are the latest AI developments?") |>
  chat(claude(), .tools = web_search)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_openai.R
\name{cancel_openai_batch}
\alias{cancel_openai_batch}
\title{Cancel an In-Progress OpenAI Batch}
\usage{
cancel_openai_batch(.batch_id, .dry_run = FALSE, .max_tries = 3, .timeout = 60)
}
\arguments{
\item{.batch_id}{Character; the unique identifier for the batch to cancel.}

\item{.dry_run}{Logical; if \code{TRUE}, returns the constructed request without executing it (default: \code{FALSE}).}

\item{.max_tries}{Integer; maximum number of retries if the request fails (default: \code{3}).}

\item{.timeout}{Integer; request timeout in seconds (default: \code{60}).}
}
\value{
A list containing the response from the OpenAI API about the cancellation status.
}
\description{
This function cancels an in-progress batch created through the OpenAI API.
The batch will be moved to a "cancelling" state and, eventually, "cancelled."
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/media.R
\name{pdf_file}
\alias{pdf_file}
\title{Create a PDF Object}
\usage{
pdf_file(.path, pages = NULL, .text_extract = FALSE, engine = NULL)
}
\arguments{
\item{.path}{The path to the PDF file on disk.}

\item{pages}{An integer vector of page numbers to include (e.g. \code{1:3}). NULL uses all pages.}

\item{.text_extract}{Logical; if TRUE always use pdftools text extraction regardless of provider
(equivalent to the old \code{.pdf} parameter behaviour). Default FALSE uses binary-first.}

\item{engine}{Optional engine name (e.g. \code{"mistral-ocr"}); a warning is issued if the provider
does not support binary PDF and the engine choice cannot be honoured.}
}
\value{
A \code{tidyllm_pdf} object.
}
\description{
Stores a reference to a local PDF file along with a pdftools text extraction.
Binary encoding is deferred to format time; providers that support binary PDF
encode on demand, others use the stored text fallback.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_openrouter.R
\name{openrouter_embedding}
\alias{openrouter_embedding}
\title{Generate Embeddings Using the OpenRouter API}
\usage{
openrouter_embedding(
  .input,
  .model = "openai/text-embedding-3-small",
  .timeout = 120,
  .dry_run = FALSE,
  .max_tries = 3
)
}
\arguments{
\item{.input}{An \code{LLMMessage} object or a character vector of texts to embed.}

\item{.model}{The embedding model ID (default: \code{"openai/text-embedding-3-small"}).}

\item{.timeout}{Request timeout in seconds (default: 120).}

\item{.dry_run}{If TRUE, returns the request object without executing it (default: FALSE).}

\item{.max_tries}{Maximum retries (default: 3).}
}
\value{
A tibble with columns \code{input} (text) and \code{embeddings} (list of numeric vectors).
}
\description{
Sends text to an embedding model accessible via OpenRouter and returns
embedding vectors. Note that embedding models are not listed in
\code{list_models(openrouter())} — specify the model ID directly.
Known supported models include \code{"openai/text-embedding-3-small"},
\code{"openai/text-embedding-3-large"}, and \code{"mistralai/mistral-embed"}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_gemini.R
\name{check_gemini_batch}
\alias{check_gemini_batch}
\title{Check the Status of a Gemini Batch Operation}
\usage{
check_gemini_batch(
  .llms = NULL,
  .batch_id = NULL,
  .timeout = 60,
  .max_tries = 3,
  .dry_run = FALSE
)
}
\arguments{
\item{.llms}{(Optional) List of LLMMessage objects with a \code{"batch_id"} attribute (as returned by \code{send_gemini_batch()}).}

\item{.batch_id}{(Optional) Character string: full batch operation name, e.g. \code{"batches/xyz123"}.
If both \code{.llms} and \code{.batch_id} are provided, \code{.batch_id} is used.}

\item{.timeout}{Integer. Request timeout in seconds. Default: 60.}

\item{.max_tries}{Integer. Maximum retry attempts. Default: 3.}

\item{.dry_run}{Logical. If TRUE, return the request object instead of making the request (for debugging). Default: FALSE.}
}
\value{
A tibble with the operation's metadata, including name, state, creation time, completion time, and done status.
}
\description{
Retrieves processing status and metadata for a Gemini batch operation.
}
\details{
You can supply either the \code{.batch_id} string (e.g. \code{"batches/xyz..."}) \strong{or}
a list of LLMMessage objects (\code{.llms}) with a \code{"batch_id"} attribute as returned by \code{send_gemini_batch()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_openai.R
\name{openai_fetch_research}
\alias{openai_fetch_research}
\title{Fetch Results from a Completed OpenAI Deep Research Job}
\usage{
openai_fetch_research(.job, .max_tries = 3)
}
\arguments{
\item{.job}{A \code{tidyllm_research_job} object. If not yet completed, polls once via
\code{openai_check_research()} and errors if still incomplete.}

\item{.max_tries}{Maximum retries per HTTP request (default: \code{3}).}
}
\value{
An updated \code{LLMMessage} with the research reply appended.
}
\description{
Extracts the assistant reply from a completed \code{tidyllm_research_job} returned
by \code{openai_deep_research(.background = TRUE)}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_gemini.R
\name{gemini_upload_file}
\alias{gemini_upload_file}
\title{Upload a File to Gemini API}
\usage{
gemini_upload_file(.file_path)
}
\arguments{
\item{.file_path}{The local file path of the file to upload.}
}
\value{
A tibble containing metadata about the uploaded file, including its name, URI, and MIME type.
}
\description{
Uploads a file to the Gemini API and returns its metadata as a tibble.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_groq.R
\name{list_groq_batches}
\alias{list_groq_batches}
\title{List Groq Batch Requests}
\usage{
list_groq_batches(
  .api_url = "https://api.groq.com/",
  .limit = 20,
  .max_tries = 3,
  .timeout = 60
)
}
\arguments{
\item{.api_url}{Base URL for the Groq API (default: "https://api.groq.com/").}

\item{.limit}{Maximum number of batches to retrieve (default: 20).}

\item{.max_tries}{Maximum retry attempts for requests (default: 3).}

\item{.timeout}{Request timeout in seconds (default: 60).}
}
\value{
A tibble with batch details including batch ID, status, creation time, and request counts.
}
\description{
Retrieves batch request details from the Groq API.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_chat_completions.R
\name{chat_completions_chat}
\alias{chat_completions_chat}
\alias{chat_completions}
\title{Chat with any OpenAI-Compatible API Endpoint}
\usage{
chat_completions_chat(
  .llm,
  .api_url,
  .api_key_env_var = NULL,
  .model = "default",
  ...
)

chat_completions(..., .called_from = NULL)
}
\arguments{
\item{.llm}{An \code{LLMMessage} object containing the conversation history.}

\item{.api_url}{Base URL for the API endpoint (required).}

\item{.api_key_env_var}{Name of the environment variable holding the API key.}

\item{.model}{The model identifier to use (default: \code{"default"}).}

\item{...}{Additional parameters passed to the underlying chat function.}

\item{.called_from}{Internal routing argument; do not set manually.}
}
\value{
An updated \code{LLMMessage} object with the assistant's response appended.
}
\description{
Provider function for interacting with any server that implements the OpenAI
Chat Completions wire format (vLLM, LiteLLM, Together, Anyscale, etc.).
Supports the \code{chat()} verb only.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_groq.R
\name{send_groq_batch}
\alias{send_groq_batch}
\title{Send a Batch of Messages to the Groq API}
\usage{
send_groq_batch(
  .llms,
  .model = "openai/gpt-oss-120b",
  .max_tokens = 1024,
  .temperature = NULL,
  .top_p = NULL,
  .frequency_penalty = NULL,
  .presence_penalty = NULL,
  .stop = NULL,
  .seed = NULL,
  .api_url = "https://api.groq.com/",
  .json_schema = NULL,
  .completion_window = "24h",
  .verbose = FALSE,
  .dry_run = FALSE,
  .overwrite = FALSE,
  .max_tries = 3,
  .timeout = 60,
  .id_prefix = "tidyllm_groq_req_"
)
}
\arguments{
\item{.llms}{A list of LLMMessage objects containing conversation histories.}

\item{.model}{Character string specifying the model to use (default: "deepseek-r1-distill-llama-70b").}

\item{.max_tokens}{Integer specifying the maximum tokens per response (default: 1024).}

\item{.temperature}{Numeric between 0 and 2 controlling response randomness.}

\item{.top_p}{Numeric between 0 and 1 for nucleus sampling.}

\item{.frequency_penalty}{Number between -2.0 and 2.0 to penalize repetition.}

\item{.presence_penalty}{Number between -2.0 and 2.0 to encourage new topics.}

\item{.stop}{One or more sequences where the API will stop generating further tokens.}

\item{.seed}{An integer for deterministic sampling.}

\item{.api_url}{Base URL for the Groq API (default: "https://api.groq.com/").}

\item{.json_schema}{A list or tidyllm schema created with \code{tidyllm_schema()} for structured JSON output (optional).}

\item{.completion_window}{Character string for the batch completion window (default: "24h").}

\item{.verbose}{Logical; if TRUE, prints a message with the batch ID (default: FALSE).}

\item{.dry_run}{Logical; if TRUE, returns the prepared request objects without executing (default: FALSE).}

\item{.overwrite}{Logical; if TRUE, allows overwriting an existing batch ID (default: FALSE).}

\item{.max_tries}{Maximum number of retries to perform the request.}

\item{.timeout}{Integer specifying the request timeout in seconds (default: 60).}

\item{.id_prefix}{Character string to specify a prefix for generating custom IDs when names in \code{.llms} are missing.}
}
\value{
An updated and named list of \code{.llms} with identifiers that align with batch responses, including a \code{batch_id} attribute.
}
\description{
This function creates and submits a batch of messages to the Groq API for asynchronous processing.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_mistral.R
\name{send_mistral_batch}
\alias{send_mistral_batch}
\title{Send a Batch of Requests to the Mistral API}
\usage{
send_mistral_batch(
  .llms,
  .model = "mistral-small-latest",
  .endpoint = "/v1/chat/completions",
  .metadata = NULL,
  .temperature = 0.7,
  .top_p = 1,
  .max_tokens = 1024,
  .min_tokens = NULL,
  .frequency_penalty = NULL,
  .logit_bias = NULL,
  .presence_penalty = NULL,
  .seed = NULL,
  .stop = NULL,
  .safe_prompt = FALSE,
  .json_schema = NULL,
  .dry_run = FALSE,
  .overwrite = FALSE,
  .max_tries = 3,
  .timeout = 60,
  .id_prefix = "tidyllm_mistral_req_"
)
}
\arguments{
\item{.llms}{A list of LLMMessage objects containing conversation histories.}

\item{.model}{The Mistral model version (default: "mistral-small-latest").}

\item{.endpoint}{The API endpoint (default: "/v1/chat/completions").}

\item{.metadata}{Optional metadata for the batch.}

\item{.temperature}{Sampling temperature to use, between \code{0.0} and \code{1.5} (default: \code{0.7}).}

\item{.top_p}{Nucleus sampling parameter, between \code{0.0} and \code{1.0} (default: \code{1}).}

\item{.max_tokens}{The maximum number of tokens to generate in the completion (default: \code{1024}).}

\item{.min_tokens}{The minimum number of tokens to generate (optional).}

\item{.frequency_penalty}{Numeric value (or NULL) for frequency penalty.}

\item{.logit_bias}{A named list modifying the likelihood of specific tokens (or NULL).}

\item{.presence_penalty}{Numeric value (or NULL) for presence penalty.}

\item{.seed}{Random seed for deterministic outputs (optional).}

\item{.stop}{Sequence(s) at which to stop generation (optional).}

\item{.safe_prompt}{Logical; if TRUE, injects a safety prompt (default: FALSE).}

\item{.json_schema}{A JSON schema object for structured output (optional).}

\item{.dry_run}{Logical; if TRUE, returns the prepared request without executing it (default: FALSE).}

\item{.overwrite}{Logical; if TRUE, allows overwriting existing custom IDs (default: FALSE).}

\item{.max_tries}{Maximum retry attempts for requests (default: 3).}

\item{.timeout}{Request timeout in seconds (default: 60).}

\item{.id_prefix}{Prefix for generating custom IDs (default: "tidyllm_mistral_req_").}
}
\value{
The prepared LLMMessage list with a batch_id attribute.
}
\description{
This function creates and submits a batch of messages to the Mistral API for asynchronous processing.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llm_verbs.R
\name{list_files}
\alias{list_files}
\title{List Files Stored on a Provider}
\usage{
list_files(.provider, ...)
}
\arguments{
\item{.provider}{A provider function call.}

\item{...}{Additional provider-specific arguments.}
}
\value{
A tibble of file metadata.
}
\description{
List Files Stored on a Provider
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_perplexity.R
\name{perplexity_deep_research}
\alias{perplexity_deep_research}
\title{Submit a Deep Research Request to Perplexity}
\usage{
perplexity_deep_research(
  .llm,
  .background = FALSE,
  .reasoning_effort = "medium",
  .search_context_size = "medium",
  .search_domain_filter = NULL,
  .search_language_filter = NULL,
  .language_preference = NULL,
  .search_recency_filter = NULL,
  .search_mode = NULL,
  .search_after_date_filter = NULL,
  .search_before_date_filter = NULL,
  .last_updated_after_filter = NULL,
  .last_updated_before_filter = NULL,
  .user_location = NULL,
  .json_schema = NULL,
  .idempotency_key = NULL,
  .api_key = Sys.getenv("PERPLEXITY_API_KEY"),
  .timeout = 300,
  .max_tries = 3
)
}
\arguments{
\item{.llm}{An \code{LLMMessage} object containing the research question.}

\item{.background}{Logical; if TRUE, returns a \code{tidyllm_research_job} immediately without waiting (default: FALSE).}

\item{.reasoning_effort}{Reasoning level: "low", "medium" (default), or "high".}

\item{.search_context_size}{Amount of search context: "low", "medium" (default), or "high".}

\item{.search_domain_filter}{Character vector of domains to allowlist/denylist (max 10; prefix with "-" to denylist).}

\item{.search_language_filter}{ISO 639-1 language code to restrict search results (e.g. "en", "de").}

\item{.language_preference}{ISO 639-1 code for preferred response language.}

\item{.search_recency_filter}{Restrict search to recent results: "hour", "day", "week", "month", or "year".}

\item{.search_mode}{Search index to use: "web" (default), "academic", or "sec".}

\item{.search_after_date_filter}{Only include content published after this date (MM/DD/YYYY).}

\item{.search_before_date_filter}{Only include content published before this date (MM/DD/YYYY).}

\item{.last_updated_after_filter}{Only include content last updated after this date (MM/DD/YYYY).}

\item{.last_updated_before_filter}{Only include content last updated before this date (MM/DD/YYYY).}

\item{.user_location}{Named list for geographic search personalisation (fields: country, city, region, latitude, longitude).}

\item{.json_schema}{A tidyllm schema created with \code{tidyllm_schema()} for structured JSON output (optional).}

\item{.idempotency_key}{Optional string; unique key to prevent duplicate submissions.}

\item{.api_key}{Character; Perplexity API key (default: from environment variable).}

\item{.timeout}{Integer; request timeout in seconds for blocking polling (default: 300).}

\item{.max_tries}{Integer; maximum retries (default: 3).}
}
\value{
If \code{.background = FALSE}, an updated \code{LLMMessage} with the research reply.
If \code{.background = TRUE}, a \code{tidyllm_research_job} object.
}
\description{
Submit a Deep Research Request to Perplexity
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llm_verbs.R
\name{delete_file}
\alias{delete_file}
\title{Delete a File from a Provider's File Store}
\usage{
delete_file(.provider, .file_id, ...)
}
\arguments{
\item{.provider}{A provider function call.}

\item{.file_id}{The file ID string to delete, or a \code{tidyllm_file} object returned by \code{upload_file()}.}

\item{...}{Additional provider-specific arguments.}
}
\value{
Invisibly NULL; prints a confirmation message.
}
\description{
Delete a File from a Provider's File Store
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_gemini.R
\name{gemini_list_files}
\alias{gemini_list_files}
\title{List Files in Gemini API}
\usage{
gemini_list_files(.page_size = 10, .page_token = NULL)
}
\arguments{
\item{.page_size}{The maximum number of files to return per page (default: 10, maximum: 100).}

\item{.page_token}{A token for fetching the next page of results (default: NULL).}
}
\value{
A tibble containing metadata for each file, including fields such as name, display name, MIME type, and URI.
}
\description{
Lists metadata for files uploaded to the Gemini API, supporting pagination.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/media.R
\name{audio_file}
\alias{audio_file}
\title{Create an Audio Object}
\usage{
audio_file(.path)
}
\arguments{
\item{.path}{The path to the audio file on disk.}
}
\value{
A \code{tidyllm_audio} object.
}
\description{
Stores a reference to a local audio file for use in multimodal messages.
The file is not encoded until the message is formatted for a provider.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_ellmer.R
\name{chat_ellmer}
\alias{chat_ellmer}
\title{Send LLM Messages to Ellmer Chat Object}
\usage{
chat_ellmer(
  .llm,
  .ellmer_chat,
  .stream = FALSE,
  .timeout = 60,
  .verbose = FALSE,
  .dry_run = FALSE,
  .tools = NULL,
  .max_tries = 3
)
}
\arguments{
\item{.llm}{An \code{LLMMessage} object containing the conversation history.}

\item{.ellmer_chat}{An ellmer chat object (e.g., from \code{ellmer::chat_anthropic()}).}

\item{.stream}{Logical; if TRUE, streams the response piece by piece (default: FALSE).}

\item{.timeout}{Request timeout in seconds (default: 60).}

\item{.verbose}{If TRUE, displays additional information about the chat (default: FALSE).}

\item{.dry_run}{If TRUE, returns the constructed request object without executing it (default: FALSE).}

\item{.tools}{Dummy for error message if called from chat with tools (default: NULL).}

\item{.max_tries}{Maximum retries to perform the request (default: 3).}
}
\value{
A new \code{LLMMessage} object containing the original messages plus the assistant's response.
}
\description{
This function converts an \code{LLMMessage} to the turns ellmer chat object.
The ellmer object is cloned and its turn history is cleared internatlly.
This maintains tidyllms stateless style.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_voyage.R
\name{voyage}
\alias{voyage}
\title{Voyage Provider Function}
\usage{
voyage(..., .called_from = NULL)
}
\arguments{
\item{...}{Parameters to be passed to the appropriate Voyage-specific function,
such as model configuration, input text, or API-specific options.}

\item{.called_from}{An internal argument specifying which action (e.g.,
\code{embed}) the function is invoked from.
This argument is automatically managed by the \code{tidyllm} verbs and should not be modified by the user.}
}
\value{
The result of the requested action, depending on the specific function invoked
}
\description{
The \code{voyage()} function acts as a provider interface for interacting with the Voyage.ai API
through \code{tidyllm}'s verbs.
It dynamically routes requests to voyage-specific functions. At the moment this is only
\code{voyage_embed()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_llamacpp.R
\name{llamacpp_list_models}
\alias{llamacpp_list_models}
\title{List Models Loaded in the llama.cpp Server}
\usage{
llamacpp_list_models(
  .server = Sys.getenv("LLAMACPP_SERVER", "http://localhost:8080"),
  .api_key = Sys.getenv("LLAMACPP_API_KEY", ""),
  .timeout = 30,
  .max_tries = 3
)
}
\arguments{
\item{.server}{Base URL of the llama.cpp server. Defaults to \code{LLAMACPP_SERVER}
env var, falling back to \code{"http://localhost:8080"}.}

\item{.api_key}{API key for the server (default: \code{LLAMACPP_API_KEY} env var).}

\item{.timeout}{Request timeout in seconds (default: 30).}

\item{.max_tries}{Maximum retries (default: 3).}
}
\value{
A tibble with columns \code{id}, \code{object}, and \code{created}.
}
\description{
Calls the \verb{/v1/models} endpoint of a running llama.cpp server and returns
the currently loaded model(s) as a tibble. In normal operation this is one
row; two rows appear when speculative decoding is active (main + draft model).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_groq.R
\name{groq}
\alias{groq}
\title{Groq API Provider Function}
\usage{
groq(..., .called_from = NULL)
}
\arguments{
\item{...}{Parameters to be passed to the Groq-specific function,
such as model configuration, input text, or API-specific options.}

\item{.called_from}{An internal argument that specifies which action (e.g.,
\code{chat}) the function is being invoked from.
This argument is automatically managed and should not be modified by the user.}
}
\value{
The result of the requested action, depending on the specific function invoked
(currently, only an updated \code{LLMMessage} object for \code{groq_chat()}).
}
\description{
The \code{groq()} function acts as an interface for interacting with the Groq API
through \code{tidyllm}'s main verbs. Currently, Groq only supports \code{groq_chat()}
for chat-based interactions and \code{groq_transcribe()} for transcription tasks.
}
\details{
Since \code{groq_transcribe()} is unique to Groq and does not have a general verb counterpart,
\code{groq()} currently routes messages only to \code{groq_chat()} when used with verbs like \code{chat()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_llamacpp.R
\name{llamacpp_health}
\alias{llamacpp_health}
\title{Check Health of the llama.cpp Server}
\usage{
llamacpp_health(
  .server = Sys.getenv("LLAMACPP_SERVER", "http://localhost:8080"),
  .timeout = 10
)
}
\arguments{
\item{.server}{Base URL of the llama.cpp server. Defaults to \code{LLAMACPP_SERVER}
env var, falling back to \code{"http://localhost:8080"}.}

\item{.timeout}{Request timeout in seconds (default: 10).}
}
\value{
A named list with at least a \code{status} element.
}
\description{
Calls the \verb{/health} endpoint of a running llama.cpp server.
Returns the status string (\code{"ok"}, \code{"loading model"}, \code{"no model loaded"},
or \code{"error"}) along with the full parsed response body as a named list.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_perplexity.R
\name{perplexity_check_research}
\alias{perplexity_check_research}
\title{Check the Status of a Perplexity Deep Research Job}
\usage{
perplexity_check_research(
  .job,
  .api_key = Sys.getenv("PERPLEXITY_API_KEY"),
  .max_tries = 3
)
}
\arguments{
\item{.job}{A \code{tidyllm_research_job} object returned by \code{perplexity_deep_research(.background = TRUE)}.}

\item{.api_key}{Character; Perplexity API key (default: from environment).}

\item{.max_tries}{Integer; maximum retries (default: 3).}
}
\value{
An updated \code{tidyllm_research_job} with a \verb{$status} field and \verb{$response} if completed.
}
\description{
Check the Status of a Perplexity Deep Research Job
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_openrouter.R
\name{openrouter_list_models}
\alias{openrouter_list_models}
\title{List Available Models on OpenRouter}
\usage{
openrouter_list_models(
  .api_url = "https://openrouter.ai",
  .timeout = 60,
  .max_tries = 3,
  .dry_run = FALSE,
  .verbose = FALSE
)
}
\arguments{
\item{.api_url}{Base URL for the OpenRouter API (default: \code{"https://openrouter.ai"}).}

\item{.timeout}{Request timeout in seconds (default: 60).}

\item{.max_tries}{Maximum retries (default: 3).}

\item{.dry_run}{If TRUE, returns the request object without executing it (default: FALSE).}

\item{.verbose}{If TRUE, displays additional information (default: FALSE).}
}
\value{
A tibble with columns \code{id}, \code{name}, \code{context_length},
\code{prompt_price_per_million}, and \code{completion_price_per_million},
or NULL if no models are found.
}
\description{
Retrieves the list of models available through the OpenRouter API, including
pricing and context window information.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_mistral.R
\name{mistral_embedding}
\alias{mistral_embedding}
\title{Generate Embeddings Using Mistral API}
\usage{
mistral_embedding(
  .input,
  .model = "mistral-embed",
  .timeout = 120,
  .max_tries = 3,
  .dry_run = FALSE
)
}
\arguments{
\item{.input}{A character vector of texts to embed or an \code{LLMMessage} object}

\item{.model}{The embedding model identifier (default: "mistral-embed").}

\item{.timeout}{Timeout for the API request in seconds (default: 120).}

\item{.max_tries}{Maximum retries to peform request}

\item{.dry_run}{If TRUE, perform a dry run and return the request object.}
}
\value{
A matrix where each column corresponds to the embedding of a message in the message history.
}
\description{
Generate Embeddings Using Mistral API
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_chat_completions.R
\name{list_openai_batches}
\alias{list_openai_batches}
\title{List OpenAI Batch Requests}
\usage{
list_openai_batches(.limit = 20, .max_tries = 3, .timeout = 60)
}
\arguments{
\item{.limit}{Maximum number of batches to retrieve (default: 20).}

\item{.max_tries}{Maximum retry attempts for requests (default: 3).}

\item{.timeout}{Request timeout in seconds (default: 60).}
}
\value{
A tibble with batch details: batch ID, status, creation time, expiration time,
and request counts (total, completed, failed).
}
\description{
Retrieves batch request details from the OpenAI Batch API.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_ollama.R
\name{ollama_chat}
\alias{ollama_chat}
\title{Interact with local AI models via the Ollama API}
\usage{
ollama_chat(
  .llm,
  .model = "qwen3.5:4b",
  .stream = FALSE,
  .seed = NULL,
  .json_schema = NULL,
  .temperature = NULL,
  .num_ctx = 2048,
  .num_predict = NULL,
  .top_k = NULL,
  .top_p = NULL,
  .min_p = NULL,
  .mirostat = NULL,
  .mirostat_eta = NULL,
  .mirostat_tau = NULL,
  .repeat_last_n = NULL,
  .repeat_penalty = NULL,
  .tools = NULL,
  .max_tool_rounds = 10,
  .tfs_z = NULL,
  .stop = NULL,
  .think = NULL,
  .ollama_server = "http://localhost:11434",
  .timeout = 120,
  .keep_alive = NULL,
  .dry_run = FALSE
)
}
\arguments{
\item{.llm}{An LLMMessage object containing the conversation history and system prompt.}

\item{.model}{Character string specifying the Ollama model to use (default: "qwen3-vl")}

\item{.stream}{Logical; whether to stream the response (default: FALSE)}

\item{.seed}{Integer; seed for reproducible generation (default: NULL)}

\item{.json_schema}{A JSON schema object as R list to enforce the output structure (default: NULL)}

\item{.temperature}{Float between 0-2; controls randomness in responses (default: NULL)}

\item{.num_ctx}{Integer; sets the context window size (default: 2048)}

\item{.num_predict}{Integer; maximum number of tokens to predict (default: NULL)}

\item{.top_k}{Integer; controls diversity by limiting top tokens considered (default: NULL)}

\item{.top_p}{Float between 0-1; nucleus sampling threshold (default: NULL)}

\item{.min_p}{Float between 0-1; minimum probability threshold (default: NULL)}

\item{.mirostat}{Integer (0,1,2); enables Mirostat sampling algorithm (default: NULL)}

\item{.mirostat_eta}{Float; Mirostat learning rate (default: NULL)}

\item{.mirostat_tau}{Float; Mirostat target entropy (default: NULL)}

\item{.repeat_last_n}{Integer; tokens to look back for repetition (default: NULL)}

\item{.repeat_penalty}{Float; penalty for repeated tokens (default: NULL)}

\item{.tools}{Either a single TOOL object or a list of TOOL objects representing the available functions for tool calls.}

\item{.max_tool_rounds}{Integer; maximum number of tool use iterations for multi-turn tool calling (default: 10).
Set to 1 for single-round tool use, or higher for multi-turn agentic loops.}

\item{.tfs_z}{Float; tail free sampling parameter (default: NULL)}

\item{.stop}{Character; custom stop sequence(s) (default: NULL)}

\item{.think}{Logical or character; controls thinking mode for supported models like Qwen3. Use FALSE to disable, TRUE to enable, or "high"/"medium"/"low" to set effort level (default: NULL - model default)}

\item{.ollama_server}{String; Ollama API endpoint (default: "http://localhost:11434")}

\item{.timeout}{Integer; API request timeout in seconds (default: 120)}

\item{.keep_alive}{Character; How long should the ollama model be kept in memory after request (default: NULL - 5 Minutes)}

\item{.dry_run}{Logical; if TRUE, returns request object without execution (default: FALSE)}
}
\value{
A new LLMMessage object containing the original messages plus the model's response
}
\description{
Interact with local AI models via the Ollama API
}
\details{
The function provides extensive control over the generation process through various parameters:
\itemize{
\item Temperature (0-2): Higher values increase creativity, lower values make responses more focused
\item Top-k/Top-p: Control diversity of generated text
\item Mirostat: Advanced sampling algorithm for maintaining consistent complexity
\item Repeat penalties: Prevent repetitive text
\item Context window: Control how much previous conversation is considered
}
}
\examples{
\dontrun{
llm_message("user", "Hello, how are you?")
response <- ollama_chat(llm, .model = "gemma2", .temperature = 0.7)

# With custom parameters
response <- ollama_chat(
  llm,
  .model = "llama2",
  .temperature = 0.8,
  .top_p = 0.9,
  .num_ctx = 4096
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_gemini.R
\name{gemini}
\alias{gemini}
\title{Google Gemini Provider Function}
\usage{
gemini(..., .called_from = NULL)
}
\arguments{
\item{...}{Parameters to be passed to the appropriate Gemini-specific function,
such as model configuration, input text, or API-specific options.}

\item{.called_from}{An internal argument specifying which action (e.g.,
\code{chat}, \code{embed}) the function is invoked from.
This argument is automatically managed by the \code{tidyllm} verbs and should not be modified by the user.}
}
\value{
The result of the requested action, depending on the specific function invoked
(e.g., an updated \code{LLMMessage} object for \code{chat()}).
}
\description{
The \code{gemini()} function acts as a provider interface for interacting with the Google Gemini API
through \code{tidyllm}'s main verbs such as \code{chat()} and \code{embed()}.
It dynamically routes requests to Gemini-specific functions
like \code{gemini_chat()} and \code{gemini_embedding()} based on the context of the call.
}
\details{
Some functions, such as \code{gemini_upload_file()} and \code{gemini_delete_file()},
are specific to Gemini and do not have general verb counterparts.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_llamacpp.R
\name{llamacpp_download_model}
\alias{llamacpp_download_model}
\title{Download a GGUF Model from Hugging Face}
\usage{
llamacpp_download_model(
  .repo,
  .filename,
  .dir = Sys.getenv("LLAMACPP_MODEL_DIR", "~/models"),
  .timeout = 3600
)
}
\arguments{
\item{.repo}{Hugging Face repository ID (e.g. \code{"Qwen/Qwen3-8B-GGUF"}).}

\item{.filename}{The exact filename to download (e.g. \code{"Qwen3-8B-Q4_K_M.gguf"}).
Use \code{list_hf_gguf_files(.repo)} to see what is available.}

\item{.dir}{Destination directory. Defaults to \code{LLAMACPP_MODEL_DIR} env var,
then \code{"~/models"}. Created if it does not exist.}

\item{.timeout}{Download timeout in seconds (default: 3600).}
}
\value{
Invisibly returns the full path of the downloaded file.
}
\description{
Downloads a single GGUF file from a Hugging Face repository with a streaming
progress bar. Use \code{list_hf_gguf_files()} first to browse available quantizations.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_openai.R
\name{openai_code_interpreter}
\alias{openai_code_interpreter}
\title{OpenAI built-in code interpreter tool (server-executed)}
\usage{
openai_code_interpreter()
}
\value{
A TOOL object declaring the code_interpreter built-in.
}
\description{
OpenAI built-in code interpreter tool (server-executed)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_groq.R
\name{groq_chat}
\alias{groq_chat}
\title{Send LLM Messages to the Groq Chat API}
\usage{
groq_chat(
  .llm,
  .model = "moonshotai/kimi-k2-instruct-0905",
  .max_tokens = 1024,
  .temperature = NULL,
  .top_p = NULL,
  .frequency_penalty = NULL,
  .presence_penalty = NULL,
  .stop = NULL,
  .seed = NULL,
  .tools = NULL,
  .tool_choice = NULL,
  .api_url = "https://api.groq.com/",
  .json_schema = NULL,
  .timeout = 60,
  .verbose = FALSE,
  .stream = FALSE,
  .dry_run = FALSE,
  .max_tries = 3,
  .max_tool_rounds = 10
)
}
\arguments{
\item{.llm}{An \code{LLMMessage} object containing the conversation history.}

\item{.model}{The identifier of the model to use (default: "llama-3.2-11b-vision-preview").}

\item{.max_tokens}{The maximum number of tokens that can be generated in the response (default: 1024).}

\item{.temperature}{Controls the randomness in the model's response. Values between 0 and 2 are allowed, where higher values increase randomness (optional).}

\item{.top_p}{Nucleus sampling parameter that controls the proportion of probability mass considered. Values between 0 and 1 are allowed (optional).}

\item{.frequency_penalty}{Number between -2.0 and 2.0. Positive values penalize repeated tokens, reducing likelihood of repetition (optional).}

\item{.presence_penalty}{Number between -2.0 and 2.0. Positive values encourage new topics by penalizing tokens that have appeared so far (optional).}

\item{.stop}{One or more sequences where the API will stop generating further tokens. Can be a string or a list of strings (optional).}

\item{.seed}{An integer for deterministic sampling. If specified, attempts to return the same result for repeated requests with identical parameters (optional).}

\item{.tools}{Either a single TOOL object or a list of TOOL objects representing the available functions for tool calls (optional).}

\item{.tool_choice}{A character string specifying the tool-calling behavior; valid values are "none", "auto", or "required" (optional).}

\item{.api_url}{Base URL for the Groq API (default: "https://api.groq.com/").}

\item{.json_schema}{A list or tidyllm schema created with \code{tidyllm_schema()} for structured JSON output (optional).}

\item{.timeout}{Request timeout in seconds (default: 60).}

\item{.verbose}{If TRUE, displays additional information after the API call, including rate limit details (default: FALSE).}

\item{.stream}{Logical; if TRUE, streams the response piece by piece (default: FALSE).}

\item{.dry_run}{If TRUE, performs a dry run and returns the constructed request object without executing it (default: FALSE).}

\item{.max_tries}{Maximum retries to peform request}

\item{.max_tool_rounds}{Integer specifying the maximum number of tool use iterations (default: 10).
Set to 1 for single-round tool use, or higher for multi-turn agentic loops.}
}
\value{
A new \code{LLMMessage} object containing the original messages plus the assistant's response.
}
\description{
This function sends a message history to the Groq Chat API and returns the assistant's reply.
}
\examples{
\dontrun{
# Basic usage
msg <- llm_message("What is Groq?")
result <- groq_chat(msg)

# With custom parameters
result2 <- groq_chat(msg, 
               .model = "llama-3.2-vision",
               .temperature = 0.5, 
               .max_tokens = 512)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidyllm_schema.R
\name{field_chr}
\alias{field_chr}
\alias{field_fct}
\alias{field_dbl}
\alias{field_lgl}
\title{Define Field Descriptors for JSON Schema}
\usage{
field_chr(.description = character(0), .vector = FALSE)

field_fct(.description = character(0), .levels, .vector = FALSE)

field_dbl(.description = character(0), .vector = FALSE)

field_lgl(.description = character(0), .vector = FALSE)
}
\arguments{
\item{.description}{A character string describing the field (optional).}

\item{.vector}{A logical value indicating if the field is a vector (default: FALSE).}

\item{.levels}{A character vector specifying allowable values (for \code{field_fct()} only).}
}
\value{
An S7 \code{tidyllm_field} object representing the field descriptor.
}
\description{
These functions create field descriptors used in \code{tidyllm_schema()} or \code{field_object()} to define JSON schema fields. They support character, factor, numeric, and logical types.
}
\examples{
field_chr("A common street name")
field_fct("State abbreviation", .levels = c("CA", "TX", "Other"))
field_dbl("House number")
field_lgl("Is residential")
field_dbl("A list of appartment numbers at the address",.vector=TRUE )
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_ollama.R
\name{ollama_embedding}
\alias{ollama_embedding}
\title{Generate Embeddings Using Ollama API}
\usage{
ollama_embedding(
  .input,
  .model = "qwen3-embedding:0.6b",
  .truncate = TRUE,
  .ollama_server = "http://localhost:11434",
  .timeout = 120,
  .dry_run = FALSE
)
}
\arguments{
\item{.input}{Aa charachter vector of texts to embed or an \code{LLMMessage} object}

\item{.model}{The embedding model identifier (default: "all-minilm").}

\item{.truncate}{Whether to truncate inputs to fit the model's context length (default: TRUE).}

\item{.ollama_server}{The URL of the Ollama server to be used (default: "http://localhost:11434").}

\item{.timeout}{Timeout for the API request in seconds (default: 120).}

\item{.dry_run}{If TRUE, perform a dry run and return the request object.}
}
\value{
A matrix where each column corresponds to the embedding of a message in the message history.
}
\description{
Generate Embeddings Using Ollama API
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_openai.R
\name{check_openai_batch}
\alias{check_openai_batch}
\title{Check Batch Processing Status for OpenAI Batch API}
\usage{
check_openai_batch(
  .llms = NULL,
  .batch_id = NULL,
  .dry_run = FALSE,
  .max_tries = 3,
  .timeout = 60
)
}
\arguments{
\item{.llms}{A list of LLMMessage objects.}

\item{.batch_id}{A manually set batch ID.}

\item{.dry_run}{Logical; if TRUE, returns the prepared request object without executing it (default: FALSE).}

\item{.max_tries}{Maximum retries to perform the request (default: 3).}

\item{.timeout}{Integer specifying the request timeout in seconds (default: 60).}
}
\value{
A tibble with information about the status of batch processing.
}
\description{
This function retrieves the processing status and other details of a specified OpenAI batch ID
from the OpenAI Batch API.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llm_verbs.R
\name{deep_research}
\alias{deep_research}
\title{Run Deep Research via a Provider}
\usage{
deep_research(.llm, .provider, .background = FALSE, ...)
}
\arguments{
\item{.llm}{An \code{LLMMessage} object containing the research question.}

\item{.provider}{A function or function call specifying the provider (e.g., \code{perplexity()}).}

\item{.background}{Logical; if TRUE, returns a \code{tidyllm_research_job} immediately (default: FALSE).}

\item{...}{Additional arguments passed to the provider's deep research function.}
}
\value{
If \code{.background = FALSE}, an \code{LLMMessage} with the research reply.
If \code{.background = TRUE}, a \code{tidyllm_research_job} for use with \code{check_job()}/\code{fetch_job()}.
}
\description{
The \code{deep_research()} function sends a message to a provider's deep research endpoint.
Currently supported: Perplexity (\code{sonar-deep-research} via async API).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_gemini.R
\name{gemini_delete_file}
\alias{gemini_delete_file}
\title{Delete a File from Gemini API}
\usage{
gemini_delete_file(.file_name)
}
\arguments{
\item{.file_name}{The file ID (e.g., "files/abc-123") to delete.}
}
\value{
Invisibly returns \code{NULL}. Prints a confirmation message upon successful deletion.
}
\description{
Deletes a specific file from the Gemini API using its file ID.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_llamacpp.R
\name{llamacpp_chat}
\alias{llamacpp_chat}
\title{Send LLM Messages to a llama.cpp Server}
\usage{
llamacpp_chat(
  .llm,
  .model = "local-model",
  .max_tokens = 1024,
  .temperature = NULL,
  .top_p = NULL,
  .stop = NULL,
  .stream = FALSE,
  .tools = NULL,
  .tool_choice = NULL,
  .json_schema = NULL,
  .grammar = NULL,
  .logprobs = FALSE,
  .top_logprobs = NULL,
  .seed = NULL,
  .thinking = NULL,
  .thinking_budget = NULL,
  .server = Sys.getenv("LLAMACPP_SERVER", "http://localhost:8080"),
  .api_key = Sys.getenv("LLAMACPP_API_KEY", ""),
  .timeout = 120,
  .verbose = FALSE,
  .dry_run = FALSE,
  .max_tries = 3,
  .max_tool_rounds = 10
)
}
\arguments{
\item{.llm}{An \code{LLMMessage} object containing the conversation history.}

\item{.model}{The model name (default: \code{"local-model"}). llama.cpp ignores this
value and serves whatever GGUF is loaded; the default is a clear placeholder.}

\item{.max_tokens}{Maximum tokens in the response (default: 1024).}

\item{.temperature}{Controls randomness (0–2, optional).}

\item{.top_p}{Nucleus sampling parameter (0–1, optional).}

\item{.stop}{One or more stop sequences (optional).}

\item{.stream}{Logical; if TRUE, streams the response (default: FALSE).}

\item{.tools}{Either a single TOOL object or a list of TOOL objects.}

\item{.tool_choice}{Tool-calling behavior: \code{"none"}, \code{"auto"}, or \code{"required"} (optional).}

\item{.json_schema}{A JSON schema as an R list to enforce structured output. Passed
as \verb{response_format = \{type: "json_schema", ...\}} — enforced at the sampler level.}

\item{.grammar}{A BNF grammar string to constrain sampling to any formal language
(e.g. \code{"root ::= [0-9]{4}"} for a 4-digit year). Takes precedence over
\code{.json_schema} when both are set; use one or the other.}

\item{.logprobs}{Logical; if TRUE, returns token log-probabilities (default: FALSE).}

\item{.top_logprobs}{Integer; number of top-alternative log-prob entries to return
per token (1–20, optional). Requires \code{.logprobs = TRUE}.}

\item{.seed}{Integer; random seed for reproducible outputs (optional).}

\item{.thinking}{Logical; if TRUE, enables extended reasoning/thinking mode for models
that support it (e.g. Qwen3). \code{NULL} (default) lets the server decide; \code{FALSE}
explicitly disables it for faster responses; \code{TRUE} forces it on. The thinking
trace is accessible via \code{get_metadata(result)$api_specific$thinking}.}

\item{.thinking_budget}{Integer; maximum tokens the model may use for thinking when
\code{.thinking = TRUE}. \code{NULL} uses the server default (optional).}

\item{.server}{Base URL of the llama.cpp server. Defaults to the \code{LLAMACPP_SERVER}
environment variable, falling back to \code{"http://localhost:8080"}.}

\item{.api_key}{API key for the llama.cpp server. Defaults to the \code{LLAMACPP_API_KEY}
environment variable. Leave unset if the server was started without \code{--api-key}.}

\item{.timeout}{Request timeout in seconds (default: 120).}

\item{.verbose}{If TRUE, displays additional information (default: FALSE).}

\item{.dry_run}{If TRUE, returns the request object without executing it (default: FALSE).}

\item{.max_tries}{Maximum retries (default: 3).}

\item{.max_tool_rounds}{Maximum tool use iterations (default: 10).}
}
\value{
A new \code{LLMMessage} object containing the original messages plus the
assistant's response.
}
\description{
Sends a message history to a local llama.cpp server using the OpenAI-compatible
Chat Completions API. Supports BNF grammar constraints (a llama.cpp-specific
feature that enforces output format at the token sampling level) and token
logprobs for uncertainty quantification.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_gemini.R
\name{gemini_list_models}
\alias{gemini_list_models}
\title{List Available Models from the Google Gemini API}
\usage{
gemini_list_models(.timeout = 60, .max_tries = 3, .dry_run = FALSE)
}
\arguments{
\item{.timeout}{Request timeout in seconds (default: 60).}

\item{.max_tries}{Maximum number of retries for the API request (default: 3).}

\item{.dry_run}{Logical; if TRUE, returns the prepared request object without executing it.}
}
\value{
A tibble containing model information with columns including \code{name}, \code{base_model_id},
\code{version}, \code{display_name}, \code{description}, \code{input_token_limit}, \code{output_token_limit},
\code{supported_generation_methods}, \code{thinking}, \code{temperature}, \code{max_temperature}, \code{top_p}, and \code{top_k},
or NULL if no models are found.
}
\description{
List Available Models from the Google Gemini API
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llm_message.R
\name{llm_message}
\alias{llm_message}
\title{Create or Update Large Language Model Message Object}
\usage{
llm_message(
  .llm = NULL,
  .prompt = NULL,
  .role = "user",
  .system_prompt = "You are a helpful assistant",
  .imagefile = NULL,
  .pdf = NULL,
  .textfile = NULL,
  .capture_plot = FALSE,
  .f = NULL
)
}
\arguments{
\item{.llm}{An existing LLMMessage object or an initial text prompt.}

\item{.prompt}{Text prompt to add to the message history.}

\item{.role}{The role of the message sender, typically "user" or "assistant".}

\item{.system_prompt}{Default system prompt if a new LLMMessage needs to be created.}

\item{.imagefile}{Path to an image file to be attached (optional).}

\item{.pdf}{Path to a PDF file to be attached (optional). Can be a character vector of length one (file path), or a list with \code{filename}, \code{start_page}, and \code{end_page}.}

\item{.textfile}{Path to a text file to be read and attached (optional).}

\item{.capture_plot}{Boolean to indicate whether a plot should be captured and attached as an image (optional).}

\item{.f}{An R function or an object coercible to a function via \code{rlang::as_function}, whose output should be captured and attached (optional).}
}
\value{
Returns an updated or new LLMMessage object.
}
\description{
This function creates a new \code{LLMMessage} object or updates an existing one.
It supports adding text prompts and various media types, such as images, PDFs, text files, or plots.
}
\seealso{
\code{\link[=df_llm_message]{df_llm_message()}}

Other Message Creation Utilities: 
\code{\link{df_llm_message}()}
}
\concept{Message Creation Utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_chat_completions.R, R/api_openai.R
\name{openai}
\alias{openai}
\title{OpenAI Provider Function}
\usage{
openai(..., .called_from = NULL)

openai(..., .called_from = NULL)
}
\arguments{
\item{...}{Parameters passed to the appropriate OpenAI-specific function.}

\item{.called_from}{Internal routing argument; do not set manually.}
}
\value{
The result of the requested action, depending on the specific function invoked
(e.g., an updated \code{LLMMessage} object for \code{chat()}, or a matrix for \code{embed()}).

Result of the requested action.
}
\description{
The \code{openai()} function acts as an interface for interacting with the OpenAI API
through main \code{tidyllm} verbs such as \code{chat()}, \code{embed()}, and
\code{send_batch()}. It dynamically routes requests to OpenAI-specific functions
like \code{openai_chat()} and \code{openai_embedding()} based on the context of the call.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_openrouter.R
\name{openrouter_chat}
\alias{openrouter_chat}
\title{Send LLM Messages to the OpenRouter Chat API}
\usage{
openrouter_chat(
  .llm,
  .model = "anthropic/claude-sonnet-4-6",
  .max_tokens = 2048,
  .temperature = NULL,
  .top_p = NULL,
  .frequency_penalty = NULL,
  .presence_penalty = NULL,
  .stop = NULL,
  .stream = FALSE,
  .json_schema = NULL,
  .tools = NULL,
  .tool_choice = NULL,
  .provider = NULL,
  .route = NULL,
  .models = NULL,
  .api_url = "https://openrouter.ai",
  .timeout = 60,
  .verbose = FALSE,
  .dry_run = FALSE,
  .max_tries = 3,
  .max_tool_rounds = 10
)
}
\arguments{
\item{.llm}{An \code{LLMMessage} object containing the conversation history.}

\item{.model}{The model identifier to use (default: \code{"google/gemini-2.5-flash"}).
Any model available on OpenRouter can be used.}

\item{.max_tokens}{Maximum number of tokens in the response (default: 2048).}

\item{.temperature}{Controls randomness (0–2, optional).}

\item{.top_p}{Nucleus sampling parameter (0–1, optional).}

\item{.frequency_penalty}{Penalizes repeated tokens (-2 to 2, optional).}

\item{.presence_penalty}{Encourages new topics (-2 to 2, optional).}

\item{.stop}{One or more stop sequences (optional).}

\item{.stream}{Logical; if TRUE, streams the response (default: FALSE).}

\item{.json_schema}{A JSON schema object for structured output (default: NULL).}

\item{.tools}{Either a single TOOL object or a list of TOOL objects for tool calls.}

\item{.tool_choice}{Tool-calling behavior: \code{"none"}, \code{"auto"}, or \code{"required"} (optional).}

\item{.provider}{A named list of OpenRouter provider preferences, e.g.
\code{list(order = c("Anthropic", "AWS Bedrock"), allow_fallbacks = TRUE)} (optional).}

\item{.route}{OpenRouter routing strategy; \code{"fallback"} routes to the next model if
the primary is unavailable (optional).}

\item{.models}{A character vector of model IDs to use as fallbacks when the primary
model is unavailable (optional). Used together with \code{.route = "fallback"}.}

\item{.api_url}{Base URL for the OpenRouter API (default: \code{"https://openrouter.ai"}).}

\item{.timeout}{Request timeout in seconds (default: 60).}

\item{.verbose}{If TRUE, displays additional information after the API call (default: FALSE).}

\item{.dry_run}{If TRUE, returns the request object without executing it (default: FALSE).}

\item{.max_tries}{Maximum retries (default: 3).}

\item{.max_tool_rounds}{Maximum number of tool use iterations (default: 10).}
}
\value{
A new \code{LLMMessage} object containing the original messages plus the assistant's response.
}
\description{
Sends a message history to the OpenRouter API, which provides access to hundreds
of models from different providers through a single OpenAI-compatible endpoint.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llm_message.R
\name{df_llm_message}
\alias{df_llm_message}
\title{Convert a Data Frame to an LLMMessage Object}
\usage{
df_llm_message(.df)
}
\arguments{
\item{.df}{A data frame with at least two rows and columns \code{role} and \code{content}.
The \code{role} column should contain "user", "assistant", or "system". The \code{content} column should contain the corresponding message text.}
}
\value{
An \code{LLMMessage} object representing the structured conversation.
}
\description{
This function converts a data frame into an \code{LLMMessage} object representing a conversation history.
The data frame must have specific columns (\code{role} and \code{content}), with each row representing a message.
}
\seealso{
\code{\link[=llm_message]{llm_message()}}

Other Message Creation Utilities: 
\code{\link{llm_message}()}
}
\concept{Message Creation Utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_claude.R
\name{claude_upload_file}
\alias{claude_upload_file}
\title{Upload a File to Claude API}
\usage{
claude_upload_file(
  .file_path,
  .api_url = "https://api.anthropic.com/",
  .timeout = 60,
  .max_tries = 3,
  .dry_run = FALSE
)
}
\arguments{
\item{.file_path}{The local file path of the file to upload.}

\item{.api_url}{Base URL for the Claude API (default: "https://api.anthropic.com/").}

\item{.timeout}{Request timeout in seconds (default: 60).}

\item{.max_tries}{Maximum retry attempts for requests (default: 3).}

\item{.dry_run}{Logical; if TRUE, returns the prepared request object without executing it.}
}
\value{
A tibble containing metadata about the uploaded file, including its file_id, name, and size.
}
\description{
Uploads a file to the Claude API and returns its metadata as a tibble.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_openai.R
\name{openai_deep_research}
\alias{openai_deep_research}
\title{Submit a Deep Research Request to OpenAI}
\usage{
openai_deep_research(
  .llm,
  .model = "o4-mini-deep-research",
  .background = FALSE,
  .reasoning_effort = "medium",
  .json_schema = NULL,
  .max_output_tokens = NULL,
  .timeout = 1800,
  .max_tries = 3
)
}
\arguments{
\item{.llm}{An \code{LLMMessage} object containing the research question.}

\item{.model}{The deep research model to use (default: \code{"o4-mini-deep-research"}).}

\item{.background}{Logical; if \code{TRUE}, returns a \code{tidyllm_research_job} immediately
without waiting for completion (default: \code{FALSE}).}

\item{.reasoning_effort}{Reasoning level for the model: \code{"low"}, \code{"medium"} (default),
or \code{"high"}.}

\item{.json_schema}{A tidyllm schema for structured JSON output (optional).}

\item{.max_output_tokens}{Maximum tokens to generate (default: \code{NULL} for model default).}

\item{.timeout}{Seconds to wait in blocking mode before giving up (default: \code{1800}).}

\item{.max_tries}{Maximum retries per HTTP request (default: \code{3}).}
}
\value{
If \code{.background = FALSE}, an updated \code{LLMMessage} with the research reply.
If \code{.background = TRUE}, a \code{tidyllm_research_job} for use with \code{check_job()}/\code{fetch_job()}.
}
\description{
Sends a research request to OpenAI using the deep research models
(\code{o3-deep-research} or \code{o4-mini-deep-research}) via the Responses API with
\code{background: true}. The model autonomously searches the web and synthesises
a long-form answer, which can take 5-30 minutes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_azure_openai.R
\name{azure_openai}
\alias{azure_openai}
\title{Azure OpenAI Endpoint Provider Function}
\usage{
azure_openai(..., .called_from = NULL)
}
\arguments{
\item{...}{Parameters to be passed to the Azure OpenAI API specific function,
such as model configuration, input text, or API-specific options.}

\item{.called_from}{An internal argument that specifies which action (e.g.,
\code{chat}) the function is being invoked from.
This argument is automatically managed and should not be modified by the user.}
}
\value{
The result of the requested action, depending on the specific function invoked
(currently, only an updated \code{LLMMessage} object for \code{azure_openai_chat()}).
}
\description{
The \code{azure_openai()} function acts as an interface for interacting with the Azure OpenAI API
through main \code{tidyllm} verbs.
}
\details{
\code{azure_openai()} currently routes messages only to \code{azure_openai_chat()} when used with \code{chat()}.

\code{send_batch()}. It dynamically routes requests to OpenAI-specific functions
like \code{azure_openai_chat()} and \code{azure_openai_embedding()} based on the context of the call.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_ollama.R
\name{ollama_list_models}
\alias{ollama_list_models}
\title{Retrieve and return model information from the Ollama API}
\usage{
ollama_list_models(.ollama_server = "http://localhost:11434")
}
\arguments{
\item{.ollama_server}{The URL of the ollama server to be used}
}
\value{
A tibble containing model information, or NULL if no models are found.
}
\description{
This function connects to the Ollama API and retrieves information
about available models, returning it as a tibble.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rate_limits.R
\name{rate_limit_info}
\alias{rate_limit_info}
\title{Get the current rate limit information for all or a specific API}
\usage{
rate_limit_info(.api_name = NULL)
}
\arguments{
\item{.api_name}{(Optional) The name of the API whose rate limit info you want to get
If not provided, the rate limit info for all APIs in the environment will be returned}
}
\value{
A tibble containing the rate limit information.
}
\description{
This function retrieves the rate limit details for the specified API,
or for all APIs stored in the .tidyllm_rate_limit_env if no API is specified.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llm_verbs.R
\name{file_info}
\alias{file_info}
\title{Get Metadata for a File Stored on a Provider}
\usage{
file_info(.provider, .file_id, ...)
}
\arguments{
\item{.provider}{A provider function call.}

\item{.file_id}{The file ID string to look up, or a \code{tidyllm_file} object returned by \code{upload_file()}.}

\item{...}{Additional provider-specific arguments.}
}
\value{
A single-row tibble of file metadata.
}
\description{
Get Metadata for a File Stored on a Provider
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llm_verbs.R
\name{send_batch}
\alias{send_batch}
\title{Send a batch of messages to a batch API}
\usage{
send_batch(
  .llms,
  .provider = getOption("tidyllm_sbatch_default"),
  .dry_run = NULL,
  .temperature = NULL,
  .timeout = NULL,
  .top_p = NULL,
  .max_tries = NULL,
  .model = NULL,
  .verbose = NULL,
  .json_schema = NULL,
  .seed = NULL,
  .stop = NULL,
  .frequency_penalty = NULL,
  .presence_penalty = NULL,
  .id_prefix = NULL
)
}
\arguments{
\item{.llms}{A list of \code{LLMMessage} objects containing conversation histories.}

\item{.provider}{A function or function call specifying the language model provider and any additional parameters.
This should be a call to a provider function like \code{openai()}, \code{claude()}, etc.
You can also set a default provider function via the \code{tidyllm_sbatch_default} option.}

\item{.dry_run}{Logical; if \code{TRUE}, simulates the request without sending it to the provider. Useful for testing.}

\item{.temperature}{Numeric; controls the randomness of the model's output (0 = deterministic).}

\item{.timeout}{Numeric; the maximum time (in seconds) to wait for a response.}

\item{.top_p}{Numeric; nucleus sampling parameter, which limits the sampling to the top cumulative probability \code{p}.}

\item{.max_tries}{Integer; the maximum number of retries for failed requests.}

\item{.model}{Character; the model identifier to use (e.g., \code{"gpt-4"}).}

\item{.verbose}{Logical; if \code{TRUE}, prints additional information about the request and response.}

\item{.json_schema}{List; A JSON schema object as R list to enforce the output structure}

\item{.seed}{Integer; sets a random seed for reproducibility.}

\item{.stop}{Character vector; specifies sequences where the model should stop generating further tokens.}

\item{.frequency_penalty}{Numeric; adjusts the likelihood of repeating tokens (positive values decrease repetition).}

\item{.presence_penalty}{Numeric; adjusts the likelihood of introducing new tokens (positive values encourage novelty).}

\item{.id_prefix}{Character string to specify a prefix for generating custom IDs when names in \code{.llms} are missing}
}
\value{
An updated and named list of \code{.llms} with identifiers that align with batch responses, including a \code{batch_id} attribute.
}
\description{
The \code{send_batch()} function allows you to send a list of \code{LLMMessage} objects
to an API.
It routes the input to the appropriate provider-specific batch API function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_openai.R
\name{openai_chat}
\alias{openai_chat}
\title{Send LLM Messages to the OpenAI Responses API}
\usage{
openai_chat(
  .llm,
  .model = "gpt-5.5",
  .max_output_tokens = NULL,
  .temperature = NULL,
  .seed = NULL,
  .stream = FALSE,
  .timeout = 60,
  .verbose = FALSE,
  .json_schema = NULL,
  .max_tries = 3,
  .dry_run = FALSE,
  .reasoning_effort = NULL,
  .tools = NULL,
  .tool_choice = NULL,
  .max_tool_rounds = 10,
  .stateful = FALSE
)
}
\arguments{
\item{.llm}{An \code{LLMMessage} object containing the conversation history.}

\item{.model}{The model identifier (default: \code{"gpt-5.5"}).}

\item{.max_output_tokens}{Maximum tokens to generate (caps reasoning + completion).}

\item{.temperature}{Sampling temperature (0-2).}

\item{.seed}{Seed for deterministic sampling.}

\item{.stream}{If TRUE, stream output to console (default: FALSE).}

\item{.timeout}{Request timeout in seconds (default: 60).}

\item{.verbose}{Print rate-limit info after the call (default: FALSE).}

\item{.json_schema}{A tidyllm schema or ellmer type for structured output.}

\item{.max_tries}{Maximum retry attempts (default: 3).}

\item{.dry_run}{If TRUE, return the request object without sending (default: FALSE).}

\item{.reasoning_effort}{For o-series models: \code{"low"}, \code{"medium"}, or \code{"high"}.}

\item{.tools}{A TOOL object or list of TOOL objects for function calling.}

\item{.tool_choice}{Tool selection behavior: \code{"none"}, \code{"auto"}, or \code{"required"}.}

\item{.max_tool_rounds}{Maximum tool-loop iterations (default: 10).}

\item{.stateful}{If \code{TRUE}, use \code{previous_response_id} to pass conversation context
server-side instead of re-sending full message history. Falls back to full-history
mode silently when no previous response ID is available (first turn or prior turn
used a different provider), and with a warning if the server rejects the stored ID
(e.g., expired context). Default: \code{FALSE}.}
}
\value{
A new \code{LLMMessage} object with the assistant's response appended.
}
\description{
Sends a message history to the OpenAI Responses API (\code{POST /v1/responses}) and
returns the assistant's reply. Supports streaming, tool use, structured output,
and reasoning models (o-series) via \code{.reasoning_effort}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pdfbatch.R
\name{pdf_page_batch}
\alias{pdf_page_batch}
\title{Batch Process PDF into LLM Messages}
\usage{
pdf_page_batch(
  .pdf,
  .general_prompt,
  .system_prompt = "You are a helpful assistant",
  .page_range = NULL,
  .prompt_fn = NULL
)
}
\arguments{
\item{.pdf}{Path to the PDF file.}

\item{.general_prompt}{A default prompt that is applied to each page if \code{.prompt_fn} is not provided.}

\item{.system_prompt}{Optional system prompt to initialize the LLMMessage (default is "You are a helpful assistant").}

\item{.page_range}{A vector of two integers specifying the start and end pages to process. If NULL, all pages are processed.}

\item{.prompt_fn}{An optional custom function that generates a prompt for each page. The function takes the page text as input
and returns a string. If NULL, \code{.general_prompt} is used for all pages.}
}
\value{
A list of LLMMessage objects, each containing the text and image for a page.
}
\description{
This function processes a PDF file page by page. For each page, it extracts the text
and converts the page into an image. It creates a list of LLMMessage objects with
the text and the image for multimodal processing. Users can specify a range of pages
to process and provide a custom function to generate prompts for each page.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LLMMessage.R
\name{LLMMessage}
\alias{LLMMessage}
\title{Large Language Model Message Class}
\usage{
LLMMessage(message_history = list(), system_prompt = character(0))
}
\arguments{
\item{message_history}{A list containing messages. Each message is a named list with keys like \code{role}, \code{content}, \code{media}, etc.}

\item{system_prompt}{A character string representing the default system prompt used for the conversation.}
}
\description{
\code{LLMMessage} is an S7 class for managing a conversation history intended for use with large language models (LLMs). Please use
\code{llm_message()}to create or modify \code{LLMMessage} objects.
}
\details{
The \code{LLMMessage} class includes the following features:
\itemize{
\item Stores message history in a structured format.
\item Supports attaching media and metadata to messages.
\item Provides generics like \code{add_message()}, \code{has_image()}, and \code{remove_message()} for interaction.
\item Enables API-specific formatting through the \code{to_api_format()} generic.
\item \code{message_history}: A list containing messages. Each message is a named list with keys like \code{role}, \code{content}, \code{media}, etc.
\item \code{system_prompt}: A character string representing the default system prompt used for the conversation.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_groq.R
\name{fetch_groq_batch}
\alias{fetch_groq_batch}
\title{Fetch Results for a Groq Batch}
\usage{
fetch_groq_batch(
  .llms,
  .batch_id = NULL,
  .api_url = "https://api.groq.com/",
  .dry_run = FALSE,
  .max_tries = 3,
  .timeout = 60
)
}
\arguments{
\item{.llms}{A list of \code{LLMMessage} objects that were part of the batch.}

\item{.batch_id}{Character; the unique identifier for the batch.}

\item{.api_url}{Character; the base URL for the Groq API (default: "https://api.groq.com/").}

\item{.dry_run}{Logical; if \code{TRUE}, returns the constructed request without executing it (default: \code{FALSE}).}

\item{.max_tries}{Integer; maximum number of retries if the request fails (default: \code{3}).}

\item{.timeout}{Integer; request timeout in seconds (default: \code{60}).}
}
\value{
A list of updated \code{LLMMessage} objects, each with the assistant's response added if successful.
}
\description{
This function retrieves the results of a completed Groq batch and updates
the provided list of \code{LLMMessage} objects with the responses.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_voyage.R
\name{voyage_rerank}
\alias{voyage_rerank}
\title{Rerank Documents Using Voyage AI API}
\usage{
voyage_rerank(
  .query,
  .documents,
  .model = "rerank-2",
  .top_k = NULL,
  .api_key = Sys.getenv("VOYAGE_API_KEY"),
  .timeout = 60,
  .max_tries = 3
)
}
\arguments{
\item{.query}{A single character string representing the search query.}

\item{.documents}{A character vector of documents to rerank.}

\item{.model}{The reranking model identifier (default: "rerank-2").}

\item{.top_k}{Integer; return only the top-k results (default: NULL, returns all).}

\item{.api_key}{Character; Voyage API key (default: from environment).}

\item{.timeout}{Integer; request timeout in seconds (default: 60).}

\item{.max_tries}{Integer; maximum retries (default: 3).}
}
\value{
A tibble with columns \code{index}, \code{document}, and \code{relevance_score}, sorted by score descending.
}
\description{
Rerank Documents Using Voyage AI API
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_openai.R
\name{openai_check_research}
\alias{openai_check_research}
\title{Check the Status of an OpenAI Background Research Job}
\usage{
openai_check_research(.job, .max_tries = 3)
}
\arguments{
\item{.job}{A \code{tidyllm_research_job} object returned by \code{openai_deep_research(.background = TRUE)}.}

\item{.max_tries}{Maximum retries per HTTP request (default: \code{3}).}
}
\value{
An updated \code{tidyllm_research_job} with \verb{$status} set. If completed, \verb{$response}
is also populated and ready for \code{fetch_job()}.
}
\description{
Polls the status of an OpenAI background response created by
\code{openai_deep_research(.background = TRUE)}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_openai.R
\name{openai_embedding}
\alias{openai_embedding}
\title{Generate Embeddings Using OpenAI API}
\usage{
openai_embedding(
  .input,
  .model = "text-embedding-3-small",
  .truncate = TRUE,
  .timeout = 120,
  .dry_run = FALSE,
  .max_tries = 3,
  .verbose = FALSE
)
}
\arguments{
\item{.input}{An existing LLMMessage object (or a character vector of texts to embed)}

\item{.model}{The embedding model identifier (default: "text-embedding-3-small").}

\item{.truncate}{Whether to truncate inputs to fit the model's context length (default: TRUE).}

\item{.timeout}{Timeout for the API request in seconds (default: 120).}

\item{.dry_run}{If TRUE, perform a dry run and return the request object.}

\item{.max_tries}{Maximum retry attempts for requests (default: 3).}

\item{.verbose}{Should information about current ratelimits be printed? (default: FALSE)}
}
\value{
A tibble with two columns: \code{input} and \code{embeddings}.
The \code{input} column contains the texts sent to embed, and the \code{embeddings} column
is a list column where each row contains an embedding vector of the sent input.
}
\description{
Generate Embeddings Using OpenAI API
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llm_verbs.R
\name{check_job}
\alias{check_job}
\title{Check the Status of a Batch or Research Job}
\usage{
check_job(.job, ...)
}
\arguments{
\item{.job}{An object with a \code{batch_id} attribute (from \code{send_batch()}) or
a \code{tidyllm_research_job} (from \code{deep_research(.background = TRUE)}).}

\item{...}{Additional arguments passed to the underlying function.}
}
\value{
Status information; type depends on \code{.job} class.
}
\description{
\code{check_job()} dispatches to \code{check_batch()} for batch objects or
\code{perplexity_check_research()} for \code{tidyllm_research_job} objects.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llm_verbs.R
\name{list_batches}
\alias{list_batches}
\title{List all Batch Requests on a Batch API}
\usage{
list_batches(.provider = getOption("tidyllm_lbatch_default"))
}
\arguments{
\item{.provider}{A function or function call specifying the language model provider and any additional parameters.
This should be a call to a provider function like \code{openai()}, \code{claude()}, etc.
You can also set a default provider function via the \code{tidyllm_lbatch_default} option.}
}
\value{
A tibble with information about the status of batch processing.
}
\description{
List all Batch Requests on a Batch API
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_azure_openai.R
\name{azure_openai_embedding}
\alias{azure_openai_embedding}
\title{Generate Embeddings Using OpenAI API on Azure}
\usage{
azure_openai_embedding(
  .input,
  .deployment = "text-embedding-3-small",
  .endpoint_url = Sys.getenv("AZURE_ENDPOINT_URL"),
  .api_version = "2023-05-15",
  .truncate = TRUE,
  .timeout = 120,
  .dry_run = FALSE,
  .max_tries = 3
)
}
\arguments{
\item{.input}{A character vector of texts to embed or an \code{LLMMesssage}object}

\item{.deployment}{The embedding model identifier (default: "text-embedding-3-small").}

\item{.endpoint_url}{Base URL for the API (default:  Sys.getenv("AZURE_ENDPOINT_URL")).}

\item{.api_version}{What API-Version othe Azure OpenAI API should be used (default: "2023-05-15")}

\item{.truncate}{Whether to truncate inputs to fit the model's context length (default: TRUE).}

\item{.timeout}{Timeout for the API request in seconds (default: 120).}

\item{.dry_run}{If TRUE, perform a dry run and return the request object.}

\item{.max_tries}{Maximum retry attempts for requests (default: 3).}
}
\value{
A tibble with two columns: \code{input} and \code{embeddings}.
The \code{input} column contains the texts sent to embed, and the \code{embeddings} column
is a list column where each row contains an embedding vector of the sent input.
}
\description{
Generate Embeddings Using OpenAI API on Azure
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llm_verbs.R
\name{check_batch}
\alias{check_batch}
\title{Check Batch Processing Status}
\usage{
check_batch(
  .llms,
  .provider = getOption("tidyllm_cbatch_default"),
  .dry_run = NULL,
  .max_tries = NULL,
  .timeout = NULL
)
}
\arguments{
\item{.llms}{A list of \code{LLMMessage} objects or a character vector with a batch ID.}

\item{.provider}{A function or function call specifying the language model provider and any additional parameters.
This should be a call to a provider function like \code{openai()}, \code{claude()}, etc.
You can also set a default provider function via the \code{tidyllm_cbatch_default} option.}

\item{.dry_run}{Logical; if TRUE, returns the prepared request object without executing it}

\item{.max_tries}{Maximum retries to perform the request}

\item{.timeout}{Integer specifying the request timeout in seconds}
}
\value{
A tibble with information about the status of batch processing.
}
\description{
This function retrieves the processing status and other details of a specified
batchid or a list of \code{LLMMessage} objects with batch attribute.
It routes the input to the appropriate provider-specific batch API function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_mistral.R
\name{mistral_chat}
\alias{mistral_chat}
\title{Send LLMMessage to Mistral API}
\usage{
mistral_chat(
  .llm,
  .model = "mistral-large-latest",
  .frequency_penalty = NULL,
  .logit_bias = NULL,
  .presence_penalty = NULL,
  .seed = NULL,
  .stop = NULL,
  .stream = FALSE,
  .temperature = 0.7,
  .top_p = 1,
  .min_tokens = NULL,
  .max_tokens = NULL,
  .json_schema = NULL,
  .safe_prompt = FALSE,
  .timeout = 120,
  .max_tries = 3,
  .dry_run = FALSE,
  .verbose = FALSE,
  .tools = NULL,
  .tool_choice = NULL,
  .max_tool_rounds = 10
)
}
\arguments{
\item{.llm}{An \code{LLMMessage} object.}

\item{.model}{The model identifier to use (default: \code{"mistral-large-latest"}).}

\item{.frequency_penalty}{Number between -2.0 and 2.0. Positive values penalize new tokens based on their existing frequency.}

\item{.logit_bias}{A named list modifying the likelihood of specified tokens appearing in the completion.}

\item{.presence_penalty}{Number between -2.0 and 2.0. Positive values penalize new tokens based on whether they appear in the text so far.}

\item{.seed}{If specified, the system will make a best effort to sample deterministically.}

\item{.stop}{Up to 4 sequences where the API will stop generating further tokens.}

\item{.stream}{If set to TRUE, the answer will be streamed to console as it comes (default: FALSE).}

\item{.temperature}{What sampling temperature to use, between 0 and 2. Higher values make the output more random.}

\item{.top_p}{An alternative to sampling with temperature, called nucleus sampling.}

\item{.min_tokens}{The minimum number of tokens to generate in the completion. Must be \verb{>= 0} (optional).}

\item{.max_tokens}{An upper bound for the number of tokens that can be generated for a completion.}

\item{.json_schema}{A JSON schema object provided by tidyllm schema or ellmer schemata.}

\item{.safe_prompt}{Whether to inject a safety prompt before all conversations (default: \code{FALSE}).}

\item{.timeout}{When should our connection time out in seconds (default: \code{120}).}

\item{.max_tries}{Maximum retries to peform request}

\item{.dry_run}{If \code{TRUE}, perform a dry run and return the request object (default: \code{FALSE}).}

\item{.verbose}{Should additional information be shown after the API call? (default: \code{FALSE})}

\item{.tools}{Either a single TOOL object or a list of TOOL objects representing the available functions for tool calls.}

\item{.tool_choice}{A character string specifying the tool-calling behavior; valid values are "none", "auto", or "required".}

\item{.max_tool_rounds}{Integer specifying the maximum number of tool use iterations (default: 10).
Set to 1 for single-round tool use, or higher for multi-turn agentic loops.}
}
\value{
Returns an updated \code{LLMMessage} object.
}
\description{
Send LLMMessage to Mistral API
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_groq.R
\name{check_groq_batch}
\alias{check_groq_batch}
\title{Check Batch Processing Status for Groq API}
\usage{
check_groq_batch(
  .llms = NULL,
  .batch_id = NULL,
  .api_url = "https://api.groq.com/",
  .dry_run = FALSE,
  .max_tries = 3,
  .timeout = 60
)
}
\arguments{
\item{.llms}{A list of LLMMessage objects with a batch_id attribute.}

\item{.batch_id}{A character string with the batch ID to check.}

\item{.api_url}{Character; base URL of the Groq API (default: "https://api.groq.com/").}

\item{.dry_run}{Logical; if TRUE, returns the prepared request object without executing (default: FALSE).}

\item{.max_tries}{Maximum retries to perform request.}

\item{.timeout}{Integer specifying the request timeout in seconds (default: 60).}
}
\value{
A tibble with information about the status of batch processing.
}
\description{
This function retrieves the processing status and other details of a specified Groq batch.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_groq.R
\name{groq_list_models}
\alias{groq_list_models}
\title{List Available Models from the Groq API}
\usage{
groq_list_models(
  .api_url = "https://api.groq.com",
  .timeout = 60,
  .max_tries = 3,
  .dry_run = FALSE,
  .verbose = FALSE
)
}
\arguments{
\item{.api_url}{Base URL for the API (default: "https://api.groq.com").}

\item{.timeout}{Request timeout in seconds (default: 60).}

\item{.max_tries}{Maximum number of retries for the API request (default: 3).}

\item{.dry_run}{Logical; if TRUE, returns the prepared request object without executing it.}

\item{.verbose}{Logical; if TRUE, prints additional information about the request.}
}
\value{
A tibble containing model information (columns include \code{id}, \code{created}, \code{owned_by}, and \code{context_window}),
or NULL if no models are found.
}
\description{
List Available Models from the Groq API
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_openrouter.R
\name{openrouter_credits}
\alias{openrouter_credits}
\title{Get OpenRouter Credit Balance}
\usage{
openrouter_credits(
  .api_url = "https://openrouter.ai",
  .timeout = 60,
  .max_tries = 3
)
}
\arguments{
\item{.api_url}{Base URL for the OpenRouter API (default: \code{"https://openrouter.ai"}).}

\item{.timeout}{Request timeout in seconds (default: 60).}

\item{.max_tries}{Maximum retries (default: 3).}
}
\value{
A tibble with columns \code{total_credits} (USD purchased), \code{total_usage}
(USD consumed so far), and \code{remaining} (USD remaining).
}
\description{
Returns the total credits purchased and total usage so far for the current
API key.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_claude.R
\name{claude_list_files}
\alias{claude_list_files}
\title{List Files in Claude API}
\usage{
claude_list_files(
  .limit = 20,
  .order = "desc",
  .api_url = "https://api.anthropic.com/",
  .timeout = 60,
  .max_tries = 3,
  .dry_run = FALSE
)
}
\arguments{
\item{.limit}{The maximum number of files to return (default: 20).}

\item{.order}{Order of results, either "asc" or "desc" (default: "desc").}

\item{.api_url}{Base URL for the Claude API (default: "https://api.anthropic.com/").}

\item{.timeout}{Request timeout in seconds (default: 60).}

\item{.max_tries}{Maximum retry attempts for requests (default: 3).}

\item{.dry_run}{Logical; if TRUE, returns the prepared request object without executing it.}
}
\value{
A tibble containing metadata for each file, including file_id, filename, size, and MIME type.
}
\description{
Lists metadata for files uploaded to the Claude API, supporting pagination.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_voyage.R
\name{voyage_embedding}
\alias{voyage_embedding}
\title{Generate Embeddings Using Voyage AI API}
\usage{
voyage_embedding(
  .input,
  .model = "voyage-4",
  .output_dimension = NULL,
  .timeout = 120,
  .dry_run = FALSE,
  .max_tries = 3,
  .verbose = FALSE
)
}
\arguments{
\item{.input}{Input to embed. Can be:
\itemize{
\item A character vector of texts
\item An \code{LLMMessage} object (all textual components will be embedded)
\item A list containing a mix of character strings and \code{tidyllm_image} objects created with \code{img()}
}}

\item{.model}{The embedding model identifier. For text-only: "voyage-3.5-lite" (default).
For multimodal inputs: "voyage-multimodal-3" is used automatically.}

\item{.output_dimension}{Optional integer to control output vector size (default: NULL, uses model default).}

\item{.timeout}{Timeout for the API request in seconds (default: 120).}

\item{.dry_run}{If TRUE, perform a dry run and return the request object without sending.}

\item{.max_tries}{Maximum retry attempts for requests (default: 3).}

\item{.verbose}{Should information about current rate limits be printed? (default: FALSE).}
}
\value{
A tibble with two columns: \code{input} and \code{embeddings}.
\itemize{
\item The \code{input} column contains the input texts or image labels
\item The \code{embeddings} column is a list column where each row contains an embedding vector
}
}
\description{
This function creates embedding vectors from text or multimodal inputs (text and images)
using the Voyage AI API. It supports three types of input:
}
\details{
\enumerate{
\item Character vector: Embeds each text string separately
\item LLMMessage object: Extracts and embeds text content from messages
\item List of mixed content: Processes a combination of text strings and image objects created with \code{img()}
}

For multimodal inputs, the function automatically switches to Voyage's multimodal API
and formats the response with appropriate labels (e.g., \code{"[IMG] image.png"}) for images.
}
\examples{
\dontrun{
# Text embeddings
voyage_embedding("How does photosynthesis work?")

# Multimodal embeddings
list("A banana", img("banana.jpg"), "Yellow fruit") |>
  voyage_embedding()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_claude.R
\name{fetch_claude_batch}
\alias{fetch_claude_batch}
\title{Fetch Results for a Claude Batch}
\usage{
fetch_claude_batch(
  .llms,
  .batch_id = NULL,
  .api_url = "https://api.anthropic.com/",
  .dry_run = FALSE,
  .max_tries = 3,
  .timeout = 60
)
}
\arguments{
\item{.llms}{A list of \code{LLMMessage} objects that were part of the batch. The list should
have names (custom IDs) set by \code{send_claude_batch()} to ensure correct alignment.}

\item{.batch_id}{Character; the unique identifier for the batch. By default this is NULL
and the function will attempt to use the \code{batch_id} attribute from \code{.llms}.}

\item{.api_url}{Character; the base URL for the Claude API (default: "https://api.anthropic.com/").}

\item{.dry_run}{Logical; if \code{TRUE}, returns the constructed request without executing it (default: \code{FALSE}).}

\item{.max_tries}{Integer; maximum number of retries if the request fails (default: \code{3}).}

\item{.timeout}{Integer; request timeout in seconds (default: \code{60}).}
}
\value{
A list of updated \code{LLMMessage} objects, each with the assistant's response added if successful.
}
\description{
This function retrieves the results of a completed Claude batch and updates
the provided list of \code{LLMMessage} objects with the responses. It aligns each
response with the original request using the \code{custom_id}s generated in \code{send_claude_batch()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidyllm_schema.R
\name{tidyllm_schema}
\alias{tidyllm_schema}
\title{Create a JSON Schema for Structured Outputs}
\usage{
tidyllm_schema(name = "tidyllm_schema", ...)
}
\arguments{
\item{name}{A character string specifying the schema name (default: "tidyllm_schema").}

\item{...}{Named arguments where each name represents a field, and each value is either a character string, a \code{tidyllm_field}, or an \code{ellmer} type.

Supported character shorthand types:
\itemize{
\item "character" or "string" for character fields
\item "logical" for boolean fields
\item "numeric" for number fields
\item "factor(...)" for enumerations
\item Use \verb{[]} to indicate vectors, e.g., "character[]"
}}
}
\value{
A list representing the JSON schema, suitable for use with \code{.json_schema} in LLM API calls.
}
\description{
This function creates a JSON schema for structured outputs, supporting both character-based shorthand and S7 \code{tidyllm_field} objects.
It also integrates with \code{ellmer} types like \code{ellmer::type_string()} if ellmer is in your namespace
}
\examples{
\dontrun{
# Example using different field types
address_schema <- tidyllm_schema(
  name = "AddressSchema",
  Street = field_chr("A common street name"),
  house_number = field_dbl(),
  City = field_chr("Name of a city"),
  State = field_fct("State abbreviation", .levels = c("CA", "TX", "Other")),
  Country = "string",
  PostalCode = "string"
)

llm_message("Imagine an address") |> chat(openai, .json_schema = address_schema)

# Example with vector field
tidyllm_schema(
  plz = field_dbl(.vector = TRUE)
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_llamacpp.R
\name{llamacpp_rerank}
\alias{llamacpp_rerank}
\title{Rerank Documents Using a llama.cpp Server}
\usage{
llamacpp_rerank(
  .query,
  .documents,
  .model = "local-model",
  .server = Sys.getenv("LLAMACPP_SERVER", "http://localhost:8080"),
  .api_key = Sys.getenv("LLAMACPP_API_KEY", ""),
  .timeout = 60,
  .max_tries = 3,
  .dry_run = FALSE
)
}
\arguments{
\item{.query}{A single query string.}

\item{.documents}{A character vector of documents to rerank.}

\item{.model}{The model name (default: \code{"local-model"}).}

\item{.server}{Base URL of the llama.cpp server. Defaults to \code{LLAMACPP_SERVER}
env var, falling back to \code{"http://localhost:8080"}.}

\item{.api_key}{API key for the server (default: \code{LLAMACPP_API_KEY} env var).}

\item{.timeout}{Request timeout in seconds (default: 60).}

\item{.max_tries}{Maximum retries (default: 3).}

\item{.dry_run}{If TRUE, returns the request object without executing it (default: FALSE).}
}
\value{
A tibble with columns \code{index} (original position), \code{document} (text),
and \code{relevance_score}, sorted by descending score.
}
\description{
Calls the \verb{/v1/reranking} endpoint of a running llama.cpp server to score
documents by relevance to a query. Useful for building fully-local RAG
pipelines (embed → cosine search → rerank → chat, all with \code{llamacpp()}).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_deepseek.R
\name{deepseek}
\alias{deepseek}
\title{Deepseek Provider Function}
\usage{
deepseek(..., .called_from = NULL)
}
\arguments{
\item{...}{Parameters to be passed to the appropriate Deepseek-specific function,
such as model configuration, input text, or API-specific options.}

\item{.called_from}{An internal argument specifying which action (e.g.,
\code{chat}, \code{embed}) the function is invoked from.
This argument is automatically managed by the \code{tidyllm} verbs and should not be modified by the user.}
}
\value{
The result of the requested action, depending on the specific function invoked
(e.g., an updated \code{LLMMessage} object for \code{chat()}).
}
\description{
The \code{deepseek()} function acts as a provider interface for interacting with the Deepseek API
through \code{tidyllm}'s \code{chat()} verb.
It dynamically routes requests to deepseek-specific function. At the moment this is only
\code{deepseek_chat()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llm_verbs.R
\name{list_models}
\alias{list_models}
\title{List Available Models for a Provider}
\usage{
list_models(.provider = getOption("tidyllm_lmodels_default"), ...)
}
\arguments{
\item{.provider}{A function or function call specifying the provider and any additional parameters.
You can also set a default provider via the \code{tidyllm_lmodels_default} option.}

\item{...}{Additional arguments to be passed to the provider-specific list_models function.}
}
\value{
A tibble containing model information.
}
\description{
The \code{list_models()} function retrieves available models from the specified provider.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_openrouter.R
\name{openrouter}
\alias{openrouter}
\title{OpenRouter Provider Function}
\usage{
openrouter(..., .called_from = NULL)
}
\arguments{
\item{...}{Parameters passed to the appropriate OpenRouter-specific function.}

\item{.called_from}{An internal argument specifying which verb invoked this function.
Managed automatically by tidyllm verbs; do not set manually.}
}
\value{
The result of the requested action (e.g., an updated \code{LLMMessage} for \code{chat()}).
}
\description{
The \code{openrouter()} function provides access to hundreds of AI models from
different providers through the OpenRouter API, using a single
OpenAI-compatible interface.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_openai.R
\name{openai_websearch}
\alias{openai_websearch}
\title{OpenAI built-in web search tool (server-executed)}
\usage{
openai_websearch()
}
\value{
A TOOL object declaring the web_search_preview built-in.
}
\description{
OpenAI built-in web search tool (server-executed)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llm_verbs.R
\name{chat}
\alias{chat}
\title{Chat with a Language Model}
\usage{
chat(
  .llm,
  .provider = getOption("tidyllm_chat_default"),
  .dry_run = NULL,
  .stream = NULL,
  .temperature = NULL,
  .timeout = NULL,
  .top_p = NULL,
  .max_tries = NULL,
  .model = NULL,
  .verbose = NULL,
  .json_schema = NULL,
  .tools = NULL,
  .max_tool_rounds = NULL,
  .seed = NULL,
  .stop = NULL,
  .frequency_penalty = NULL,
  .presence_penalty = NULL
)
}
\arguments{
\item{.llm}{An \code{LLMMessage} object containing the message or conversation history to send to the language model.}

\item{.provider}{A function or function call specifying the language model provider and any additional parameters.
This should be a call to a provider function like \code{openai()}, \code{claude()}, etc.
You can also set a default provider function via the \code{tidyllm_chat_default} option.}

\item{.dry_run}{Logical; if \code{TRUE}, simulates the request without sending it to the provider. Useful for testing.}

\item{.stream}{Logical; if \code{TRUE}, streams the response from the provider in real-time.}

\item{.temperature}{Numeric; controls the randomness of the model's output (0 = deterministic).}

\item{.timeout}{Numeric; the maximum time (in seconds) to wait for a response.}

\item{.top_p}{Numeric; nucleus sampling parameter, which limits the sampling to the top cumulative probability \code{p}.}

\item{.max_tries}{Integer; the maximum number of retries for failed requests.}

\item{.model}{Character; the model identifier to use (e.g., \code{"gpt-4"}).}

\item{.verbose}{Logical; if \code{TRUE}, prints additional information about the request and response.}

\item{.json_schema}{List; A JSON schema object as R list to enforce the output structure}

\item{.tools}{Either a single TOOL object or a list of TOOL objects representing the available functions for tool calls.}

\item{.max_tool_rounds}{Integer; the maximum number of tool use iterations for multi-turn tool calling (default varies by provider).}

\item{.seed}{Integer; sets a random seed for reproducibility.}

\item{.stop}{Character vector; specifies sequences where the model should stop generating further tokens.}

\item{.frequency_penalty}{Numeric; adjusts the likelihood of repeating tokens (positive values decrease repetition).}

\item{.presence_penalty}{Numeric; adjusts the likelihood of introducing new tokens (positive values encourage novelty).}
}
\value{
An updated \code{LLMMessage} object containing the response from the language model.
}
\description{
The \code{chat()} function sends a message to a language model via a specified provider and returns the response.
It routes the provided \code{LLMMessage} object to the appropriate provider-specific chat function,
while allowing for the specification of common arguments applicable across different providers.
}
\details{
The \code{chat()} function provides a unified interface for interacting with different language model providers.
Common arguments such as \code{.temperature}, \code{.model}, and \code{.stream} are supported by most providers and can be
passed directly to \code{chat()}. If a provider does not support a particular argument, an error will be raised.

Advanced provider-specific configurations can be accessed via the provider functions.
}
\examples{
\dontrun{
# Basic usage with OpenAI provider
llm_message("Hello World") |>
   chat(ollama(.ollama_server = "https://my-ollama-server.de"),.model="mixtral")
   
   chat(mistral,.model="mixtral")

# Use streaming with Claude provider
llm_message("Tell me a story") |>
   chat(claude(),.stream=TRUE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_mistral.R
\name{list_mistral_batches}
\alias{list_mistral_batches}
\title{List Mistral Batch Requests}
\usage{
list_mistral_batches(
  .limit = 100,
  .max_tries = 3,
  .timeout = 60,
  .status = NULL,
  .created_after = NULL
)
}
\arguments{
\item{.limit}{Maximum number of batches to retrieve (default: 20).}

\item{.max_tries}{Maximum retry attempts for requests (default: 3).}

\item{.timeout}{Request timeout in seconds (default: 60).}

\item{.status}{Filter by status. (default: NULL)}

\item{.created_after}{created after a string specifiying a date-time  (default: NULL)}
}
\value{
A tibble with batch details for all batches fitting the request
}
\description{
Retrieves batch request details from the OpenAI Batch API.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_ellmer.R
\name{ellmer}
\alias{ellmer}
\title{Alias for the Ellmer Provider Function}
\usage{
ellmer(..., .called_from = NULL)
}
\arguments{
\item{...}{Additional parameters to pass through (for consistency with other providers).}

\item{.called_from}{An internal argument specifying which action (e.g.,
\code{chat}) the function is invoked from.}
}
\value{
A provider function that can be used with \code{tidyllm::chat()}.
}
\description{
The \code{chat_ellmer()} function acts as a provider interface for interacting with
ellmer chat objects through \code{tidyllm}'s  verb interface
}
\details{
This function allows you to use any ellmer chat object (e.g., from
\code{ellmer::chat_anthropic()}, \code{ellmer::chat_openai()}, etc.) as a stateless
backend for tidyllm. The ellmer object is cloned for each interaction to
maintain tidyllm's stateless approach.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/message_retrieval.R
\name{get_reply}
\alias{get_reply}
\alias{last_reply}
\title{Retrieve Assistant Reply as Text}
\usage{
get_reply(.llm, .index = NULL)

last_reply(.llm)
}
\arguments{
\item{.llm}{An \code{LLMMessage} object containing the message history.}

\item{.index}{A positive integer indicating the index of the assistant reply to retrieve.
Defaults to \code{NULL}, which retrieves the last reply.}
}
\value{
Returns a character string containing the assistant's reply, or \code{NA_character_} if no reply exists.
}
\description{
Extracts the plain text content of the assistant's reply from an \code{LLMMessage} object.
Use \code{\link[=get_reply_data]{get_reply_data()}} for structured replies in JSON format.
}
\details{
This function is the core utility for retrieving assistant replies by index.
For convenience, \code{\link[=last_reply]{last_reply()}} is provided as a wrapper to retrieve the
latest assistant reply.
}
\seealso{
\code{\link[=get_reply_data]{get_reply_data()}}, \code{\link[=last_reply]{last_reply()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_llamacpp.R
\name{llamacpp_embedding}
\alias{llamacpp_embedding}
\title{Generate Embeddings Using a llama.cpp Server}
\usage{
llamacpp_embedding(
  .input,
  .model = "local-model",
  .server = Sys.getenv("LLAMACPP_SERVER", "http://localhost:8080"),
  .api_key = Sys.getenv("LLAMACPP_API_KEY", ""),
  .timeout = 120,
  .dry_run = FALSE,
  .max_tries = 3
)
}
\arguments{
\item{.input}{A character vector of texts to embed, or an \code{LLMMessage} object.}

\item{.model}{The model name (default: \code{"local-model"}). llama.cpp ignores this
and serves the loaded embedding model.}

\item{.server}{Base URL of the llama.cpp server. Defaults to the \code{LLAMACPP_SERVER}
environment variable, falling back to \code{"http://localhost:8080"}.}

\item{.api_key}{API key for the server (default: \code{LLAMACPP_API_KEY} env var).}

\item{.timeout}{Request timeout in seconds (default: 120).}

\item{.dry_run}{If TRUE, returns the request object without executing it (default: FALSE).}

\item{.max_tries}{Maximum retries (default: 3).}
}
\value{
A tibble with columns \code{input} (text) and \code{embeddings} (list of numeric vectors).
}
\description{
Sends text to the \verb{/v1/embeddings} endpoint of a running llama.cpp server
and returns embedding vectors.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_azure_openai.R
\name{list_azure_openai_batches}
\alias{list_azure_openai_batches}
\title{List Azure OpenAI Batch Requests}
\usage{
list_azure_openai_batches(
  .endpoint_url = Sys.getenv("AZURE_ENDPOINT_URL"),
  .limit = 20,
  .max_tries = 3,
  .timeout = 60
)
}
\arguments{
\item{.endpoint_url}{Base URL for the API (default:  Sys.getenv("AZURE_ENDPOINT_URL")).}

\item{.limit}{Maximum number of batches to retrieve (default: 20).}

\item{.max_tries}{Maximum retry attempts for requests (default: 3).}

\item{.timeout}{Request timeout in seconds (default: 60).}
}
\value{
A tibble with batch details: batch ID, status, creation time, expiration time,
and request counts (total, completed, failed).
}
\description{
Retrieves batch request details from the Azure OpenAI Batch API.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_deepseek.R
\name{deepseek_chat}
\alias{deepseek_chat}
\title{Send LLM Messages to the DeepSeek Chat API}
\usage{
deepseek_chat(
  .llm,
  .model = "deepseek-chat",
  .thinking = NULL,
  .max_tokens = 2048,
  .temperature = NULL,
  .top_p = NULL,
  .frequency_penalty = NULL,
  .presence_penalty = NULL,
  .stop = NULL,
  .stream = FALSE,
  .logprobs = NULL,
  .top_logprobs = NULL,
  .tools = NULL,
  .tool_choice = NULL,
  .api_url = "https://api.deepseek.com/",
  .timeout = 60,
  .verbose = FALSE,
  .dry_run = FALSE,
  .max_tries = 3,
  .max_tool_rounds = 10
)
}
\arguments{
\item{.llm}{An \code{LLMMessage} object containing the conversation history.}

\item{.model}{The identifier of the model to use (default: "deepseek-chat").}

\item{.thinking}{If TRUE, switches to the \code{deepseek-reasoner} model and captures the reasoning trace (default: NULL).}

\item{.max_tokens}{The maximum number of tokens that can be generated in the response (default: 2048).}

\item{.temperature}{Controls the randomness in the model's response. Values between 0 and 2 are allowed (optional).}

\item{.top_p}{Nucleus sampling parameter that controls the proportion of probability mass considered (optional).}

\item{.frequency_penalty}{Number between -2.0 and 2.0. Penalizes repeated tokens to reduce repetition (optional).}

\item{.presence_penalty}{Number between -2.0 and 2.0. Encourages new topics by penalizing tokens that have appeared so far (optional).}

\item{.stop}{One or more sequences where the API will stop generating further tokens (optional).}

\item{.stream}{Logical; if TRUE, streams the response piece by piece (default: FALSE).}

\item{.logprobs}{If TRUE, returns log probabilities of each output token (default: FALSE).}

\item{.top_logprobs}{Number between 0 and 5 specifying the number of top log probabilities to return (optional).}

\item{.tools}{Either a single TOOL object or a list of TOOL objects representing the available functions for tool calls.}

\item{.tool_choice}{A character string specifying the tool-calling behavior; valid values are \code{"none"}, \code{"auto"}, or \code{"required"} (optional).}

\item{.api_url}{Base URL for the DeepSeek API (default: \code{"https://api.deepseek.com/"}).}

\item{.timeout}{Request timeout in seconds (default: 60).}

\item{.verbose}{If TRUE, displays additional information after the API call (default: FALSE).}

\item{.dry_run}{If TRUE, returns the constructed request object without executing it (default: FALSE).}

\item{.max_tries}{Maximum retries to perform the request (default: 3).}

\item{.max_tool_rounds}{Integer specifying the maximum number of tool use iterations (default: 10).
Set to 1 for single-round tool use, or higher for multi-turn agentic loops.}
}
\value{
A new \code{LLMMessage} object containing the original messages plus the assistant's response.
}
\description{
This function sends a message history to the DeepSeek Chat API and returns the assistant's reply.
Currently tool calls cause problems on the DeepSeek API
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_llamacpp.R
\name{llamacpp_list_local_models}
\alias{llamacpp_list_local_models}
\title{List Local GGUF Model Files}
\usage{
llamacpp_list_local_models(
  .path = Sys.getenv("LLAMACPP_MODEL_DIR", "~/models")
)
}
\arguments{
\item{.path}{Directory to scan (default: \code{LLAMACPP_MODEL_DIR} env var, then
\code{"~/models"}). The path is expanded with \code{path.expand()}.}
}
\value{
A tibble with columns \code{filename}, \code{size_gb}, \code{modified}, and \code{path}.
}
\description{
Scans a directory for \code{.gguf} files and returns a tibble with their names,
sizes, and modification times. No server needed.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_mistral.R
\name{mistral}
\alias{mistral}
\title{Mistral Provider Function}
\usage{
mistral(..., .called_from = NULL)
}
\arguments{
\item{...}{Parameters to be passed to the appropriate Mistral-specific function,
such as model configuration, input text, or API-specific options.}

\item{.called_from}{An internal argument that specifies which action (e.g.,
\code{chat}, \code{embed}, \code{send_batch}) the function is being invoked from.
This argument is automatically managed and should not be modified by the user.}
}
\value{
The result of the requested action, depending on the specific function invoked
(e.g., an updated \code{LLMMessage} object for \code{chat()}, or a matrix for \code{embed()}).
}
\description{
The \code{mistral()} function acts as an interface for interacting with the Mistral API
through main \code{tidyllm} verbs such as \code{chat()} and \code{embed()}.
It dynamically routes requests to Mistral-specific functions
like \code{mistral_chat()} and \code{mistral_embedding()} based on the context of the call.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidyllm_schema.R
\name{field_object}
\alias{field_object}
\title{Define a nested object field}
\usage{
field_object(.description = character(0), ..., .vector = FALSE)
}
\arguments{
\item{.description}{A character string describing the field (optional).}

\item{...}{Named fields to include in the object definition (required).}

\item{.vector}{A logical value indicating if the field is a list of objects (default: FALSE).}
}
\value{
An S7 \code{tidyllm_field} object of type "object" containing nested fields.
}
\description{
Define a nested object field
}
\examples{
# Define an address object with nested fields
address <- field_object("A mailing address",
  street = field_chr("Street name"),
  city = field_chr("City name"),
  zipcode = field_chr("Postal code")
)

# Create a vector of objects
addresses <- field_object("List of addresses", 
  street = field_chr("Street name"),
  city = field_chr("City name"),
  .vector = TRUE
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_perplexity.R
\name{perplexity_fetch_research}
\alias{perplexity_fetch_research}
\title{Fetch Results from a Completed Perplexity Deep Research Job}
\usage{
perplexity_fetch_research(
  .job,
  .api_key = Sys.getenv("PERPLEXITY_API_KEY"),
  .max_tries = 3
)
}
\arguments{
\item{.job}{A \code{tidyllm_research_job} object. Must have status "completed" or will poll once.}

\item{.api_key}{Character; Perplexity API key (default: from environment).}

\item{.max_tries}{Integer; maximum retries (default: 3).}
}
\value{
An updated \code{LLMMessage} with the research reply appended.
}
\description{
Fetch Results from a Completed Perplexity Deep Research Job
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_llamacpp.R
\name{list_hf_gguf_files}
\alias{list_hf_gguf_files}
\title{List GGUF Files Available in a Hugging Face Repository}
\usage{
list_hf_gguf_files(.repo, .timeout = 30)
}
\arguments{
\item{.repo}{Hugging Face repository ID in \code{"owner/model"} format
(e.g. \code{"Qwen/Qwen3-8B-GGUF"}).}

\item{.timeout}{Request timeout in seconds (default: 30).}
}
\value{
A tibble with columns \code{filename}, \code{size_gb}, and \code{url}.
}
\description{
Calls the Hugging Face Hub API to list all \code{.gguf} files in a repository,
along with their sizes. Useful for choosing a quantization level before
calling \code{llamacpp_download_model()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_groq.R
\name{groq_transcribe}
\alias{groq_transcribe}
\title{Transcribe an Audio File Using Groq transcription API}
\usage{
groq_transcribe(
  .audio_file,
  .model = "playai-tts",
  .language = NULL,
  .prompt = NULL,
  .temperature = 0,
  .api_url = "https://api.groq.com/openai/v1/audio/transcriptions",
  .dry_run = FALSE,
  .verbose = FALSE,
  .max_tries = 3
)
}
\arguments{
\item{.audio_file}{The path to the audio file (required). Supported formats include flac, mp3, mp4, mpeg, mpga, m4a, ogg, wav, or webm.}

\item{.model}{The model to use for transcription (default: "whisper-large-v3").}

\item{.language}{The language of the input audio, in ISO-639-1 format (optional).}

\item{.prompt}{A prompt to guide the transcription style. It should match the audio language (optional).}

\item{.temperature}{Sampling temperature, between 0 and 1, with higher values producing more randomness (default: 0).}

\item{.api_url}{Base URL for the API (default: "https://api.groq.com/openai/v1/audio/transcriptions").}

\item{.dry_run}{Logical; if TRUE, performs a dry run and returns the request object without making the API call (default: FALSE).}

\item{.verbose}{Logical; if TRUE, rate limiting info is displayed after the API request (default: FALSE).}

\item{.max_tries}{Maximum retries to peform request}
}
\value{
A character vector containing the transcription.
}
\description{
This function reads an audio file and sends it to the Groq transcription API for transcription.
}
\examples{
\dontrun{
# Basic usage
groq_transcribe(.audio_file = "example.mp3")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_gemini.R
\name{send_gemini_batch}
\alias{send_gemini_batch}
\title{Submit a list of LLMMessage objects to Gemini's batch API}
\usage{
send_gemini_batch(
  .llms,
  .model = "gemini-2.5-flash",
  .temperature = NULL,
  .max_output_tokens = NULL,
  .top_p = NULL,
  .top_k = NULL,
  .presence_penalty = NULL,
  .frequency_penalty = NULL,
  .stop_sequences = NULL,
  .safety_settings = NULL,
  .json_schema = NULL,
  .grounding_threshold = NULL,
  .timeout = 120,
  .dry_run = FALSE,
  .max_tries = 3,
  .display = "tidyllm_batch",
  .id_prefix = "tidyllm_gemini_req_"
)
}
\arguments{
\item{.llms}{List of LLMMessage objects (named or unnamed).}

\item{.model}{The model identifier (default: "gemini-1.5-flash").}

\item{.temperature}{Controls randomness (default: NULL, range: 0-2).}

\item{.max_output_tokens}{Maximum tokens in the response (default: NULL).}

\item{.top_p}{Nucleus sampling (default: NULL, range: 0-1).}

\item{.top_k}{Diversity in token selection (default: NULL).}

\item{.presence_penalty}{Penalizes new tokens (default: NULL, -2 to 2).}

\item{.frequency_penalty}{Penalizes frequent tokens (default: NULL, -2 to 2).}

\item{.stop_sequences}{Character vector or NULL of up to 5.}

\item{.safety_settings}{Optional list of safety settings (default: NULL).}

\item{.json_schema}{Optional schema to enforce output structure.}

\item{.grounding_threshold}{Optional grounding threshold (0-1) to enable Google Search.}

\item{.timeout}{Timeout in seconds (default: 120).}

\item{.dry_run}{If TRUE, returns the constructed request (default: FALSE).}

\item{.max_tries}{Maximum retry attempts (default: 3).}

\item{.display}{Display name for this batch (default: "tidyllm_batch").}

\item{.id_prefix}{Prefix for message IDs (default: "tidyllm_gemini_req_").}
}
\value{
Named list of LLMMessage objects with attributes \code{batch_id} and \code{json}
}
\description{
Returns a named list (same as input) with batch_id and json attributes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_ollama.R
\name{send_ollama_batch}
\alias{send_ollama_batch}
\title{Send a Batch of Messages to Ollama API}
\usage{
send_ollama_batch(
  .llms,
  .model = "qwen3.5:4b",
  .stream = FALSE,
  .seed = NULL,
  .json_schema = NULL,
  .temperature = NULL,
  .num_ctx = 2048,
  .num_predict = NULL,
  .top_k = NULL,
  .top_p = NULL,
  .min_p = NULL,
  .mirostat = NULL,
  .mirostat_eta = NULL,
  .mirostat_tau = NULL,
  .repeat_last_n = NULL,
  .repeat_penalty = NULL,
  .tfs_z = NULL,
  .stop = NULL,
  .ollama_server = "http://localhost:11434",
  .timeout = 120,
  .keep_alive = NULL,
  .dry_run = FALSE
)
}
\arguments{
\item{.llms}{A list of LLMMessage objects containing conversation histories.}

\item{.model}{Character string specifying the Ollama model to use (default: "qwen3-vl")}

\item{.stream}{Logical; whether to stream the response (default: FALSE)}

\item{.seed}{Integer; seed for reproducible generation (default: NULL)}

\item{.json_schema}{A JSON schema object as R list to enforce the output structure (default: NULL)}

\item{.temperature}{Float between 0-2; controls randomness in responses (default: NULL)}

\item{.num_ctx}{Integer; sets the context window size (default: 2048)}

\item{.num_predict}{Integer; maximum number of tokens to predict (default: NULL)}

\item{.top_k}{Integer; controls diversity by limiting top tokens considered (default: NULL)}

\item{.top_p}{Float between 0-1; nucleus sampling threshold (default: NULL)}

\item{.min_p}{Float between 0-1; minimum probability threshold (default: NULL)}

\item{.mirostat}{Integer (0,1,2); enables Mirostat sampling algorithm (default: NULL)}

\item{.mirostat_eta}{Float; Mirostat learning rate (default: NULL)}

\item{.mirostat_tau}{Float; Mirostat target entropy (default: NULL)}

\item{.repeat_last_n}{Integer; tokens to look back for repetition (default: NULL)}

\item{.repeat_penalty}{Float; penalty for repeated tokens (default: NULL)}

\item{.tfs_z}{Float; tail free sampling parameter (default: NULL)}

\item{.stop}{Character; custom stop sequence(s) (default: NULL)}

\item{.ollama_server}{String; Ollama API endpoint (default: "http://localhost:11434")}

\item{.timeout}{Integer; API request timeout in seconds (default: 120)}

\item{.keep_alive}{Character; How long should the ollama model be kept in memory after request (default: NULL - 5 Minutes)}

\item{.dry_run}{Logical; if TRUE, returns request object without execution (default: FALSE)}
}
\value{
A list of updated \code{LLMMessage} objects, each with the assistant's response added if successful.
}
\description{
This function creates and submits a batch of messages to the Ollama API
Contrary to other batch functions, this functions waits for the batch to finish and receives requests.
The advantage compared to sending single messages via \code{chat()} is that Ollama handles large parallel
requests quicker than many individual chat requests.
}
\details{
The function provides extensive control over the generation process through various parameters:
\itemize{
\item Temperature (0-2): Higher values increase creativity, lower values make responses more focused
\item Top-k/Top-p: Control diversity of generated text
\item Mirostat: Advanced sampling algorithm for maintaining consistent complexity
\item Repeat penalties: Prevent repetitive text
\item Context window: Control how much previous conversation is considered
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_chat_completions.R, R/api_openai.R
\name{chatgpt}
\alias{chatgpt}
\title{Alias for the OpenAI Provider Function}
\usage{
chatgpt(..., .called_from = NULL)

chatgpt(..., .called_from = NULL)
}
\arguments{
\item{...}{Parameters passed to the appropriate OpenAI-specific function.}

\item{.called_from}{Internal routing argument; do not set manually.}
}
\value{
The result of the requested action, depending on the specific function invoked
(e.g., an updated \code{LLMMessage} object for \code{chat()}, or a matrix for \code{embed()}).
}
\description{
The \code{chatgpt} function is an alias for the \code{openai()} provider function.
It provides a convenient way to interact with the OpenAI API for tasks such
as sending chat messages, generating embeddings, and handling batch operations
using \code{tidyllm} verbs like \code{chat()}, \code{embed()}, and \code{send_batch()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_azure_openai.R
\name{send_azure_openai_batch}
\alias{send_azure_openai_batch}
\title{Send a Batch of Messages to Azure OpenAI Batch API}
\usage{
send_azure_openai_batch(
  .llms,
  .endpoint_url = Sys.getenv("AZURE_ENDPOINT_URL"),
  .deployment = "gpt-4o-mini",
  .api_version = "2024-10-01-preview",
  .max_completion_tokens = NULL,
  .reasoning_effort = NULL,
  .frequency_penalty = NULL,
  .logit_bias = NULL,
  .presence_penalty = NULL,
  .seed = NULL,
  .stop = NULL,
  .temperature = NULL,
  .top_p = NULL,
  .logprobs = NULL,
  .top_logprobs = NULL,
  .dry_run = FALSE,
  .overwrite = FALSE,
  .max_tries = 3,
  .timeout = 60,
  .verbose = FALSE,
  .json_schema = NULL,
  .id_prefix = "tidyllm_azure_openai_req_"
)
}
\arguments{
\item{.llms}{A list of LLMMessage objects containing conversation histories.}

\item{.endpoint_url}{Base URL for the API (default: Sys.getenv("AZURE_ENDPOINT_URL")).}

\item{.deployment}{The identifier of the model that is deployed (default: "gpt-4o-mini").}

\item{.api_version}{Which version of the API is deployed (default: "2024-10-01-preview")}

\item{.max_completion_tokens}{Integer specifying the maximum tokens per response (default: NULL).}

\item{.reasoning_effort}{How long should reasoning models reason (can either be "low","medium" or "high")}

\item{.frequency_penalty}{Number between -2.0 and 2.0. Positive values penalize new tokens based on their existing frequency.}

\item{.logit_bias}{A named list modifying the likelihood of specified tokens appearing in the completion.}

\item{.presence_penalty}{Number between -2.0 and 2.0. Positive values penalize new tokens based on whether they appear in the text so far.}

\item{.seed}{If specified, the system will make a best effort to sample deterministically.}

\item{.stop}{Up to 4 sequences where the API will stop generating further tokens.}

\item{.temperature}{What sampling temperature to use, between 0 and 2. Higher values make the output more random.}

\item{.top_p}{An alternative to sampling with temperature, called nucleus sampling.}

\item{.logprobs}{If TRUE, get the log probabilities of each output token (default: NULL).}

\item{.top_logprobs}{If specified, get the top N log probabilities of each output token (0-5, default: NULL).}

\item{.dry_run}{Logical; if TRUE, returns the prepared request object without executing it (default: FALSE).}

\item{.overwrite}{Logical; if TRUE, allows overwriting an existing batch ID (default: FALSE).}

\item{.max_tries}{Maximum number of retries to perform the request (default: 3).}

\item{.timeout}{Integer specifying the request timeout in seconds (default: 60).}

\item{.verbose}{Logical; if TRUE, additional info about the requests is printed (default: FALSE).}

\item{.json_schema}{A JSON schema object provided by tidyllm schema or ellmer schemata (default: NULL).}

\item{.id_prefix}{Character string to specify a prefix for generating custom IDs when names in \code{.llms} are missing (default: "tidyllm_azure_openai_req_").}
}
\value{
An updated and named list of \code{.llms} with identifiers that align with batch responses, including a \code{batch_id} attribute.
}
\description{
This function creates and submits a batch of messages to the Azure OpenAI Batch API for asynchronous processing.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llm_verbs.R
\name{fetch_batch}
\alias{fetch_batch}
\title{Fetch Results from a Batch API}
\usage{
fetch_batch(
  .llms,
  .provider = getOption("tidyllm_fbatch_default"),
  .dry_run = NULL,
  .max_tries = NULL,
  .timeout = NULL
)
}
\arguments{
\item{.llms}{A list of \code{LLMMessage} objects containing conversation histories.}

\item{.provider}{A function or function call specifying the language model provider and any additional parameters.
This should be a call to a provider function like \code{openai()}, \code{claude()}, etc.
You can also set a default provider function via the \code{tidyllm_fbatch_default} option.}

\item{.dry_run}{Logical; if \code{TRUE}, returns the constructed request without executing it}

\item{.max_tries}{Integer; maximum number of retries if the request fails}

\item{.timeout}{Integer; request timeout in seconds}
}
\value{
A list of updated \code{LLMMessage} objects, each with the assistant's response added if successful.
}
\description{
This function retrieves the results of a completed  batch and updates
the provided list of \code{LLMMessage} objects with the responses. It aligns each
response with the original request using the \code{custom_id}s generated in \code{send_batch()}.
}
\details{
The function routes the input to the appropriate provider-specific batch API function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llm_verbs.R
\name{upload_file}
\alias{upload_file}
\title{Upload a File to a Provider's File Store}
\usage{
upload_file(.provider, .path, ...)
}
\arguments{
\item{.provider}{A provider function call (e.g. \code{claude()}, \code{gemini()}, \code{openai()}).}

\item{.path}{Path to the local file to upload.}

\item{...}{Additional provider-specific arguments.}
}
\value{
A \code{tidyllm_file} object.
}
\description{
Upload a File to a Provider's File Store
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_gemini.R
\name{list_gemini_batches}
\alias{list_gemini_batches}
\title{List Recent Gemini Batch Operations}
\usage{
list_gemini_batches(
  .filter = NULL,
  .page_size = 20,
  .timeout = 60,
  .max_tries = 3,
  .dry_run = FALSE
)
}
\arguments{
\item{.filter}{Optional filter expression for batch listing (see Gemini API docs).}

\item{.page_size}{Integer. Maximum number of results to return. Default: 20.}

\item{.timeout}{Integer. Request timeout in seconds. Default: 60.}

\item{.max_tries}{Integer. Maximum retry attempts. Default: 3.}

\item{.dry_run}{Logical. If TRUE, returns the request object (for debugging). Default: FALSE.}
}
\value{
A tibble with columns: name, state, done, create_time, complete_time.
}
\description{
Returns a tibble with recent Gemini batch operations and their metadata.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llm_verbs.R
\name{embed}
\alias{embed}
\title{Generate text embeddings}
\usage{
embed(
  .input,
  .provider = getOption("tidyllm_embed_default"),
  .model = NULL,
  .truncate = NULL,
  .timeout = NULL,
  .dry_run = NULL,
  .max_tries = NULL
)
}
\arguments{
\item{.input}{A character vector of texts v, a list of texts and image objects,  or an \code{LLMMessage} object}

\item{.provider}{A function or function call specifying the language model provider and any additional parameters.
This should be a call to a provider function like \code{openai()}, \code{ollama()}, etc.
You can also set a default provider function via the \code{tidyllm_embed_default} option.}

\item{.model}{The embedding model to use}

\item{.truncate}{Whether to truncate inputs to fit the model's context length}

\item{.timeout}{Timeout for the API request in seconds}

\item{.dry_run}{If TRUE, perform a dry run and return the request object.}

\item{.max_tries}{Maximum retry attempts for requests}
}
\value{
A tibble with two columns: \code{input} and \code{embeddings}.
The \code{input} column contains the texts sent to embed, and the \code{embeddings} column
is a list column where each row contains an embedding vector of the sent input.
}
\description{
The \code{embed()} function allows you to embed a text via a specified provider.
It routes the input to the appropriate provider-specific embedding function.
}
\examples{
\dontrun{
c("What is the meaning of life, the universe and everything?",
 "How much wood would a woodchuck chuck?",
 "How does the brain work?") |>
 embed(gemini)
 }
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_azure_openai.R
\name{fetch_azure_openai_batch}
\alias{fetch_azure_openai_batch}
\title{Fetch Results for an Azure OpenAI Batch}
\usage{
fetch_azure_openai_batch(
  .llms,
  .endpoint_url = Sys.getenv("AZURE_ENDPOINT_URL"),
  .batch_id = NULL,
  .dry_run = FALSE,
  .max_tries = 3,
  .timeout = 60
)
}
\arguments{
\item{.llms}{A list of \code{LLMMessage} objects that were part of the batch.}

\item{.endpoint_url}{Base URL for the API (default:  Sys.getenv("AZURE_ENDPOINT_URL")).}

\item{.batch_id}{Character; the unique identifier for the batch. By default this is NULL
and the function will attempt to use the \code{batch_id} attribute from \code{.llms}.}

\item{.dry_run}{Logical; if \code{TRUE}, returns the constructed request without executing it (default: \code{FALSE}).}

\item{.max_tries}{Integer; maximum number of retries if the request fails (default: \code{3}).}

\item{.timeout}{Integer; request timeout in seconds (default: \code{60}).}
}
\value{
A list of updated \code{LLMMessage} objects, each with the assistant's response added if successful.
}
\description{
This function retrieves the results of a completed Azure OpenAI batch and updates
the provided list of \code{LLMMessage} objects with the responses. It aligns each
response with the original request using the \code{custom_id}s generated in \code{send_azure_openai_batch()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/message_retrieval.R
\name{get_reply_data}
\alias{get_reply_data}
\alias{last_reply_data}
\title{Retrieve Assistant Reply as Structured Data}
\usage{
get_reply_data(.llm, .index = NULL)

last_reply_data(.llm)
}
\arguments{
\item{.llm}{An \code{LLMMessage} object containing the message history.}

\item{.index}{A positive integer indicating the index of the assistant reply to retrieve.
Defaults to \code{NULL}, which retrieves the last reply.}
}
\value{
Returns the parsed data from the assistant's reply, or \code{NULL} if parsing fails.
}
\description{
Parses the assistant's reply as JSON and returns the corresponding structured data.
If the reply is not marked as JSON, attempts to extract and parse JSON content from the text.
}
\details{
For convenience, \code{\link[=last_reply_data]{last_reply_data()}} is provided as a wrapper to retrieve the
latest assistant reply's data.
}
\seealso{
\code{\link[=get_reply]{get_reply()}}, \code{\link[=last_reply_data]{last_reply_data()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/media.R
\name{img}
\alias{img}
\title{Create an Image Object}
\usage{
img(.path)
}
\arguments{
\item{.path}{The path to the image file on disk.}
}
\value{
A \code{tidyllm_image} object.
}
\description{
This function reads an image file from disk, encodes it in base64,
and returns a \code{tidyllm_image} object that can be used in multimodal
embedding requests or attached to messages via \code{.media}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_ollama.R
\name{ollama}
\alias{ollama}
\title{Ollama API Provider Function}
\usage{
ollama(..., .called_from = NULL)
}
\arguments{
\item{...}{Parameters to be passed to the appropriate Ollama-specific function,
such as model configuration, input text, or API-specific options.}

\item{.called_from}{An internal argument specifying the verb (e.g., \code{chat}, \code{embed})
the function is invoked from. This argument is automatically managed by \code{tidyllm} and
should not be set by the user.}
}
\value{
The result of the requested action:
\itemize{
\item For \code{chat()}: An updated \code{LLMMessage} object containing the model's response.
\item For \code{embed()}: A matrix where each column corresponds to an embedding.
}
}
\description{
The \code{ollama()} function acts as an interface for interacting with local AI models via the Ollama API.
It integrates seamlessly with the main \code{tidyllm} verbs such as \code{chat()} and \code{embed()}.
}
\details{
Some functionalities, like \code{ollama_download_model()} or \code{ollama_list_models()}
are unique to the Ollama API and do not have a general verb counterpart.
These functions can be only accessed directly.

Supported Verbs:
\itemize{
\item \strong{\code{chat()}}: Sends a message to an Ollama model and retrieves the model's response.
\item \strong{\code{embed()}}: Generates embeddings for input texts using an Ollama model.
\item \strong{\code{send_batch()}}: Behaves different than the other \code{send_batch()} verbs since it immediately processes the answers
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_claude.R
\name{claude_chat}
\alias{claude_chat}
\title{Interact with Claude AI models via the Anthropic API}
\usage{
claude_chat(
  .llm,
  .model = "claude-sonnet-4-6",
  .max_tokens = 2048,
  .temperature = NULL,
  .top_k = NULL,
  .top_p = NULL,
  .metadata = NULL,
  .stop_sequences = NULL,
  .tools = NULL,
  .json_schema = NULL,
  .file_ids = NULL,
  .api_url = "https://api.anthropic.com/",
  .verbose = FALSE,
  .max_tries = 3,
  .timeout = 60,
  .stream = FALSE,
  .dry_run = FALSE,
  .thinking = FALSE,
  .thinking_budget = 1024,
  .max_tool_rounds = 10
)
}
\arguments{
\item{.llm}{An LLMMessage object containing the conversation history and system prompt.}

\item{.model}{Character string specifying the Claude model version (default: "claude-sonnet-4-6").}

\item{.max_tokens}{Integer specifying the maximum number of tokens in the response (default: 1024).}

\item{.temperature}{Numeric between 0 and 1 controlling response randomness.}

\item{.top_k}{Integer controlling diversity by limiting the top K tokens.}

\item{.top_p}{Numeric between 0 and 1 for nucleus sampling.}

\item{.metadata}{List of additional metadata to include with the request.}

\item{.stop_sequences}{Character vector of sequences that will halt response generation.}

\item{.tools}{List of additional tools or functions the model can use.}

\item{.json_schema}{A schema to enforce an output structure}

\item{.file_ids}{Character; A vector of file IDs for files that were uploaded to Anthropics Servers}

\item{.api_url}{Base URL for the Anthropic API (default: "https://api.anthropic.com/").}

\item{.verbose}{Logical; if TRUE, displays additional information about the API call (default: FALSE).}

\item{.max_tries}{Maximum retries to peform request}

\item{.timeout}{Integer specifying the request timeout in seconds (default: 60).}

\item{.stream}{Logical; if TRUE, streams the response piece by piece (default: FALSE).}

\item{.dry_run}{Logical; if TRUE, returns the prepared request object without executing it (default: FALSE).}

\item{.thinking}{Logical; if TRUE, enables Claude's thinking mode for complex reasoning tasks (default: FALSE).}

\item{.thinking_budget}{Integer specifying the maximum tokens Claude can spend on thinking (default: 1024). Must be at least 1024.}

\item{.max_tool_rounds}{Integer specifying the maximum number of tool use iterations (default: 10).
Set to 1 for single-round tool use, or higher for multi-turn agentic loops.}
}
\value{
A new LLMMessage object containing the original messages plus Claude's response.
}
\description{
Interact with Claude AI models via the Anthropic API
}
\examples{
\dontrun{
# Basic usage
msg <- llm_message("What is R programming?")
result <- claude_chat(msg)

# With custom parameters
result2 <- claude_chat(msg, 
                 .temperature = 0.7, 
                 .max_tokens = 1000)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidyllm-package.R
\docType{package}
\name{tidyllm-package}
\alias{tidyllm}
\alias{tidyllm-package}
\title{tidyllm: Tidy Integration of Large Language Models}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

A tidy interface for integrating large language model (LLM) APIs such as 'Claude', 'Openai', 'Gemini','Mistral' and local models via 'Ollama' into R workflows. The package supports text and media-based interactions, interactive message history, batch request APIs, and a tidy, pipeline-oriented interface for streamlined integration into data workflows. Web services are available at \url{https://www.anthropic.com}, \url{https://openai.com}, \url{https://aistudio.google.com/}, \url{https://mistral.ai/} and \url{https://ollama.com}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://edubruell.github.io/tidyllm/}
  \item Report bugs at \url{https://github.com/edubruell/tidyllm/issues}
}

}
\author{
\strong{Maintainer}: Eduard Brüll \email{eduard.bruell@zew.de}

Other contributors:
\itemize{
  \item Jia Zhang [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_gemini.R
\name{gemini_chat}
\alias{gemini_chat}
\title{Send LLMMessage to Gemini API}
\usage{
gemini_chat(
  .llm,
  .model = "gemini-2.5-flash",
  .fileid = NULL,
  .temperature = NULL,
  .max_output_tokens = NULL,
  .top_p = NULL,
  .top_k = NULL,
  .grounding_threshold = NULL,
  .presence_penalty = NULL,
  .frequency_penalty = NULL,
  .stop_sequences = NULL,
  .safety_settings = NULL,
  .json_schema = NULL,
  .tools = NULL,
  .thinking_budget = NULL,
  .timeout = 120,
  .dry_run = FALSE,
  .max_tries = 3,
  .verbose = FALSE,
  .stream = FALSE,
  .max_tool_rounds = 10
)
}
\arguments{
\item{.llm}{An existing LLMMessage object or an initial text prompt.}

\item{.model}{The model identifier (default: "gemini-1.5-flash").}

\item{.fileid}{Optional vector of file IDs uploaded via \code{gemini_upload_file()} (default: NULL).}

\item{.temperature}{Controls randomness in generation (default: NULL, range: 0.0-2.0).}

\item{.max_output_tokens}{Maximum tokens in the response (default: NULL).}

\item{.top_p}{Controls nucleus sampling (default: NULL, range: 0.0-1.0).}

\item{.top_k}{Controls diversity in token selection (default: NULL, range: 0 or more).}

\item{.grounding_threshold}{A grounding threshold between 0 and 1. With lower
grounding thresholds  Gemini will use Google to search for relevant information
before answering.  (default: NULL).}

\item{.presence_penalty}{Penalizes new tokens (default: NULL, range: -2.0 to 2.0).}

\item{.frequency_penalty}{Penalizes frequent tokens (default: NULL, range: -2.0 to 2.0).}

\item{.stop_sequences}{Optional character sequences to stop generation (default: NULL, up to 5).}

\item{.safety_settings}{A list of safety settings (default: NULL).}

\item{.json_schema}{A schema to enforce an output structure}

\item{.tools}{Either a single TOOL object or a list of TOOL objects representing the available functions for tool calls.}

\item{.thinking_budget}{Token budget for internal reasoning (default: NULL). Works with \code{gemini-2.5-flash} and \code{gemini-2.5-pro}.}

\item{.timeout}{When should our connection time out (default: 120 seconds).}

\item{.dry_run}{If TRUE, perform a dry run and return the request object.}

\item{.max_tries}{Maximum retries to perform request (default: 3).}

\item{.verbose}{Should additional information be shown after the API call.}

\item{.stream}{Should the response be streamed (default: FALSE).}

\item{.max_tool_rounds}{Integer specifying the maximum number of tool use iterations (default: 10).
Set to 1 for single-round tool use, or higher for multi-turn agentic loops.}
}
\value{
A new \code{LLMMessage} object containing the original messages plus the assistant's response.
}
\description{
Send LLMMessage to Gemini API
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_ollama.R
\name{ollama_download_model}
\alias{ollama_download_model}
\title{Download a model from the Ollama API}
\usage{
ollama_download_model(.model, .ollama_server = "http://localhost:11434")
}
\arguments{
\item{.model}{The name of the model to download.}

\item{.ollama_server}{The base URL of the Ollama API (default is "http://localhost:11434").}
}
\description{
This function sends a request to the Ollama API to download a specified model
from Ollama's large online library of models.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_claude.R
\name{send_claude_batch}
\alias{send_claude_batch}
\title{Send a Batch of Messages to Claude API}
\usage{
send_claude_batch(
  .llms,
  .model = "claude-sonnet-4-6",
  .max_tokens = 1024,
  .temperature = NULL,
  .top_k = NULL,
  .top_p = NULL,
  .stop_sequences = NULL,
  .json_schema = NULL,
  .thinking = FALSE,
  .thinking_budget = 1024,
  .api_url = "https://api.anthropic.com/",
  .verbose = FALSE,
  .dry_run = FALSE,
  .overwrite = FALSE,
  .max_tries = 3,
  .timeout = 60,
  .id_prefix = "tidyllm_claude_req_"
)
}
\arguments{
\item{.llms}{A list of LLMMessage objects containing conversation histories.}

\item{.model}{Character string specifying the Claude model version (default: "claude-sonnet-4-6").}

\item{.max_tokens}{Integer specifying the maximum tokens per response (default: 1024).}

\item{.temperature}{Numeric between 0 and 1 controlling response randomness.}

\item{.top_k}{Integer for diversity by limiting the top K tokens.}

\item{.top_p}{Numeric between 0 and 1 for nucleus sampling.}

\item{.stop_sequences}{Character vector of sequences that halt response generation.}

\item{.json_schema}{A schema to enforce an output structure}

\item{.thinking}{Logical; if TRUE, enables Claude's thinking mode for complex reasoning tasks (default: FALSE).}

\item{.thinking_budget}{Integer specifying the maximum tokens Claude can spend on thinking (default: 1024). Must be at least 1024.

Defaults to "tidyllm_claude_req_".}

\item{.api_url}{Base URL for the Claude API (default: "https://api.anthropic.com/").}

\item{.verbose}{Logical; if TRUE, prints a message with the batch ID (default: FALSE).}

\item{.dry_run}{Logical; if TRUE, returns the prepared request object without executing it (default: FALSE).}

\item{.overwrite}{Logical; if TRUE, allows overwriting an existing batch ID associated with the request (default: FALSE).}

\item{.max_tries}{Maximum number of retries to perform the request.}

\item{.timeout}{Integer specifying the request timeout in seconds (default: 60).}

\item{.id_prefix}{Character string to specify a prefix for generating custom IDs when names in \code{.llms} are missing.}
}
\value{
An updated and named list of \code{.llms} with identifiers that align with batch responses, including a \code{batch_id} attribute.
}
\description{
This function creates and submits a batch of messages to the Claude API for asynchronous processing.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_gemini.R
\name{gemini_file_metadata}
\alias{gemini_file_metadata}
\title{Retrieve Metadata for a File from Gemini API}
\usage{
gemini_file_metadata(.file_name)
}
\arguments{
\item{.file_name}{The file ID (e.g., "files/abc-123") to retrieve metadata for.}
}
\value{
A tibble containing metadata fields such as name, display name, MIME type, size, and URI.
}
\description{
Retrieves metadata for a specific file uploaded to the Gemini API.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/message_retrieval.R
\name{get_logprobs}
\alias{get_logprobs}
\title{Retrieve Log Probabilities from Assistant Replies}
\usage{
get_logprobs(.llm, .index = NULL)
}
\arguments{
\item{.llm}{An \code{LLMMessage} object containing the message history.}

\item{.index}{A positive integer specifying which assistant reply's log probabilities to extract.
If \code{NULL} (default), log probabilities for all replies are returned.}
}
\value{
A tibble containing log probabilities for the specified assistant reply or all replies.
}
\description{
Extracts token log probabilities from assistant replies within an \code{LLMMessage} object.
Each row represents a token with its log probability and top alternative tokens.
}
\details{
An empty tibble is output if no logprobs were requested. Works with \code{openai_chat()}, \code{llamacpp_chat()}, and other providers that support logprobs.

Columns include:
\itemize{
\item \code{reply_index}: The index of the assistant reply in the message history.
\item \code{token}: The generated token.
\item \code{logprob}: The log probability of the generated token.
\item \code{bytes}: The byte-level encoding of the token.
\item \code{top_logprobs}: A list column containing the top alternative tokens with their log probabilities.
}
}
\seealso{
\code{\link[=get_metadata]{get_metadata()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_openrouter.R
\name{openrouter_generation}
\alias{openrouter_generation}
\title{Get Details for an OpenRouter Generation}
\usage{
openrouter_generation(
  .id,
  .api_url = "https://openrouter.ai",
  .timeout = 60,
  .max_tries = 3
)
}
\arguments{
\item{.id}{The generation ID string (e.g. \code{"gen-..."}) returned by OpenRouter.}

\item{.api_url}{Base URL for the OpenRouter API (default: \code{"https://openrouter.ai"}).}

\item{.timeout}{Request timeout in seconds (default: 60).}

\item{.max_tries}{Maximum retries (default: 3).}
}
\value{
A named list with generation details including \code{id}, \code{model},
\code{total_cost}, \code{tokens_prompt}, \code{tokens_completion}, and \code{latency}.
}
\description{
Fetches cost, latency, and token details for a past generation using its ID.
The generation ID is available in \code{get_metadata(result)$api_specific$id}
after a \code{chat()} call.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_claude.R
\name{claude_websearch}
\alias{claude_websearch}
\title{Builtin Claude Web Search Tool}
\usage{
claude_websearch()
}
\description{
Returns a TOOL object for Claude's builtin web_search tool.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_gemini.R
\name{gemini_embedding}
\alias{gemini_embedding}
\title{Generate Embeddings Using the Google Gemini API}
\usage{
gemini_embedding(
  .input,
  .model = "gemini-embedding-2-preview",
  .truncate = TRUE,
  .timeout = 120,
  .dry_run = FALSE,
  .max_tries = 3
)
}
\arguments{
\item{.input}{A character vector of texts to embed or an \code{LLMMessage} object}

\item{.model}{The embedding model identifier (default: "text-embedding-3-small").}

\item{.truncate}{Whether to truncate inputs to fit the model's context length (default: TRUE).}

\item{.timeout}{Timeout for the API request in seconds (default: 120).}

\item{.dry_run}{If TRUE, perform a dry run and return the request object.}

\item{.max_tries}{Maximum retry attempts for requests (default: 3).}
}
\value{
A matrix where each column corresponds to the embedding of a message in the message history.
}
\description{
Generate Embeddings Using the Google Gemini API
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_llamacpp.R
\name{llamacpp}
\alias{llamacpp}
\title{llama.cpp Provider Function}
\usage{
llamacpp(..., .called_from = NULL)
}
\arguments{
\item{...}{Parameters passed to the appropriate llama.cpp-specific function.}

\item{.called_from}{An internal argument specifying which verb invoked this
function. Managed automatically by tidyllm verbs; do not set manually.}
}
\value{
The result of the requested action (e.g., an updated \code{LLMMessage}
for \code{chat()}, a tibble for \code{embed()} or \code{list_models()}).
}
\description{
The \code{llamacpp()} provider connects tidyllm to a locally running
\href{https://github.com/ggml-org/llama.cpp}{llama.cpp} server. It exposes
the same verb/provider pattern as every other tidyllm provider while
also offering llama.cpp-specific features: BNF grammar constraints
(\code{.grammar}), token logprobs (\code{.logprobs}), and model management helpers.

The server must be started separately before calling \code{llamacpp()}. See
\code{llamacpp_health()} to verify the server is running, and
\code{llamacpp_download_model()} / \code{list_hf_gguf_files()} to obtain models.
}
